#include "glm.h"

namespace tadah_fussy
{
  namespace statistics
  {
    using math_lib::SumOp;
    using math_lib::Sum;
    using math_lib::SumAll;
    using math_lib::OpBind;
    using math_lib::solver::GradientMethod;

    using matrix_lib::PrintArray;
    using matrix_lib::PrintMatrix;

    using error_lib::LE;
    using error_lib::EQ;
    using error_lib::PNum;
    using error_lib::CheckArrayLength;
    using error_lib::CheckLinearModel;

    using std::vector;
    using std::cerr;
    using std::endl;
    using std::string;

    using stat_type = distribution::Base::stat_type;
    using counter_type = distribution::Base::counter_type;

    namespace
    {
      /*
        行列に対する比較用関数オブジェクト
      */
      template< class MATRIX >
      struct LessMatrixIterator
      {
        bool operator()( const typename MATRIX::const_iterator& i1, const typename MATRIX::const_iterator& i2 ) const
        {
          auto i1_ = i1;
          auto i2_ = i2;

          for ( ; i1_ != i1.end() && i2_ != i2.end() ; ++i1_ , ++i2_ )
            if ( *i1_ != *i2_ ) return( *i1_ < *i2_ );

          // 先頭が一致していれば false とする ( 要素数はチェックしない )
          return( false );
        }
      };

    }

    namespace glm
    {
      /*
        Generic::calcTerm : 期待値 μ から g'(μ) と w = V[η]・g'(μ)^2 を求める

        η は E[η] = μ を満たす確率変数である。
      */
      void Generic::calcTerm()
      {
        auto dxi_p = dxi_.begin();
        auto w_p = w_.begin();

        for ( auto mu_s = mu_.begin() ; mu_s != mu_.end() ; ++mu_s ) {
          *dxi_p = g_->df( *mu_s );
          stat_type eta = pdf_->aveInv( *mu_s );
          stat_type var = pdf_->variance( eta );
          *w_p = var * std::pow( *dxi_p, 2.0 );
          ++w_p;
          ++dxi_p;
        }
      }

      /*
        Generic::createLES : 連立方程式の構築
      */
      void Generic::createLES( SquareMatrix< stat_type >* sa, vector< stat_type >* sb )
      {
        vector< stat_type > ri( n_ ); // 右辺の x に掛ける係数

        // 係数行列の計算 : w を重み付け係数とした内積 ( x[j], x[k] )
        for ( size_type j = 0 ; j < p_ ; ++j ) {
          for ( size_type k = j ; k < p_ ; ++k ) {
            (*sa)[j][k] = x_[0][j] * x_[0][k] / w_[0];
            for ( size_type i = 1 ; i < n_ ; ++i )
              (*sa)[j][k] += x_[i][j] * x_[i][k] / w_[i];
            (*sa)[k][j] = (*sa)[j][k];
          }
        }

        // 右辺の計算
        for ( size_type i = 0 ; i < n_ ; ++i )
          ri[i] = ( (*g_)( mu_[i] ) + ( y_[i] - mu_[i] ) * dxi_[i] ) / w_[i];
        for ( size_type j = 0 ; j < p_ ; ++j ) {
          auto col = x_.column( j );
          (*sb)[j] = std::inner_product( col, col.end(), ri.begin(), stat_type() );
        }
      }

      /*
        Generic::init : 係数計算のための連立方程式 s と係数を保持する配列 a を初期化する
      */
      void Generic::init( SquareMatrix< stat_type >* sa, vector< stat_type >* sb, vector< stat_type >* a )
      {
        n_ = x_.rows(); // 独立変数ベクトル xi の数
        p_ = x_.cols(); // 独立変数ベクトルの要素数

        mu_.assign( y_.begin(), y_.end() ); // y の期待値 (yで初期化)
        dxi_.resize( n_ );// g'(mu)
        w_.resize( n_ );  // 対角行列の要素

        a->assign( p_, stat_type() ); // 係数の初期化
        sa->assign( p_ ); // 連立方程式の初期化
        sb->resize( p_ );

#ifdef DEBUG
        cerr << "N = " << n_ << "; p = " << p_ << endl << endl;

        PrintMatrix( cerr, x_, "x = " );
        cerr << endl << endl;
        PrintArray( cerr, y_.begin(), y_.end(), "y = " );
        cerr << endl;
        PrintArray( cerr, mu_.begin(), mu_.end(), "mu = " );
        cerr << endl << endl;
#endif
      }

      /*
        Generic::create : 係数計算のための連立方程式 s を作成する
      */
      void Generic::create( SquareMatrix< stat_type >* sa, vector< stat_type >* sb )
      {
        calcTerm();     // g'(mu) と対角行列の計算

#ifdef DEBUG
        PrintArray( cerr, dxi_.begin(), dxi_.end(), "g'(mu) = " );
        cerr << endl;
        PrintArray( cerr, w_.begin(), w_.end(), "w = " );
        cerr << endl << endl;
#endif

        createLES( sa, sb ); // 連立方程式の構築

#ifdef DEBUG
        cerr << "Equation System :" << endl;
        PrintMatrix( cerr, *sa );
        cerr << endl;
        PrintArray( cerr, sb->begin(), sb->end() );
        cerr << endl;
#endif
      }

      /*
        Generic::update : 係数計算後の更新処理を行う
      */
      void Generic::update( const vector< stat_type >& a )
      {
        for ( size_type i = 0 ; i < n_ ; ++i )
          mu_[i] = g_->invf( std::inner_product( a.begin(), a.end(), x_.row( i ), stat_type() ) );

#ifdef DEBUG
        PrintArray( cerr, a.begin(), a.end(), "a = " );
        cerr << endl;
        PrintArray( cerr, mu_.begin(), mu_.end(), "mu = " );
        cerr << endl << endl;
#endif
      }

      /*
        Generic::fit : 当てはめ値の計算
      */
      void Generic::fit()
      {
        size_type n = x_.rows(); // 独立変数の数

        yExp_.resize( n );

        for ( size_type i = 0 ; i < n ; ++i ) {
          auto row = x_.row( i );
          yExp_[i] = predict( row, row.end() );
        }
      }

      /*
        Generic::calcLa : 当てはめ値に対する最大対数尤度の計算
      */
      void Generic::calcLa()
      {
        using size_type = std::vector< stat_type >::size_type;

        la_ = stat_type();
        for ( size_type i = 0 ; i < y_.size() ; ++i ) {
          stat_type eta = pdf_->aveInv( yExp_[i] );
          la_ += pdf_->logLikelihood( eta, y_[i] );
        }
      }

      /*
        Generic::calcLmax : 最大モデルに対する最大対数尤度の計算
      */
      void Generic::calcLmax()
      {
        using size_type = std::vector< stat_type >::size_type;

        lmax_ = stat_type();
        for ( size_type i = 0 ; i < y_.size() ; ++i ) {
          stat_type eta = pdf_->aveInv( y_[i] );
          lmax_ += pdf_->logLikelihood( eta, y_[i] );
        }
      }

      /*
        Generic::calcLmin : 最小モデルに対する最大対数尤度の計算
      */
      void Generic::calcLmin()
      {
        using size_type = std::vector< stat_type >::size_type;

        stat_type eta = pdf_->aveInv( MeanAll< stat_type >( y_ ) );
        lmin_ = stat_type();
        for ( size_type i = 0 ; i < y_.size() ; ++i ) {
          lmin_ += pdf_->logLikelihood( eta, y_[i] );
        }
      }

      /*
        BinomialLogistic::calcLmin : 最小モデルに対する最大対数尤度の計算
      */
      void BinomialLogistic::calcLmin()
      {
        using size_type = std::vector< stat_type >::size_type;

        stat_type mean = SumAll< stat_type >( y_ ) / SumAll< stat_type >( n_ );

        // BinomialLogistic では y_ のサイズ分関数を呼び出さないと
        // 辻褄が合わなくなることに注意
        lmin_ = stat_type();
        for ( size_type i = 0 ; i < y_.size() ; ++i ) {
          stat_type eta = pdf_->aveInv( n_[i] * mean );
          lmin_ += pdf_->logLikelihood( eta, y_[i] );
        }
      }

      /*
        ピアソン・カイ二乗統計量 X^2 を求める
      */
      stat_type BinomialLogistic::pearsonChiSq() const
      {
        auto yi = y_.begin();
        auto ni = n_.begin();

        stat_type xi2 = 0;
        for ( auto npi = yExp_.begin() ; npi != yExp_.end() ; ++npi ) {
          xi2 += *ni * std::pow( *yi - *npi, 2.0 ) / ( *npi * ( *ni - *npi ) );
          ++yi;
          ++ni;
        }

        return( xi2 );
      }

      /*
        多項分布モデル用計算ルーチン

        pi : 確率の当てはめ値
        ni : 各独立変数ごとの総数
        x : 独立変数
        w : 対角行列 W_kk'
        s : 連立方程式
        c : カテゴリ数
        p : 独立変数ベクトルの要素数
        n : 独立変数ベクトルの数
      */
      namespace
      {
        /*
          GLM_ML_calcDiagMatrix : 多項分布モデル用対角行列計算

          pi : 確率の当てはめ値
          ni : 各独立変数ごとの総数
          w : 対角行列 W_kk'
          c : カテゴリ数
          n : 独立変数ベクトルの数
        */
        void GLM_ML_calcDiagMatrix
        ( const Matrix< stat_type >& pi, const vector< stat_type >& ni, vector< stat_type >* w,
          size_t c, size_t n )
        {
          auto it = w->begin();
          for ( size_t rk = 0 ; rk < c - 1 ; ++rk ) {
            // 対角成分にあたる分割行列を先に処理する
            for ( size_t i = 0 ; i < n ; ++i )
              *it++ = pi[i][rk + 1] * ni[i] * ( 1.0 - pi[i][rk + 1] );
            // その他の分割行列を処理
            for ( size_t ck = rk + 1 ; ck < c - 1 ; ++ck )
              for ( size_t i = 0 ; i < n ; ++i )
                *it++ = -pi[i][rk + 1] * pi[i][ck + 1] * ni[i];
          }
        }

        /*
          GLM_ML_calcCoef : 多項分布モデル用係数計算

          w : 対角行列 W_kk'
          x : 独立変数
          sa : 係数行列へのポインタ
          c : カテゴリ数
          p : 独立変数ベクトルの要素数
          n : 独立変数ベクトルの数
          rk, ck : 分割行列の行列番号
          rj, cj : 分割行列内の計算対象要素の行列番号
        */
        template< template< class > class MATRIX >
        void GLM_ML_calcCoef
        ( const vector< stat_type >& w, const Matrix< stat_type >& x,
          MATRIX< stat_type >* sa,
          size_t c, size_t p, size_t n,
          size_t rk, size_t rj, size_t ck, size_t cj )
        {
          (*sa)[rk * p + rj][ck * p + cj] = 0;

          auto cit = w.begin() + n * ( ( c * ( c - 1 ) - ( c - rk - 1 ) * ( c - rk ) ) / 2 + ck - rk );

          for ( size_t i = 0 ; i < n ; ++i )
            (*sa)[rk * p + rj][ck * p + cj] += x[i][rj] * x[i][cj] * *( cit + i );
        }

        /*
          GLM_ML_calcCoefMatrix : 多項分布モデル用係数行列の計算

          w : 対角行列 W_kk'
          x : 独立変数
          sa : 係数行列へのポインタ
          c : カテゴリ数
          p : 独立変数ベクトルの要素数
          n : 独立変数ベクトルの数
        */
        template< template< class > class MATRIX >
        void GLM_ML_calcCoefMatrix
        ( const vector< stat_type >& w, const Matrix< stat_type >& x,
          MATRIX< stat_type >* sa,
          size_t c, size_t p, size_t n )
        {
          for ( size_t rk = 0 ; rk < c - 1 ; ++rk ) {
            for ( size_t rj = 0 ; rj < p ; ++rj ) {
              // 分割行列もその要素も対角成分
              GLM_ML_calcCoef( w, x, sa, c, p, n, rk, rj, rk, rj );
              // 分割行列が対角成分
              for ( size_t cj = rj + 1 ; cj < p ; ++cj ) {
                GLM_ML_calcCoef( w, x, sa, c, p, n, rk, rj, rk, cj );
                (*sa)[rk * p + cj][rk * p + rj] = (*sa)[rk * p + rj][rk * p + cj];
              }
            }
            for ( size_t ck = rk + 1 ; ck < c - 1 ; ++ck ) {
              for ( size_t rj = 0 ; rj < p ; ++rj ) {
                // 分割行列は非対角で要素は対角成分
                GLM_ML_calcCoef( w, x, sa, c, p, n, rk, rj, ck, rj );
                (*sa)[ck * p + rj][rk * p + rj] = (*sa)[rk * p + rj][ck * p + rj];
                // 分割行列もその要素も非対角
                for ( size_t cj = rj + 1 ; cj < p ; ++cj ) {
                  GLM_ML_calcCoef( w, x, sa, c, p, n, rk, rj, ck, cj );
                  (*sa)[ck * p + cj][rk * p + rj] = (*sa)[ck * p + rj][rk * p + cj] =
                    (*sa)[rk * p + cj][ck * p + rj] = (*sa)[rk * p + rj][ck * p + cj];
                }
              }
            }
          }
        }

        /*
          GLM_ML_calcRSide : 多項分布モデル用連立方程式の右辺の計算

          w : 対角行列 W_kk'
          x : 独立変数
          y : 従属変数(各カテゴリの発生回数)
          ni : 各独立変数ごとの総数
          pi : 確率の当てはめ値
          sb : 右辺へのポインタ
          c : カテゴリ数
          p : 独立変数ベクトルの要素数
          n : 独立変数ベクトルの数
        */
        void GLM_ML_calcRSide
        ( const vector< stat_type >& w, const Matrix< stat_type >& x, const Matrix< stat_type >& y,
          const vector< stat_type >& ni, const Matrix< stat_type >& pi, std::vector< stat_type >* sb,
          size_t c, size_t p, size_t n )
        {
          for ( size_t k = 0 ; k < c - 1 ; ++k ) {
            for ( size_t j = 0 ; j < p ; ++j ) {
              (*sb)[k * p + j] = 0;
              for ( size_t k2 = 0 ; k2 < c - 1 ; ++k2 ) {
                size_t rk = ( k > k2 ) ? k2 : k; // 分割行列の行番号
                size_t ck = ( k > k2 ) ? k : k2; // 分割行列の列番号
                auto cit = w.begin() + n * ( ( c * ( c - 1 ) - ( c - rk - 1 ) * ( c - rk ) ) / 2 + ck - rk );
                for ( size_t i = 0 ; i < n ; ++i )
                  (*sb)[k * p + j] += x[i][j] * std::log( pi[i][k2 + 1] / pi[i][0] ) * *( cit + i );
              }
              stat_type u_kj = 0; // Ukj の値
              for ( size_t i = 0 ; i < n ; ++i )
                u_kj += ( y[i][k + 1] - ni[i] * pi[i][k + 1] ) * x[i][j];
              (*sb)[k * p + j] += u_kj;
            }
          }
        }

        /*
          GLM_ML_initPi : ni, pi の初期化

          y の行ごとに、総和 ni とその相対比率 pi を計算する
        */
        void GLM_ML_initPi( const Matrix< stat_type >& y, vector< stat_type >* ni, Matrix< stat_type >* pi )
        {
          using size_type = Matrix< stat_type >::size_type;

          size_type n = y.rows();
          size_type c = y.cols();

          ni->resize( n );
          for ( size_type i = 0 ; i < n ; ++i ) {
            Matrix< stat_type >::const_iterator row = y.row( i );
            (*ni)[i] = Sum< stat_type >( row, row.end() );
          }

          pi->assign( n, c ); // 確率の当てはめ値 ( n x c )
          // piは y_ik / ni_i で初期化
          for ( size_type i = 0 ; i < n ; ++i )
            for ( size_type k = 0 ; k < c ; ++k )
              (*pi)[i][k] = y[i][k] / (*ni)[i];
        }

        /*
          GLM_CalcFitValue : 当てはめ値の計算用ヘルパ関数
          (名義・順序ロジスティック回帰専用)
        */
        void GLM_CalcFitValue( const Matrix< stat_type >& pi, const vector< stat_type >& ni,
                               Matrix< stat_type >* fitVal )
        {
          *fitVal = pi;
          for ( Matrix< stat_type >::size_type r = 0 ; r < fitVal->rows() ; ++r )
            for ( Matrix< stat_type >::size_type c = 0 ; c < fitVal->cols() ; ++c )
              (*fitVal)[r][c] *= ni[r];
        }

        /*
          GLM_CalcLa : 当てはめ値から最大対数尤度 lα を求めるヘルパ関数
          (名義・順序ロジスティック回帰専用)
        */
        template< class In1, class In2 >
        stat_type GLM_CalcLa( In1 ys, In1 ye, In2 pis )
        {
          stat_type l = stat_type();

          for ( ; ys != ye ; ++ys, ++pis )
            l += *ys * std::log( *pis );

          return( l );
        }

        /*
          GLM_CalcLMax : 最大モデルの最大対数尤度を求めるヘルパ関数
          (名義・順序ロジスティック回帰専用)
        */
        stat_type GLM_CalcLMax( const Matrix< stat_type >& y, const vector< stat_type >& ni )
        {
          using size_type = Matrix< stat_type >::size_type;

          stat_type l = stat_type();

          for ( size_type r = 0 ; r < y.rows() ; ++r )
            for ( size_type c = 0 ; c < y.cols() ; ++c )
              if ( y[r][c] != 0 )
                l += y[r][c] * std::log( y[r][c] / ni[r] );

          return( l );
        }

        /*
          GLM_CalcLMin : 最小モデルの最大対数尤度を求めるヘルパ関数
          (名義・順序ロジスティック回帰専用)
        */
        stat_type GLM_CalcLMin( const Matrix< stat_type >& y )
        {
          using size_type = Matrix< stat_type >::size_type;

          stat_type ySum = SumAll< stat_type >( y );
          vector< stat_type > yc( y.cols() );
          for ( vector< stat_type >::size_type c = 0 ;
                c < yc.size() ; ++c ) {
            auto col = y.cbegin( c );
            yc[c] = Sum< stat_type >( col, col.end() );
          }

          stat_type l = stat_type();

          for ( size_type r = 0 ; r < y.rows() ; ++r )
            for ( size_type c = 0 ; c < y.cols() ; ++c )
              if ( yc[c] != 0 )
                l += y[r][c] * std::log( yc[c] / ySum );

          return( l );
        }

        /*
          GLM_PearsonResiduals : ピアソン残差を求める
        */
        void GLM_PearsonResiduals( const Matrix< stat_type >& obj, const Matrix< stat_type >& exp,
                                   Matrix< stat_type >* res )
        {
          assert( res != nullptr );

          PearsonResiduals< stat_type, stat_type > pr;

          // スコア法が実行されていなかった場合、exp のサイズはゼロ
          // res は obj のサイズで初期化し、要素の代入は exp のサイズ分行う
          res->assign( obj.rows(), obj.cols() );
          for ( Matrix< stat_type >::size_type r = 0 ; r < exp.rows() ; ++r ) {
            for ( Matrix< stat_type >::size_type c = 0 ; c < exp.cols() ; ++c ) {
              (*res)[r][c] = pr( obj[r][c], exp[r][c] );
            }
          }
        }

        /*
          GLM_DevianceResiduals : 逸脱度残差を求める

          平方根の内部はゼロ以上である必要があるが、それは保証されないためこの関数は利用しない。
          名義・順序ロジスティックでは逸脱度残差は公開しないものとする
        */
        /*
        void GLM_DevianceResiduals( const Matrix< stat_type >& y, const std::vector< stat_type >& ni,
                                    const Matrix< stat_type >& pi, Matrix< stat_type >* res )
        {
          using size_type = Matrix< stat_type >::size_type;

          assert( res != nullptr );

          // スコア法実行前は pi は空。
          // 範囲外アクセスを避けるため pi の範囲で処理する。
          res->assign( pi.rows(), pi.cols(), stat_type() );

          for ( size_type r = 0 ; r < pi.rows() ; ++r ) {
            for ( size_type c = 0 ; c < pi.cols() ; ++c ) {
              if ( y[r][c] != 0 )
                (*res)[r][c] = std::sqrt( 2 * y[r][c] * ( std::log( y[r][c] / ni[r] ) - std::log( pi[r][c] ) ) );
            }
          }
        }
        */

      } // namespace

      /*
        MultinomialLogistic コンストラクタ : 独立変数 x と従属変数 y を指定して構築

        N は独立変数ベクトルの数、P は独立変数ベクトルの要素数、C はカテゴリの数としたとき、
        x は N x P 行列、y は N x C 行列である必要がある。
      */
      MultinomialLogistic::MultinomialLogistic
      ( const Matrix< stat_type >& x, const Matrix< stat_type >& y )
        : x_( x ), y_( y ), la_( stat_type() ), lmax_( stat_type() ),
          lmin_( stat_type() )
      {
        const std::string FUNC_NAME = "tadah_fussy::statistics::glm::MultinomialLogistic constructor";

        error_lib::CheckLinearModel( x, y,
                                     FUNC_NAME + " : size of x " + error_message::NOT_POSITIVE_LENGTH,
                                     FUNC_NAME + " : size of y " + error_message::NOT_POSITIVE_LENGTH,
                                     FUNC_NAME + " : size of x " + error_message::NOT_EQUAL_LENGTH + " size of y" );

        n_ = x_.rows(); // 独立変数ベクトル x_i の数
        p_ = x_.cols(); // 独立変数ベクトルの要素数
        c_ = y_.cols(); // カテゴリ数
      }

      /*
        MultinomialLogistic::init : 多項分布モデル用変数の初期化
      */
      void MultinomialLogistic::init( SquareMatrix< stat_type >* sa, vector< stat_type >* sb, vector< stat_type >* a )
      {
        GLM_ML_initPi( y_, &ni_, &pi_ ); // ni, pi の初期化

        w_.resize( n_ * c_ * ( c_ - 1 ) / 2 ); // 対角行列 W_kk'

        sa->assign( ( c_ - 1 ) * p_ ); // 連立方程式の初期化
        sb->resize( ( c_ - 1 ) * p_ );
        a->resize( ( c_ - 1 ) * p_ );  // 係数の初期化

#ifdef DEBUG
        cerr << "N = " << n_ << " ; p = " << p_ << " ; c = " << c_ << endl << endl;

        PrintMatrix( cerr, x_, "x = " );
        cerr << endl << endl;
        PrintMatrix( cerr, y_, "y = " );
        cerr << endl << endl;
#endif
      }

      /*
        MultinomialLogistic::create : 多項分布モデル用連立方程式の構築
      */
      void MultinomialLogistic::create( SquareMatrix< stat_type >* sa, vector< stat_type >* sb )
      {
        // 対角行列の計算 W_kk' の計算
        GLM_ML_calcDiagMatrix( pi_, ni_, &w_, c_, n_ );
        // 係数行列の計算
        GLM_ML_calcCoefMatrix( w_, x_, sa, c_, p_, n_ );
        // 右辺の計算
        GLM_ML_calcRSide( w_, x_, y_, ni_, pi_, sb, c_, p_, n_ );

#ifdef DEBUG
        PrintMatrix( cerr, pi_, "pi = " );
        cerr << endl << endl;
        cerr << "Equation System :" << endl;
        PrintMatrix( cerr, *sa );
        cerr << endl;
        PrintArray( cerr, sb->begin(), sb->end() );
        cerr << endl;
#endif
      }

      /*
        MultinomialLogistic::update : 多項分布モデル用 ρ 値の更新
      */
      void MultinomialLogistic::update( const vector< stat_type >& a )
      {
        // カテゴリごとの確率計算
        vector< stat_type > rho( c_ - 1 );
        for ( size_type i = 0 ; i < n_ ; ++i ) {
          auto a_row = a.begin();
          for ( size_type k = 0 ; k < c_ - 1 ; ++k ) {
            auto x_row = x_.row( i );
            rho[k] = std::exp( std::inner_product( x_row, x_row.end(), a_row, stat_type() ) );
            a_row += p_;
          }
          pi_[i][0] = 1.0 / ( 1.0 + SumAll< stat_type >( rho ) );
          for ( size_t k = 1 ; k < c_ ; ++k )
            pi_[i][k] = rho[k - 1] * pi_[i][0];
        }

#ifdef DEBUG
        for ( vector< stat_type >::size_type k = 0 ; k < c_ - 1 ; ++k ) {
          std::ostringstream oss;
          oss << "a[" << k << "] = ";
          PrintArray( cerr, a.begin() + k * p_, a.begin() + ( k + 1 ) * p_, oss.str() );
          cerr << endl;
        }
        cerr << endl;
#endif
      }

      /*
        MultinomialLogistic::cleanUp : 反復処理完了後に必要な処理を行う
      */
      void MultinomialLogistic::cleanUp()
      {
        GLM_CalcFitValue( pi_, ni_, &yExp_ );
        la_ = GLM_CalcLa( y_.begin(), y_.end(), pi_.begin() );
        lmax_ = GLM_CalcLMax( y_, ni_ );
        lmin_ = GLM_CalcLMin( y_ );
      }

      /*
        MultinomialLogistic::pearsonResiduals : ピアソン残差を求める
      */
      void MultinomialLogistic::pearsonResiduals( Matrix< stat_type >* res ) const
      { GLM_PearsonResiduals( y_, yExp_, res ); }

      /*
        CumulativeLogitModel_Base コンストラクタ

        x : 独立変数
        y : 従属変数(各カテゴリの発生回数)

        n は独立変数ベクトルの数
        p は独立変数ベクトルの要素数
        c はカテゴリの数

        x は n x p 行列
        y は n x c 行列
      */
      CumulativeLogitModel_Base::CumulativeLogitModel_Base( const Matrix< stat_type >& x, const Matrix< stat_type >& y )
        : x_( x ), y_( y ), la_( stat_type() ), lmax_( stat_type() ), lmin_( stat_type() )
      {
        const string FUNC_NAME = "tadah_fussy::statistics::glm::CumulativeLogitModel_Base constructor";

        error_lib::CheckLinearModel( x, y,
                                     FUNC_NAME + " : size of x " + error_message::NOT_POSITIVE_LENGTH,
                                     FUNC_NAME + " : size of y " + error_message::NOT_POSITIVE_LENGTH,
                                     FUNC_NAME + " : size of x " + error_message::NOT_EQUAL_LENGTH + " size of y" );

        // 独立変数ベクトル x_i の数
        n_ = x_.rows();
        // 独立変数ベクトルの要素数
        p_ = x_.cols();
        // カテゴリ数
        c_ = y_.cols();
      }

      namespace
      {
        /*
          GLM_CLM_initRho : 累積ロジットモデル・比例オッズモデル用 ρ の初期化 ( y から計算する )
        */
        void GLM_CLM_initRho( const Matrix< stat_type >& y, vector< stat_type>* ni, Matrix< stat_type >* rho, Matrix< stat_type >* pi )
        {
          using size_type = Matrix< stat_type >::size_type;

          size_type n = y.rows();
          size_type c = y.cols();

          // pi の計算 ( y から求める )
          GLM_ML_initPi( y, ni, pi );

          rho->assign( n, c - 1 );
          for ( size_type i = 0 ; i < n ; ++i ) {
            stat_type sumPi = 0; // i までの累積確率
            for ( size_type k = 0 ; k < c - 1 ; ++k ) {
              sumPi += (*pi)[i][k];
              if ( sumPi == 0 ) sumPi = 1E-9; // ゼロ除算防止
              if ( sumPi > 1 ) sumPi = 1;
              (*rho)[i][k] = ( 1.0 - sumPi ) / sumPi;
            }
          }
        }

      }

      /*
        CumulativeLogitModel_Base::getPi : 累積ロジットモデル・比例オッズモデル用 ρ から π を求める
      */
      void CumulativeLogitModel_Base::calcPi()
      {
        pi_.assign( n_, c_ ); // 確率の当てはめ値 ( n x c )
        for ( size_type i = 0 ; i < n_ ; ++i ) {
          pi_[i][0] = 1.0 / ( 1.0 + rho_[i][0] );
          for ( size_type k = 1 ; k < c_ - 1 ; ++k )
            pi_[i][k] = 1.0 / ( 1.0 + rho_[i][k] ) - 1.0 / ( 1.0 + rho_[i][k - 1] );
          if ( c_ > 1 )
            pi_[i][c_ - 1] = 1.0 - 1.0 / ( 1.0 + rho_[i][c_ - 2] );
        }
      }

      /*
        CumulativeLogitModel_Base::init : 累積ロジットモデル・比例オッズモデル用変数の初期化
      */
      void CumulativeLogitModel_Base::init( SquareMatrix< stat_type >* sa, vector< stat_type >* sb, vector< stat_type >* a )
      {
        sa->assign( coefSize() );
        sb->resize( coefSize() );
        a->resize( coefSize() );
        GLM_CLM_initRho( y_, &ni_, &rho_, &pi_ );

#ifdef DEBUG
        cerr << "N = " << xSize() << " ; p = " << varSize() << " ; c = " << categorySize() << endl << endl;
        PrintMatrix( cerr, pi_, "pi = " );
        cerr << endl << endl;
        PrintMatrix( cerr, rho_, "rho = " );
        cerr << endl << endl;
#endif
      }

      /*
        CumulativeLogitModel_Base::create : 累積ロジットモデル・比例オッズモデル用連立方程式の構築
      */
      void CumulativeLogitModel_Base::create( SquareMatrix< stat_type >* sa, vector< stat_type >* sb )
      {
        calcCoefMatrix( sa ); // 係数行列の計算
        calcRSide( sb );      // 右辺の計算

#ifdef DEBUG
        cerr << "Equation System :" << endl;
        PrintMatrix( cerr, *sa );
        cerr << endl;
        PrintArray( cerr, sb->begin(), sb->end() );
        cerr << endl;
#endif
      }

      /*
        CumulativeLogitModel_Base::update : 累積ロジットモデル・比例オッズモデル用 ρ 値の更新
      */
      void CumulativeLogitModel_Base::update( const vector< stat_type >& a )
      {
        coef2Rho( a );

#ifdef DEBUG
        PrintMatrix( cerr, pi_, "pi = " );
        cerr << endl << endl;
        PrintMatrix( cerr, rho_, "rho = " );
        cerr << endl << endl;
#endif
      }

      /*
        CumulativeLogitModel_Base::cleanUp : 反復処理完了後に必要な処理を行う
      */
      void CumulativeLogitModel_Base::cleanUp()
      {
        calcPi();
        GLM_CalcFitValue( pi_, ni_, &yExp_ );
        la_ = GLM_CalcLa( y_.begin(), y_.end(), pi_.begin() );
        lmax_ = GLM_CalcLMax( y_, ni_ );
        lmin_ = GLM_CalcLMin( y_ );
      }

      /*
        CumulativeLogitModel_Base::pearsonResiduals : ピアソン残差を求める
      */
      void CumulativeLogitModel_Base::pearsonResiduals( Matrix< stat_type >* res ) const
      { GLM_PearsonResiduals( y_, yExp_, res ); }

      /*
        ProportionalOddsModel::coef2Rho : 求めた係数から rho を計算する
      */
      void ProportionalOddsModel::coef2Rho( const vector< stat_type >& a )
      {
        auto ai = a.begin() + c_ - 1;

        for ( size_type i = 0 ; i < n_ ; ++i ) {
          for ( size_type k = 0 ; k < c_ - 1 ; ++k ) {
            auto xi = x_.row( i );
            rho_[i][k] = std::exp( std::inner_product( xi, xi.end(), ai, a[k] ) );
          }
        }
      }

      /*
        ProportionalOddsModel::calcH11 : H11 の計算
      */
      void ProportionalOddsModel::calcH11( SquareMatrix< stat_type >* sa )
      {
        for ( size_type k = 0 ; k < c_ - 1 ; ++k ) {
          // 対角成分を計算
          for ( size_type i = 0 ; i < n_ ; ++i ) {
            (*sa)[k][k] += ( y_[i][k] + y_[i][k + 1] ) * rho_[i][k] / std::pow( 1.0 + rho_[i][k], 2.0 );
            if ( k + 1 < c_ - 1 )
              (*sa)[k][k] += y_[i][k + 1] * rho_[i][k] * rho_[i][k + 1] / std::pow( rho_[i][k] - rho_[i][k + 1], 2.0 );
            if ( k > 0 )
              (*sa)[k][k] += y_[i][k] * rho_[i][k] * rho_[i][k - 1] / std::pow( rho_[i][k] - rho_[i][k - 1], 2.0 );
          }
          if ( k + 1 == c_ - 1 ) break;

          // 対角成分の右隣を計算
          for ( size_type i = 0 ; i < n_ ; ++i )
            (*sa)[k][k + 1] -= y_[i][k + 1] * rho_[i][k] * rho_[i][k + 1] / std::pow( rho_[i][k] - rho_[i][k + 1], 2.0 );
          (*sa)[k + 1][k] = (*sa)[k][k + 1]; // 対称な成分にコピー
        }
      }

      /*
        ProportionalOddsModel::calcH22 : H22 の計算
      */
      void ProportionalOddsModel::calcH22( SquareMatrix< stat_type >* sa )
      {
        // k ごとの和の部分を先に計算
        vector< stat_type > sum( n_, stat_type() );
        for ( size_type i = 0 ; i < n_ ; ++i )
          for ( size_type k = 0 ; k < c_ - 1 ; ++k )
            sum[i] += ( y_[i][k] + y_[i][k + 1] ) * rho_[i][k] / std::pow( 1.0 + rho_[i][k], 2.0 );

        for ( size_type sr = c_ - 1 ; sr < coefSize() ; ++sr ) { // sr は係数行列の行番号
          size_type jr = sr - ( c_ - 1 ); // x_ に対する行番号
          for ( size_type sc = sr ; sc < coefSize() ; ++sc ) { // sc は係数行列の列番号(対角成分からスタート)
            size_type jc = sc - ( c_ - 1 ); // x_ に対する列番号
            for ( size_type i = 0 ; i < n_ ; ++i )
              (*sa)[sr][sc] += x_[i][jr] * x_[i][jc] * sum[i];
            // 対角成分(左下)はコピー
            if ( sc != sr )
              (*sa)[sc][sr] = (*sa)[sr][sc];
          }
        }
      }

      /*
        ProportionalOddsModel::calcH12 : H12(H21) の計算
      */
      void ProportionalOddsModel::calcH12( SquareMatrix< stat_type >* sa )
      {
        // 計算は右上側(H12)に対して行う
        for ( size_type k = 0 ; k < c_ - 1 ; ++k ) {
          for ( size_type sc = c_ - 1 ; sc < coefSize() ; ++sc ) {
            size_type j = sc - ( c_ - 1 ); // x_ に対する列番号
            for ( size_type i = 0 ; i < n_ ; ++i )
              (*sa)[k][sc] += x_[i][j] * ( y_[i][k] + y_[i][k + 1] ) * rho_[i][k] / std::pow( 1.0 + rho_[i][k], 2.0 );
            (*sa)[sc][k] = (*sa)[k][sc]; // 対角成分(左下)はコピー
          }
        }
      }

      /*
        ProportionalOddsModel::calcCoefMatrix : 係数行列を計算する
      */
      void ProportionalOddsModel::calcCoefMatrix( SquareMatrix< stat_type >* sa )
      {
        sa->assign( coefSize() ); // ゼロクリアするためリサイズ
        calcH11( sa ); // H11 の計算
        calcH22( sa ); // H22 の計算
        calcH12( sa ); // H12(H21) の計算
      }

      /*
        ProportionalOddsModel::calcRSide : 連立方程式の右辺を計算する
      */
      void ProportionalOddsModel::calcRSide( vector< stat_type >* sb )
      {
        vector< stat_type > u_sum( n_, 0 ); // 対数尤度の偏微分 u の計算に必要な項
        vector< stat_type > h_sum( n_, 0 ); // 係数行列と係数の積 Ha の計算に必要な項
        for ( size_type i = 0 ; i < n_ ; ++i ) {
          for ( size_type k = 0 ; k < c_ - 1 ; ++k ) {
            u_sum[i] += ( y_[i][k] + y_[i][k + 1] ) / ( 1.0 + rho_[i][k] ) - y_[i][k];
            h_sum[i] += ( y_[i][k] + y_[i][k + 1] ) * rho_[i][k] * std::log( rho_[i][k] )
              / std::pow( 1.0 + rho_[i][k], 2.0 );
          }
        }

        vector< stat_type > u( coefSize(), 0 ); // u の値
        vector< stat_type > h( coefSize(), 0 ); // Ha の値
        for ( size_type k = 0 ; k < c_ - 1 ; ++k ) {
          for ( size_type i = 0 ; i < n_ ; ++i ) {
            u[k] += ( y_[i][k] + y_[i][k + 1] ) / ( 1.0 + rho_[i][k] );
            h[k] -= ( y_[i][k] + y_[i][k + 1] ) * rho_[i][k] * std::log( rho_[i][k] ) / std::pow( 1.0 + rho_[i][k], 2 );
            if ( k + 1 < c_ - 1 ) {
              u[k] += y_[i][k + 1] * rho_[i][k + 1] / ( rho_[i][k] - rho_[i][k + 1] );
              h[k] += y_[i][k + 1] * rho_[i][k] * rho_[i][k + 1]
                * ( std::log( rho_[i][k + 1] ) - std::log( rho_[i][k] ) ) / std::pow( rho_[i][k] - rho_[i][k + 1], 2 );
            }
            if ( k > 0 ) {
              u[k] += y_[i][k] * rho_[i][k - 1] / ( rho_[i][k] - rho_[i][k - 1] );
              h[k] += y_[i][k] * rho_[i][k] * rho_[i][k - 1]
                * ( std::log( rho_[i][k - 1] ) - std::log( rho_[i][k] ) ) / std::pow( rho_[i][k] - rho_[i][k - 1], 2 );
            } else {
              u[k] -= y_[i][k];
            }
          }
        }
        for ( size_type j = 0 ; j < p_ ; ++j ) {
          for ( size_type i = 0 ; i < n_ ; ++i ) {
            u[j + c_ - 1] += x_[i][j] * u_sum[i];
            h[j + c_ - 1] -= x_[i][j] * h_sum[i];
          }
        }
        for ( size_type k = 0 ; k < coefSize() ; ++k )
          (*sb)[k] = -h[k] + u[k]; // 右辺 = -Ha + u
      }

      /*
        CumulativeLogitModel::coef2Rho : 求めた係数から &rho; を計算する
      */
      void CumulativeLogitModel::coef2Rho( const vector< stat_type >& a )
      {
        for ( size_type i = 0 ; i < n_ ; ++i ) {
          for ( size_type k = 0 ; k < c_ - 1 ; ++k ) {
            auto ai = a.begin() + k * p_;
            auto xi = x_.row( i );
            rho_[i][k] = std::exp( std::inner_product( xi, xi.end(), ai, stat_type() ) );
          }
        }
      }

      /*
        CumulativeLogitModel::calcCoefDiagBlock : 係数行列の対角部分行列を計算する
      */
      void CumulativeLogitModel::calcCoefDiagBlock( SquareMatrix< stat_type >* sa, size_type k )
      {
        // k に依存する部分を先に計算
        vector< stat_type > sum( n_, 0 );
        for ( size_type i = 0 ; i < n_ ; ++i ) {
          sum[i] += ( y_[i][k] + y_[i][k + 1] ) * rho_[i][k] / std::pow( 1.0 + rho_[i][k], 2.0 );
          if ( k + 1 < c_ - 1 )
            sum[i] += y_[i][k + 1] * rho_[i][k] * rho_[i][k + 1] / std::pow( rho_[i][k] - rho_[i][k + 1], 2.0 );
          if ( k > 0 )
            sum[i] += y_[i][k] * rho_[i][k] * rho_[i][k - 1] / std::pow( rho_[i][k] - rho_[i][k - 1], 2.0 );
        }

        for ( size_type jr = 0 ; jr < p_ ; ++jr ) {
          for ( size_type jc = 0 ; jc < p_ ; ++jc ) {
            for ( size_type i = 0 ; i < n_ ; ++i )
              (*sa)[k * p_ + jr][k * p_ + jc] += x_[i][jr] * x_[i][jc] * sum[i];
          }
        }
      }

      /*
        CumulativeLogitModel::calcCoefNonDiagBlock : 係数行列の非対角部分行列(対角成分の右側)を計算する
      */
      void CumulativeLogitModel::calcCoefNonDiagBlock( SquareMatrix< stat_type >* sa, size_type k )
      {
        // k に依存する部分を先に計算
        vector< stat_type > sum( n_, 0 );
        for ( size_type i = 0 ; i < n_ ; ++i )
          sum[i] += y_[i][k + 1] * rho_[i][k] * rho_[i][k + 1] / std::pow( rho_[i][k] - rho_[i][k + 1], 2.0 );

        for ( size_type jr = 0 ; jr < p_ ; ++jr ) {
          for ( size_type jc = 0 ; jc < p_ ; ++jc ) {
            for ( size_type i = 0 ; i < n_ ; ++i )
              (*sa)[k * p_ + jr][( k + 1 ) * p_ + jc] -= x_[i][jr] * x_[i][jc] * sum[i];
            // 対称成分にコピー
            (*sa)[( k + 1 ) * p_ + jc][k * p_ + jr] = (*sa)[k * p_ + jr][( k + 1 ) * p_ + jc];
          }
        }
      }

      /*
        CumulativeLogitModel::calcCoefMatrix : 係数行列を計算する
      */
      void CumulativeLogitModel::calcCoefMatrix( SquareMatrix< stat_type >* sa )
      {
        sa->assign( coefSize() );
        for ( size_type k = 0 ; k < c_ - 1 ; ++k ) {
          calcCoefDiagBlock( sa, k );
          if ( k + 1 < c_ - 1 ) calcCoefNonDiagBlock( sa, k );
        }
      }

      /*
        CumulativeLogitModel::calcRSide : 連立方程式の右辺を計算する
      */
      void CumulativeLogitModel::calcRSide( vector< stat_type >* sb )
      {
        Matrix< stat_type > u_sum( n_, c_ - 1 ); // 対数尤度の偏微分 u の計算に必要な項
        Matrix< stat_type > h_sum( n_, c_ - 1 ); // 係数行列と係数の積 Ha の計算に必要な項
        // k に依存する部分を先に計算しておく
        for ( size_type k = 0 ; k < c_ - 1 ; ++k ) {
          for ( size_type i = 0 ; i < n_ ; ++i ) {
            h_sum[i][k] = -( y_[i][k] + y_[i][k + 1] ) * rho_[i][k] * std::log( rho_[i][k] )
              / std::pow( 1.0 + rho_[i][k], 2.0 );
            u_sum[i][k] = ( y_[i][k] + y_[i][k + 1] ) / ( 1.0 + rho_[i][k] );
            if ( k > 0 ) {
              h_sum[i][k] += y_[i][k] * rho_[i][k] * rho_[i][k - 1]
                * ( std::log( rho_[i][k - 1] ) - std::log( rho_[i][k] ) ) / std::pow( rho_[i][k] - rho_[i][k - 1], 2.0 );
              u_sum[i][k] += y_[i][k] * rho_[i][k - 1] / ( rho_[i][k] - rho_[i][k - 1] );
            } else {
              u_sum[i][k] -= y_[i][k];
            }
            if ( k + 1 < c_ - 1 ) {
              h_sum[i][k] += y_[i][k + 1] * rho_[i][k] * rho_[i][k + 1]
                * ( std::log( rho_[i][k + 1] ) - std::log( rho_[i][k] ) ) / std::pow( rho_[i][k] - rho_[i][k + 1], 2.0 );
              u_sum[i][k] += y_[i][k + 1] * rho_[i][k + 1] / ( rho_[i][k] - rho_[i][k + 1] );
            }
          }
        }

        for ( size_type k = 0 ; k < c_ - 1 ; ++k ) {
          for ( size_type j = 0 ; j < p_ ; ++j ) {
            (*sb)[k * p_ + j] = 0;
            for ( size_type i = 0 ; i < n_ ; ++i )
              (*sb)[k * p_ + j] -= x_[i][j] * ( h_sum[i][k] - u_sum[i][k] );
          }
        }
      }

      /*
        PoissonRegression::calcLmin : 最小モデルに対する最大対数尤度の計算
      */
      void PoissonRegression::calcLmin()
      {
        using size_type = std::vector< stat_type >::size_type;

        stat_type mean = SumAll< stat_type >( y_ ) /
          SumAll< stat_type >( lf_.ni_ );

        lmin_ = stat_type();
        for ( size_type i = 0 ; i < y_.size() ; ++i ) {
          stat_type eta = pdf_->aveInv( lf_.ni_[i] * mean );
          //lmin_ += eta * y_[i] - pdf_->a( eta );
          //std::cout << "pdf_->logLikelihood( " << eta << ", " << y_[i] << " ) = " << pdf_->logLikelihood( eta, y_[i] ) << endl;
          lmin_ += pdf_->logLikelihood( eta, y_[i] );
        }
      }

      /*
        SurvivalAnalysis::init : スコア法での初期化を行う

        ここでは係数の初期化(ゼロクリア)と連立方程式用インスタンスのリサイズを行う。
      */
      void SurvivalAnalysis::init( SquareMatrix< stat_type >* sa, vector< stat_type >* sb, vector< stat_type >* a )
      {
        a->assign( coefSize(), stat_type() );   // 係数の初期化
        sa->assign( coefSize() ); // 連立方程式のリサイズ
        sb->resize( coefSize() );
      }

      namespace
      {
        /*
          MultCoefVec : 連立方程式の係数行列と解ベクトルの積を求めて xa に返す
        */
        void MultCoefVec( const SquareMatrix< stat_type >& sa, const vector< stat_type >& sb, vector< stat_type >* xa )
        {
          for ( vector< stat_type >::size_type i = 0 ; i < sb.size() ; ++i )
            (*xa)[i] = std::inner_product( sb.begin(), sb.end(), sa[i], stat_type() );
        }

        /*
          GetParam : 独立変数ベクトル x と係数ベクトル a = [ as, ae ) から exp( ( x, a ) ) を求めて param に返す
        */
        template< class In >
        void GetParam( const Matrix< stat_type >& x, In as, In ae, vector< stat_type >* param )
        {
          for ( Matrix< stat_type >::size_type i = 0 ; i < x.rows() ; ++i )
            (*param)[i] = std::exp( std::inner_product( as, ae, x[i], stat_type() ) );
        }

      }

      /*
        SurvivalAnalysis_Exponential::init : スコア法での初期化を行う
      */
      void SurvivalAnalysis_Exponential::init( SquareMatrix< stat_type >* sa, vector< stat_type >* sb, vector< stat_type >* a )
      {
        SurvivalAnalysis::init( sa, sb, a );
        mu_.resize( xSize() );
        GetParam( x_, a->begin(), a->end(), &mu_ );
      }

      /*
        SurvivalAnalysis_Exponential::calcCoef : 連立方程式の係数行列を計算する

        coef_j,k = Σ -t[i] * x[i][j] * x[i][k] / μ[i]
      */
      void SurvivalAnalysis_Exponential::calcCoef( SquareMatrix< stat_type >* sa ) const
      {
        for ( size_type k = 0 ; k < coefSize() ; ++k ) {
          for ( size_type j = k ; j < coefSize() ; ++j ) {
            (*sa)[k][j] = 0;
            for ( size_type i = 0 ; i < xSize() ; ++i )
              (*sa)[k][j] -= -t_[i].time * x_[i][j] * x_[i][k] / mu_[i];
            (*sa)[j][k] = (*sa)[k][j];
          }
        }
      }

      /*
        SurvivalAnalysis_Exponential::calcUj : u_j = ∂l / ∂a[j] を計算する

        u_j = Σ ( t[i] - δ[i] * μ[i] ) * x[i][j] / μ[i]
      */
      void SurvivalAnalysis_Exponential::calcUj( vector< stat_type >* uj ) const
      {
        for ( size_type j = 0 ; j < coefSize() ; ++j ) {
          (*uj)[j] = 0;
          for ( size_type i = 0 ; i < xSize() ; ++i )
            (*uj)[j] += ( t_[i].time - ( ( t_[i].censored ) ? 0 : mu_[i] ) ) * x_[i][j] / mu_[i];
        }
      }

      /*
        SurvivalAnalysis_Exponential::calcRSide : 連立方程式の右辺を計算する
      */
      void SurvivalAnalysis_Exponential::calcRSide( const SquareMatrix< stat_type >& sa, vector< stat_type >* sb ) const
      {
        vector< stat_type > uj( coefSize() );
        calcUj( &uj );

        vector< stat_type > xa( coefSize() );
        MultCoefVec( sa, *sb, &xa );

        for ( size_type i = 0 ; i < coefSize() ; ++i )
          (*sb)[i] = xa[i] + uj[i];
      }

      /*
        SurvivalAnalysis_Exponential::create : スコア法の連立方程式を作成する
      */
      void SurvivalAnalysis_Exponential::create( SquareMatrix< stat_type >* sa, vector< stat_type >* sb )
      {
        calcCoef( sa );
        calcRSide( *sa, sb );

#ifdef DEBUG
        cerr << "Equation System :" << endl;
        PrintMatrix( cerr, *sa );
        cerr << endl;
        PrintArray( cerr, sb->begin(), sb->end() );
        cerr << endl;
#endif
      }

      /*
        SurvivalAnalysis_Exponential::update : 係数を更新する
      */
      void SurvivalAnalysis_Exponential::update( const std::vector< stat_type >& a )
      {
        GetParam( x_, a.begin(), a.end(), &mu_ );

#ifdef DEBUG
        PrintArray( cerr, a.begin(), a.end(), "a = " );
        cerr << endl << endl;
#endif
      }

      /*
        SurvivalAnalysis_Exponential::calcLa : 当てはめ値に対する最大対数尤度を計算する
      */
      void SurvivalAnalysis_Exponential::calcLa()
      {
        la_ = stat_type();
        for ( vector< stat_type >::size_type i = 0 ;
              i < mu_.size() ; ++i ) {
          la_ -= t_[i].time / mu_[i];
          if ( ! t_[i].censored )
            la_ -= std::log( mu_[i] );
        }
      }

      namespace
      {
        /*
          SA_Exp_Summary : 独立変数の等しい要素ごとに生存時間を集計する
          (SurvivalAnalysis_Exponential専用)

          キーに独立変数の開始反復子、値に第一要素を打ち切りなしの個数、第二要素を合計時間とするペアを持つ
          std::map の派生クラスとする
        */
        struct SA_Exp_Summary :
          public std::map< Matrix< stat_type >::const_iterator,
                           std::pair< stat_type, stat_type >,
                           LessMatrixIterator< Matrix< stat_type > >
                           >
        {
          SA_Exp_Summary( const Matrix< stat_type >& x, const std::vector< SurvivalTime >& t )
          {
            for ( Matrix< stat_type >::size_type r = 0 ;
                  r < x.rows() ; ++r ) {
              iterator i = this->find( x[r] );
              if ( i == this->end() )
                i = this->insert
                  ( std::make_pair( x[r],
                                    std::make_pair( stat_type(), stat_type() )
                                    ) ).first;
              if ( ! t[r].censored )
                ( i->second ).first += 1;
              ( i->second ).second += t[r].time;
            }
          }
        };

      }

      /*
        SurvivalAnalysis_Exponential::calcLmax : 最大モデルに対する最大対数尤度を計算する
      */
      void SurvivalAnalysis_Exponential::calcLmax()
      {
        SA_Exp_Summary summary( x_, t_ );
        lmax_ = stat_type();
        for ( auto ci = summary.begin() ; ci != summary.end() ; ++ci ) {
          lmax_ += ( ci->second ).first * ( std::log( ( ci->second ).first / ( ci->second ).second ) - 1 );
        }
      }

      /*
        SurvivalAnalysis_Exponential::calcLmin : 最小モデルに対する最大対数尤度を計算する
      */
      void SurvivalAnalysis_Exponential::calcLmin()
      {
        stat_type tSum = stat_type();
        counter_type cnt = counter_type();

        for ( std::vector< SurvivalTime >::size_type i = 0 ;
              i < t_.size() ; ++i ) {
          if ( ! t_[i].censored ) ++cnt;
          tSum += t_[i].time;
        }

        lmin_ = cnt * std::log( cnt / tSum ) - cnt;
      }

      /*
        SurvivalAnalysis_Weibull::init : スコア法での初期化を行う

        λ成分の初期値がゼロの場合処理に失敗するため 1 を代入する
      */
      void SurvivalAnalysis_Weibull::init( SquareMatrix< stat_type >* sa, vector< stat_type >* sb, vector< stat_type >* a )
      {
        SurvivalAnalysis::init( sa, sb, a );
        a->back() = 1;
        (*sb)[coefSize() - 1] = 1;

        phi_.resize( xSize() ); // Φ[i]
        d1_.resize( xSize() );  // Φ[i] * t[i]^λ
        d2_.resize( xSize() );  // ln( t[i] / θ[i] )

        GetParam( x_, a->begin(), a->end() - 1, &phi_ );

        for ( size_type i = 0 ; i < xSize() ; ++i ) {
          d1_[i] = phi_[i] * std::pow( t_[i].time, a->back() );
          d2_[i] = std::log( t_[i].time / std::pow( phi_[i], -1 / a->back() ) );
        }
      }

      /*
        SurvivalAnalysis_Weibull::calcCoef : 係数行列を計算する

        coef_j,k = Σ -Φ[i] * t[i]^λ * x[i][j] * x[i][k]
        coef_j,λ = Σ -Φ[i] * t[i]^λ * x[i][j] * ln( t[i] / θ[i] )
        coef_λ,λ = Σ -δ[i] / λ^2 - Φ[i] * t[i]^λ * ln( t[i] / θ[i] )^2

        d1 = Φ[i] * t[i]^λ ; d2 = ln( t[i] / θ[i] )
      */
      void SurvivalAnalysis_Weibull::calcCoef( stat_type lambda, SquareMatrix< stat_type >* sa ) const
      {
        // 線形式の係数どうし ( a[j] と a[k] ) に対する成分 coef_j,k = ∂u_j / ∂a[k] を求める
        for ( size_type k = 0 ; k < coefSize() - 1 ; ++k ) {
          for ( size_type j = k ; j < coefSize() - 1 ; ++j ) {
            (*sa)[k][j] = 0;
            for ( size_type i = 0 ; i < xSize() ; ++i )
              (*sa)[k][j] -= -d1_[i] * x_[i][j] * x_[i][k];
            (*sa)[j][k] = (*sa)[k][j];
          }
        }

        // 線形式の係数 a[j] と λ に対する成分 coef_j,λ = ∂u_j / ∂λ を求める
        for ( size_type j = 0 ; j < coefSize() - 1 ; ++j ) {
          (*sa)[coefSize() - 1][j] = 0;
          for ( size_type i = 0 ; i < xSize() ; ++i )
            (*sa)[coefSize() - 1][j] -= -d1_[i] * x_[i][j] * d2_[i];
          (*sa)[j][coefSize() - 1] = (*sa)[coefSize() - 1][j];
        }

        // λ に対する成分 coef_λ,λ = ∂u_λ / ∂λ を求める
        (*sa)[coefSize() - 1][coefSize() - 1] = 0;
        for ( size_type i = 0 ; i < xSize() ; ++i )
          (*sa)[coefSize() - 1][coefSize() - 1] -= ( ( t_[i].censored ) ? 0 : -1.0 / std::pow( lambda, 2.0 ) ) - d1_[i] * std::pow( d2_[i], 2.0 );
      }

      /*
        SurvivalAnalysis_Weibull::calcUj : u[j] = ∂l / ∂a[j] を計算する

        u_j = Σ ( δ[i] - Φ[i] * t[i]^λ ) * x[i][j]
        u_λ = Σ δ[i] * [ 1 / λ + ln( t[i] / θ[i] ) ] - Φ[i] * t[i]^λ * ln( t[i] / θ[i] )

        d1 = Φ[i] * t[i]^λ ; d2 = ln( t[i] / θ[i] )
      */
      void SurvivalAnalysis_Weibull::calcUj( stat_type lambda, vector< stat_type >* uj ) const
      {
        for ( size_type j = 0 ; j < coefSize() - 1 ; ++j ) {
          (*uj)[j] = 0;
          for ( size_type i = 0 ; i < xSize() ; ++i )
            (*uj)[j] += ( ( ( t_[i].censored ) ? 0 : 1 ) - d1_[i] ) * x_[i][j];
        }
        (*uj)[coefSize() - 1] = 0;
        for ( size_type i = 0 ; i < xSize() ; ++i )
          (*uj)[coefSize() - 1] += ( ( t_[i].censored ) ? 0 : 1 / lambda + d2_[i] ) - d1_[i] * d2_[i];
      }

      /*
        SurvivalAnalysis_Weibull::calcRSide : 連立方程式の右辺を計算する

        d1 = Φ[i] * t[i]^λ ; d2 = ln( t[i] / θ[i] )
      */
      void SurvivalAnalysis_Weibull::calcRSide( stat_type lambda, const SquareMatrix< stat_type >& sa, vector< stat_type >* sb ) const
      {
        vector< stat_type > uj( coefSize() );
        calcUj( lambda, &uj );

        vector< stat_type > xa( coefSize() );
        MultCoefVec( sa, *sb, &xa );

        for ( size_type i = 0 ; i < coefSize() ; ++i )
          (*sb)[i] = xa[i] + uj[i];
      }

      /*
        SurvivalAnalysis_Weibull::create : スコア法の連立方程式を作成する
      */
      void SurvivalAnalysis_Weibull::create( SquareMatrix< stat_type >* sa, vector< stat_type >* sb )
      {
        calcCoef( (*sb)[coefSize() - 1], sa );
        calcRSide( (*sb)[coefSize() - 1], *sa, sb );

#ifdef DEBUG
        cerr << "Equation System :" << endl;
        PrintMatrix( cerr, *sa );
        cerr << endl;
        PrintArray( cerr, sb->begin(), sb->end() );
        cerr << endl;
#endif
      }

      /*
        SurvivalAnalysis_Weibull::update : 係数を更新する
      */
      void SurvivalAnalysis_Weibull::update( const std::vector< stat_type >& a )
      {
        GetParam( x_, a.begin(), a.end() - 1, &phi_ );

        for ( size_type i = 0 ; i < xSize() ; ++i ) {
          d1_[i] = phi_[i] * std::pow( t_[i].time, a.back() );
          d2_[i] = std::log( t_[i].time / std::pow( phi_[i], -1 / a.back() ) );
        }

#ifdef DEBUG
        PrintArray( cerr, a.begin(), a.end(), "a = " );
        cerr << endl << endl;
#endif
      }

      namespace
      {
        // t^λを計算する
        class SA_Weibull_TL
        {
          stat_type lambda_;

        public:

          SA_Weibull_TL( stat_type lambda ) : lambda_( lambda ) {}

          stat_type operator()( SurvivalTime t )
          { return( std::pow( t.time, lambda_ ) ); }
        };

        // t^λlntを計算する
        class SA_Weibull_TL_LnT
        {
          stat_type lambda_;

        public:

          SA_Weibull_TL_LnT( stat_type lambda ) : lambda_( lambda ) {}

          stat_type operator()( SurvivalTime t )
          { return( std::pow( t.time, lambda_ ) * std::log( t.time ) ); }
        };

        // δを返す
        struct SA_Weibull_Delta
        {
          stat_type operator()( SurvivalTime t )
          { return( ( t.censored ) ? 0 : 1 ); }
        };

        // δlntを計算する
        struct SA_Weibull_Delta_LnT
        {
          stat_type operator()( SurvivalTime t )
          { return( ( t.censored ) ? 0 : std::log( t.time ) ); }
        };

        // 1/λ - ( Σt^λlnt / Σt^λ - Σδlnt / Σδ ) の計算
        class SA_Weibull_Diff
        {
          stat_type d2_; // Σδlnt / Σδ
          const vector< SurvivalTime >& t_; // tへの参照

        public:

          SA_Weibull_Diff( stat_type d2, const vector< SurvivalTime >& t ) : d2_( d2 ), t_( t ) {}

          stat_type operator()( stat_type lambda )
          {
            SA_Weibull_TL tl( lambda );
            SA_Weibull_TL_LnT tl_lnt( lambda );
            // Σt^λlnt / Σt^λ
            stat_type d1 =
              std::accumulate( t_.begin(), t_.end(), stat_type(),
                               OpBind< double >( tl_lnt, SumOp< stat_type >() )
                               ) /
              std::accumulate( t_.begin(), t_.end(), stat_type(),
                               OpBind< double >( tl, SumOp< stat_type >() )
                               );
            return( 1 / lambda - ( d1 - d2_ ) );
          }
        };

      } // namespace

      /*
        SurvivalAnalysis_Weibull::calcLa : 当てはめ値に対する最大対数尤度を求める
      */
      void SurvivalAnalysis_Weibull::calcLa()
      {
        la_ = stat_type();
        for ( std::vector< SurvivalTime >::size_type i = 0 ;
              i < t_.size() ; ++i ) {
          la_ -= d1_[i];
          if ( ! t_[i].censored ) {
            la_ += ( lambda() - 1 ) * std::log( t_[i].time ) + std::log( lambda() ) + std::log( phi_[i] );
          }
        }
      }

      namespace
      {
        /*
          SA_Weibull_Summary1 : 独立変数の等しい要素ごとに Σt^λ・lnt / Σt^λ を計算する
          (SurvivalAnalysis_Weibull専用)

          キーに独立変数の開始反復子、値に第一要素をΣti^λ・lnti、第二要素をΣti^λとするペアを持つ
          std::map の派生クラスとする
        */
        struct SA_Weibull_Summary1 :
          public std::map< Matrix< stat_type >::const_iterator,
                           std::pair< stat_type, stat_type >,
                           LessMatrixIterator< Matrix< stat_type > >
                           >
        {
          SA_Weibull_Summary1( const Matrix< stat_type >& x, const std::vector< SurvivalTime >& t,
                               stat_type lambda )
          {
            SA_Weibull_TL tl( lambda );
            SA_Weibull_TL_LnT tl_lnt( lambda );

            for ( Matrix< stat_type >::size_type r = 0 ;
                  r < x.rows() ; ++r ) {
              iterator i = this->find( x[r] );
              if ( i == this->end() )
                i = this->insert
                  ( std::make_pair( x[r],
                                    std::make_pair( stat_type(), stat_type() )
                                    ) ).first;
              ( i->second ).first += tl_lnt( t[r] );
              ( i->second ).second += tl( t[r] );
            }
          }
        };

        /*
          SA_Weibull_Summary2 : 独立変数の等しい要素ごとに Σδ / Σt^λ を計算する
          (SurvivalAnalysis_Weibull専用)

          キーに独立変数の開始反復子、値に第一要素をΣδ、第二要素をΣti^λとするペアを持つ
          std::map の派生クラスとする
        */
        struct SA_Weibull_Summary2 :
          public std::map< Matrix< stat_type >::const_iterator,
                           std::pair< stat_type, stat_type >,
                           LessMatrixIterator< Matrix< stat_type > >
                           >
        {
          SA_Weibull_Summary2( const Matrix< stat_type >& x, const std::vector< SurvivalTime >& t,
                               stat_type lambda )
          {
            SA_Weibull_Delta delta;
            SA_Weibull_TL tl( lambda );

            for ( Matrix< stat_type >::size_type r = 0 ;
                  r < x.rows() ; ++r ) {
              iterator i = this->find( x[r] );
              if ( i == this->end() )
                i = this->insert
                  ( std::make_pair( x[r],
                                    std::make_pair( stat_type(), stat_type() )
                                    ) ).first;
              ( i->second ).first += delta( t[r] );
              ( i->second ).second += tl( t[r] );
            }
          }
        };

        /*
          SA_Weibull_DlDlambda : ∂l/∂λを計算する
          (SurvivalAnalysis_Weibull専用)
        */
        class SA_Weibull_DlDlambda
        {
          const Matrix< stat_type >& x_;
          const std::vector< SurvivalTime >& t_;

        public:

          SA_Weibull_DlDlambda( const Matrix< stat_type >& x, const std::vector< SurvivalTime >& t )
            : x_( x ), t_( t ) {}

          stat_type operator()( stat_type lambda )
          {
            SA_Weibull_Summary1 summary( x_, t_, lambda );
            stat_type res = stat_type();
            for ( std::vector< SurvivalTime >::size_type i = 0 ;
                  i < t_.size() ; ++i ) {
              if ( t_[i].censored ) continue;
              res -= summary[x_[i]].first / summary[x_[i]].second;
              res += std::log( t_[i].time ) + 1 / lambda;
            }
            return( res );
          }
        };

      } // namespace

      /*
        SurvivalAnalysis_Weibull::calcLmax : 最大モデルに対する最大対数尤度を求める
      */
      void SurvivalAnalysis_Weibull::calcLmax()
      {
        stat_type h = std::sqrt( DEFAULT_THRESHOLD );
        stat_type e = DEFAULT_THRESHOLD;
        SA_Weibull_DlDlambda dldl( x_, t_ );
        stat_type init = 1;

        // 逆に進む場合は h の符号を反転する
        if ( ( dldl( init ) > dldl( init + e ) && dldl( init ) < 0 ) ||
             ( dldl( init ) < dldl( init + e ) && dldl( init ) > 0 ) )
          h = -h;

        stat_type lambda = GradientMethod( dldl, init, h, e );
        SA_Weibull_Summary2 summary( x_, t_, lambda );

        lmax_ = stat_type();
        for ( vector< SurvivalTime >::size_type i = 0 ;
              i < t_.size() ; ++i ) {
          stat_type theta = summary[x_[i]].first / summary[x_[i]].second;
          lmax_ -=  theta * std::pow( t_[i].time, lambda );
          if ( ! t_[i].censored )
            lmax_ += ( lambda - 1 ) * std::log( t_[i].time ) + std::log( lambda ) + std::log( theta );
        }
      }

      /*
        SurvivalAnalysis_Weibull::calcLminParam : 最小モデルに対するθ^(-λ)、λを求める
      */
      void SurvivalAnalysis_Weibull::calcLminParam( stat_type* lambda, stat_type* theta )
      {
        // Σδ
        stat_type sumDelta =
          std::accumulate( t_.begin(), t_.end(), stat_type(),
                           OpBind< double >( SA_Weibull_Delta(), SumOp< stat_type >() )
                           );

        // Σδlnt / Σδ
        stat_type d2 =
          std::accumulate( t_.begin(), t_.end(), stat_type(),
                           OpBind< double >( SA_Weibull_Delta_LnT(), SumOp< stat_type >() )
                           ) / sumDelta;

        stat_type h = std::sqrt( DEFAULT_THRESHOLD );
        stat_type e = DEFAULT_THRESHOLD;
        SA_Weibull_Diff diff( d2, t_ );
        stat_type init = 1;

        // 逆に進む場合は h の符号を反転する
        if ( ( diff( init ) > diff( init + e ) && diff( init ) < 0 ) ||
             ( diff( init ) < diff( init + e ) && diff( init ) > 0 ) )
          h = -h;

        *lambda = GradientMethod( diff, init, h, e );
        *theta =
          sumDelta /
          std::accumulate( t_.begin(), t_.end(), stat_type(),
                           OpBind< double >( SA_Weibull_TL( *lambda ), SumOp< stat_type >() )
                           );
      }

      /*
        SurvivalAnalysis_Weibull::calcLmin : 最小モデルに対する最大対数尤度を求める
      */
      void SurvivalAnalysis_Weibull::calcLmin()
      {
        stat_type lambda; // λ
        stat_type theta;  // θ^(-λ)
        calcLminParam( &lambda, &theta );

        lmin_ = stat_type();
        for ( std::vector< SurvivalTime >::size_type i = 0 ;
              i < t_.size() ; ++i ) {
          lmin_ -= theta * std::pow( t_[i].time, lambda );
          if ( ! t_[i].censored )
            lmin_ += ( lambda - 1 ) * std::log( t_[i].time ) + std::log( lambda ) + std::log( theta );
        }
      }

    } // namespace glm

  } // namespace statistics

} // namespace tadah_fussy
