namespace tadah_fussy
{
  namespace statistics
  {
    namespace glm
    {
      /*
        Base::scoringMethod : スコア法による係数の推定

        a : 求めた係数を保持する配列へのポインタ
        var_a : 係数の分散を保持する配列へのポインタ
        solver : 連立方程式を解くための関数オブジェクト
        test : 収束テスト用関数オブジェクト
        maxCount : 反復処理の最大回数
      */
      template< class Solver, class Test >
      void Base::scoringMethod
      ( std::vector< stat_type >* a, std::vector< stat_type >* var_a,
        Solver solver, Test test, uint32_t maxCount )
      {
        const std::string FUNC_NAME = "tadah_fussy::statistics::glm::Base::scoringMethod";

        assert( a != nullptr );

        SquareMatrix< stat_type > sa; // 連立方程式計算用係数行列
        std::vector< stat_type > sb;  // 連立方程式計算用解ベクトル

#ifdef DEBUG
        std::cerr << "----- initializing -----" << std::endl << std::endl;
#endif

        init( &sa, &sb, a );

        /* 係数の計算 */
        uint32_t cnt = 0;
        for ( ; cnt < maxCount ; ++cnt ) {

#ifdef DEBUG
          std::cerr << "----- cnt = " << cnt << " -----" << std::endl << std::endl;
#endif

          create( &sa, &sb ); // 連立方程式の構築
          solver( sa, sb.begin() );  // 連立方程式の計算

          // 各係数が収束しているかを確認する
          test.init();
          for ( std::vector< stat_type >::size_type j = 0 ; j < sb.size() ; ++j ) {
            test( (*a)[j], sb[j] );
            (*a)[j] = sb[j];
          }

          update( *a );

          if ( test.isConvergent() ) break;
        }

        if ( cnt >= maxCount )
          throw exception::ExcessLimit( FUNC_NAME + " : " + error_message::EXCESS_LIMIT , maxCount );

        /* 係数の分散を計算 */
        if ( var_a != 0 ) {

#ifdef DEBUG
          std::cerr << "----- calculating variance of a -----" << std::endl << std::endl;
#endif

          SquareMatrix< stat_type > inv;
          create( &sa, &sb ); // 連立方程式の構築
          matrix_lib::Inverse( sa, &inv, solver );

#ifdef DEBUG
          std::cerr << "Covariance matrix :" << std::endl;
          matrix_lib::PrintMatrix( std::cerr, inv );
          std::cerr << std::endl << std::endl;
#endif

          var_a->resize( inv.size() );
          for ( SquareMatrix< stat_type >::size_type i = 0 ; i < inv.size() ; ++i )
            (*var_a)[i] = inv[i][i];
        }

        /* 最後の後始末 */
        cleanUp();
      }

      /*
        GeneralizedLinearModel::initY : 従属変数 y の初期化
      */
      template< class In >
      void Generic::initY( In ys, In ye )
      {
        // コンストラクタのみから呼び出すので関数名はコンストラクタ
        const std::string FUNC_NAME = "tadah_fussy::statistics::glm::Generic constructor";

        error_lib::CheckLinearModel( x_, ys, ye,
                                     FUNC_NAME + " : size of x " + error_message::NOT_POSITIVE_LENGTH,
                                     FUNC_NAME + " : size of y " + error_message::NEGATIVE_LENGTH,
                                     FUNC_NAME + " : size of x " + error_message::NOT_EQUAL_LENGTH + " size of y" );
        y_.assign( ys, ye );
      }

      /*
        GeneralizedLinearModel::fit : [ xs, xe ) を独立変数としたときの予測値の計算
      */
      template< class In >
      distribution::Base::stat_type Generic::predict( In xs, In xe ) const
      {
        const std::string FUNC_NAME = "tadah_fussy::statistics::glm::Generic::predict";

        assert( g_ != nullptr ); // 連結関数は指定されているはず

        error_lib::CheckBivariable( xs, xe, a_.begin(), a_.end(),
                                    FUNC_NAME + " : [ xs, xe ) " + error_message::NEGATIVE_LENGTH,
                                    FUNC_NAME + " : size of coefficient " + error_message::NEGATIVE_LENGTH,
                                    FUNC_NAME + " : [ xs, xe ) " + error_message::NOT_EQUAL_LENGTH + " size of coefficient" );

        return( fit( xs, xe, a_.begin() ) );
      }

      /*
        逸脱度残差 d を求める
      */
      template< class Out >
      void Generic::devianceResiduals( Out out ) const
      {
        using size_type = std::vector< stat_type >::size_type;

        std::vector< stat_type > d( yExp_.size(), stat_type() );
        // 最大モデルの最大対数尤度の項
        for ( size_type i = 0 ; i < yExp_.size() ; ++i ) {
          stat_type eta = pdf_->aveInv( y_[i] );
          d[i] = pdf_->logLikelihood( eta, y_[i] );
        }
        // 当てはめ値による最大対数尤度の項
        for ( size_type i = 0 ; i < yExp_.size() ; ++i ) {
          stat_type eta = pdf_->aveInv( yExp_[i] );
          d[i] = std::sqrt( ( d[i] - pdf_->logLikelihood( eta, y_[i] ) ) * 2 );
          if ( y_[i] < yExp_[i] ) d[i] *= -1;
        }
        std::copy( d.begin(), d.end(), out );
      }

      /*
        ピアソン残差 Xi を求める
      */
      template< class Out >
      void Generic::pearsonResiduals( Out out ) const
      {
        using size_type = std::vector< stat_type >::size_type;

        PearsonResiduals< stat_type, stat_type > pr;
        for ( size_type i = 0 ; i < yExp_.size() ; ++i )
          *out++ = pr( y_[i], yExp_[i] );
      }

    } // namespace glm

    /*
      Bin2Cnt コンストラクタ : 独立変数列 [ xs, xe ) と二値データ列 [ bin_s, bin_e ) を指定して構築
    */
    template< class X, class Less >
    template< class In1, class In2 >
    Bin2Cnt< X, Less >::Bin2Cnt( In1 xs, In1 xe, In2 bin_s, In2 bin_e )
    {
      using counter_type = distribution::Base::counter_type;

      const std::string FUNC_NAME = "tadah_fussy::statistics::Bin2Cnt constructor";

      error_lib::CheckBivariable( xs, xe, bin_s, bin_e,
                                  FUNC_NAME + " : [ xs, xe ) " + error_message::NEGATIVE_LENGTH,
                                  FUNC_NAME + " : [ bin_s, bin_e ) " + error_message::NEGATIVE_LENGTH,
                                  FUNC_NAME + " : [ xs, xe ) " + error_message::NOT_EQUAL_LENGTH + " [ bin_s, bin_e )" );

      // 計数処理
      while ( xs != xe ) {
        iterator it = this->find( *xs );
        if ( it == this->end() ) {
          (*this)[*xs] = typename std::map< X, std::pair< counter_type, counter_type >, Less >::mapped_type( 1, ( *bin_s ) ? 1 : 0 );
        } else {
          ++( ( it->second ).first );
          if ( *bin_s ) ++( ( it->second ).second );
        }
        ++xs;
        ++bin_s;
      }
    }

    /*
      Bin2Cnt コンストラクタ : 独立変数と二値データのペア列 [ s, e ) を指定して構築

      データ列の要素は、独立変数をキー、二値データを値とした std::pair であることを想定している。
    */
    template< class X, class Less >
    template< class In >
    Bin2Cnt< X, Less >::Bin2Cnt( In s, In e )
    {
      using counter_type = distribution::Base::counter_type;

      const std::string FUNC_NAME = "tadah_fussy::statistics::Bin2Cnt constructor";

      error_lib::CheckArrayLength( s, e,
                                   FUNC_NAME + " : [ s, e ) " + error_message::NEGATIVE_LENGTH );
      // 計数処理
      for ( ; s != e ; ++s ) {
        iterator it = this->find( s->first );
        if ( it == this->end() ) {
          (*this)[s->first] = typename std::map< X, std::pair< counter_type, counter_type >, Less >::mapped_type( 1, ( s->second ) ? 1 : 0 );
        } else {
          ++( ( it->second ).first );
          if ( s->second ) ++( ( it->second ).second );
        }
      }
    }

    namespace glm
    {
      /*
        BinomialLogistic コンストラクタ

        二項分布を確率分布とした一般化線形モデル(ロジスティック回帰)

        x 独立変数ベクトルへの参照
        n 各試行の総試行回数への参照
        y 各試行の成功回数への参照
        g 連結関数への参照
        h 微小値(π=0,1のときに計算結果がゼロになるのを防ぐためのしきい値)
      */
      template< class In1, class In2 >
      BinomialLogistic::BinomialLogistic( const Matrix< stat_type >& x, In1 ns, In1 ne, In2 ys, In2 ye,
                                          const LinkFunction_IF& g, stat_type h )
        : Generic( x, ys, ye ), sf_( g )
      {
        const std::string FUNC_NAME = "tadah_fussy::statistics::glm::BinomialLogistic constructor";

        // y は基底クラスでチェック済。n のみ確認
        error_lib::CheckLinearModel( x, ns, ne,
                                     FUNC_NAME + " : size of x " + error_message::NOT_POSITIVE_LENGTH,
                                     FUNC_NAME + " : size of n " + error_message::NEGATIVE_LENGTH,
                                     FUNC_NAME + " : size of x " + error_message::NOT_EQUAL_LENGTH + " size of n" );

        auto y = y_.begin();
        In1 n = ns;
        for ( ; y != y_.end() ; ++y , ++n ) {
          error_lib::LE< stat_type >( error_lib::NNNum( *y, FUNC_NAME + " : all elements of y " + error_message::NNNUM ),
                                      error_lib::PNum( *n, FUNC_NAME + " : all elements of n " + error_message::PNUM ),
                                      FUNC_NAME + " : y[i] " + error_message::LE + " n[i]" );
          if ( *y == 0 ) *y = h * *n;
          if ( *y == *n ) *y -= h * *n;
        }

        n_.assign( ns, ne );
        bin_.assign( ns, ne );
        sf_.assign( ns, ne );

        Generic::setExpFamily( bin_ );
        Generic::setLinkFunc( sf_ );
      }

      /*
        BinomialLogistic::pi : 推定確率を求める
      */
      template< class Out >
      void BinomialLogistic::pi( Out out ) const
      {
        for ( std::vector< stat_type >::size_type i = 0 ;
              i < yExp_.size() ; ++i )
          *out++ = yExp_[i] / n_[i];
      }

      /*
        ピアソン残差 Xi を求める
      */
      template< class Out >
      void BinomialLogistic::pearsonResiduals( Out out ) const
      {
        auto yi = y_.begin();
        auto ni = n_.begin();

        for ( auto npi = yExp_.begin() ; npi != yExp_.end() ; ++npi ) {
          *out++ = ( *yi - *npi ) * std::sqrt( *ni / ( *npi * ( *ni - *npi ) ) );
          ++yi;
          ++ni;
        }
      }

      /*
        HosmerLemeshowTest コンストラクタ

        n : 試行回数
        y : y の実測値
        pi : 確率の当てはめ値
        grpCnt : グループ数
      */
      template< class RanN, class RanY, class RanPi >
      HosmerLemeshowTest::HosmerLemeshowTest( RanN ns, RanN ne, RanY ys, RanY ye, RanPi pis, RanPi pie,
                                              counter_type grpCnt )
        : chisqHL_( 0 )
      {
        const std::string FUNC_NAME = "tadah_fussy::statistics::glm::HosmerLemeshowTest constructor";

        // 要素数の取得とチェック
        auto sz = error_lib::ArrayLength( ns, ne );
        error_lib::CheckArrayLength( ys, ye, sz,
                                     FUNC_NAME + " : size of [ ys, ye ) " + error_message::NEGATIVE_LENGTH,
                                     FUNC_NAME + " : size of [ ys, ye ) " + error_message::NOT_EQUAL_LENGTH + " size of [ ns, ne )" );
        error_lib::CheckArrayLength( pis, pie, sz,
                                     FUNC_NAME + " : size of [ pis, pie ) " + error_message::NEGATIVE_LENGTH,
                                     FUNC_NAME + " : size of [ pis, pie ) " + error_message::NOT_EQUAL_LENGTH + " size of [ ns, ne )" );
        error_lib::PNum( grpCnt, FUNC_NAME + " : grpCnt " + error_message::PNUM );

        // 計数データの合計 totalCnt を求める
        counter_type totalCnt = 0;
        RanN n = ns;
        RanY y = ys;
        for ( ; n != ne ; ++n, ++y ) {
          error_lib::LE( error_lib::NNNum( *y, FUNC_NAME + " : all elements of y " + error_message::NNNUM ),
                         error_lib::PNum( *n, FUNC_NAME + " : all elements of n " + error_message::PNUM ),
                         FUNC_NAME + " : y[i] " + error_message::LE + " n[i]" );
          totalCnt += *n;
        }
        error_lib::PNum( totalCnt, FUNC_NAME + " : total of n " + error_message::PNUM );

        counter_type grpSz = totalCnt / grpCnt; // 1 グループあたりの要素数の目安

        // pi と要素番号をペアとする map の作成
        using difference_type = typename std::iterator_traits< RanN >::difference_type;
        using MMap = std::multimap< stat_type, difference_type >;
        MMap mmapIndex;
        for ( difference_type i = 0 ; i < sz ; ++i )
          mmapIndex.insert( std::make_pair( pis[i], i ) );

        stat_type se = 0, so = 0, fe = 0, fo = 0; // グループごとの成功・失敗回数(実測・期待値)
        counter_type cnt = 0;   // グループの総数
        counter_type grpNo = 1; // グループ番号
        auto cit = mmapIndex.begin();
        stat_type pi_min = cit->first; // グループの確率の最小値
        while ( cit != mmapIndex.end() ) {
          difference_type i = cit->second; // 要素番号を取得
          so += ys[i];
          fo += ns[i] - ys[i];
          se += ns[i] * pis[i];
          fe += ns[i] * ( 1.0 - pis[i] );
          cnt += ns[i];
          stat_type pi_max = cit->first;
          ++cit;
          if ( ( ( cnt >= grpSz ) && ( grpNo < grpCnt ) ) || // グループの要素数が目安を超えた(かつ最後のグループではない)
               ( ( cnt > 0 ) && cit == mmapIndex.end() ) ) { // 最後のグループ(全データ集計完了)
            pi_.push_back( std::make_pair( pi_min, pi_max ) );
            se_.push_back( se );
            so_.push_back( so );
            fe_.push_back( fe );
            fo_.push_back( fo );
            ++grpNo;
            pi_min = cit->first;
            chisqHL_ += std::pow( so - se, 2.0 ) / se;
            chisqHL_ += std::pow( fo - fe, 2.0 ) / fe;
            so = fo = se = fe = 0;
            cnt = 0;
          }
        }
      }

      /*
        PoissonRegression コンストラクタ : ポアソン回帰

        x : 独立変数
        ni : 試行回数
        y : 発生回数(従属変数)
      */
      template< class In1, class In2 >
      PoissonRegression::PoissonRegression( const Matrix< stat_type >& x, In1 ns, In1 ne, In2 ys, In2 ye )
        : Generic( x, ys, ye )
      {
        const std::string FUNC_NAME = "tadah_fussy::statistics::glm::PoissonRegression constructor";

        Generic::setExpFamily( poisson_ );
        Generic::setLinkFunc( lf_ );

        // x と ni のサイズ比較 ( y は基底クラスでチェック済 )
        error_lib::CheckLinearModel( x, ns, ne,
                                     FUNC_NAME + " : size of x " + error_message::NOT_POSITIVE_LENGTH,
                                     FUNC_NAME + " : size of [ ns, ne ) " + error_message::NEGATIVE_LENGTH,
                                     FUNC_NAME + " : row size of x " + error_message::NOT_EQUAL_LENGTH + " size of [ ns, ne )" );

        In1 y = ys;
        In2 n = ns;
        for ( ; y != ye ; ++y , ++n ) {
          error_lib::LE( error_lib::NNNum( *y, FUNC_NAME + " : all elements of y " + error_message::NNNUM ),
                         error_lib::PNum( *n, FUNC_NAME + " : all elements of n " + error_message::PNUM ),
                         FUNC_NAME + " : y[i] " + error_message::LE + " n[i]" );
        }

#ifdef DEBUG
        matrix_lib::PrintArray( std::cerr, ns, ne, "ni = " );
        std::cerr << std::endl;
#endif

        // 連結関数の作成
        lf_.ni_.assign( ns, ne );
      }

      /*
        SurvivalAnalysis コンストラクタ : 独立変数ベクトル x と従属変数(生存時間) t を指定して構築
      */
      template< class In >
      SurvivalAnalysis::SurvivalAnalysis( const Matrix< stat_type >& x, In ts, In te )
        : x_( x )
      {
        const std::string FUNC_NAME = "tadah_fussy::statistics::glm::SurvivalAnalysis constructor";

        error_lib::CheckLinearModel( x_, ts, te,
                                     FUNC_NAME + " : size of x " + error_message::NOT_POSITIVE_LENGTH,
                                     FUNC_NAME + " : size of [ ts, te ) " + error_message::NEGATIVE_LENGTH,
                                     FUNC_NAME + " : row size of x " + error_message::NOT_EQUAL_LENGTH + " size of [ ts, te )" );

        t_.assign( ts, te );
      }

      /*
        SurvivalAnalysis_Exponential::residuals : 補正のない残差を返す
      */
      template< class Out >
      void SurvivalAnalysis_Exponential::residuals( Out out ) const
      {
        for ( std::vector< stat_type >::size_type i = 0 ;
              i < mu_.size() ; ++i )
          *out++ = t_[i].time / mu_[i];
      }

      /*
        SurvivalAnalysis_Exponential::coxSnellResiduals : Cox-Snell 残差を返す
      */
      template< class Out >
      void SurvivalAnalysis_Exponential::coxSnellResiduals( Out out, stat_type delta ) const
      {
        for ( std::vector< stat_type >::size_type i = 0 ;
              i < mu_.size() ; ++i )
          *out++ = t_[i].time / mu_[i] + ( ( t_[i].censored ) ? delta : 0 );
      }

      /*
        SurvivalAnalysis_Exponential::martingaleResiduals : Martingale 残差を返す
      */
      template< class Out >
      void SurvivalAnalysis_Exponential::martingaleResiduals( Out out ) const
      {
        for ( std::vector< stat_type >::size_type i = 0 ;
              i < mu_.size() ; ++i )
          *out++ = ( ( t_[i].censored ) ? 0 : 1 ) - t_[i].time / mu_[i];
      }

      /*
        SurvivalAnalysis_Exponential::devianceResiduals : 逸脱度残差を返す
      */
      template< class Out >
      void SurvivalAnalysis_Exponential::devianceResiduals( Out out ) const
      {
        for ( std::vector< stat_type >::size_type i = 0 ;
              i < mu_.size() ; ++i ) {
          // 補正なし残差
          stat_type r = t_[i].time / mu_[i];
          // Martingale残差
          stat_type rm = ( ( t_[i].censored ) ? 0 : 1 ) - r;
          *out++ = ( ( rm < 0 ) ? -1 : 1 ) * std::sqrt( -2 * ( rm + ( ( t_[i].censored ) ? 0 : 1 ) * std::log( r ) ) );
        }
      }

      /*
        SurvivalAnalysis_Weibull::fitValue : 当てはめ値を返す
      */
      template< class Out >
      void SurvivalAnalysis_Weibull::fitValue( Out out ) const
      {
        stat_type recl = -1 / lambda();
        for ( std::vector< stat_type >::size_type i = 0 ;
              i < phi_.size() ; ++i ) {
          *out++ = std::pow( phi_[i], recl );
        }
      }

      /*
        SurvivalAnalysis_Weibull::residuals : 補正のない残差を返す
      */
      template< class Out >
      void SurvivalAnalysis_Weibull::residuals( Out out ) const
      {
        for ( std::vector< stat_type >::size_type i = 0 ;
              i < phi_.size() ; ++i )
          *out++ = d1_[i];
      }

      /*
        SurvivalAnalysis_Weibull::coxSnellResiduals : Cox-Snell 残差を返す
      */
      template< class Out >
      void SurvivalAnalysis_Weibull::coxSnellResiduals( Out out, stat_type delta ) const
      {
        for ( std::vector< stat_type >::size_type i = 0 ;
              i < phi_.size() ; ++i )
          *out++ = d1_[i] + ( ( t_[i].censored ) ? delta : 0 );
      }

      /*
        SurvivalAnalysis_Weibull::martingaleResiduals : Martingale 残差を返す
      */
      template< class Out >
      void SurvivalAnalysis_Weibull::martingaleResiduals( Out out ) const
      {
        for ( std::vector< stat_type >::size_type i = 0 ;
              i < phi_.size() ; ++i )
          *out++ = ( ( t_[i].censored ) ? 0 : 1 ) - d1_[i];
      }

      /*
        SurvivalAnalysis_Weibull::devianceResiduals : 逸脱度残差を返す
      */
      template< class Out >
      void SurvivalAnalysis_Weibull::devianceResiduals( Out out ) const
      {
        for ( std::vector< stat_type >::size_type i = 0 ;
              i < phi_.size() ; ++i ) {
          // 補正なし残差
          stat_type r = d1_[i];
          // Martingale残差
          stat_type rm = ( ( t_[i].censored ) ? 0 : 1 ) - r;
          *out++ = ( ( rm < 0 ) ? -1 : 1 ) * std::sqrt( -2 * ( rm + ( ( t_[i].censored ) ? 0 : 1 ) * std::log( r ) ) );
        }
      }

      /*
        KMEstimate コンストラクタ : 生存時間の配列 [ s, e ) を指定して構築
      */
      template< class In >
      KMEstimate::KMEstimate( In s, In e )
      {
        using stat_type = distribution::Base::stat_type;
        using counter_type = distribution::Base::counter_type;

        // 配列の長さ(要素数)
        counter_type n = error_lib::ArrayLength( s, e );

        // 生存時間ごとの要素数をカウントするためのバッファ
        // 値は std::pair 型であり、first が死亡数、second が打ち切りを含めた死亡数を表す
        typedef std::map< SurvivalTime, std::pair< counter_type, counter_type > > Buffer;
        Buffer buff;

        // 要素数のカウント
        for ( ; s != e ; ++s ) {
          auto it = buff.find( *s );
          if ( it == buff.end() ) {
            std::pair< counter_type, counter_type > death( ( s->censored ) ? 0 : 1, 1 );
            buff.insert( typename Buffer::value_type( *s, death ) );
          } else {
            if ( ! s->censored ) ( it->second ).first += 1;
            ( it->second ).second += 1;
          }
        }

        // カウンタからカプラン・マイヤー推定値を算出
        stat_type d = 1.0;
        for ( auto it = buff.begin() ; it != buff.end() ; ++it ) {
          // ( 直前の生存数 - 死亡数 ) / 直前の生存数 を算出
          d *= static_cast< stat_type >( n - ( it->second ).first ) / n;
          (*this)[it->first] = d;
          // 要素数からは打ち切りも含めて減算する
          n -= ( it->second ).second;
        }
      }

    } // namespace glm

  } // namespace statistics

} // namespace tadah_fussy
