#include "quantification.h"

namespace tadah_fussy
{
  namespace statistics
  {
    namespace quantification
    {
      using std::vector;
      using std::multimap;
      using std::string;
      using std::sqrt;

      namespace
      {
        /*
          CalcFi : 要素 i が反応するカテゴリの数 fi を求める

          data : 対象データ
          fi : 求めた fi を保持する配列へのポインタ
        */
        void CalcFi( const Matrix< bool >& data, vector< stat_type >* fi )
        {
          fi->clear();

          for ( Matrix< bool >::size_type i = 0 ; i < data.rows() ; ++i ) {
            stat_type total = 0;
            for ( Matrix< bool >::size_type j = 0 ; j < data.cols() ; ++j )
              if ( data[i][j] ) total += 1;
            fi->push_back( total );
          }
        }

        /*
          CalcGj : カテゴリ j に反応した要素の数 gj を求める

          data : 対象データ
          gj : 求めた gj を保持する配列へのポインタ
        */
        void CalcGj( const Matrix< bool >& data, vector< stat_type >* gj )
        {
          gj->clear();

          for ( Matrix< bool >::size_type j = 0 ; j < data.cols() ; ++j ) {
            stat_type total = 0;
            for ( Matrix< bool >::size_type i = 0 ; i < data.rows() ; ++i )
              if ( data[i][j] ) total += 1;
            gj->push_back( total );
          }
        }

        /*
          CreateMatrixJ : yj に対する固有値問題の対称行列を計算する

          data : 対象データ
          fi : 要素 i が反応するカテゴリの数
          gj : カテゴリ j に反応した要素の数
          mat : 求めた行列を保持する変数へのポインタ
        */
        void CreateMatrixJ( const Matrix< bool >& data,
                            const vector< stat_type >& fi, const vector< stat_type >& gj,
                            SymmetricMatrix< stat_type >* mat )
        {
          mat->assign( data.cols() );
          for ( Matrix< bool >::size_type j = 0 ; j < data.cols() ; ++j ) {
            for ( Matrix< bool >::size_type v = j ; v < data.cols() ; ++v ) {
              (*mat)[j][v] = 0;
              for ( Matrix< bool >::size_type i = 0 ; i < data.rows() ; ++i ) {
                if ( data[i][j] && data[i][v] )
                  (*mat)[j][v] += 1.0 / ( fi[i] * sqrt( gj[j] ) * sqrt( gj[v] ) );
              }
            }
          }
        }

        /*
          CreateMatrixI : xi に対する固有値問題の対称行列を計算する

          data : 対象データ
          fi : 要素 i が反応するカテゴリの数
          gj : カテゴリ j に反応した要素の数
          mat : 求めた行列を保持する変数へのポインタ
        */
        void CreateMatrixI( const Matrix< bool >& data,
                            const vector< stat_type >& fi, const vector< stat_type >& gj,
                            SymmetricMatrix< stat_type >* mat )
        {
          mat->assign( data.rows() );
          for ( Matrix< bool >::size_type i = 0 ; i < data.rows() ; ++i ) {
            for ( Matrix< bool >::size_type u = i ; u < data.rows() ; ++u ) {
              (*mat)[i][u] = 0;
              for ( Matrix< bool >::size_type j = 0 ; j < data.cols() ; ++j ) {
                if ( data[i][j] && data[u][j] )
                  (*mat)[i][u] += 1.0 / ( gj[j] * sqrt( fi[i] ) * sqrt( fi[u] ) );
              }
            }
          }
        }

      } // namespace

      /*
        Class3 コンストラクタ : データ data を指定して構築
      */
      Class3::Class3( const Matrix< bool >& data )
      {
        const stat_type THRESHOLD = 1E-6; // 固有値計算用しきい値
        const uint32_t MAX_CNT = 10000;   // 固有値計算用最大反復回数

        // fi, gj の計算
        vector< stat_type > fi, gj;

        CalcFi( data, &fi );
        CalcGj( data, &gj );

        // 固有値分解を行う行列の計算
        SymmetricMatrix< stat_type > mat;
        if ( data.cols() < data.rows() )
          CreateMatrixJ( data, fi, gj, &mat );
        else
          CreateMatrixI( data, fi, gj, &mat );

        // 固有値分解
        SquareMatrix< stat_type > r, q;
        try {
          math_lib::eigen_lib::Householder_DoubleShiftQR( mat, &r, &q, THRESHOLD, MAX_CNT );
        } catch( exception::ExcessLimit& ee ) {
          SymmetricMatrix< stat_type > eigen( mat );
          math_lib::eigen_lib::Jacobi( &eigen, &q, THRESHOLD, MAX_CNT );
          r = eigen;
        }

        // 固有値の取得
        multimap< stat_type, SquareMatrix< stat_type >::size_type, std::greater< stat_type > > l; // 固有値用バッファ
        for ( SquareMatrix< stat_type >::size_type i = 0 ; i < r.rows() ; ++i )
          l.insert( std::make_pair( r[i][i], i ) );
        l.erase( l.begin() ); // 最大値(=1)は無効解なので除外

        // xi, yj の取得
        if ( data.cols() < data.rows() ) {
          for ( auto lp = l.begin() ; lp != l.end() ; ++lp ) {
            lambda_.push_back( lp->first );
            yj_.push_back( vector< stat_type >( q.column( lp->second ), q.column( lp->second ).end() ) );
            xi_.push_back( vector< stat_type >( fi.size(), 0 ) );
            auto& y = yj_.back();
            auto& x = xi_.back();
            for ( vector< stat_type >::size_type j = 0 ; j < gj.size() ; ++j )
              y[j] /= sqrt( gj[j] );
            for ( vector< stat_type >::size_type i = 0 ; i < fi.size() ; ++i ) {
              for ( vector< stat_type >::size_type j = 0 ; j < gj.size() ; ++j )
                if ( data[i][j] )
                  x[i] += y[j] / fi[i];
              x[i] /= lp->first;
            }
          }
        } else {
          for ( auto lp = l.begin() ; lp != l.end() ; ++lp ) {
            lambda_.push_back( lp->first );
            xi_.push_back( vector< stat_type >( q.column( lp->second ), q.column( lp->second ).end() ) );
            yj_.push_back( vector< stat_type >( gj.size(), 0 ) );
            auto& x = xi_.back();
            auto& y = yj_.back();
            for ( vector< stat_type >::size_type i = 0 ; i < fi.size() ; ++i )
              x[i] /= sqrt( fi[i] );
            for ( vector< stat_type >::size_type j = 0 ; j < gj.size() ; ++j )
              for ( vector< stat_type >::size_type i = 0 ; i < fi.size() ; ++i ) {
                if ( data[i][j] )
                  y[j] += x[i] / gj[j];
                y[j] /= lp->first;
              }
          }
        }

#ifdef DEBUG
        matrix_lib::PrintArray( std::cout, lambda_.begin(), lambda_.end(), "lambda = " );
        std::cout << std::endl;
        std::cout << "x :" << std::endl;
        for ( auto x = xi_.begin() ; x != xi_.end() ; ++x ) {
          matrix_lib::PrintArray( std::cout, x->begin(), x->end(), " " );
          std::cout << std::endl;
        }
        std::cout << "y :" << std::endl;
        for ( auto y = yj_.begin() ; y != yj_.end() ; ++y ) {
          matrix_lib::PrintArray( std::cout, y->begin(), y->end(), " " );
          std::cout << std::endl;
        }
#endif
      }

      /*
        Class3::dataSize : 要素数を返す
      */
      size_type Class3::dataSize() const
      {
        if ( xi_.size() == 0 ) return( 0 );

        return( xi_[0].size() );
      }

      /*
        Class3::categorySize : カテゴリ数を返す
      */
      size_type Class3::categorySize() const
      {
        if ( yj_.size() == 0 ) return( 0 );

        return( yj_[0].size() );
      }

      namespace
      {
        /*
          CreateH : 親近性からなる行列 E から H を作成する

          e : 親近性からなる行列 E
          h : 作成した行列 H を保持する変数へのポインタ
        */
        void CreateH( const SquareMatrix< stat_type >& e, SymmetricMatrix< stat_type >* h )
        {
          h->assign( e.size() );

          // 非対角成分の計算
          for ( SquareMatrix< stat_type >::size_type r = 0 ; r < e.size() ; ++r ) {
            for ( SquareMatrix< stat_type >::size_type c = r + 1 ; c < e.size() ; ++c ) {
              ( *h )[r][c] = e[r][c] + e[c][r];
            }
          }

          // 対角成分の計算
          for ( SquareMatrix< stat_type >::size_type r = 0 ; r < e.size() ; ++r ) {
            stat_type total = 0;
            for ( SquareMatrix< stat_type >::size_type c = 0 ; c < e.size() ; ++c ) {
              if ( c == r ) continue;
              total += ( *h )[r][c];
            }
            ( *h )[r][r] = -total;
          }
        }

      } // namespace

      /*
        Class4 コンストラクタ : 親近性からなる行列 E を指定して構築

        e : 親近性からなる行列 E
      */
      Class4::Class4( const SquareMatrix< stat_type >& e )
      {
        const stat_type THRESHOLD = 1E-6; // 固有値計算用しきい値
        const uint32_t MAX_CNT = 10000;   // 固有値計算用最大反復回数

        // 行列 H の作成
        SymmetricMatrix< stat_type > h;
        CreateH( e, &h );

        // 固有値分解
        SquareMatrix< stat_type > r, q;
        try {
          math_lib::eigen_lib::Householder_DoubleShiftQR( h, &r, &q, THRESHOLD, MAX_CNT );
        } catch( exception::ExcessLimit& ee ) {
          SymmetricMatrix< stat_type > eigen( h );
          math_lib::eigen_lib::Jacobi( &eigen, &q, THRESHOLD, MAX_CNT );
          r = eigen;
        }

        // 固有値の取得
        multimap< stat_type, SquareMatrix< stat_type >::size_type, std::greater< stat_type > > l; // 固有値用バッファ
        for ( SquareMatrix< stat_type >::size_type i = 0 ; i < r.rows() ; ++i )
          l.insert( std::make_pair( r[i][i], i ) );

        // 数量の取得
        for ( auto lp = l.begin() ; lp != l.end() ; ++lp ) {
          lambda_.push_back( lp->first );
          xi_.push_back( vector< stat_type >( q.column( lp->second ), q.column( lp->second ).end() ) );
        }
      }

      /*
        Class4::dataSize : 要素数を返す
      */
      size_type Class4::dataSize() const
      {
        if ( xi_.size() == 0 ) return( 0 );

        return( xi_[0].size() );
      }

    } // namespace quantification

  } // namespace statistics

} // namespace tadah_fussy
