namespace tadah_fussy
{
  namespace statistics
  {
    namespace quantification
    {
      /*
        Class1 専用ヘルパ関数
      */

      /*
        GetCategory : カテゴリ毎のリストを作成する

        [ s, e ) : カテゴリデータ
        category : 生成するカテゴリリストへのポインタ
      */
      template< class In, class X >
      void GetCategory( In s, In e, item_type< X, uint32_t >* category )
      {
        for ( ; s != e ; ++s ) {
          auto i = category->find( *s );
          if ( i == category->end() )
            category->insert( std::make_pair( *s, 1 ) );
          else
            ++( i->second ); // カテゴリに属する要素数をカウント
        }
      }

      /*
        GetCategoryList : カテゴリ毎のリストを作成する

        x : アイテムごとのカテゴリデータ ( 列:アイテム 行:要素 )
        category : 生成するカテゴリリストへのポインタ
      */
      template< class X >
      void GetCategoryList( const Matrix< X >& x, list_type< X, uint32_t >* category )
      {
        using size_type = typename Matrix< X >::size_type;

        for ( size_type c = 0 ; c < x.cols() ; ++c ) {
          category->push_back( item_type< X, uint32_t >() );
          auto& map = category->back();
          GetCategory( x.column( c ), x.column( c ).end(), &map );
        }
      }

      /*
        GetDupMatched : 指定した二つのアイテム・カテゴリを両方持つ要素の数をカウントする

        x : アイテムごとのカテゴリデータ ( 列:アイテム 行:要素 )
        j, u : アイテムの番号 ( x の列番号 )
        k, v : カテゴリ

        戻り値 : 計算したカウント
      */
      template< class X >
      stat_type GetDupMatched( const Matrix< X >& x,
                               typename Matrix< X >::size_type j, const X& k,
                               typename Matrix< X >::size_type u, const X& v )
      {
        using size_type = typename Matrix< X >::size_type;

        // 同一アイテム内で異なるカテゴリが重複する事はない
        if ( j == u && k != v ) return( 0 );

        stat_type total = 0;
        for ( size_type r = 0 ; r < x.rows() ; ++r )
          if ( x[r][j] == k && x[r][u] == v )
            total += 1;

        return( total );
      }

      /*
        GetMatchedYSum : 指定したアイテム・カテゴリを持つ要素の従属変数の和を求める

        x : アイテムごとのカテゴリデータ ( 列:アイテム 行:要素 )
        ys : 従属変数のリストの先頭
        j : アイテムの番号 ( x の列番号 )
        k : カテゴリ

        戻り値 : 計算した和
      */
      template< class X, class In >
      stat_type GetMatchedYSum( const Matrix< X >& x, In ys,
                                typename Matrix< X >::size_type j, const X& k )
      {
        using size_type = typename Matrix< X >::size_type;

        stat_type total = 0;
        for ( size_type r = 0 ; r < x.rows() ; ++r ) {
          if ( x[r][j] == k ) total += *ys;
          ++ys;
        }

        return( total );
      }

      /*
        CreateLES : 数量化法 I 類用の正規方程式の係数計算

        x : アイテムごとのカテゴリデータ ( 列:アイテム 行:要素 )
        ys : 従属変数のリストの先頭
        category : カテゴリのリスト
        coef : 生成する係数行列へのポインタ
        a : 生成する解ベクトルへのポインタ
      */
      template< class X, class In >
      void CreateLES( const Matrix< X >& x, In ys, const list_type< X, uint32_t >& category,
                      SquareMatrix< stat_type >* coef, std::vector< stat_type >* a )
      {
        using SM = SquareMatrix< stat_type >;
        using List = list_type< X, uint32_t >;

        SM buff; // 線形独立の形にする前の係数行列

        // カテゴリの総数を求める
        typename List::size_type total = 0;
        for ( auto i = category.begin() ; i != category.end() ; ++i )
          total += i->size();

        // 配列・行列のリサイズ
        buff.assign( total, total );
        a->resize( total - category.size() + 1 );
        coef->assign( total - category.size() + 1, total - category.size() + 1 );

        // 係数の計算
        SM::size_type startR = 0;
        SM::size_type c = 0;
        for ( typename List::size_type j = 0 ; j < category.size() ; ++j ) {
          for ( auto k = category[j].begin() ; k != category[j].end() ; ++k ) {
            SM::size_type r = startR;
            typename List::size_type u = j;
            for ( auto v = k ; v != category[j].end() ; ++v ) {
              buff[r][c] = buff[c][r] = GetDupMatched( x, j, k->first, u, v->first );
              ++r;
            }
            while ( ++u < category.size() ) {
              for ( auto v = category[u].begin() ; v != category[u].end() ; ++v ) {
                buff[r][c] = buff[c][r] = GetDupMatched( x, j, k->first, u, v->first );
                ++r;
              }
            }
            ++c;
            ++startR;
          }
        }

        // 解ベクトルの計算
        auto ai = a->begin();
        for ( typename List::size_type j = 0 ; j < category.size() ; ++j ) {
          for ( auto k = category[j].begin() ; k != category[j].end() ; ++k ) {
            if ( j > 0 && k == category[j].begin() ) continue; // 各アイテムの最初のカテゴリは除外(最初のアイテムを除く)
            *ai++ = GetMatchedYSum( x, ys, j, k->first );
          }
        }

        // 線形独立の形になるように行列を削除する
        std::set< SM::size_type > ignoreRC; // 削除する行列番号
        SM::size_type i = 0;
        for ( typename List::size_type j = 0 ; j + 1 < category.size() ; ++j ) {
          i += category[j].size();
          ignoreRC.insert( i );
        }
        SM::size_type oR = 0; // 出力行
        for ( SM::size_type r = 0 ; r < total ; ++r ) {
          if ( ignoreRC.find( r ) != ignoreRC.end() ) continue;
          SM::size_type oC = 0; // 出力列
          for ( SM::size_type c = 0 ; c < total ; ++c ) {
            if ( ignoreRC.find( c ) != ignoreRC.end() ) continue;
            (*coef)[oR][oC] = buff[r][c];
            ++oC;
          }
          ++oR;
        }
      }

      /*
        NormalizeCoef : 係数の標準化

        x : アイテムごとのカテゴリデータ ( 列:アイテム 行:要素 )
        category : カテゴリのリスト
        a :  標準化する係数
      */
      template< class X >
      void NormalizeCoef( const Matrix< X >& x, const list_type< X, uint32_t >& category,
                          list_type< X, stat_type >* a )
      {
        using size_type = typename list_type< X, uint32_t >::size_type;

        for ( size_type j = 0 ; j < category.size() ; ++j ) {
          // 減算する値の計算 Δ = Σ( a_jk * n_jk ) / N
          stat_type delta = 0;
          auto& aj = (*a)[j];
          for ( auto k = category[j].begin() ; k != category[j].end() ; ++k )
            delta += aj[k->first] * k->second;
          delta /= x.rows();
          // 標準化
          for ( auto k = category[j].begin() ; k != category[j].end() ; ++k )
            aj[k->first] -= delta;
        }
      }

      /*
        CalcSjj : x の分散値を求める

        x : アイテムごとのカテゴリデータ ( 列:アイテム 行:要素 )
        j : アイテム番号
        category : カテゴリのリスト
        aj : 対象アイテムの係数

        戻り値 : 求めた分散値
      */
      template< class X >
      stat_type CalcSjj( const Matrix< X >& x, typename Matrix< X >::size_type j,
                         const list_type< X, uint32_t >& category,
                         const item_type< X, stat_type >& aj )
      {
        stat_type t1 = 0;
        stat_type t2 = 0;
        for ( auto k = category[j].begin() ; k != category[j].end() ; ++k ) {
          auto ajk = aj.find( k->first );
          t1 += std::pow( ajk->second, 2 ) * k->second; // Σ( a_jk * a_jk * n_jk )
          t2 += ajk->second * k->second;                // Σ( a_jk * n_jk )
        }

        return( ( t1 - std::pow( t2, 2 ) / x.rows() ) / x.rows() );
      }

      /*
        CalcSyj : x と y の共分散を求める

        x : アイテムごとのカテゴリデータ ( 列:アイテム 行:要素 )
        j : アイテム番号
        category : カテゴリのリスト
        ys : y のリストの先頭
        my : y の平均値
        aj : 対象アイテムの係数

        戻り値 : 求めた共分散値
      */
      template< class X, class In >
      stat_type CalcSyj( const Matrix< X >& x, typename Matrix< X >::size_type j,
                         const list_type< X, uint32_t >& category, In ys, stat_type my,
                         const item_type< X, stat_type >& aj )
      {
        stat_type t1 = 0;
        stat_type t2 = 0;
        for ( auto k = category[j].begin() ; k != category[j].end() ; ++k ) {
          auto ajk = aj.find( k->first );
          t1 += ajk->second * GetMatchedYSum( x, ys, j, k->first ); // Σ( a_jk * Σy_iδ_ijk )
          t2 += ajk->second * k->second;                            // Σ( a_jk * n_jk )
        }

        return( ( t1 - my * t2 ) / x.rows() );
      }

      /*
        CalcSju : x 同士の共分散を求める

        x : アイテムごとのカテゴリデータ ( 列:アイテム 行:要素 )
        j, u : アイテム番号
        category : カテゴリのリスト
        aj, au : 対象アイテムの係数

        戻り値 : 求めた共分散値
      */
      template< class X >
      stat_type CalcSju( const Matrix< X >& x,
                         typename Matrix< X >::size_type j, typename Matrix< X >::size_type u,
                         const list_type< X, uint32_t >& category,
                         const item_type< X, stat_type >& aj, const item_type< X, stat_type >& au )
      {
        const auto& cj = category[j];
        const auto& cu = category[u];

        stat_type t1 = 0;
        for ( auto cjk = cj.begin() ; cjk != cj.end() ; ++cjk ) {
          for ( auto cuv = cu.begin() ; cuv != cu.end() ; ++cuv ) {
            auto ajk = aj.find( cjk->first );
            auto auv = au.find( cuv->first );
            // Σ( a_jk * a_uv * Σ( δ_ijk * δ_iuv ) )
            t1 += ajk->second * auv->second * GetDupMatched( x, j, cjk->first, u, cuv->first );
          }
        }

        stat_type t2 = 0;
        for ( auto cjk = cj.begin() ; cjk != cj.end() ; ++cjk ) {
          auto ajk = aj.find( cjk->first );
          t2 += ajk->second * cjk->second; // Σ( a_jk * n_jk )
        }

        stat_type t3 = 0;
        for ( auto cuv = cu.begin() ; cuv != cu.end() ; ++cuv ) {
          auto auv = au.find( cuv->first );
          t3 += auv->second * cuv->second; // Σ( a_uv * n_uv )
        }

        return( ( t1 - t2 * t3 / x.rows() ) / x.rows() );
      }

      /*
        CreateCovariance : 共分散行列の計算

        x : アイテムごとのカテゴリデータ ( 列:アイテム 行:要素 )
        category : カテゴリのリスト
        ys, ye : y の先頭と終端の次の位置
        my : y の平均値
        a : 係数
        s : 共分散行列を保持する変数のポインタ
      */
      template< class X, class In >
      void CreateCovariance( const Matrix< X >& x, const list_type< X, uint32_t >& category,
                             In ys, In ye, stat_type my, const list_type< X, stat_type >& a,
                             SquareMatrix< stat_type >* s )
      {
        using size_type = SquareMatrix< stat_type >::size_type;

        s->assign( x.cols() + 1 );

        // syy の計算
        (*s)[0][0] = SampleVariance< stat_type >( ys, ye );

        // sjj の計算
        for ( size_type j = 1 ; j < s->rows() ; ++j )
          (*s)[j][j] = CalcSjj( x, j - 1, category, a[j - 1] );

        // syj の計算
        for ( size_type j = 1 ; j < s->rows() ; ++j )
          (*s)[j][0] = (*s)[0][j] = CalcSyj( x, j - 1, category, ys, my, a[j - 1] );

        // sju の計算
        for ( size_type j = 1 ; j < s->rows() ; ++j )
          for ( size_type u = j + 1 ; u < s->rows() ; ++u )
            (*s)[j][u] = (*s)[u][j] = CalcSju( x, j - 1, u - 1, category, a[j - 1], a[u - 1] );
      }

      /*
        Class1 コンストラクタ

        x : アイテムごとのカテゴリデータ ( 列:アイテム 行:要素 )
        ys, ye : y の先頭と終端の次の位置
        solver : 連立方程式を解くためのソルバ
      */
      template< class X >
      template< class In, class Solver >
      Class1< X >::Class1( const Matrix< X >& x, In ys, In ye, Solver solver )
      {
        const std::string FUNC_NAME = "tadah_fussy::statistics::quantification::Class1 constructor";

        error_lib::CheckLinearModel( x, ys, ye,
                                  FUNC_NAME + " : x " + error_message::NOT_POSITIVE_LENGTH,
                                  FUNC_NAME + " : [ ys, ye ) " + error_message::NEGATIVE_LENGTH,
                                  FUNC_NAME + " : size of x " + error_message::NOT_EQUAL_LENGTH + " [ ys, ye )" );

        // カテゴリリストの作成
        GetCategoryList( x, &category_ );

        // y の平均値を求める
        my_ = Mean< stat_type >( ys, ye );

#ifdef DEBUG
        for ( size_type i = 0 ; i < category_.size() ; ++i ) {
          std::cout << "category[" << i << "]";
          for ( auto j = category_[i].begin() ; j != category_[i].end() ; ++j )
            std::cout << " " << j->first;
          std::cout << std::endl;
        }
#endif

        // 正規方程式の作成
        SquareMatrix< stat_type > coef;
        std::vector< stat_type > aBuff;
        CreateLES( x, ys, category_, &coef, &aBuff );

#ifdef DEBUG
        matrix_lib::PrintMatrix( std::cout, coef, "coef = " );
        std::cout << std::endl;
#endif

        // 係数の計算
        solver( coef, aBuff.begin() );

        // 足りない係数(ゼロとおいた係数)を加える
        auto j = category_.begin();
        auto a = aBuff.begin();
        a_.push_back( a_item_type() );
        for ( auto k = j->begin() ; k != j->end() ; ++k )
          ( a_.back() )[k->first] = *a++;
        for ( ++j ; j != category_.end() ; ++j ) {
          auto k = j->begin();
          a_.push_back( a_item_type() );
          ( a_.back() )[k->first] = 0;
          for ( ++k ; k != j->end() ; ++k )
            ( a_.back() )[k->first] = *a++;
        }

#ifdef DEBUG
        std::cout << "--- a ( not normalized ) ---" << std::endl;
        for ( typename x_list_type::size_type i = 0 ; i < category_.size() ; ++i ) {
          const auto& x = category_[i];
          const auto& a = a_[i];
          for ( auto j = x.begin() ; j != x.end() ; ++j )
            std::cout << j->first << ":" << ( a.find( j->first ) )->second << " ";
          std::cout << std::endl;
        }
#endif

        // 係数の標準化
        NormalizeCoef( x, category_, &a_ );

#ifdef DEBUG
        std::cout << "--- a ( normalized ) ---" << std::endl;
        for ( typename x_list_type::size_type i = 0 ; i < category_.size() ; ++i ) {
          const auto& x = category_[i];
          const auto& a = a_[i];
          for ( auto j = x.begin() ; j != x.end() ; ++j )
            std::cout << j->first << ":" << ( a.find( j->first ) )->second << " ";
          std::cout << std::endl;
        }
#endif

        // 共分散行列の計算
        CreateCovariance( x, category_, ys, ye, my_, a_, &s_ );
        matrix_lib::Inverse( s_, &invS_, solver );

#ifdef DEBUG
        matrix_lib::PrintMatrix( std::cout, s_, "S = " );
        std::cout << std::endl;
        matrix_lib::PrintMatrix( std::cout, invS_, "invS = " );
        std::cout << std::endl;
#endif
      }

      /*
        Class1< X >::a : 指定したアイテム番号・カテゴリに対する係数 ( 数量 ) を返す

        item : アイテム番号
        category : カテゴリ
      */
      template< class X >
      stat_type Class1< X >::a( size_type item, const X& category ) const
      {
        const std::string FUNC_NAME = "tadah_fussy::statistics::quantification::Class1< X >::a";

        // アイテム番号が範囲外なら out_of_range 例外を投げる
        if ( item >= category_.size() )
          throw std::out_of_range( FUNC_NAME + " : item index is out of range" );

        // カテゴリが見つからなかったら out_of_range 例外を投げる
        auto i = a_[item].find( category );
        if ( i == a_[item].end() )
          throw std::out_of_range( FUNC_NAME + " : category not found" );

        return( i->second );
      }

      /*
        Class1< X >::category : 指定したアイテム・カテゴリ番号のカテゴリの値を返す

        item : アイテム番号
        category : カテゴリ番号

        戻り値 : カテゴリの値
      */
      template< class X >
      X Class1< X >::category( size_type item, size_type category ) const
      {
        const std::string FUNC_NAME = "tadah_fussy::statistics::quantification::Class1< X >::category";

        // 範囲外なら out_of_range 例外を投げる
        if ( item >= category_.size() )
          throw std::out_of_range( FUNC_NAME + " : item index is out of range" );
        if ( category >= category_[item].size() )
          throw std::out_of_range( FUNC_NAME + " : category index is out of range" );

        const auto& categoryList = category_[item];
        auto ans = categoryList.begin();
        std::advance( ans, category );

        return( ans->first );
      }

      /*
        Class1< X >::expY : 予測値を返す

        xs xe : 独立変数の開始と終了の次の位置

        戻り値 : 予測値
      */
      template< class X >
      template< class In >
      stat_type Class1< X >::expY( In xs, In xe ) const
      {
        const std::string FUNC_NAME = "tadah_fussy::statistics::quantification::Class1< X >::expY";

        error_lib::CheckArrayLength( xs, xe, itemSize(),
                                     FUNC_NAME + " : [ xs, xe ) " + error_message::NEGATIVE_LENGTH,
                                     FUNC_NAME + " : length of [ xs, xe ) " + error_message::NOT_EQUAL_LENGTH + "itemSize()" );

        stat_type ans = my_;
        for ( size_type item = 0 ; xs != xe ; ++xs, ++item ) {
          auto ajk = a_[item].find( *xs );
          if ( ajk == a_[item].end() )
            throw std::out_of_range( FUNC_NAME + " : category not found" );
          ans += ajk->second;
        }

        return( ans );
      }

      /*
        Quantification1< X >::partialCC : 指定したアイテム番号と y の間の偏相関係数を返す

        item : アイテム番号
      */
      template< class X >
      stat_type Class1< X >::partialCC( size_type item ) const
      {
        const std::string FUNC_NAME = "tadah_fussy::statistics::quantification::Class1< X >::partialCC";

        // アイテム番号が範囲外なら out_of_range 例外を投げる
        if ( item >= category_.size() )
          throw std::out_of_range( FUNC_NAME + " : item index is out of range" );

        return( -invS_[0][item + 1] / std::sqrt( invS_[0][0] * invS_[item + 1][item + 1] ) );
      }

      /*
        Class2 専用ヘルパ関数
      */

      /*
        CalcNijk : あるカテゴリに属する要素の中で、指定したアイテム・カテゴリの独立変数を持つものをカウントする

        x : 独立変数
        ys : 従属変数の開始位置
        j : 対象のアイテム番号
        cjk : 対象のカテゴリ
        y : 要素の対象カテゴリ

        戻り値 : カウントした結果
      */
      template< class X, class In >
      stat_type CalcNijk( const Matrix< X >& x, In ys,
                          typename Matrix< X >::size_type j, const X& cjk,
                          typename std::iterator_traits< In >::value_type y )
      {
        using size_type = typename Matrix< X >::size_type;

        stat_type total = 0;
        for ( size_type r = 0 ; r < x.rows() ; ++r ) {
          if ( *ys++ != y ) continue;
          if ( x[r][j] == cjk )
            total += 1;
        }

        return( total );
      }

      /*
        CalcFjk_uv : f( jk, uv ) の計算

        x : 独立変数
        j, u : 対象のアイテム番号
        k, v : 対象のカテゴリ反復子

        戻り値 : f( jk, uv ) の値
      */
      template< class X >
      stat_type CalcFjk_uv( const Matrix< X >& x,
                            typename list_type< X, uint32_t >::size_type j, typename item_type< X, uint32_t >::const_iterator k,
                            typename list_type< X, uint32_t >::size_type u, typename item_type< X, uint32_t >::const_iterator v )
      {
        stat_type t1 = GetDupMatched( x, j, k->first, u, v->first ); // Σ( δ_ijk * δ_iuv )
        stat_type t2 = k->second * v->second / x.rows();             // n_jk * n_uv / N

        return( t1 - t2 );
      }

      /*
        CalcGjk_uv : g( jk, uv ) の計算

        x : 独立変数
        ys : 従属変数の開始位置
        j, u : 対象のアイテム番号
        k, v : 対象のカテゴリ反復子
        y : 要素のカテゴリリスト

        戻り値 : g( jk, uv ) の値
      */
      template< class X, class In >
      stat_type CalcGjk_uv( const Matrix< X >& x, In ys,
                            typename list_type< X, uint32_t >::size_type j, typename item_type< X, uint32_t >::const_iterator k,
                            typename list_type< X, uint32_t >::size_type u, typename item_type< X, uint32_t >::const_iterator v,
                            const item_type< typename std::iterator_traits< In >::value_type, uint32_t >& y )
      {
        stat_type t1 = 0;
        for ( auto i = y.begin() ; i != y.end() ; ++i ) {
          stat_type n_ijk = CalcNijk( x, ys, j, k->first, i->first );
          stat_type n_iuv = CalcNijk( x, ys, u, v->first, i->first );
          t1 += n_ijk * n_iuv / i->second; // Σ( n_ijk * n_iuv / n_i )
        }
        stat_type t2 = k->second * v->second / x.rows(); // n_jk * n_uv / N

        return( t1 - t2 );
      }

      /*
        CreateMatrix : 行列 F, G の計算

        x : 独立変数
        ys : 従属変数の開始位置
        xCategory : カテゴリリスト(独立変数用)
        yCategory : カテゴリリスト(従属変数用)
        f : 行列 F へのポインタ
        g : 行列 G へのポインタ
      */
      template< class X, class In >
      void CreateMatrix( const Matrix< X >& x, In ys,
                         const list_type< X, uint32_t >& xCategory,
                         const item_type< typename std::iterator_traits< In >::value_type, uint32_t >& yCategory,
                         SymmetricMatrix< stat_type >* f, SymmetricMatrix< stat_type >* g )
      {
        using SM = SymmetricMatrix< stat_type >;
        using List = list_type< X, uint32_t >;

        SM fBuff, gBuff; // 線形独立の形にする前の係数行列

        // カテゴリの総数を求める
        typename List::size_type total = 0;
        for ( auto i = xCategory.begin() ; i != xCategory.end() ; ++i )
          total += i->size();

        // 行列のリサイズ
        fBuff.assign( total, total );
        gBuff.assign( total, total );
        f->assign( total - xCategory.size(), total - xCategory.size() );
        g->assign( total - xCategory.size(), total - xCategory.size() );

        // 行列の要素を求める
        typename List::size_type j = 0;
        auto k = xCategory[j].begin();
        for ( SM::size_type r = 0 ; r < fBuff.rows() ; ++r ) {
          typename List::size_type u = j;
          auto v = k;
          for ( SM::size_type c = r ; c < fBuff.cols() ; ++c ) {
            fBuff[r][c] = CalcFjk_uv( x, j, k, u, v );
            gBuff[r][c] = CalcGjk_uv( x, ys, j, k, u, v, yCategory );
            if ( ++v == xCategory[u].end() ) {
              ++u;
              v = xCategory[u].begin();
            }
          }
          if ( ++k == xCategory[j].end() ) {
            ++j;
            k = xCategory[j].begin();
          }
        }

        // 線形独立の形になるように行列を削除する
        std::set< SM::size_type > ignoreRC; // 削除する行列番号
        SM::size_type i = 0;
        for ( typename List::size_type j = 0 ; j < xCategory.size() ; ++j ) {
          ignoreRC.insert( i );
          i += xCategory[j].size();
        }
        SM::size_type oR = 0; // 出力行
        for ( SM::size_type r = 0 ; r < total ; ++r ) {
          if ( ignoreRC.find( r ) != ignoreRC.end() ) continue;
          SM::size_type oC = oR; // 出力列
          for ( SM::size_type c = r ; c < total ; ++c ) {
            if ( ignoreRC.find( c ) != ignoreRC.end() ) continue;
            (*f)[oR][oC] = fBuff[r][c];
            (*g)[oR][oC] = gBuff[r][c];
            ++oC;
          }
          ++oR;
        }
      }

      /*
        GeneralEigenvalueProblem : 一般固有値問題 ( G - λF )a = 0 を解く

        f : 行列 F
        g : 行列 G
        r : 固有値を保持する行列へのポインタ (対角成分が固有値になる)
        q : 固有ベクトルを保持する行列へのポインタ
        e : 収束判定のためのしきい値
        maxCnt : 収束しなかった場合の最大処理回数
        solver : 連立方程式の計算オブジェクト ( 逆行列計算用 )
      */
      template< class Solver >
      void GeneralEigenvalueProblem( const SymmetricMatrix< stat_type >& f, const SymmetricMatrix< stat_type >& g,
                                     SquareMatrix< stat_type >* r, SquareMatrix< stat_type >* q,
                                     stat_type e, uint32_t maxCnt, Solver solver )
      {
        // F を固有値分解
        try {
          math_lib::eigen_lib::Householder_DoubleShiftQR( f, r, q, e, maxCnt );
        } catch( exception::ExcessLimit& ee ) {
          SymmetricMatrix< stat_type > eigen( f );
          math_lib::eigen_lib::Jacobi( &eigen, q, e, maxCnt );
          *r = eigen;
        }

        // P = Q√R の計算
        SquareMatrix< stat_type > p;
        p.assign( r->rows() );
        for ( SquareMatrix< stat_type >::size_type i = 0 ; i < r->rows() ; ++i )
          for ( SquareMatrix< stat_type >::size_type j = 0 ; j < q->rows() ; ++j )
            p[j][i] = (*q)[j][i] * std::sqrt( (*r)[i][i] );

        // P の転置行列
        SquareMatrix< stat_type > pt( p );
        pt.transpose();

        // P, Pt の逆行列
        SquareMatrix< stat_type > invP, invPt;
        matrix_lib::Inverse( p, &invP, solver );
        matrix_lib::Inverse( pt, &invPt, solver );

        // P^-1GPt^-1 の計算
        SquareMatrix< stat_type > buff;
        matrix_lib::mult( invP, g, &buff );
        matrix_lib::mult( buff, invPt, &buff );

        SymmetricMatrix< stat_type > pmGPtm;
        pmGPtm.assign( buff.size() );
        for ( SymmetricMatrix< stat_type >::size_type i = 0 ; i < buff.rows() ; ++i )
          for ( SymmetricMatrix< stat_type >::size_type j = i ; j < buff.cols() ; ++j )
            pmGPtm[i][j] = buff[i][j];

        // P^-1GPt^-1 を固有値分解
        try {
          math_lib::eigen_lib::Householder_DoubleShiftQR( pmGPtm, r, q, e, maxCnt );
        } catch( exception::ExcessLimit& ee ) {
          math_lib::eigen_lib::Jacobi( &pmGPtm, q, e, maxCnt );
          *r = pmGPtm;
        }
        for ( SquareMatrix< stat_type >::size_type j = 0 ; j < q->cols() ; ++j )
          matrix_lib::multMV( invPt, q->column( j ), q->column( j ) );
      }

      /*
        CalcBi : 従属変数のカテゴリごとに Y の平均を求める

        x : 独立変数
        ys : 従属変数の開始位置
        yCategory : カテゴリリスト(従属変数用)
        a : 求めた係数
        bi : Y の平均を保持する配列へのポインタ
      */
      template< class X, class In >
      void CalcBi( const Matrix< X >& x,
                   In ys, const item_type< typename std::iterator_traits< In >::value_type, uint32_t >& yCategory,
                   const list_type< X, stat_type >& a,
                   item_type< typename std::iterator_traits< In >::value_type, stat_type >* bi )
      {
        // bi の初期化
        for ( auto i = yCategory.begin() ; i != yCategory.end() ; ++i )
          bi->insert( std::make_pair( i->first, 0 ) );

        // Y の和をカテゴリごとに求める
        for ( typename Matrix< X >::size_type c = 0 ; c < x.cols() ; ++c ) {
          const auto& ac = a[c];
          In y = ys;
          for ( typename Matrix< X >::size_type r = 0 ; r < x.rows() ; ++r ) {
            auto i = ac.find( x[r][c] );
            (*bi)[*y++] += i->second;
          }
        }

        // カテゴリごとの個数で割って平均とする
        for ( auto i = bi->begin() ; i != bi->end() ; ++i ) {
          auto y = yCategory.find( i->first );
          i->second /= y->second;
        }
      }

      /*
        CalcSyy : y の分散値を求める ( 数量化法 II 専用 )

        x : 独立変数
        xCategory : カテゴリリスト(独立変数用)
        ys : 従属変数の開始位置
        yCategory : カテゴリリスト(従属変数用)
        a : 求めた係数値
        bi : 従属変数のカテゴリごとの y の平均
      */
      template< class X, class In >
      stat_type CalcSyy( const Matrix< X >& x, const list_type< X, uint32_t >& xCategory,
                         In ys, const item_type< typename std::iterator_traits< In >::value_type, uint32_t >& yCategory,
                         const list_type< X, stat_type >& a,
                         const item_type< typename std::iterator_traits< In >::value_type, stat_type >& bi )
      {
        stat_type t1 = 0;
        stat_type t2 = 0;
        for ( auto i = bi.begin() ; i != bi.end() ; ++i ) {
          auto j = yCategory.find( i->first );
          t1 += std::pow( i->second, 2 ) * j->second; // n_i * b_i * b_i
          t2 += i->second * j->second;                // n_i * b_i
        }

        return( ( t1 - std::pow( t2, 2 ) / x.rows() ) / x.rows() );
      }

      /*
        CalcSyj : x と y の共分散を求める

        x : アイテムごとのカテゴリデータ ( 列:アイテム 行:要素 )
        j : アイテム番号
        xCategory : カテゴリのリスト ( 独立変数用 )
        ys : y のリストの先頭
        yCategory : カテゴリのリスト ( 従属変数用 )
        aj : 対象アイテムの係数
        bi : 従属変数のカテゴリごとの y の平均
      */
      template< class X, class In >
      stat_type CalcSyj( const Matrix< X >& x, typename Matrix< X >::size_type j,
                         const list_type< X, uint32_t >& xCategory,
                         In ys, const item_type< typename std::iterator_traits< In >::value_type, uint32_t >& yCategory,
                         const item_type< X, stat_type >& aj,
                         const item_type< typename std::iterator_traits< In >::value_type, stat_type >& bi )
      {
        stat_type t1 = 0;
        stat_type t2 = 0;
        for ( auto k = xCategory[j].begin() ; k != xCategory[j].end() ; ++k ) {
          stat_type sum = 0;
          for ( auto i = bi.begin() ; i != bi.end() ; ++i )
            sum += i->second * CalcNijk( x, ys, j, k->first, i->first );
          auto ajk = aj.find( k->first );
          t1 += sum * ajk->second;       // ΣkΣi( a_jk * b_i * n_ijk )
          t2 += ajk->second * k->second; // Σk( a_jk * n_jk )
        }
        stat_type t3 = 0;
        for ( auto i = yCategory.begin() ; i != yCategory.end() ; ++i ) {
          t3 += ( bi.find( i->first ) )->second * i->second; // Σi( n_i * b_i )
        }

        return( ( t1 - t2 * t3 / x.rows() ) / x.rows() );
      }

      /*
        CreateCovariance : 共分散行列の計算

        x : アイテムごとのカテゴリデータ ( 列:アイテム 行:要素 )
        xCategory : カテゴリのリスト ( 独立変数用 )
        ys : y のリストの先頭
        yCategory : カテゴリのリスト ( 従属変数用 ) 
        a : 求めた係数
        bi : 従属変数のカテゴリごとの y の平均
        s : 共分散行列を保持する変数のポインタ
      */
      template< class X, class In >
      void CreateCovariance( const Matrix< X >& x, const list_type< X, uint32_t >& xCategory,
                             In ys, const item_type< typename std::iterator_traits< In >::value_type, uint32_t >& yCategory,
                             const list_type< X, stat_type >& a, const item_type< typename std::iterator_traits< In >::value_type, stat_type >& bi,
                             SquareMatrix< stat_type >* s )
      {
        using size_type = SquareMatrix< stat_type >::size_type;

        s->assign( x.cols() + 1 );

        // syy の計算
        (*s)[0][0] = CalcSyy( x, xCategory, ys, yCategory, a, bi );

        // sjj の計算
        for ( size_type j = 1 ; j < s->rows() ; ++j )
          (*s)[j][j] = CalcSjj( x, j - 1, xCategory, a[j - 1] );

        // syj の計算
        for ( size_type j = 1 ; j < s->rows() ; ++j )
          (*s)[j][0] = (*s)[0][j] = CalcSyj( x, j - 1, xCategory, ys, yCategory, a[j - 1], bi );

        // sju の計算
        for ( size_type j = 1 ; j < s->rows() ; ++j )
          for ( size_type u = j + 1 ; u < s->rows() ; ++u )
            (*s)[j][u] = (*s)[u][j] = CalcSju( x, j - 1, u - 1, xCategory, a[j - 1], a[u - 1] );
      }

      /*
        Class2 コンストラクタ

        x : アイテムごとのカテゴリデータ ( 列:アイテム 行:要素 )
        ys, ye : y の先頭と終端の次の位置
        solver : 連立方程式を解くためのソルバ
      */
      template< class X, class Y >
      template< class In, class Solver >
      Class2< X, Y >::Class2( const Matrix< X >& x, In ys, In ye, Solver solver )
      {
        const stat_type THRESHOLD = 1E-6;
        const uint32_t MAX_CNT = 10000;

        const std::string FUNC_NAME = "tadah_fussy::statistics::quantification::Class2 constructor";

        error_lib::CheckLinearModel( x, ys, ye,
                                     FUNC_NAME + " : x " + error_message::NOT_POSITIVE_LENGTH,
                                     FUNC_NAME + " : [ ys, ye ) " + error_message::NEGATIVE_LENGTH,
                                     FUNC_NAME + " : size of x " + error_message::NOT_EQUAL_LENGTH + " [ ys, ye )" );

        GetCategoryList( x, &xCategory_ );
        GetCategory( ys, ye, &yCategory_ );

        SymmetricMatrix< stat_type > f, g;
        CreateMatrix( x, ys, xCategory_, yCategory_, &f, &g );

#ifdef DEBUG
        matrix_lib::PrintMatrix( std::cout, f, "F = " );
        std::cout << std::endl;
        matrix_lib::PrintMatrix( std::cout, g, "G = " );
        std::cout << std::endl;
#endif

        SquareMatrix< stat_type > r, q;
        GeneralEigenvalueProblem( f, g, &r, &q, THRESHOLD, MAX_CNT, solver );

#ifdef DEBUG
        matrix_lib::PrintMatrix( std::cout, r, "R = " );
        std::cout << std::endl;
        matrix_lib::PrintMatrix( std::cout, q, "Q = " );
        std::cout << std::endl;
#endif

        // 最大固有値の固有ベクトルを抽出する
        if ( r.size() == 0 )
          throw std::runtime_error( FUNC_NAME + " : no eigenvalue found" );
        stat_type maxEigenValue = r[0][0];
        SquareMatrix< stat_type >::size_type maxEigenVector = 0;
        for ( SquareMatrix< stat_type >::size_type i = 1 ; i < r.size() ; ++i ) {
          if ( maxEigenValue < r[i][i] ) {
            maxEigenValue = r[i][i];
            maxEigenVector = i;
          }
        }
        auto c = q.column( maxEigenVector );
        for ( auto i = xCategory_.begin() ; i != xCategory_.end() ; ++i ) {
          auto j = i->begin();
          a_.push_back( a_item_type() );
          ( a_.back() )[j->first] = 0;
          for ( ++j ; j != i->end() ; ++j )
            ( a_.back() )[j->first] = *c++;
        }

#ifdef DEBUG
        std::cout << "--- a ( not normalized ) ---" << std::endl;
        for ( typename x_list_type::size_type i = 0 ; i < xCategory_.size() ; ++i ) {
          const auto& x = xCategory_[i];
          const auto& a = a_[i];
          for ( auto j = x.begin() ; j != x.end() ; ++j )
            std::cout << j->first << ":" << ( a.find( j->first ) )->second << " ";
          std::cout << std::endl;
        }
#endif

        // 係数の標準化
        NormalizeCoef( x, xCategory_, &a_ );

#ifdef DEBUG
        std::cout << "--- a ( normalized ) ---" << std::endl;
        for ( typename x_list_type::size_type i = 0 ; i < xCategory_.size() ; ++i ) {
          const auto& x = xCategory_[i];
          const auto& a = a_[i];
          for ( auto j = x.begin() ; j != x.end() ; ++j )
            std::cout << j->first << ":" << ( a.find( j->first ) )->second << " ";
          std::cout << std::endl;
        }
#endif

        CalcBi( x, ys, yCategory_, a_, &bi_ );

        CreateCovariance( x, xCategory_, ys, yCategory_, a_, bi_, &s_ );
        matrix_lib::Inverse( s_, &invS_, solver );

#ifdef DEBUG
        matrix_lib::PrintMatrix( std::cout, s_, "S = " );
        std::cout << std::endl;
        matrix_lib::PrintMatrix( std::cout, invS_, "invS = " );
        std::cout << std::endl;
#endif
      }

      /*
        Class2< X, Y >::a : 指定したアイテム番号・カテゴリに対する係数 ( 数量 ) を返す

        item : アイテム番号
        category : カテゴリ
      */
      template< class X, class Y >
      stat_type Class2< X, Y >::a( size_type item, const X& category ) const
      {
        const std::string FUNC_NAME = "tadah_fussy::statistics::quantification::Class2< X, Y >::a";

        // アイテム番号が範囲外なら out_of_range 例外を投げる
        if ( item >= xCategory_.size() )
          throw std::out_of_range( FUNC_NAME + " : item index is out of range" );

        // カテゴリが見つからなかったら out_of_range 例外を投げる
        auto i = a_[item].find( category );
        if ( i == a_[item].end() )
          throw std::out_of_range( FUNC_NAME + " : category not found" );

        return( i->second );
      }

      /*
        Class2< X, Y >::category : 指定したアイテム・カテゴリ番号のカテゴリの値を返す

        item : アイテム番号
        category : カテゴリ番号

        戻り値 : カテゴリの値
      */
      template< class X, class Y >
      X Class2< X, Y >::category( size_type item, size_type category ) const
      {
        const std::string FUNC_NAME = "tadah_fussy::statistics::quantification::Class2< X, Y >::category";

        // 範囲外なら out_of_range 例外を投げる
        if ( item >= xCategory_.size() )
          throw std::out_of_range( FUNC_NAME + " : item index is out of range" );
        if ( category >= xCategory_[item].size() )
          throw std::out_of_range( FUNC_NAME + " : category index is out of range" );

        const auto& categoryList = xCategory_[item];
        auto ans = categoryList.begin();
        std::advance( ans, category );

        return( ans->first );
      }

      /*
        Quantification2< X, Y >::expY : 独立変数から y の予測値を返す

        s, e : 独立変数の先頭と末尾の次の位置
      */
      template< class X, class Y >
      template< class In >
      stat_type Class2< X, Y >::expY( In s, In e ) const
      {
        const std::string FUNC_NAME = "tadah_fussy::statistics::quantification::Class2< X, Y >::expY";

        error_lib::CheckArrayLength( s, e, a_.size() );

        stat_type y = 0;
        for ( typename a_list_type::size_type j = 0 ; j < a_.size() ; ++j ) {
          const auto& aj = a_[j];
          const auto& ajk = aj.find( *s++ );
          if ( ajk == aj.end() )
            throw std::out_of_range( FUNC_NAME + " : category not found" );
          y += ajk->second;
        }

        return( y );
      }

      /*
        Class2< X, Y >::discriminant : 独立変数から y のカテゴリを判別する

        s, e : 独立変数の先頭と末尾の次の位置
      */
      template< class X, class Y >
      template< class In >
      Y Class2< X, Y >::discriminant( In s, In e ) const
      {
        const std::string FUNC_NAME = "tadah_fussy::statistics::quantification::Class2< X, Y >::discriminant";

        // y の予測値に最も近い bi を探す
        stat_type ey = expY( s, e );
        auto minIndex = bi_.begin();
        if ( minIndex == bi_.end() )
          return( distribution::Base::nan() );
        stat_type minDiff = std::abs( minIndex->second - ey );
        auto i = minIndex;
        while ( ++i != bi_.end() ) {
          stat_type diff = std::abs( i->second - ey );
          if ( minDiff > diff ) {
            minDiff = diff;
            minIndex = i;
          }
        }

        return( minIndex->first );
      }

      /*
        Class2< X, Y >::partialCC : 指定したアイテム番号と y の間の偏相関係数を返す

        item : アイテム番号
      */
      template< class X, class Y >
      stat_type Class2< X, Y >::partialCC( size_type item ) const
      {
        const std::string FUNC_NAME = "tadah_fussy::statistics::quantification::Class2< X, Y >::partialCC";

        // アイテム番号が範囲外なら out_of_range 例外を投げる
        if ( item >= xCategory_.size() )
          throw std::out_of_range( FUNC_NAME + " : item index is out of range" );

        return( -invS_[0][item + 1] / std::sqrt( invS_[0][0] * invS_[item + 1][item + 1] ) );
      }

    } // namespace quantification

  } // namespace statistics

} // namespace tadah_fussy
