/**
   @file regression.h
   @brief 回帰分析ルーチン定義用ヘッダ

   @author tadah_fussy
   @date 2016/06/04 PLS 回帰のみを実装
   @date 2021/03/25 名前空間の変更
**/
#ifndef REGRESSION_20210325_H
#define REGRESSION_20210325_H

#include "statistics.h"

namespace tadah_fussy
{
  namespace statistics
  {
    namespace regression
    {
      /**
         @brief PLS 回帰 (NIPALS)
      **/
      class NIPALS
      {
        using stat_type = distribution::Base::stat_type;

        Matrix< stat_type > x_; // 独立変数 X
        Matrix< stat_type > y_; // 従属変数 Y

        std::vector< stat_type > b_; // 主成分 t, w での回帰係数
        Matrix< stat_type > pT_;     // p = Xt・t を列ベクトルとする行列の転置
        Matrix< stat_type > qT_;     // Q = Yt・w を列ベクトルとする行列の転置

        // 反復処理により u, t, v, w を求める関数
        void iterProc( Matrix< stat_type >* x, Matrix< stat_type >* y,
                       size_t latentVecIdx, stat_type e, uint32_t maxCnt );

      public:

        /// @brief 独立変数 X と従属変数 Y を指定して構築
        ///
        /// @param x 独立変数 X
        /// @param y 従属変数 Y
        template< class T, class U >
        NIPALS( const Matrix< T >& x, const Matrix< U >& y );

        /// @brief PLS回帰を行い、回帰係数を a に返す
        ///
        /// @param a 回帰係数を得るための配列へのポインタ
        /// @param latentVecs 回帰係数(計算する潜在ベクトル)の個数
        /// @param e 反復処理でベクトルが収束したかを判定するためのしきい値
        /// @param maxCnt 反復処理の最大処理回数(反復処理は各回帰係数の算出で繰り返されるため、それぞれの処理での最大回数を意味する)
        /// @return なし
        void operator()( Matrix< distribution::Base::stat_type >* a, size_t latentVecs, distribution::Base::stat_type e, uint32_t maxCnt );
      };

      /**
         @brief PLS 回帰 (SIMPLS)
      **/
      class SIMPLS
      {
        using stat_type = distribution::Base::stat_type;

        Matrix< stat_type > x_; // 独立変数 X
        Matrix< stat_type > y_; // 従属変数 Y

        Matrix< stat_type > u_;  // S = Xt・Y の左特異ベクトルを列ベクトルとする行列
        Matrix< stat_type > pT_; // p = X・t を列ベクトルとする行列の転置
        Matrix< stat_type > qT_; // Q = Y・t を列ベクトルとする行列の転置

        // 反復処理により u, t, v, w を求める関数
        void iterProc( Matrix< stat_type >* x, Matrix< stat_type >* y,
                       size_t latentVecIdx, stat_type e, uint32_t maxCnt );

      public:

        /// @brief 独立変数 X と従属変数 Y を指定して構築
        ///
        /// @param x 独立変数 X
        /// @param y 従属変数 Y
        template< class T, class U >
        SIMPLS( const Matrix< T >& x, const Matrix< U >& y );

        /// @brief PLS回帰を行い、回帰係数を a に返す
        ///
        /// @param a 回帰係数を得るための配列へのポインタ
        /// @param latentVecs 回帰係数(計算する潜在ベクトル)の個数
        /// @param e 反復処理でベクトルが収束したかを判定するためのしきい値
        /// @param maxCnt 反復処理の最大処理回数(反復処理は各回帰係数の算出で繰り返されるため、それぞれの処理での最大回数を意味する)
        /// @return なし
        void operator()( Matrix< distribution::Base::stat_type >* a, size_t latentVecs, distribution::Base::stat_type e, uint32_t maxCnt );
      };

    } // namespace regression

  } // namespace statistics

} // namespace tadah_fussy

#include "regression_inline.h"

#endif
