#include "statistics.h"

namespace tadah_fussy
{
  namespace statistics
  {
    using error_message::NNNUM;
    using error_message::GT;
    using error_message::LE;
    using error_message::EQ;
    using error_message::IN01;

    using math_lib::BinCoef;
    using math_lib::OpBind;
    using math_lib::Sum;
    using math_lib::SumAll;
    using math_lib::Pi;

    using matrix_lib::PrintArray;
    using matrix_lib::PrintMatrix;

    using std::string;
    using std::vector;
    using std::accumulate;

    using stat_type = distribution::Base::stat_type;
    using counter_type = distribution::Base::counter_type;

    namespace distribution
    {
      /**
         Discrete : 離散確率分布
      **/

      /*
        Discrete::p : 区間 [s,e] における確率を返す

        s > e の場合は反転して計算する
      */
      stat_type Discrete::p( var_type s, var_type e ) const
      {
        if ( s > e ) std::swap( s, e );

        math_lib::KahanSumOp< stat_type > kahanSum;
        stat_type sum = stat_type();
        for ( ; s <= e ; ++s )
          sum = kahanSum( sum, (*this)[s] );

        return( sum );
      }

      /**
         Binomial : 二項分布
      **/

      /*
        Binomial::average : 平均値
      */
      stat_type Binomial::average() const
      {
        return( n_ * p_ );
      }

      /*
        Binomial::variance : 分散
      */
      stat_type Binomial::variance() const
      {
        return( n_ * p_ * ( 1.0 - p_ ) );
      }

      /*
        Binomial::operator[] : 確率変数 r に対する確率密度を返す
      */
      stat_type Binomial::operator[]( var_type r ) const
      {
        // r < 0 または r > n_ なら戻り値は 0
        if ( r < 0 || r > n_ ) return( 0 );

        // p_ = 0_ なら常に r = 0 しか発生しない
        if ( p_ == 0 ) return( ( r == 0 ) ? 1 : 0 );
        // p_ = 1 なら常に r = n_ しか発生しない
        if ( p_ == 1 ) return( ( r == n_ ) ? 1 : 0 );

        // n_ = 0 かつ r = 0 なら戻り値は 1
        if ( n_ == 0 && r == 0 ) return( 1 );

        stat_type c = BinCoef< stat_type >( n_, r );

        return( c * std::pow( p_, r ) * std::pow( 1.0 - p_, n_ - r ) );
      }

      /**
         Poisson : ポアソン分布
      **/

      /*
        Poisson::average : 平均値
      */
      stat_type Poisson::average() const
      {
        return( lambda_ );
      }

      /*
        Poisson::variance : 分散
      */
      stat_type Poisson::variance() const
      {
        return( lambda_ );
      }

      /*
        Poisson::operator[] : 確率変数 r に対する確率密度を返す
      */
      stat_type Poisson::operator[]( var_type r ) const
      {
        if ( r < 0 ) return( 0 );

        // λ = 0 なら事象の確率 p = 0 を意味する
        if ( lambda_ == 0 ) return( ( r == 0 ) ? 1 : 0 );

        return( std::exp( -lambda_ ) * std::pow( lambda_, r ) / std::tgamma( static_cast< stat_type >( r + 1.0 ) ) );
      }

      /**
         NegativeBinomial : 負の二項分布
      **/

      /*
        NegativeBinomial::average : 平均値
      */
      stat_type NegativeBinomial::average() const
      {
        return( k_ * ( 1 - p_ ) / p_ );
      }

      /*
        NegativeBinomial::variance : 分散
      */
      stat_type NegativeBinomial::variance() const
      {
        return( k_ * ( 1 - p_ ) / ( p_ * p_ ) );
      }

      /*
        NegativeBinomial::operator[] : 確率変数 r に対する確率密度を返す
      */
      stat_type NegativeBinomial::operator[]( var_type r ) const
      {
        // 確率変数がゼロより小さければ確率は常にゼロ
        if ( r < 0 ) return( 0 );

        // 事象の起こらない確率がゼロなら r = 0 のとき 1 でそれ以外は 0
        if ( p_ == 1 ) return( ( r == 0 ) ? 1 : 0 );

        stat_type c = BinCoef< stat_type >( r + k_ - 1, r );

        return( c * std::pow( p_, k_ ) * std::pow( 1.0 - p_, r ) );
      }

      /*
        NegativeBinomial_N::average : 平均値
      */
      stat_type NegativeBinomial_N::average() const
      {
        return( k_ / p_ );
      }

      /*
        NegativeBinomial_N::variance : 分散
      */
      stat_type NegativeBinomial_N::variance() const
      {
        return( k_ * ( 1 - p_ ) / ( p_ * p_ ) );
      }

      /*
        NegativeBinomial_N::operator[] : 確率変数 n に対する確率密度を返す
      */
      stat_type NegativeBinomial_N::operator[]( var_type n ) const
      {
        // 試行回数がゼロ以下か k_ に満たない場合は確率ゼロ
        if ( n <= 0 || n < k_ ) return( 0 );

        // 事象の起こらない確率がゼロなら n = k_ のとき 1 でそれ以外は 0
        if ( p_ == 1 ) return( ( n == k_ ) ? 1 : 0 );

        stat_type c = BinCoef< stat_type >( n - 1, k_ - 1 );

        return( c * std::pow( p_, k_ ) * std::pow( 1.0 - p_, n - k_ ) );
      }

      /**
         Hypergeometric : 超幾何分布
      **/

      /*
        Hypergeometric::average : 平均値を返す
      */
      stat_type Hypergeometric::average() const
      {
        // 母数がゼロなら無効値
        if ( a_ + b_ == 0 ) return( nan() );

        return( static_cast< stat_type >( a_ * k_ ) / ( a_ + b_ ) );
      }

      /*
        Hypergeometric::variance : 分散を返す
      */
      stat_type Hypergeometric::variance() const
      {
        // 母数が 1 以下なら無効値
        if ( a_ + b_ <= 1 ) return( nan() );

        return( static_cast< stat_type >( a_ * b_ * k_ * ( a_ + b_ - k_ ) ) / ( ( a_ + b_ ) * ( a_ + b_ ) * ( a_ + b_ - 1 ) ) );
      }

      /*
        Hypergeometric::operator[] : 確率変数 r に対する確率密度を返す
      */
      stat_type Hypergeometric::operator[]( var_type r ) const
      {
        if ( r < 0 || r > k_ ) return( 0 );

        stat_type c1 = BinCoef< stat_type >( a_, r );
        stat_type c2 = BinCoef< stat_type >( b_, k_ - r );
        stat_type c3 = BinCoef< stat_type >( a_ + b_, k_ );

        return( c1 * c2 / c3 );
      }

      /**
         U : U-分布
      **/

      /*
        U::calcP : U-分布 の確率分布 Pm,n(u) を求める
      */
      stat_type U::calcP( var_type m, var_type n, var_type u ) const
      {
        if ( u < 0 || u > m * n )
          return( 0 );

        // 左右対称のため u の値は mn / 2 以下にできる
        if ( u > m * n / 2 )
          u = m * n - u;

        // Pm,0(U) = P0,n(U) = 1 ( u = 0 ) or 0 ( u ≠ 0 )
        if ( m == 0 || n == 0 )
          return( ( u == 0 ) ? 1 : 0 );

        stat_type res = getP( m, n, u );
        if ( res >= 0 ) return( res );

        res = ( m * calcP( m - 1, n, u - n ) + n * calcP( m, n - 1, u ) ) / ( m + n );
        setP( m, n, u, res );

        return( res );
      }

      /*
        U::getP : バッファから U-分布の確率 Pm,n(u) を取得する

        未登録の場合は負数を返す。
      */
      stat_type U::getP( var_type m, var_type n, var_type u ) const
      {
        UKey key( m, n ); // UMap内の対象キー
        auto cit = buff_.find( key ); // 対象キーに対するuMapの反復子
        if ( cit == buff_.end() )
          return( -1 );

        // 左右対称のため u の値は mn / 2 以下にできる
        if ( u > m * n / 2 )
          u = m * n - u;

        const vector< stat_type >& vec = cit->second; // 値を取得する対象配列
        if ( static_cast< vector< stat_type >::size_type >( u ) >= vec.size() )
          return( 0 );

        return( vec[u] );
      }

      /*
        U::setP : バッファに U-分布の確率 Pm,n(u) をセットする
      */
      void U::setP( var_type m, var_type n, var_type u, stat_type p ) const
      {
        UKey key( m, n ); // UMap内の対象キー

        // 対象キーが見つからない場合は新規追加
        if ( buff_.find( key ) == buff_.end() ) {
          // あらかじめ負数で初期化しておく
          buff_.insert( std::make_pair( key, vector< stat_type >( m * n / 2 + 1, -1 ) ) );
        }

        // 左右対称のため u の値は mn / 2 以下にできる
        if ( u > m * n / 2 )
          u = m * n - u;

        vector< stat_type >& vec = buff_[key]; // 値をセットする対象配列
        if ( static_cast< vector< stat_type >::size_type >( u ) < vec.size() )
          vec[u] = p;
      }

      /*
        U コンストラクタ
      */
      U::U( var_type m, var_type n )
        : m_( m ), n_( n )
      {
        const string FUNC_NAME = "tadah_fussy::statistics::distribution::U constructor";

        error_lib::NNNum( m, FUNC_NAME + " : m " + NNNUM );
        error_lib::NNNum( n, FUNC_NAME + " : n " + NNNUM );
      }

      // U( m, n ) = u となる確率を返す
      stat_type U::operator[]( var_type u ) const
      {
        return( calcP( m_, n_, u ) );
      }

      // 平均値
      stat_type U::average() const
      {
        return( m_ * n_ / 2.0 );
      }

      // 分散値
      stat_type U::variance() const
      {
        return( m_ * n_ * ( m_ + n_ + 1 ) / 12.0 );
      }

      /**
         SignedRank : 符号付き順位和の確率分布
      **/

      /*
        SignedRank::calcP : 符号付き順位和 sRank の確率分布 Pn(sRank) を求める
      */
      stat_type SignedRank::calcP( var_type n, var_type sRank ) const
      {
        if ( n == 0 ) return( ( sRank == 0 ) ? 1 : 0 );

        var_type sRankMax = n * ( n + 1 ) / 2;
        if ( std::abs( sRank ) > sRankMax )
          return( 0 );

        stat_type res = getP( n, sRank );
        if ( res >= 0 ) return( res );

        res = ( calcP( n - 1, sRank - n ) + calcP( n - 1, sRank + n ) ) / 2.0;
        setP( n, sRank, res );

        return( res );
      }

      /*
        SignedRank::getP : バッファから符号付き順位和 sRank の確率分布 Pn(sRank) を取得する

        未登録の場合は負数を返す。
      */
      stat_type SignedRank::getP( var_type n, var_type sRank ) const
      {
        auto cit = buff_.find( n ); // 対象キーに対する sRankMap の反復子
        if ( cit == buff_.end() )
          return( -1 );

        const ProbMap& sMap = cit->second; // 値を取得する対象配列
        auto sPair = sMap.find( sRank );
        if ( sPair == sMap.end() )
          return( -1 );

        return( sPair->second );
      }

      /*
        SignedRank::setP : バッファに符号付き順位和 sRank の確率分布 Pn(sRank) をセットする
      */
      void SignedRank::setP( var_type n, var_type sRank, stat_type p ) const
      {
        ProbMap& sRankDist = buff_[n];
        sRankDist[sRank] = p;
      }

      /*
        SignedRank::operator[] : Pn(sRank) となる確率を返す
      */
      stat_type SignedRank::operator[]( var_type sRank ) const
      {
        return( calcP( n_, sRank ) );
      }

      /*
        SignedRank::average : 平均値
      */
      stat_type SignedRank::average() const
      {
        return( 0 );
      }

      /*
        SignedRank::variance : 分散値
      */
      stat_type SignedRank::variance() const
      {
        return( static_cast< stat_type >( n_ * ( n_ + 1 ) * ( 2 * n_ + 1 ) ) / 6.0 );
      }

      /**
         Sg : Sg/σ^2 の正確な確率分布
      **/

      /*
        Sg::calcRankSqSum : 順列のリスト permList 内の、各集団の順列に対する添字のリスト classIdx で
        指定された順列の各要素毎の和を計算し、その二乗和を求める

        permListの第一添字を集団のインデックス、第二添字を集団内の要素のインデックスとしたとき
        classIdx[i] は i 番目の集団の順列に対する添字を表し
        permList[classIdx[i]] は i 番目の集団に対する順列を表し
        permList[classIdx[i]][j] は 対象の集団に対する順列の j 番目の要素の順位を表す
      */
      stat_type Sg::calcRankSqSum( const vector< vector< Discrete::var_type > >& permList,
                                   const vector< GetPerm::class_index_type >& classIdx )
      {
        typedef stat_type ( *func )( stat_type, stat_type );
        using namespace std::placeholders;

        GetPerm::data_index_type dataCnt = permList[0].size(); // 集団内の要素数

        vector< stat_type > sq( dataCnt ); // 順列の和を保持するバッファ
        for ( GetPerm::data_index_type i = 0 ; i < dataCnt ; ++i )
          sq[i] = SumAll< stat_type >( classIdx, OpBind( GetPerm( permList, i ) ) );

        return( SumAll< stat_type >( sq, OpBind( std::bind( std::ptr_fun( static_cast< func >( std::pow ) ), _1, 2.0 ) ) ) );
      }

      /*
        Sg::nextClassIdx : classIdx を順列の次の組み合わせに更新する

        index : 現在着目している最小添字
        maxPerm : 順列の個数(指定できる順列の番号の最大値)

        これ以上組み合わせが存在しない場合は false を返す
      */
      bool Sg::nextClassIdx( vector< GetPerm::class_index_type >::size_type index,
                             GetPerm::class_index_type maxPerm, vector< GetPerm::class_index_type >* classIdx )
      {
        // 最小添字が順列の番号リストのサイズを超えたらこれ以上組み合わせはない
        if ( index >= classIdx->size() ) return( false );

        // index が指す番号が最大値より小さい場合
        if ( (*classIdx)[index] < maxPerm - 1 ) {
          ++( (*classIdx)[index] );
          return( true );
        }

        // index が指す番号が最大値に達した場合
        // 番号をリセットして次の添字の値を更新
        (*classIdx)[index] = 0;

        return( nextClassIdx( index + 1, maxPerm, classIdx ) );
      }

      /*
        Sg コンストラクタ : Sg / σ^2 の正確な確率分布を求める

        dataCnt : 集団内の要素数
        classCnt : 集団の数
      */
      Sg::Sg( Discrete::var_type dataCnt, Discrete::var_type classCnt, Base::stat_type threshold )
        : std::map< Base::stat_type, Base::stat_type, numeric_lib::ApproxLess< Base::stat_type > >
        ( numeric_lib::ApproxLess< Base::stat_type >( threshold ) )
      {
        const string FUNC_NAME = "tadah_fussy::statistics::distribution::Sg constructor";

        error_lib::NNNum( dataCnt, FUNC_NAME + " : dataCnt " + NNNUM );
        error_lib::NNNum( classCnt, FUNC_NAME + " : classCnt " + NNNUM );

        if ( dataCnt == 0 || classCnt == 0 ) return;

        // 集団内の要素数個の順列を初期化
        vector< Discrete::var_type > perm( dataCnt );
        for ( Discrete::var_type i = 1 ; i <= dataCnt ; ++i )
          perm[i - 1] = i;

        // 全ての順列からなるリストを作成
        vector< vector< Discrete::var_type > > permList;
        do {
          permList.push_back( perm );
        } while ( next_permutation( perm.begin(), perm.end() ) );

        vector< GetPerm::class_index_type > classIdx( classCnt, GetPerm::class_index_type() ); // 各集団に対する順列の番号

        stat_type offset = static_cast< stat_type >( classCnt * ( dataCnt + 1 ) ) * 3.0;   // オフセット値( 3n( k + 1 ) )
        stat_type var = static_cast< stat_type >( classCnt * dataCnt * ( dataCnt + 1 ) ) / 12.0; // 分散値( nk( k + 1 ) / 12 )

        // 計算開始
        do {
          // Sg / var - offset (確率変数)
          stat_type x = calcRankSqSum( permList, classIdx ) / var - offset;
          iterator it = find( x );
          if ( it == end() )
            (*this)[x] = 1;
          else
            it->second += 1;
        } while ( nextClassIdx( 0, permList.size(), &classIdx ) );

        // 和を確率密度に変換(組み合わせ総数は ( k! )^n )
        stat_type total = std::pow( std::tgamma( static_cast< stat_type >( dataCnt + 1 ) ), classCnt );
        for ( iterator it = begin() ; it != end() ; ++it )
          it->second /= total;
      }

    } // namespace distribution

    /**
       確率密度を求めるヘルパ関数
    **/

    /*
      連続分布 dist に従う確率変数を n 個抽出した時、j 番目の順序統計量が x になる確率密度を返す
    */
    stat_type OrderedPDF( stat_type x, counter_type n, counter_type j, distribution::Continuous& dist )
    {
      // j が範囲外なら確率密度はゼロとする
      if ( j == 0 || n < j ) return( 0 );

      // k = N! / ( j - 1 )!( N - j )!
      stat_type k = std::tgamma( static_cast< stat_type >( n + 1 ) ) /
        ( std::tgamma( static_cast< stat_type >( j ) ) * std::tgamma( static_cast< stat_type >( n - j + 1 ) ) );
      // distの下側確率
      stat_type lowerP = dist.lower_p( x );

      return(
             k * std::pow( lowerP, j - 1 ) *
             std::pow( 1.0 - lowerP, n - j ) * dist[x]
             );
    }

    namespace distribution
    {
      /**
         Continuous : 連続確率分布
      **/

      /*
        binSearch : 確率が prob になる区間 [0,t] に対して t を求める(二部探索)

        threshold はしきい値(pを求める精度)を表す。
      */
      stat_type Continuous::binSearch( stat_type prob, stat_type threshold ) const
      {
        const stat_type delta = 1.0; // 範囲を探索するときの単位
        stat_type t = delta;         // t は delta で初期化
        stat_type gap = delta / 2.0; // t を調整するための差分

        // delta 刻みで範囲を探索
        for ( ; prob > p( 0, t ) ; t += 1.0 );

        // t - delta から t の間で二部探索
        t -= gap;
        gap /= 2.0;
        while ( gap >= prob * threshold ) {
          stat_type prob_t = p( 0, t );
          if ( std::abs( prob_t - prob ) < std::numeric_limits<stat_type>::epsilon() )
            break;
          t = ( prob_t > prob ) ? t - gap : t + gap;
          gap /= 2.0;
        }

        return( t );
      }

      /**
         Cauchy : コーシー分布
      **/

      // 確率変数 x における確率密度を返す
      stat_type Cauchy::operator[]( stat_type x ) const
      {
        return( r_ / ( math_lib::Pi< stat_type >() * ( ( x - m_ ) * ( x - m_ ) + r_ * r_ ) ) );
      }

      // 区間 ( -∞, a ] における確率を返す
      stat_type Cauchy::lower_p( stat_type a ) const
      {
        return( std::atan( ( a - m_ ) / r_ ) / math_lib::Pi< stat_type >() + 1.0 / 2.0 );
      }

      // 区間  [ a, b ] または [ b, a ] における確率を返す
      stat_type Cauchy::p( stat_type a, stat_type b ) const
      {
        return( ( ( b > a ) ?
                  std::atan( ( b - m_ ) / r_ ) - std::atan( ( a - m_ ) / r_ ) :
                  std::atan( ( a - m_ ) / r_ ) - std::atan( ( b - m_ ) / r_ ) )
                / math_lib::Pi< stat_type >() );
      }

      stat_type Cauchy::average() const { return( nan() ); }  // 平均値
      stat_type Cauchy::variance() const { return( nan() ); } // 分散

      /**
         Exponential : 指数分布
      **/

      stat_type Exponential::average() const  // 平均値
      {
        return( 1.0 / lambda_ );
      }

      stat_type Exponential::variance() const // 分散
      {
        return( 1.0 / std::pow( lambda_, 2.0 ) );
      }

      /*
        Exponential::operator[] : 確率変数 t に対する確率密度を返す
      */
      stat_type Exponential::operator[]( stat_type t ) const
      {
        if ( t < 0 ) return( 0 ); // 負の領域はゼロ

        return( lambda_ * std::exp( -lambda_ * t ) );
      }

      /*
        Exponential::lower_p : 区間 ( -∞, a ] における確率を返す
      */
      stat_type Exponential::lower_p( stat_type a ) const
      {
        if ( a < 0 ) return( 0 ); // 区間が負の領域ならゼロを返す

        return( 1.0 - std::exp( -lambda_ * a ) );
      }

      /**
         Uniform : 一様分布
      **/

      stat_type Uniform::average() const // 平均値
      {
        return( ( a_ + b_ ) / 2.0 );
      }

      stat_type Uniform::variance() const // 分散
      {
        return( std::pow( b_ - a_, 2.0 ) / 12.0 );
      }

      /*
        Uniform::operator[] : 確率変数 x における確率密度を返す
      */
      stat_type Uniform::operator[]( stat_type x ) const
      {
        if ( x < a_ || x > b_ ) return( 0 );

        return( ( a_ == b_ ) ? infinity() : 1.0 / ( b_ - a_ ) );
      }

      /*
        Uniform::lower_p : 区間 ( -∞, x ] における確率を返す
      */
      stat_type Uniform::lower_p( stat_type x ) const
      {
        if ( x < a_ ) return( 0 );
        if ( x >= b_ ) return( 1 );

        return( ( x - a_ ) / ( b_ - a_ ) );
      }

      /**
         Normal : 正規分布
      **/

      // 確率変数 x における確率密度を返す
      stat_type Normal::operator[]( stat_type x ) const
      {
        return( std::exp( -std::pow( x - mu_, 2.0 ) / ( var_ * 2.0 ) ) /
                std::sqrt( 2.0 * math_lib::Pi< stat_type >() * var_ )
                );
      }

      /*
        Normal::lower_p : 区間 ( -∞, a ] における確率を返す
      */
      stat_type Normal::lower_p( stat_type a ) const
      {
        a = ( a - mu_ ) / std::sqrt( var_ * 2.0 );
        stat_type p = std::erf( std::abs( a ) ) / 2.0;

        if ( a < 0 )
          p = 0.5 - p;
        else
          p += 0.5;

        return( p );
      }

      /*
        Normal::p : 区間 [ a, b ] における確率を返す
      */
      stat_type Normal::p( stat_type a, stat_type b ) const
      {
        if ( a == b ) return( 0 );
        if ( a > b ) std::swap( a, b );

        a = ( a - mu_ ) / std::sqrt( var_ * 2.0 );
        b = ( b - mu_ ) / std::sqrt( var_ * 2.0 );

        if ( a > 0 && b > 0 ) {
          return( ( std::erf( b ) - std::erf( a ) ) / 2 );
        } else if ( a < 0 && b > 0 ) {
          return( ( std::erf( b ) + std::erf( -a ) ) / 2 );
        } else {
          return( ( std::erf( -a ) - std::erf( -b ) ) / 2 );
        }
      }

      stat_type Normal::average() const { return( mu_ ); }   // 平均値
      stat_type Normal::variance() const { return( var_ ); } // 分散

      /**
         ChiSquare : カイ二乗分布
      **/

      namespace
      {
        /*
          IncompleteGamma : 不完全ガンマ関数 Γ( n / 2, x ) = ∫{0→x} t^(n/2) e^-t dt の計算
        */
        stat_type IncompleteGamma( Base::counter_type n, stat_type x )
        {
          // x < 0 なら未定義
          if ( x < 0 ) return( Base::nan() );
          // x → +∞ ならガンマ関数そのもの
          if ( std::isinf( x ) ) return( std::tgamma( static_cast< stat_type >( n / 2.0 ) ) );

          if ( n == 1 ) {
            return( std::sqrt( math_lib::Pi< stat_type >() ) * std::erf( std::sqrt( x ) ) );
          } else if ( n == 2 ) {
            return( 1.0 - std::exp( -x ) );
          } else if ( n > 2 ) {
            stat_type a = n / 2.0 - 1.0;
            return( a * IncompleteGamma( n - 2, x ) - std::pow( x, a ) * std::exp( -x ) );
          }

          return( 0 );
        }

      }

      /*
        ChiSquare::operator[] : 確率変数 x における確率密度を返す
      */
      stat_type ChiSquare::operator[]( stat_type x ) const
      {
        if ( x == 0 && n_ == 1 ) return( infinity() );
        if ( x < 0 ) return( 0 );

        return( std::pow( x, ( n_ - 2.0 ) / 2.0 ) * std::exp( -x / 2.0 ) /
                ( std::pow( 2.0, n_ / 2.0 ) * std::tgamma( static_cast< stat_type >( n_ / 2.0 ) ) ) );
      }

      /*
        ChiSquare::lower_p : 区間 (-∞,a] における確率を返す
      */
      stat_type ChiSquare::lower_p( stat_type a ) const
      {
        if ( a <= 0 ) return( 0 );

        return( IncompleteGamma( n_, a / 2.0 ) /
                std::tgamma( static_cast< stat_type >( n_ / 2.0 ) ) );
      }

      stat_type ChiSquare::average() const  // 平均値
      {
        return( n_ );
      }

      stat_type ChiSquare::variance() const // 分散
      {
        return( 2 * n_ );
      }

      /**
         F : F-分布
      **/

      namespace
      {
        /*
          SinPowerInteg : sin 関数のべき乗の積分値 I_n(x) = ∫{0→arcsin(√x)} sinθ^n dθ の計算
        */
        stat_type SinPowerInteg( Base::counter_type n, stat_type x )
        {
          if ( x < 0 || x > 1 ) return( Base::nan() );

          if ( n == 0 ) {
            return( std::asin( std::sqrt( x ) ) );
          } else if ( n == 1 ) {
            return( 1.0 - std::sqrt( 1.0 - x ) );
          } else {
            stat_type d = -std::pow( x, ( n - 1.0 ) / 2.0 ) * std::sqrt( 1.0 - x ) / n;
            return( d + ( n - 1.0 ) * SinPowerInteg( n - 2, x ) / n );
          }
        }

        /*
          IncompleteBeta : 不完全ベータ関数 Bx( m / 2, n / 2 ) = ∫{0→x} t^(m/2-1) ( 1 - t )^(n/2-1) dt の計算
        */
        stat_type IncompleteBeta( Base::counter_type m, Base::counter_type n, stat_type x )
        {
          if ( x < 0 || x > 1 ) return( Base::nan() );
          if ( m == 0 || n == 0 ) return( Base::nan() );

          stat_type a = m / 2.0;
          stat_type b = n / 2.0;

          if ( n == 1 ) {
            return( SinPowerInteg( m - 1, x ) * 2.0 );
          } else if ( n == 2 ) {
            return( std::pow( x, a ) / a );
          } else {
            return( ( b - 1.0 ) * IncompleteBeta( m + 2, n - 2, x ) / a + std::pow( x, a ) * std::pow( 1.0 - x, b - 1.0 ) / a );
          }
        }

      }

      /*
        F::operator[] : 確率変数 x における確率密度を返す
      */
      stat_type F::operator[]( stat_type x ) const
      {
        if ( x == 0 && m_ == 1 ) return( infinity() );
        if ( x == 0 && m_ == 2 )
          return( 2.0 * std::tgamma( static_cast< stat_type >( n_ / 2.0 + 1.0 ) ) /
                  ( n_ * std::tgamma( static_cast< stat_type >( n_ / 2.0 ) ) ) );
        if ( x <= 0 ) return( 0 );

        stat_type a = m_ / 2.0;
        stat_type b = n_ / 2.0;
        stat_type s = ( m_ + n_ ) / 2.0;

        stat_type beta = std::tgamma( a ) * std::tgamma( b ) / std::tgamma( s );

        return( std::pow( m_, a ) * std::pow( n_, b ) * std::pow( x, a - 1 ) /
                ( beta * std::pow( m_ * x + n_, s ) ) );
      }

      /*
        F::lower_p : 区間 ( -∞, a ] における確率を返す
      */
      stat_type F::lower_p( stat_type a ) const
      {
        if ( a <= 0 ) return( 0 );
        if ( std::isinf( a ) ) return( 1 );

        stat_type beta = std::tgamma( static_cast< stat_type >( m_ / 2.0 ) ) * std::tgamma( static_cast< stat_type >( n_ / 2.0 ) ) /
          std::tgamma( static_cast< stat_type >( ( m_ + n_ ) / 2.0 ) );

        stat_type r = a / ( a + static_cast<stat_type>( n_ ) / static_cast<stat_type>( m_ ) );
        if ( m_ >= n_ )
          return( IncompleteBeta( m_, n_, r ) / beta );
        else
          return( 1.0 - IncompleteBeta( n_, m_, 1.0 - r ) / beta );
      }

      /*
        F::average : 平均値を返す
      */
      stat_type F::average() const
      {
        stat_type n = n_;

        return( ( n_ > 2 ) ? n / ( n - 2 ) : nan() );
      }

      /*
        F::variance : 分散を返す
      */
      stat_type F::variance() const
      {
        if ( n_ <= 4 ) return( nan() );

        stat_type m = m_;
        stat_type n = n_;

        return( 2.0 * std::pow( n, 2.0 ) * ( m + n - 2.0 ) / ( m * std::pow( n - 2.0, 2.0 ) * ( n - 4.0 ) ) );
      }

      /**
         T : t-分布
      **/

      /*
        T::operator[] : 確率変数 x における確率密度を返す
      */
      stat_type T::operator[]( stat_type x ) const
      {
        stat_type beta = std::tgamma( static_cast< stat_type >( ( n_ + 1.0 ) / 2.0 ) ) /
          ( std::tgamma( static_cast< stat_type >( n_ / 2.0 ) ) * std::sqrt( Pi< stat_type >() ) );

        return( beta / ( std::sqrt( n_ ) * std::pow( 1.0 + std::pow( x, 2.0 ) / n_, ( n_ + 1.0 ) / 2.0 ) ) );
      }

      /*
        T::lower_p : 区間 (-∞,a] における確率を返す
      */
      stat_type T::lower_p( stat_type a ) const
      {
        stat_type beta = 2.0 * ( std::tgamma( static_cast< stat_type >( n_ / 2.0 ) ) * std::sqrt( Pi< stat_type >() ) ) /
          std::tgamma( static_cast< stat_type >( ( n_ + 1.0 ) / 2.0 ) );
        stat_type p = IncompleteBeta( n_, 1, n_ / ( n_ + std::pow( a, 2.0 ) ) ) / beta;
        if ( a > 0 ) p = 1.0 - p;

        return( p );
      }

      stat_type T::average() const // 平均値
      {
        return( ( n_ > 1 ) ? 0 : nan() );
      }
      stat_type T::variance() const // 分散
      {
        stat_type n = n_;

        return( ( n_ > 2 ) ? n / ( n - 2 ) : ( ( n_ == 2 ) ? infinity() : nan() ) );
      }

      /**
         多変数確率分布
      **/

      /*
        MultiDiscrete::getProb : 計算済の確率変数 p を元に再起的に確率を求める

        s, e : 和を求める確率変数の範囲への参照( s[i]:開始 e[i]:終了 )
        curIdx : 和を求める範囲の開始位置( curIdx以降の和を求める )
      */
      MultiDiscrete::stat_type MultiDiscrete::getProb( array_type& p, const array_type& s, const array_type& e, size_type curIdx ) const
      {
        stat_type d = 0;

        for ( p[curIdx] = s[curIdx] ; p[curIdx] <= e[curIdx] ; ++( p[curIdx] ) ) {
          if ( curIdx == size() - 1 ) {
            d += (*this)[p];                     // 末尾に対してはその範囲の和を計算
          } else {
            d += getProb( p, s, e, curIdx + 1 ); // そうでなければ再起処理
          }
        }

        return( d );
      }

      /*
        MultiDiscrete::p : 指定した範囲の確率を求める

        s, e : 和を求める確率変数の範囲( s[i]:開始 e[i]:終了 )
      */
      MultiDiscrete::stat_type MultiDiscrete::p( array_type s, array_type e ) const
      {
        size_type n = size();

        if ( s.size() != n || e.size() != n ) return( nan() );

        // 開始と終了の大小関係が逆なら反転する
        for ( size_type i = 0 ; i < n ; ++i )
          if ( s[i] > e[i] ) std::swap( s[i], e[i] );
        array_type p( n );

        return( getProb( p, s, e, 0 ) );
      }

      /**
         Multinomial : 多項分布
      **/

      /*
        Multinomial::operator[] : 確率変数ベクトル r に対する確率密度を返す
      */
      MultiDiscrete::stat_type Multinomial::operator[]( const array_type& r ) const
      {
        if ( p_.size() != r.size() + 1 ) return( nan() ); // 確率変数の数は余事象を含めると事象ごとの確率の数に一致しなければならない

        // 発生回数に負数はないか
        var_type n = 0; // 余事象を除く事象の発生回数
        for ( size_type i = 0 ; i < r.size() ; ++i ) {
          if ( r[i] < 0 ) return( 0 );
          n += r[i];
        }
        if ( n > n_ ) return( nan() ); // 発生回数の総計が総試行回数より大きい

        // 階乗 r! の積を保持する変数(余事象で初期化)
        stat_type fact_r = std::tgamma( static_cast< stat_type >( n_ - n + 1 ) );
        // p^r の積を保持する変数(余事象で初期化)
        stat_type p = ( p_.back() == 0 && ( n_ - n ) == 0 ) ?
          1 : std::pow( p_.back(), n_ - n );

        for ( size_type i = 0 ; i < r.size() ; ++i ) {
          // 確率ゼロの試行の回数がゼロでなければ事象の確率はゼロ
          if ( p_[i] == 0 && r[i] != 0 ) return( 0 );

          // 試行回数がゼロなら p^r = 1 かつ r! = 1 ( p != 0 はチェック済 )
          if ( r[i] != 0 ) {
            p *= std::pow( p_[i], r[i] );
            fact_r *= std::tgamma( static_cast< stat_type >( r[i] + 1 ) );
          }
        }

        return( std::tgamma( static_cast< stat_type >( n_ + 1 ) ) * p / fact_r );
      }

      /*
        Multinomial::average : 平均値
      */
      MultiDiscrete::stat_type Multinomial::average( size_type i ) const
      {
        return( ( i < size() ) ? n_ * p_[i]: nan() );
      }

      /*
        Multinomial::variance : 分散
      */
      MultiDiscrete::stat_type Multinomial::variance( size_type i ) const
      {
        return( ( i < size() ) ? n_ * p_[i] * ( 1.0 - p_[i] ) : nan() );
      }

      /*
        Multinomial::covariance : i, j 番目の確率変数に対する共分散を返す
      */
      MultiDiscrete::stat_type Multinomial::covariance( size_type i, size_type j ) const
      {
        if ( ( i >= size() ) || ( j >= size() ) )
          return( nan() );

        return( -static_cast< stat_type >( n_ ) * p_[i] * p_[j] );
      }

      /*
        Multinomial::size : 変数の数
      */
      MultiDiscrete::size_type Multinomial::size() const
      {
        return( p_.size() );
      }

      /*
        NegativeMultinomial::operator[] : 確率変数(多変量) r に対する確率密度を返す
      */
      MultiDiscrete::stat_type NegativeMultinomial::operator[]( const array_type& r ) const
      {
        if ( p_.size() != r.size() ) return( nan() ); // 各事象の回数は、確率のサイズと一致しなければならない

        // 発生回数に負数はないか
        var_type n = 0; // 余事象を除く事象の発生回数
        for ( size_type i = 0 ; i < r.size() ; ++i ) {
          if ( r[i] < 0 ) return( 0 );
          n += r[i];
        }

        stat_type p = std::pow( p0_, s_ );          // p^r の総積
        stat_type rFact = std::tgamma( static_cast< stat_type >( s_ ) ); // r! の総積
        for ( size_type i = 0 ; i < r.size() ; ++i ) {
          // 確率ゼロの試行の回数がゼロでなければ事象の確率はゼロ
          if ( p_[i] == 0 && r[i] != 0 ) return( 0 );

          // 試行回数がゼロなら p^r = 1 かつ r! = 1 ( p != 0 はチェック済 )
          if ( r[i] != 0 ) {
            p *= std::pow( p_[i], r[i] );
            rFact *= std::tgamma( static_cast< stat_type >( r[i] + 1 ) );
          }
        }

        return( std::tgamma( static_cast< stat_type >( n + s_ ) ) * p / rFact );
      }

      /*
        NegativeMultinomial::average : 平均値
      */
      MultiDiscrete::stat_type NegativeMultinomial::average( size_type i ) const
      {
        return( ( i >= size() ) ? nan() : s_ * p_[i] / p0_ );
      }

      /*
        NegativeMultinomial::variance : 分散
      */
      MultiDiscrete::stat_type NegativeMultinomial::variance( size_type i ) const
      {
        return( ( i >= size() ) ? nan() : s_ * p_[i] / p0_ + s_ * std::pow( p_[i] / p0_, 2.0 ) );
      }

      /*
        NegativeMultinomial::covariance : i, j 番目の確率変数に対する共分散を返す
      */
      MultiDiscrete::stat_type NegativeMultinomial::covariance( size_type i, size_type j ) const
      {
        if ( i >= size() || j >= size() ) return( nan() );

        return( s_ * p_[i] * p_[j] / std::pow( p0_, 2.0 ) );
      }

      /*
        変数の数
      */
      MultiDiscrete::size_type NegativeMultinomial::size() const
      {
        return( p_.size() );
      }

    } // namespace distribution

      /**
         区間推定
      **/

    namespace interval_estimation
    {
      /*
        Mean_TDist : 信頼度 b で t-分布により平均値を区間推定する
      */
      ConfidenceInterval Mean_TDist::operator()
        ( stat_type b, stat_type threshold ) const
      {
        const string FUNC_NAME = "tadah_fussy::statistics::interval_estimation::Mean_TDist::operator()";

        error_lib::In01( b, FUNC_NAME + " : b " + IN01 );

        distribution::T tDist( summary_.count() - 1 );

        // 不偏分散の平方根
        stat_type sigma = std::sqrt( summary_.unbiasedVariance() );
        // t 値
        stat_type ta = tDist.binSearch( b / 2.0, threshold );
        // 区間の幅
        stat_type d = sigma * ta / std::sqrt( static_cast< stat_type >( summary_.count() ) );

        return( ConfidenceInterval( summary_.mean() - d, summary_.mean() + d ) );
      }

      /*
        Mean_TDist::p : t 値の絶対値が t より大きくなる確率 ( p 値 ) を求める
      */
      stat_type Mean_TDist::p( stat_type t ) const
      {
        distribution::T tDist( summary_.count() - 1 );

        return( 0.5 - tDist.p( 0, std::abs( t ) ) );
      }

      /*
        Mean_NormDist::operator() : 信頼度 b で標準正規分布により平均値を区間推定する
      */
      ConfidenceInterval Mean_NormDist::operator()
        ( stat_type b, stat_type threshold ) const
      {
        const string FUNC_NAME = "tadah_fussy::statistics::interval_estimation::Mean_NormDist::operator()";

        error_lib::In01( b, FUNC_NAME + " : b " + IN01 );

        // z 値
        stat_type za = nDist_.binSearch( b / 2.0, threshold );
        // 区間の幅を求める
        stat_type d = sigma_ * za / std::sqrt( static_cast< stat_type >( summary_.count() ) );

        return( ConfidenceInterval( summary_.mean() - d, summary_.mean() + d ) );
      }

      /*
        MeanDiff::operator() : 信頼度 b で平均差を区間推定する
      */
      ConfidenceInterval MeanDiff::operator()
        ( stat_type b, stat_type threshold ) const
      {
        const string FUNC_NAME = "tadah_fussy::statistics::interval_estimation::MeanDiff::operator()";

        error_lib::In01( b, FUNC_NAME + " : b " + IN01 );

        counter_type dof = count1() + count2() - 2;
        distribution::F fDist( 1, dof );

        stat_type fh = fDist.binSearch( b, threshold ); // f 値
        stat_type invSum = 1.0 / count1() + 1.0 / count2(); // データ数の逆数の和
        stat_type ns1 = count1() * variance1(); // データ 1 のデータ数と標本分散の積
        stat_type ns2 = count2() * variance2(); // データ 2 のデータ数と標本分散の積

        // 区間の幅
        stat_type d = std::sqrt( fh * invSum * ( ns1 + ns2 ) / dof );

        return( ConfidenceInterval( std::abs( mean1() - mean2() ) - d, std::abs( mean1() - mean2() ) + d ) );
      }

      namespace
      {
        /*
          IEst_Var : 偏差の平方和 sumSq から信頼度 b で分散を区間推定する

          chiDist は推定に使うカイ二乗分布を表す。
        */
        ConfidenceInterval IEst_Var( stat_type sumSq, distribution::ChiSquare& chiDist,
                                     stat_type b, stat_type threshold )
        {
          stat_type aHalf = ( 1.0 - b ) / 2.0;    // a/2 = ( 1 - b ) / 2

          return( ConfidenceInterval
                  ( sumSq / chiDist.binSearch( 1.0 - aHalf, threshold ),
                    sumSq / chiDist.binSearch( aHalf, threshold ) )
                  );
        }

      }

      /*
        VarN_1::operator() : 信頼度 b で分散を区間推定する
      */
      ConfidenceInterval VarN_1::operator()
        ( stat_type b, stat_type threshold ) const
      {
        const string FUNC_NAME = "tadah_fussy::statistics::interval_estimation::VarN_1::operator()";

        error_lib::In01( b, FUNC_NAME + " : b " + IN01 );

        distribution::ChiSquare chiDist( summary_.count() - 1 );

        return( IEst_Var( summary_.devSq(), chiDist, b, threshold ) );
      }

      /*
        VarN::operator() : 信頼度 b で分散を区間推定する
      */
      ConfidenceInterval VarN::operator()
        ( stat_type b, stat_type threshold ) const
      {
        const string FUNC_NAME = "tadah_fussy::statistics::interval_estimation::VarN::operator()";

        error_lib::In01( b, FUNC_NAME + " : b " + IN01 );

        distribution::ChiSquare chiDist( count_ );

        return( IEst_Var( sumSq_, chiDist, b, threshold ) );
      }

      /*
        VarRatio::operator() : 信頼度 b で分散比を区間推定する
      */
      ConfidenceInterval VarRatio::operator()
        ( stat_type b, stat_type threshold ) const
      {
        const string FUNC_NAME = "tadah_fussy::statistics::interval_estimation::VarRatio::operator()";

        error_lib::In01( b, FUNC_NAME + " : b " + IN01 );

        stat_type a = ( 1.0 - b ) / 2.0; // 危険率

        distribution::F fDist( n1_ - 1, n2_ - 1 );

        return( ConfidenceInterval( fDist.binSearch( a, threshold ), fDist.binSearch( 1.0 - a, threshold ) ) );
      }

      /*
        EstPercent_SS : 百分率の区間推定(小標本)

        総試行回数 N = cnt のうち r 回成功したときの信頼区間を信頼度 b で求める。
        信頼区間は F-分布を使って求められる。
        求めた信頼区間は interval に登録する。
      */
      ConfidenceInterval Percent_SS::operator()
        ( stat_type b, stat_type threshold ) const
      {
        const string FUNC_NAME = "tadah_fussy::statistics::interval_estimation::Percent_SS::operator()";

        error_lib::In01( b, FUNC_NAME + " : b " + IN01 );

        stat_type aHalf = ( 1.0 - b ) / 2.0; // a / 2 = ( 1 - b ) / 2
        ConfidenceInterval ci;

        // 百分率の下限 ta を求める
        counter_type m = 2 * r();                   // 自由度 m = 2r
        counter_type n = 2 * ( count() - r() + 1 ); // 自由度 n = 2( N - r + 1 )
        distribution::F fDist_a( m, n );
        stat_type ta = fDist_a.binSearch( aHalf, threshold );

        ci.lower = m * ta / ( m * ta + n );

        // 百分率の上限 tb を求める
        m = 2 * ( r() + 1 );      // 自由度 m = 2( r + 1 )
        n = 2 * ( count() - r() ); // 自由度 n = 2( N - r )
        distribution::F fDist_b( m, n );
        stat_type tb = fDist_b.binSearch( 1.0 - aHalf, threshold );

        ci.upper = m * tb / ( m * tb + n );

        return( ci );
      }

      /*
        EstPercent_SS : 百分率の区間推定(大標本)

        総試行回数 N = cnt のうち r 回成功したときの信頼区間を信頼度 b で求める。
        信頼区間は正規分布を使って求められる。
        求めた信頼区間は interval に登録する。
      */
      ConfidenceInterval Percent_LS::operator()
        ( stat_type b, stat_type threshold ) const
      {
        const string FUNC_NAME = "tadah_fussy::statistics::interval_estimation::Percent_LS::operator()";

        error_lib::In01( b, FUNC_NAME + " : b " + IN01 );

        distribution::Normal nDist( 0, 1 );
        stat_type ta = nDist.binSearch( b / 2.0, threshold );

        stat_type mean = static_cast< stat_type >( r() ) / count();
        stat_type range = ta * std::sqrt( r() * ( count() - r() ) / std::pow( count(), 3 ) );

        return( ConfidenceInterval( mean - range, mean + range ) );
      }

    } // namespace interval_estimation

    namespace test
    {
      /**
         検定
      **/

      /* 一標本の平均値検定 */

      namespace
      {
        /*
          OneSideLess 危険率 a で平均値 m0 を片側検定 ( μ < m0 ) する。

          mp : 母集団平均の信頼区間上限を代入する変数へのポインタ(NULLならば無視)
        */
        template< class IntervalEstimation_Mean >
        bool OneSideLess( const IntervalEstimation_Mean& iEst, stat_type m0, stat_type a,
                          stat_type* mp,
                          stat_type threshold )
        {
          interval_estimation::ConfidenceInterval ci =
            iEst( std::abs( 1.0 - 2.0 * a ), threshold );

          stat_type m = ( a > 0.5 ) ? ci.lower : ci.upper;
          if ( mp != 0 ) *mp = m;

          return( m < m0 );
        }

        /*
          OneSideMore : 危険率 a で平均値 m0 を片側検定 ( μ > m0 ) する。

          mp : 母集団平均の信頼区間下限を代入する変数へのポインタ(NULLならば無視)
        */
        template< class IntervalEstimation_Mean >
        bool OneSideMore( const IntervalEstimation_Mean& iEst, stat_type m0, stat_type a,
                          stat_type* mp,
                          stat_type threshold )
        {
          interval_estimation::ConfidenceInterval ci =
            iEst( std::abs( 1.0 - 2.0 * a ), threshold );

          stat_type m = ( a > 0.5 ) ? ci.upper : ci.lower;
          if ( mp != 0 ) *mp = m;

          return( m > m0 );
        }

        /*
          TwoSide : 危険率 a で平均値 m0 を両側検定 ( μ ≠ m0 ) する。

          cip : 母集団平均の信頼区間を代入する変数へのポインタ(NULLならば無視)
        */
        template< class IntervalEstimation_Mean >
        bool TwoSide( const IntervalEstimation_Mean& iEst, stat_type m0, stat_type a,
                      interval_estimation::ConfidenceInterval* cip, stat_type threshold )
        {
          interval_estimation::ConfidenceInterval ci =
            iEst( 1.0 - a, threshold );

          if ( cip != 0 ) cip->lower = ci.lower;
          if ( cip != 0 ) cip->upper = ci.upper;

          return( m0 < ci.lower || m0 > ci.upper );
        }

      } // namespace

      /* t-分布を使った一標本の平均値検定 */

      /*
        OneSampleMean_TDist::oneSideLess : 危険率 a で平均値 m0 を片側検定 ( μ < m0 ) する
      */
      bool OneSampleMean_TDist::oneSideLess( stat_type m0, stat_type a,
                                             stat_type* mp,
                                             stat_type threshold )
      {
        const string FUNC_NAME = "tadah_fussy::statistics::test::OneSampleMean_TDist::oneSideLess";

        error_lib::In01( a, FUNC_NAME + " : a " + IN01 );

        return( OneSideLess( iEst_, m0, a, mp, threshold ) );
      }

      /*
        OneSampleMean_TDist::oneSideMore : 危険率 a で平均値 m0 を片側検定 ( μ > m0 ) する
      */
      bool OneSampleMean_TDist::oneSideMore( stat_type m0, stat_type a,
                                             stat_type* mp,
                                             stat_type threshold )
      {
        const string FUNC_NAME = "tadah_fussy::statistics::test::OneSampleMean_TDist::oneSideMore";

        error_lib::In01( a, FUNC_NAME + " : a " + IN01 );

        return( OneSideMore( iEst_, m0, a, mp, threshold ) );
      }

      /*
        OneSampleMean_TDist::twoSide : 危険率 a で平均値 m0 を両側検定 ( μ ≠ m0 ) する
      */
      bool OneSampleMean_TDist::twoSide( stat_type m0, stat_type a,
                                         interval_estimation::ConfidenceInterval* cip,
                                         stat_type threshold )
      {
        const string FUNC_NAME = "tadah_fussy::statistics::test::OneSampleMean_TDist::twoSide";

        error_lib::In01( a, FUNC_NAME + " : a " + IN01 );

        return( TwoSide( iEst_, m0, a, cip, threshold ) );
      }

      /* 標準正規分布を使った一標本の平均値検定 */

      /*
        OneSampleMean_NormDist::oneSideLess : 危険率 a で平均値 m0 を片側検定 ( μ < m0 ) する
      */
      bool OneSampleMean_NormDist::oneSideLess( stat_type m0, stat_type a,
                                                stat_type* mp,
                                                stat_type threshold )
      {
        const string FUNC_NAME = "tadah_fussy::statistics::test::OneSampleMean_NormDist::oneSideLess";

        error_lib::In01( a, FUNC_NAME + " : a " + IN01 );

        return( OneSideLess( iEst_, m0, a, mp, threshold ) );
      }

      /*
        OneSampleMean_NormDist::oneSideMore : 危険率 a で平均値 m0 を片側検定 ( μ > m0 ) する
      */
      bool OneSampleMean_NormDist::oneSideMore( stat_type m0, stat_type a,
                                                stat_type* mp,
                                                stat_type threshold )
      {
        const string FUNC_NAME = "tadah_fussy::statistics::test::OneSampleMean_NormDist::oneSideMore";

        error_lib::In01( a, FUNC_NAME + " : a " + IN01 );

        return( OneSideMore( iEst_, m0, a, mp, threshold ) );
      }

      /*
        OneSampleMean_NormDist::twoSide : 危険率 a で平均値 m0 を両側検定 ( μ ≠ m0 ) する
      */
      bool OneSampleMean_NormDist::twoSide( stat_type m0, stat_type a,
                                            interval_estimation::ConfidenceInterval* cip,
                                            stat_type threshold )
      {
        const string FUNC_NAME = "tadah_fussy::statistics::test::OneSampleMean_NormDist::twoSide";

        error_lib::In01( a, FUNC_NAME + " : a " + IN01 );

        return( TwoSide( iEst_, m0, a, cip, threshold ) );
      }

      /* 二標本の検定 */

      /*
        TwoSample::testVarRatio : 二標本の不偏分散比を危険率 a で検定する

        cip : 不偏分散比の信頼区間を代入する変数へのポインタ(NULLならば無視)
      */
      bool TwoSample::testVarRatio( stat_type a, interval_estimation::ConfidenceInterval* cip,
                                    stat_type threshold )
      {
        stat_type f = summary1_.unbiasedVariance() / summary2_.unbiasedVariance(); // 不偏分散比
        interval_estimation::ConfidenceInterval ci = ( interval_estimation::VarRatio( summary1_.count(), summary2_.count() ) )( 1.0 - a, threshold );
        if ( cip != nullptr ) *cip = ci;

        return( f < ci.lower || ci.upper < f );
      }

      /*
        TwoSample::testMean : 二標本の平均差を危険率 a で検定する

        cip : 平均差の信頼区間を代入する変数へのポインタ(NULLならば無視)
      */
      bool TwoSample::testMeanDiff( stat_type a, interval_estimation::ConfidenceInterval* cip,
                                    stat_type threshold )
      {
        interval_estimation::ConfidenceInterval ci = ( interval_estimation::MeanDiff( summary1_, summary2_ ) )( 1.0 - a, threshold );
        if ( cip != nullptr ) *cip = ci;

        return( ci.lower > 0 || ci.upper < 0 );
      }

      /* 二項検定 */

      /*
        Binomial::test : 百分率の区間推定 iEst を使って p を危険率 a で検定する

        cip : 百分率の信頼区間を代入する変数へのポインタ(NULLならば無視)
      */
      bool Binomial::test( interval_estimation::Percent& iEst,
                           stat_type p, stat_type a, interval_estimation::ConfidenceInterval* cip,
                           stat_type threshold )
      {
        interval_estimation::ConfidenceInterval ci = iEst( 1.0 - a, threshold );
        if ( cip != nullptr ) *cip = ci;

        return( p < ci.lower || ci.upper < p );
      }

      /*
        Binomial::testSS : 危険率 a で百分率 p を二項検定(小標本)する

        cip : 百分率の信頼区間を代入する変数へのポインタ(NULLならば無視)
      */
      bool Binomial::testSS( stat_type p, stat_type a, interval_estimation::ConfidenceInterval* cip,
                             stat_type threshold )
      {
        const string FUNC_NAME = "tadah_fussy::statistics::test::Binomial::testSS";

        error_lib::In01( p, FUNC_NAME + " : p " + IN01 );
        error_lib::In01( a, FUNC_NAME + " : a " + IN01 );

        interval_estimation::Percent_SS iEst( count_, r_ );

        return( test( iEst, p, a, cip, threshold ) );
      }

      /*
        Binomial::testSS : 危険率 a で百分率 p を二項検定(大標本)する

        cip : 百分率の信頼区間を代入する変数へのポインタ(NULLならば無視)
      */
      bool Binomial::testLS( stat_type p, stat_type a, interval_estimation::ConfidenceInterval* cip,
                             stat_type threshold )
      {
        const string FUNC_NAME = "tadah_fussy::statistics::test::Binomial::testLS";

        error_lib::In01( p, FUNC_NAME + " : p " + IN01 );
        error_lib::In01( a, FUNC_NAME + " : a " + IN01 );

        interval_estimation::Percent_LS iEst( count_, r_ );

        return( test( iEst, p, a, cip, threshold ) );
      }

      /* カイ二乗検定(適合度検定) */

      /*
        Test_ChiSquare コンストラクタ ( 離散確率分布用 )

        dist は比較対象の離散確率分布、hist は各事象の頻度データをそれぞれ表す。
      */
      ChiSquare::ChiSquare( const distribution::Discrete& dist, const Histogram< distribution::Discrete::var_type, counter_type >& hist )
      {
        typedef Histogram< distribution::Discrete::var_type, counter_type >::const_iterator const_iterator;

        const string FUNC_NAME = "tadah_fussy::statistics::test::ChiSquare constructor";

        error_lib::GT< counter_type >
          ( hist.size(), 1, FUNC_NAME + " : size of hist " + GT + " 1" );

        counter_type total = 0; // 総標本数

        for ( const_iterator i = hist.begin() ; i != hist.end() ; ++i ) {
          obj_.push_back( i->second );
          total += i->second;
        }

        for ( const_iterator i = hist.begin() ; i != hist.end() ; ++i )
          exp_.push_back( dist[i->first] * total );

        //calcChiSquare();
        chiSquare_ = PearsonChiSq( obj_.begin(), obj_.end(), exp_.begin(), exp_.end() );
        calcP();
      }

      /*
        ChiSquare コンストラクタ ( 連続確率分布用 )

        dist は比較対象の連続確率分布、hist は各事象(範囲)の頻度データをそれぞれ表す。
      */
      ChiSquare::ChiSquare( const distribution::Continuous& dist,
                            const Histogram< histogram::RangeClass< distribution::Continuous::stat_type >, counter_type >& hist )
      {
        typedef Histogram< histogram::RangeClass< distribution::Continuous::stat_type >, counter_type >::const_iterator const_iterator;

        const string FUNC_NAME = "tadah_fussy::statistics::test::ChiSquare constructor";

        error_lib::GT< counter_type >
          ( hist.size(), 1, FUNC_NAME + " : size of hist " + GT + " 1" );

        counter_type total = 0; // 総標本数

        for ( const_iterator i = hist.begin() ; i != hist.end() ; ++i ) {
          obj_.push_back( i->second );
          total += i->second;
        }

        for ( const_iterator i = hist.begin() ; i != hist.end() ; ++i )
          exp_.push_back( dist.p( ( i->first ).lower, ( i->first ).upper ) * total );

        //calcChiSquare();
        chiSquare_ = PearsonChiSq( obj_.begin(), obj_.end(), exp_.begin(), exp_.end() );
        calcP();
      }

      /*
        ChiSquare::calcP : p 値を求める
      */
      void ChiSquare::calcP()
      {
        counter_type n = obj_.size(); // 級数

        p_ = 1.0 - distribution::ChiSquare( n - 1 ).p( 0, chiSquare_ );
      }

      /*
        ChiSquare::operator() : カイ二乗検定により危険率 a で適合度検定する
      */
      bool ChiSquare::operator()( stat_type a, stat_type threshold )
      {
        const string FUNC_NAME = "tadah_fussy::statistics::test::ChiSquare::operator()";

        counter_type n = obj_.size(); // 級数

        error_lib::In01( a, FUNC_NAME + " : a " + IN01 );

        distribution::ChiSquare chiDist( n - 1 );
        stat_type rej = chiDist.binSearch( 1.0 - a, threshold ); // 棄却域の定義域

        return( chiSquare_ > rej );
      }

      /* カイ二乗検定(独立性検定) */

      /*
        ChiSquare_Independence コンストラクタ
      */
      ChiSquare_Independence::ChiSquare_Independence( const Matrix< counter_type >& data )
      {
        const string FUNC_NAME = "tadah_fussy::statistics::test::ChiSquare_Independence constructor";

        error_lib::GT< counter_type >
          ( data.rows(), 1, FUNC_NAME + " : data.rows() " + GT + " 1" );
        error_lib::GT< counter_type >
          ( data.cols(), 1, FUNC_NAME + " : data.cols() " + GT + " 1" );

        exp_.assign( data.rows(), data.cols() );

        calcExp( data );
        p_ = 1.0 - distribution::ChiSquare( ( exp_.rows() - 1 ) * ( exp_.cols() - 1 ) ).p( 0, chiSquare_ );
      }

      /*
        ChiSquare_Independence::calcExp : 期待値の計算
      */
      void ChiSquare_Independence::calcExp( const Matrix< counter_type >& data )
      {
        typedef Matrix< counter_type >::size_type size_type;
        typedef Matrix< counter_type >::const_iterator const_iterator;

        // 各合計値の計算
        vector< stat_type > rowCnt( data.rows(), 0 ); // 一行の合計
        vector< stat_type > colCnt( data.cols(), 0 ); // 一列の合計
        stat_type total = 0; // 度数の総計

        // 行・列ごとの合計を求める
        for ( size_type r = 0 ; r < data.rows() ; ++r ) {
          const_iterator cit = data.row( r );
          rowCnt[r] = Sum< stat_type >( cit, cit.end() );
          total += rowCnt[r];
        }
        for ( size_type c = 0 ; c < data.cols() ; ++c ) {
          const_iterator cit = data.column( c );
          colCnt[c] = Sum< stat_type >( cit, cit.end() );
        }

        // 期待値 ( 行の合計 x 列の合計 / 総計 ) の計算
        for ( size_type r = 0 ; r < data.rows() ; ++r ) {
          for ( size_type c = 0 ; c < data.cols() ; ++c ) {
            exp_[r][c] = ( rowCnt[r] * colCnt[c] ) / total;
          }
        }

        // カイ二乗値を計算する
        //chiSquare_ = std::inner_product( data.begin(), data.end(), exp_.begin(), stat_type(), SumOp< stat_type >(), CalcChiSquare() );
        chiSquare_ = PearsonChiSq( data.begin(), data.end(), exp_.begin(), exp_.end() );
      }

      /*
        ChiSquare_Independence::operator() : カイ二乗検定により危険率 a で独立性検定を行う
      */
      bool ChiSquare_Independence::operator()( stat_type a, stat_type threshold )
      {
        const string FUNC_NAME = "tadah_fussy::statistics::test::ChiSquare_Independence::operator()";

        error_lib::In01( a, FUNC_NAME + " : a " + IN01 );

        distribution::ChiSquare chiDist( ( exp_.rows() - 1 ) * ( exp_.cols() - 1 ) );
        stat_type rej = chiDist.binSearch( 1.0 - a, threshold ); // 棄却域の定義域

        return ( chiSquare_ > rej );
      }

      /* フィッシャーの直接確率検定 */

      /*
        FishersExact コンストラクタ : フィッシャーの直接確率検定(Fisher's Exact Test)

        以下のような 2x2 対応表の頻度から p 値を計算する

        B0  B1
        A0 d00 d01
        A1 d10 d11
      */
      FishersExact::FishersExact( counter_type d00, counter_type d10, counter_type d01, counter_type d11 )
        : rSum0_( d00 + d01 ), rSum1_( d10 + d11 ), cSum0_( d00 + d10 ), cSum1_( d01 + d11 ),
          total_( d00 + d10 + d01 + d11 )
      {
        // d00 以上となる確率の和を求める
        stat_type pSum = 1.0 / mulOfFact( d00, d10, d01, d11 );
        while ( d10 > 0 && d01 > 0 ) {
          ++d00;
          --d10;
          --d01;
          ++d11;
          pSum += 1.0 / mulOfFact( d00, d10, d01, d11 );
        }

        p_ = mulOfFact( rSum0_, rSum1_, cSum0_, cSum1_ ) * pSum / std::tgamma( static_cast< stat_type >( total_ + 1 ) );
      }

      /* 一元配置分散分析( One-way ANOVA) */

      /*
        OneWayANOVA コンストラクタ : 一元配置分散分析 ( One-way ANOVA )

        data は vector を要素とする vector (要素数の異なる二次元配列)
      */
      OneWayANOVA::OneWayANOVA( const vector< vector< stat_type > >& data )
        : sc_( distribution::Base::nan() ), se_( distribution::Base::nan() ), f_( distribution::Base::nan() )
      {
        const string FUNC_NAME = "tadah_fussy::statistics::test::OneWayANOVA constructor";

        // 集団は 1 個より多くなければならない
        classCnt_= error_lib::GT< counter_type >
          ( data.size(), 1, FUNC_NAME + " : size of data " + GT + " 1" );

        // 各集団と全体の集計値を求める
        for ( auto cit = data.begin() ; cit != data.end() ; ++cit )
          classSum_.push_back( Moment1( cit->begin(), cit->end() ) );
        totalSum_ = SumAll< Moment1 >( classSum_ );

        // 各集団と全体の変動を求める
        vector< stat_type > scVec( classCnt_ ); // 各集団間の変動(平方和)の項
        vector< stat_type > seVec( classCnt_ ); // 集団内の変動(平方和)の項
        for ( counter_type i = 0 ; i < classCnt_ ; ++i ) {
          error_lib::PNum( classSum_[i].count(), FUNC_NAME + " : all class must have at least one data" ); // データのない集団があってはいけない
          Deviation< stat_type > dev( classSum_[i].mean() );
          scVec[i] = classSum_[i].count() * std::pow( classSum_[i].mean() - totalSum_.mean(), 2.0 );
          seVec[i] = SumAll< stat_type >( data[i], OpBind( dev ) );
        }
        sc_ = SumAll< stat_type >( scVec );
        se_ = SumAll< stat_type >( seVec );

        // 全データの総数は、集団数より多くなければならない
        // 各集団に対してデータが一つしかない場合はこれが成り立たなくなる
        error_lib::GT( totalSum_.count(), classCnt_,
                       FUNC_NAME + " : size of data " + GT + " size of class" );

        // F 値の計算
        f_ = ( sc_ == 0 && se_ == 0 ) ? distribution::Base::nan() : ( sc_ * dofW() ) / ( se_ * dofB() );
      }

      /*
        OneWayANOVA::cr_f : 危険率 a での f 値の棄却域を返す
      */
      stat_type OneWayANOVA::cr_f( stat_type a, stat_type threshold ) const
      {
        const string FUNC_NAME = "tadah_fussy::statistics::test::OneWayANOVA::cr_f";

        error_lib::In01( a, FUNC_NAME + " : a " + IN01 );

        distribution::F fDist( dofB(), dofW() );
        return( fDist.binSearch( 1.0 - a, threshold ) );
      }

      /* 二元配置分散分析 (Two-way ANOVA) */

      /*
        TwoWayANOVA コンストラクタ : 行列 data を指定して実行
      */
      TwoWayANOVA::TwoWayANOVA( const Matrix< stat_type >& data )
      {
        const string FUNC_NAME = "tadah_fussy::statistics::test::TwoWayANOVA constructor";

        rows_ = error_lib::GT< counter_type >
          ( data.rows(), 1, FUNC_NAME + " : data.rows() " + GT + " 1" );
        cols_ = error_lib::GT< counter_type >
          ( data.cols(), 1, FUNC_NAME + " : data.cols() " + GT + " 1" );

        mr_.resize( rows() );
        mc_.resize( cols() );

        calcMean( data );
        calcVar( data );
      }

      /*
        TwoWayANOVA::calcMean : 平均値の算出
      */
      void TwoWayANOVA::calcMean( const Matrix< stat_type >& data )
      {
        for ( size_type r = 0 ; r < data.rows() ; ++r ) {
          Matrix< stat_type >::const_iterator row = data.row( r );
          mr_[r] = Mean< stat_type >( row, row.end() );
        }
        for ( size_type c = 0 ; c < data.cols() ; ++c ) {
          Matrix< stat_type >::const_iterator col = data.column( c );
          mc_[c] = Mean< stat_type >( col, col.end() );
        }
        m_ = Moment2( data.begin(), data.end() );
      }

      /*
        TwoWayANOVA::calcVar : 変動の算出
      */
      void TwoWayANOVA::calcVar( const Matrix< stat_type >& data )
      {
        Deviation< stat_type > dev( m_.mean() );
        sr_ = SumAll< stat_type >( mr_, OpBind( dev ) ) * cols(); // 行間変動
        sc_ = SumAll< stat_type >( mc_, OpBind( dev ) ) * rows(); // 列間変動
        st_ = m_.devSq();       // 全体の変動
        se_ = st_ - sr_ - sc_;  // 誤差変動
        if ( se_ < 0 ) se_ = 0; // 誤差により負数になることを防ぐため
      }

      /*
        TwoWayANOVA::cr_f : 危険率 a での f 値の棄却域を返す

        dof は F-分布の一番目の自由度で、行間の f 値なら dofR()、列間の f 値なら dofC() になる。
        threshold : BinSearch で棄却域を求める時のしきい値
      */
      stat_type TwoWayANOVA::cr_f( counter_type dof, stat_type a, stat_type threshold ) const
      {
        const string FUNC_NAME = "tadah_fussy::statistics::test::TwoWayANOVA::cr_f";

        error_lib::In01( a, FUNC_NAME + " : a " + IN01 );

        distribution::F fDist( dof, dofE() );

        return( fDist.binSearch( 1.0 - a, threshold ) );
      }

      /*
        TwoWayANOVA::p : F 値以上になる確率 ( p 値 ) を返す

        dof は F-分布の一番目の自由度で、行間の f 値なら dofR()、列間の f 値なら dofC() になる。
        f は F 値で、行間の f 値なら fR()、列間の f 値なら fC() になる。
      */
      stat_type TwoWayANOVA::p( counter_type dof, stat_type f ) const
      {
        if ( std::isnan( f ) )
          return( distribution::Base::nan() );

        distribution::F fDist( dof, dofE() );

        return( 1.0 - fDist.lower_p( f ) );
      }

      /* 反復のある二元配置分散分析 ( Two-way Repeated-Measures ANOVA ) */

      /*
        RepMeasANOVA::calcMean : 平均値の算出
      */
      void RepMeasANOVA::calcMean()
      {
        m_ = 0;
        for ( counter_type r = 0 ; r < rows() ; ++r ) {
          auto row = data_.row( r );
          mr_[r] = Mean< stat_type >( row, row.end() );
          m_ = accumulate( row, row.end(), m_ );
        }
        if ( rows() * data_.cols() > 0 )
          m_ /= rows() * data_.cols();

        for ( counter_type c = 0 ; c < cols() ; ++c ) {
          mc_[c] = 0;
          for ( counter_type n = 0 ; n < classCnt() ; ++n ) {
            auto col = data_.column( n * cols() + c );
            mc_[c] = accumulate( col, col.end(), mc_[c] );
          }
          mc_[c] /= rows() * classCnt();
        }

        for ( counter_type r = 0 ; r < rows() ; ++r ) {
          for ( counter_type c = 0 ; c < cols() ; ++c ) {
            auto i = data_.row( r, c, cols() );
            mrc_[r][c] = Mean< stat_type >( i, i.end() );
          }
        }
      }

      /*
        RepMeasANOVA::calcVar : 変動の算出
      */
      void RepMeasANOVA::calcVar()
      {
        // 行間・列間変動の算出
        Deviation< stat_type > dev( m_ );
        sr_ = SumAll< stat_type >( mr_, OpBind( dev ) ) * cols() * classCnt(); // 行間変動
        sc_ = SumAll< stat_type >( mc_, OpBind( dev ) ) * rows() * classCnt(); // 列間変動

        // 全変動・誤差変動の算出
        st_ = se_ = 0;
        for ( counter_type r = 0 ; r < rows() ; ++r ) {
          for ( counter_type c = 0 ; c < cols() ; ++c ) {
            Deviation< stat_type > dev_rc( mrc_[r][c] );
            auto i = data_.row( r, c, cols() );
            st_ = accumulate( i, i.end(), st_, OpBind( dev ) );
            se_ = accumulate( i, i.end(), se_, OpBind( dev_rc ) );
          }
        }
        src_ = st_ - sr_ - sc_ - se_; // 交互作用
        if ( src_ < 0 ) src_ = 0;     // 誤差により負数になることを防ぐため
      }

      /*
        RepMeasANOVA::cr_f : 危険率 a での f 値の棄却域を返す

        dof は F-分布の一番目の自由度で、行間の f 値なら dofR()、列間の f 値なら dofC() になる。
      */
      stat_type RepMeasANOVA::cr_f( counter_type dof, stat_type a, stat_type threshold ) const
      {
        const string FUNC_NAME = "tadah_fussy::statistics::test::RepMeasANOVA::cr_f";

        error_lib::In01( a, FUNC_NAME + " : a " + IN01 );

        distribution::F fDist( dof, dofE() );

        return( fDist.binSearch( 1.0 - a, threshold ) );
      }

      /*
        RepMeasANOVA::p : F 値以上になる確率 ( p 値 ) を返す

        dof は F-分布の一番目の自由度で、行間の f 値なら dofR()、列間の f 値なら dofC()、交互作用なら dofRC() になる。
        f は F 値で、行間の f 値なら fR()、列間の f 値なら fC()、交互作用なら fRC() になる。
      */
      stat_type RepMeasANOVA::p( counter_type dof, stat_type f ) const
      {
        if ( std::isnan( f ) )
          return( distribution::Base::nan() );

        distribution::F fDist( dof, dofE() );

        return( 1.0 - fDist.lower_p( f ) );
      }

      /* 共分散分析 (ANCOVA) */

      namespace
      {
        /*
          CalcSb : ANCOVA の集団間変動を求めるための関数オブジェクト
        */
        class CalcSb
        {
          stat_type m_; // 全データの平均

        public:

          // コンストラクタ
          CalcSb( stat_type m )
            : m_( m ) {}

          // 集団間変動 ssB の各要素を計算する
          //
          //   ssB = Σi( n[i] * ( my[i] - m )^2 )
          //    n[i]  : 各集団のデータ数
          //    my[i] : 各集団の従属変数の標本平均
          stat_type operator()( const MomentPair2& mtp2 )
          { return( std::pow( mtp2.mean().second - m_, 2 ) * mtp2.count() ); }
        };

        /*
          CalcSi : ANCOVA の交互作用の変動を求めるための関数オブジェクト
        */
        struct CalcSi
        {
          // 交互作用の変動 ssI の各要素を計算する
          //
          //   ssI = Σi( n[i]・sxy[i]^2 / vx[i] ) - ssC
          //       = Σi( Sxy[i]^2 / Sx[i] ) - ssC
          //    n[i]   : 各集団のデータ数
          //    sxy[i] : 各集団の標本共分散 ( = Sxy[i] / n[i] )
          //    vx[i]  : 各集団の標本分散 ( = Sx[i] / n[i] )
          stat_type operator()( const MomentPair2& mtp2 )
          { return( std::pow( mtp2.proDev(), 2 ) / mtp2.devSq1() ); }
        };

      } // namespace

      /*
        Test_ANCOVA コンストラクタ
      */
      ANCOVA::ANCOVA( const vector< vector< value_type > >& data )
      {
        const string FUNC_NAME = "tadah_fussy::statistics::test::ANCOVA constructor";

        // 集計値の計算
        classCnt_ = data.size();
        error_lib::GT< counter_type >( classCnt_, 1, FUNC_NAME + " : size of class " + GT + " 1" );
        for ( counter_type i = 0 ; i < classCnt_ ; ++i )
          classSum_.push_back( MomentPair2( data[i].begin(), data[i].end() ) );
        totalSum_ = SumAll< MomentPair2 >( classSum_ );
        error_lib::GT< counter_type >( totalSum_.count(), 2 * classCnt_,
                                       FUNC_NAME + " : size of data " + GT + " 2 * ( size of class )." );

        // 変動の計算
        sb_ = SumAll< stat_type >( classSum_, OpBind( CalcSb( totalSum_.mean().second ) ) );
        sc_ = std::pow( SumAll< stat_type >( classSum_, OpBind( std::mem_fun_ref< stat_type, MomentPair2 >( &MomentPair2::proDev ) ) ), 2 ) /
          SumAll< stat_type >( classSum_, OpBind( std::mem_fun_ref< stat_type, MomentPair2 >( &MomentPair2::devSq1 ) ) );
        si_ = SumAll< stat_type >( classSum_, OpBind( CalcSi() ) ) - sc_;
      }

      /*
        ANCOVA::cr_f : 危険率 a での F 値の棄却域を返す

        dof は F-分布の一番目の自由度で、集団内の F 値なら dofC()、交互作用の F 値なら dofI()、補正した集団間の F 値なら dofB() になる。
      */
      stat_type ANCOVA::cr_f( counter_type dof, stat_type a, stat_type threshold ) const
      {
        const string FUNC_NAME = "tadah_fussy::statistics::test::ANCOVA::cr_f";

        error_lib::In01( a, FUNC_NAME + " : a " + IN01 );

        distribution::F fDist( dof, dofE() );

        return( fDist.binSearch( 1.0 - a, threshold ) );
      }

      /*
        ANCOVA::p : F 値以上になる確率 ( p 値 ) を返す

        dof は F-分布の一番目の自由度で、集団内の F 値なら dofC()、交互作用の F 値なら dofI()、補正した集団間の F 値なら dofB() になる。
        f は F 値で、集団内の F 値なら fC()、交互作用の F 値なら fI()、補正した集団間の変動なら fB() になる。
      */
      stat_type ANCOVA::p( counter_type dof, stat_type f ) const
      {
        if ( std::isnan( f ) )
          return( distribution::Base::nan() );

        distribution::F fDist( dof, dofE() );

        return( 1.0 - fDist.lower_p( f ) );
      }

      /* マン・ホイットニーの U-検定 */

      /*
        MannWhitney::p : U 値に対する p 値を返す

        m_ と n_ のいずれかが 10 を超えていた場合は、U-分布を正規分布で近似する。
      */
      stat_type MannWhitney::p() const
      {
        if ( m_ <= 10 && n_ <= 10 ) {
          distribution::U uDist( m_, n_ );
          return( uDist.p( 0, u_ ) + ( u_ - floor( u_ ) ) * uDist[u_ + 1] );
        } else {
          distribution::Normal nd
            ( static_cast< stat_type >( m_ * n_ ) / 2.0,
              std::sqrt( m_ * n_ * ( m_ + n_ + 1 ) / 12.0 )
              );
          return( nd.lower_p( u_ ) );
        }
      }

      /* ウィルコクソンの符号順位検定 ( Wilcoxon Signed-rank Test ) */

      /*
        Wilcoxon::p : 符号付き順位和に対する p 値を返す

        n_ が 10 を超えていた場合は、符号付き順位和に対する確率密度関数を正規分布で近似する。
      */
      stat_type Wilcoxon::p() const
      {
        stat_type sRank = std::abs( sRank_ );

        if ( n_ == 0 ) {
          return( 1.0 ); // データがないときは常に 1.0
        } else if ( n_ <= 10 ) {
          distribution::SignedRank sRankDist( n_ );
          return( 1.0 - 2 * ( ( ( sRank < 1 ) ? 0 : sRankDist.p( 0, sRank - 1 ) ) + ( sRank - floor( sRank ) ) * sRankDist[sRank] ) );
        } else {
          distribution::Normal nd( 0, std::sqrt( n_ * ( n_ + 1 ) * ( 2 * n_ + 1 ) / 6.0 ) );
          return( 2 * nd.lower_p( -sRank ) );
        }
      }

      /* クラスカル・ワリス検定 ( Kruskal Wallis Test ) */

      /*
        KruskalWallis::calcH : 順位データ rank から H 値を求める
      */
      void KruskalWallis::calcH( const vector< vector< stat_type > >& rank )
      {
        // Σ( Ti^2 / ni ) の計算
        h_ = SumAll< stat_type >( rank, OpBind( CalcAveSq() ) );

        // H の計算
        h_ -= totalCnt_ * std::pow( totalCnt_ + 1, 2.0 ) / 4.0;
        h_ /= static_cast< stat_type >( totalCnt_ * ( totalCnt_ + 1 ) ) / 12.0;
        if ( r_ == 0 )
          h_ = distribution::Base::infinity();
        else
          h_ /= r_;
      }

      /* フリードマン検定 ( Friedman Test ) */

      /*
        Friedman::calcX2 : カイ二乗値の計算

        集団数・要素数のいずれか一方がゼロなら x2_ は計算しない(コンストラクタでの初期値のまま)
      */
      void Friedman::calcX2()
      {
        if ( classCnt_ == 0 || dataCnt_ == 0 ) return;

        // Σ( Tg^2 / var ) の計算
        x2_ = 0;
        for ( counter_type i = 0 ; i < dataCnt_ ; ++i ) {
          auto it = rank_.column( i );
          x2_ += std::pow( Sum< stat_type >( it, it.end() ), 2.0 );
        }

        // x2 の計算
        x2_ = x2_ * 12.0 / ( classCnt_ * dataCnt_ * ( dataCnt_ + 1 ) ) -
          classCnt_ * ( dataCnt_ + 1 ) * 3.0;
        // R 値で補正
        if ( r_ == 0 )
          x2_ = distribution::Base::infinity();
        else
          x2_ /= r_;
      }

    } // namespace test

    namespace regression
    {
      /* 単回帰分析 */

      /*
        Singular::iEst_RegCoef :
        確率密度関数 dist を使い回帰直線の係数を信頼度 b で区間推定して interval_a(傾き) と interval_b(切片) に返す

        var : 誤差項の分散
        threshold : binSearch で t を求める時のしきい値
      */
      void Singular::iEst_RegCoef( const distribution::Continuous& dist, stat_type var, stat_type b,
                                   interval_estimation::ConfidenceInterval* interval_a, interval_estimation::ConfidenceInterval* interval_b,
                                   stat_type threshold ) const
      {
        const string FUNC_NAME = "tadah_fussy::statistics::regression::Singular::iEst_RegCoef";

        error_lib::In01( b, FUNC_NAME + " : b " + IN01 );

        stat_type t = dist.binSearch( b / 2.0, threshold ); // 確率分布の片側信頼区間

        stat_type diff_a = t * std::sqrt( var / ( summary_.count() * summary_.var1() ) );
        stat_type diff_b = t * std::sqrt( var * ( 1.0 / summary_.count() + std::pow( summary_.mean1(), 2 ) / ( summary_.count() * summary_.var1() ) ) );

        if ( interval_a != nullptr ) {
          interval_a->lower = est_a_ - diff_a;
          interval_a->upper = est_a_ + diff_a;
        }
        if ( interval_b != nullptr ) {
          interval_b->lower = est_b_ - diff_b;
          interval_b->upper = est_b_ + diff_b;
        }
      }

      /*
        Singular::iEst_DepVar : の区間推定
        確率密度関数 dist を使い従属変数(またはその期待値)を信頼度 b で区間推定して interval_y に返す

        const distribution::Continuous& dist : 確率密度関数(左右対称を前提)
        stat_type var : 誤差項の分散
        stat_type b : 信頼度
        stat_type x0 : 独立変数
        stat_type expFlag : 期待値を求める場合は 1, そうでなければ 0
        stat_type threshold : BinSearchでtを求める時のしきい値

        戻り値 : 信頼区間
      */
      interval_estimation::ConfidenceInterval Singular::iEst_DepVar( const distribution::Continuous& dist, stat_type var, stat_type b, stat_type x,
                                                                     stat_type expFlag, stat_type threshold ) const
      {
        const string FUNC_NAME = "tadah_fussy::statistics::regression::Singular::iEst_DepVar";

        error_lib::In01( b, FUNC_NAME + " : b " + IN01 );

        stat_type t = dist.binSearch( b / 2.0, threshold ); // 確率分布の片側信頼区間

        stat_type diff_y = t * std::sqrt
          ( var * ( ( expFlag * summary_.count() + 1 ) * summary_.var1() + std::pow( x - summary_.mean1(), 2 ) ) / ( summary_.count() * summary_.var1() ) );

        return( interval_estimation::ConfidenceInterval( expY( x ) - diff_y, expY( x ) + diff_y ) );
      }

      /* 重回帰分析 */

      /*
        Multiple::regCoef_iEst : 回帰直線の区間推定

        const distribution::Continuous& dist : 確率密度関数(左右対称を前提)
        stat_type var : 誤差項の分散
        stat_type b : 信頼度
        vector< ConfidenceInterval >* interval_a : 回帰係数 ai の信頼区間
        stat_type threshold : BinSearchでtを求める時のしきい値

        戻り値 : True ... 成功 , False ... 利用不可, 信頼度が不正
      */
      void Multiple::iEst_RegCoef( const distribution::Continuous& dist, stat_type var, stat_type b,
                                   vector< interval_estimation::ConfidenceInterval >* interval_a,
                                   stat_type threshold ) const
      {
        assert( interval_a != nullptr );

        const string FUNC_NAME = "tadah_fussy::statistics::regression::Multiple::iEst_RegCoef";

        error_lib::In01( b, FUNC_NAME + " : b " + IN01 );

        stat_type t = dist.binSearch( b / 2.0, threshold ); // 確率分布の片側信頼区間

        interval_a->resize( varSize() + 1 );
        for ( counter_type i = 1 ; i <= varSize() ; ++i ) {
          stat_type diff_a = t * std::sqrt( s_inv_[i - 1][i - 1] * var / sampleSize() );
          ( *interval_a )[i].lower = a( i ) - diff_a;
          ( *interval_a )[i].upper = a( i ) + diff_a;
        }

        stat_type sum = 0;
        for ( counter_type j = 0 ; j < varSize() ; ++j )
          for ( counter_type i = 0 ; i < varSize() ; ++i )
            sum += summary_x_[j].mean() * summary_x_[i].mean() * s_inv_[j][i];
        stat_type diff_a0 = t * std::sqrt( var * ( sum + 1 ) / sampleSize() );

        ( *interval_a )[0].lower = est_a0_ - diff_a0;
        ( *interval_a )[0].upper = est_a0_ + diff_a0;
      }

      /*
        Multiple::iEst_DepVar : の区間推定
        確率密度関数 dist を使い従属変数(またはその期待値)を信頼度 b で区間推定して返す

        const distribution::Continuous& dist : 確率密度関数(左右対称を前提)
        stat_type var : 誤差項の分散
        stat_type b : 信頼度
        vector< stat_type >& x : 独立変数ベクトル
        stat_type expFlag : 期待値を求める場合は 1, そうでなければ 0
        stat_type threshold : BinSearchでtを求める時のしきい値

        戻り値 : 信頼区間
      */
      interval_estimation::ConfidenceInterval
      Multiple::iEst_DepVar( const distribution::Continuous& dist, stat_type var, stat_type b, const vector< stat_type >& x,
                             stat_type expFlag, stat_type threshold ) const
      {
        const string FUNC_NAME = "tadah_fussy::statistics::regression::Multiple::iEst_DepVar";

        error_lib::In01( b, FUNC_NAME + " : b " + IN01 );

        stat_type t = dist.binSearch( b / 2.0, threshold ); // 確率分布の片側信頼区間

        stat_type d0Sq = 0;
        for ( counter_type j = 0 ; j < varSize() ; ++j )
          for ( counter_type i = 0 ; i < varSize() ; ++i )
            d0Sq += ( x[i] - summary_x( i ).mean() ) * ( x[j] - summary_x( j ).mean() ) * s_inv_[i][j];

        stat_type exp_y = expY( x.begin(), x.end() );
        stat_type diff_y = t * std::sqrt
          ( var * ( expFlag * sampleSize() + d0Sq + 1 ) / sampleSize() );

        return( interval_estimation::ConfidenceInterval( exp_y - diff_y, exp_y + diff_y ) );
      }

    } // namespace regression

    namespace interval_estimation
    {
      /*
        CorrCoef::operator() : 相関係数の区間推定

        stat_type b : 信頼度
        stat_type threshold : BinSearchでtを求める時のしきい値

        戻り値 : 信頼区間
      */
      ConfidenceInterval CorrCoef::operator()( stat_type b, stat_type threshold )
      {
        const string FUNC_NAME = "tadah_fussy::statistics::interval_estimation::CorrCoef::operator()";

        counter_type n = summary_.count();

        error_lib::GT< counter_type >( n, 3, FUNC_NAME + " : size of summary " + GT + " 3" );
        error_lib::In01( b, FUNC_NAME + " : b " + IN01 );

        distribution::Normal nDist( 0, 1 / std::sqrt( n - 3 ) );

        stat_type t = distribution::Normal( 0, 1 / std::sqrt( n - 3 ) ).binSearch( b / 2.0, threshold ); // 確率分布の片側信頼区間

        return( ConfidenceInterval
                ( ( std::exp( ( z() - t ) * 2.0 ) - 1.0 ) / ( std::exp( ( z() - t ) * 2.0 ) + 1.0 ),
                  ( std::exp( ( z() + t ) * 2.0 ) - 1.0 ) / ( std::exp( ( z() + t ) * 2.0 ) + 1.0 ) )
                );
      }

    }

    /*
      RankCorrCoef::spearmanRankCorrCoef : スピアマンの順位相関係数を求める
    */
    stat_type RankCorrCoef::spearmanRankCorrCoef() const
    {
      size_type n = x_.size(); // データ数
      if ( n <= 1 ) return( distribution::Base::nan() );

      // 順位差の二乗を求める
      stat_type d_sum = 0;
      for ( size_type i = 0 ; i < n ; ++i )
        d_sum += std::pow( x_[i] - y_[i], 2 );

      stat_type denom = n * ( std::pow( n, 2 ) - 1.0 ); // 分母( N( N^2 -1 ) )

      return( ( denom - 6.0 * d_sum ) / denom );
    }

    /*
      RankCorrCoef::kendallRankCorrCoef : ケンドールの順位相関係数を求める
    */
    stat_type RankCorrCoef::kendallRankCorrCoef() const
    {
      size_type n = x_.size(); // データ数
      if ( n <= 1 ) return( distribution::Base::nan() );

      int p = 0; // 順位の大小関係が x と y で同じ数
      int q = 0; // 順位の大小関係が x と y で異なる数
      for ( size_type i = 0 ; i < n - 1 ; ++i ) {
        for ( size_type j = i + 1 ; j < n ; ++j ) {
          if ( x_[i] == x_[j] || y_[i] == y_[j] )
            continue;
          if ( ( x_[i] > x_[j] && y_[i] > y_[j] ) ||
               ( x_[i] < x_[j] && y_[i] < y_[j] ) ) {
            ++p;
          } else {
            ++q;
          }
        }
      }

      return( ( p - q ) * 2.0 / ( n * ( n - 1 ) ) );
    }

    /*
      FactorAnalysis コンストラクタ : 因子分析用データセットの構築

      data : 対象のデータ列 (列方向に同種のデータ列が並ぶ行列)
      qrThreshold : 固有値算出時のしきい値(値の精度)
      qrMaxCnt : 固有値算出時の最大反復回数
    */
    FactorAnalysis::FactorAnalysis( const Matrix< stat_type >& data, stat_type qrThreshold, uint32_t qrMaxCnt )
      : data_( data )
    {
      typedef Matrix< stat_type >::size_type size_type;
      typedef Matrix< stat_type >::iterator iterator;

      const string FUNC_NAME = "tadah_fussy::statistics::FactorAnalysis constructor";

      // データの正規化
      for ( size_type c = 0 ; c < data_.cols() ; ++c ) {
        iterator col = data_.column( c );
        if ( ! Normalize( col, col.end() ) )
          throw std::runtime_error( FUNC_NAME + " : Failed to normalize data ( standard deviation is zero )." );
      }

      // 共分散行列の生成
      CovarianceMatrix( data_, &cov_ );

      // 固有値計算と抽出
      SquareMatrix< stat_type > buff;
      math_lib::eigen_lib::Householder_DoubleShiftQR< stat_type >( cov_, &buff, 0, qrThreshold, qrMaxCnt );

      eigenValue_.resize( buff.size() );
      for ( size_t i = 0 ; i < eigenValue_.size() ; ++i )
        eigenValue_[i] = buff[i][i];
    }

    /*
      FactorAnalysis::estimateFactorLoading : 因子負荷量の推定(主因子分析法)

      因子数を m として、因子負荷量と独自因子の分散を求め、それぞれ a, d に返す。
      d にはあらかじめ初期値を登録しておく必要がある。

      iterThreshold : 反復法のしきい値
      qrThreshold : 固有値計算時のしきい値
      iterMaxCnt : 反復法の最大処理回数
      qrMaxCnt : 固有値計算時の最大計算回数
    */
    bool FactorAnalysis::estimateFactorLoading( size_type m, Matrix< stat_type >* a, vector< stat_type >* d,
                                                stat_type iterThreshold, stat_type qrThreshold,
                                                uint32_t iterMaxCnt, uint32_t qrMaxCnt ) const
    {
      assert( a != nullptr && d != nullptr );

      const string FUNC_NAME = "tadah_fussy::statistics::FactorAnalysis::estimateFactorLoading";

      size_type p = cov_.size(); // データ種の数

      error_lib::LE( m, p, FUNC_NAME + " : m " + LE + " p" );
      error_lib::EQ( p, d->size(), FUNC_NAME + " : p " + EQ + " size of d" );

      // 因子負荷量のサイズの初期化
      a->assign( p, m );

      SquareMatrix< stat_type > q; // 固有値分解後の固有ベクトルからなる直交行列 Q

      // 反復法の開始
      for ( uint32_t iterCnt = 0 ; iterCnt < iterMaxCnt ; ++iterCnt ) {
        // 共分散行列をコピーして対角成分から独自因子成分を減算 AAt = V - D
        SymmetricMatrix< stat_type > covBuff( cov_ );
        for ( size_type i = 0 ; i < p ; ++i )
          covBuff[i][i] -= (*d)[i];

        // 固有値分解 AAt = QD'Q
        SquareMatrix< stat_type > buff;
        math_lib::eigen_lib::Householder_DoubleShiftQR< stat_type >( covBuff, &buff, &q, qrThreshold, qrMaxCnt );
        //EV_Jacobi< stat_type > jacobi( qrThreshold, qrMaxCnt );
        //jacobi( &covBuff, &q );
        //SquareMatrix< stat_type > buff( covBuff );

        // D'の対角成分から固有値を抽出
        std::multimap< stat_type, size_type > mmapEigenValue; // 固有値と列番号からなるマップ
        for ( size_type i = 0 ; i < p ; ++i )
          mmapEigenValue.insert( std::pair< stat_type, size_type >( buff[i][i], i ) );

        // 固有値とその固有ベクトルから AAt の対角成分を計算
        vector< stat_type > aat( p, 0 ); // AAt の対角成分
        auto cit_ev = mmapEigenValue.rbegin();
        for ( size_type i = 0 ; i < m ; ++i ) {
          size_type j = cit_ev->second;
          stat_type eigenValue = cit_ev->first;
          for ( size_type r = 0 ; r < p ; ++r ) {
            (*a)[r][i] = std::sqrt( std::abs( eigenValue ) ) * q[r][j];
            aat[r] += std::pow( (*a)[r][i], 2 );
          }
          ++cit_ev;
        }

        // 新たな独自因子成分 D'' を計算し、前回のものと比較
        bool isConverged = true;
        for ( size_type i = 0 ; i < p ; ++i ) {
          stat_type new_d = cov_[i][i] - aat[i];
          if ( std::abs( (*d)[i] - new_d ) >= iterThreshold )
            isConverged = false;
          (*d)[i] = new_d; // 新たな D の成分を代入
        }

        if ( isConverged ) return( true );
      }

      return( false );
    }

    /*
      FactorAnalysis::normalizeByCommunality : 共通性 ( = 1 - d[i] ) による因子負荷量 a の標準化
    */
    void FactorAnalysis::normalizeByCommunality( const vector< stat_type >& d, Matrix< stat_type >* a )
    {
      assert( a != nullptr );

      for ( size_type r = 0 ; r < a->rows() ; ++r ) {
        if ( r >= d.size() ) break;
        stat_type h = std::sqrt( 1.0 - d[r] ); // 共通性の平方根
        if ( h == 0 ) continue;
        for ( size_type c = 0 ; c < a->cols() ; ++c )
          (*a)[r][c] /= h;
      }
    }

    /*
      FactorAnalysis::transpose : 行列 a を転置して at に返す。共通性の平方根 h で除算する処理も同時に行う。
    */
    void FactorAnalysis::transpose( const Matrix< stat_type >& a, const std::vector< stat_type >& h, Matrix< stat_type >* at )
    {
      size_type p = a.rows(); // データ種の数
      size_type m = a.cols(); // 因子の数

      // 転置処理開始(共通性での除算を実施)
      at->assign( m, p );
      for ( size_type i = 0 ; i < m ; ++i ) {
        for ( size_type j = 0 ; j < p ; ++j ) {
          if ( h[j] != 0 ) (*at)[i][j] = a[j][i] / h[j];
        }
      }
    }

    /*
      FactorAnalysis::calcMaxVarAngle : 分散を最大にする回転角を求めて返す

      at : 因子負荷量 A の転置行列を共通性の平方根で割った行列
      i0, i1 : 回転対象の行列位置
    */
    FactorAnalysis::stat_type
    FactorAnalysis::calcMaxVarAngle( const Matrix< stat_type >& at, size_type i0, size_type i1 )
    {
      if ( at.rows() == 0 ) return( distribution::Base::nan() );
      if ( i0 >= at.rows() || i1 >= at.rows() ) return( distribution::Base::nan() );

      size_type p = at.cols(); // データ種のサイズ

      // 回転角を求めるための項(A,B,C,D)を求めるための計算
      vector< stat_type > vec_a( p ), vec_b( p ), vec_c( p ), vec_d( p );
      for ( vector< stat_type >::size_type i = 0 ; i < p ; ++i ) {
        vec_a[i] = ( at[i0][i] + at[i1][i] ) * ( at[i0][i] - at[i1][i] );
        vec_b[i] = at[i0][i] * at[i1][i] * 2;
        vec_c[i] = ( vec_a[i] + vec_b[i] ) * ( vec_a[i] - vec_b[i] );
        vec_d[i] = vec_a[i] * vec_b[i] * 2;
      }

      // 回転角を求めるための項(A,B,C,D)
      stat_type asum = SumAll< stat_type >( vec_a );
      stat_type bsum = SumAll< stat_type >( vec_b );
      stat_type csum = SumAll< stat_type >( vec_c );
      stat_type dsum = SumAll< stat_type >( vec_d );

      stat_type numerator = dsum - 2 * asum * bsum / p; // D - 2AB / p
      stat_type theta = std::atan( numerator / ( ( asum + bsum ) * ( asum - bsum ) / p - csum ) ); // 回転角(4θ)

      // 符号が逆の場合、π 加える
      if ( numerator * sin( theta ) > 0 )
        theta += math_lib::Pi< stat_type >();

      return( theta / 4.0 );
    }

    /*
      FactorAnalysis::varimax : 因子負荷量 a の回転 ( 基準Varimax法 )

      threshold : 反復法のしきい値
      maxCnt : 反復法の最大処理回数
    */
    bool FactorAnalysis::varimax( Matrix< stat_type >* a, stat_type threshold, uint32_t maxCnt )
    {
      assert( a != nullptr );

      //typedef std::pointer_to_binary_function< stat_type, stat_type, stat_type > binfun;
      //typedef std::binder2nd< binfun > binder2nd;
      typedef stat_type ( *func )( stat_type, stat_type );

      size_type p = a->rows(); // データ種の数
      if ( p < 1 ) return( true );
      size_type m = a->cols(); // 潜在的因子の数

      //binder2nd pow4 = std::bind2nd( std::ptr_fun( static_cast< func >( std::pow ) ), 4.0 ); // 四乗和計算用関数オブジェクト
      //binder2nd pow2 = std::bind2nd( std::ptr_fun( static_cast< func >( std::pow ) ), 2.0 ); // 二乗和計算用関数オブジェクト
      using namespace std::placeholders;
      auto pow4 = std::bind( static_cast< func >( std::pow ), _1, 4.0 );
      auto pow2 = std::bind( static_cast< func >( std::pow ), _1, 2.0 );

      vector< stat_type > h( p ); // 共通性の平方根
      for ( size_type i = 0 ; i < p ; ++i ) {
        auto row = a->row( i );
        h[i] = std::sqrt( Sum< stat_type >( row, row.end(), OpBind( pow2 ) ) );
      }

      Matrix< stat_type > at; // a の二乗の転置行列
      transpose( *a, h, &at );

      vector< stat_type > vk( m ); // 各列の分散
      for ( size_type i = 0 ; i < m ; ++i ) {
        auto row = at.row( i );
        vk[i] = Sum< stat_type >( row, row.end(), OpBind( pow4 ) ) / p + std::pow( Sum< stat_type >( row, row.end(), OpBind( pow2 ) ) / p, 2.0 );
      }
      stat_type v = SumAll< stat_type >( vk ); // 分散の合計

      // 反復処理開始
      for ( uint32_t cnt = 0 ; cnt < maxCnt ; ++cnt ) {
        for ( size_type i = 0 ; i < m ; ++i ) {
          for ( size_type j = i + 1 ; j < m ; ++j ) {
            stat_type theta = calcMaxVarAngle( at, i, j ); // 分散を最大にする回転角
            //std::cout << "theta=" << theta * 360 / ( 2 *M_PI ) << std::endl;
            for ( size_type k = 0 ; k < p ; ++k ) {
              stat_type ai = (*a)[k][i] * cos( theta ) - (*a)[k][j] * sin( theta ); // k 行 i 列の更新
              stat_type aj = (*a)[k][i] * sin( theta ) + (*a)[k][j] * cos( theta ); // k 行 j 列の更新
              (*a)[k][i] = ai;
              (*a)[k][j] = aj;
              if ( h[k] != 0 ) {
                at[i][k] = ai / h[k];
                at[j][k] = aj / h[k];
              }
            }
          }
        }
        for ( size_type i = 0 ; i < m ; ++i ) {
          auto row = at.row( i );
          vk[i] = Sum< stat_type >( row, row.end(), OpBind( pow4 ) ) / p + std::pow( Sum< stat_type >( row, row.end(), OpBind( pow2 ) ) / p, 2.0 );
        }
        stat_type new_v = SumAll< stat_type >( vk ); // 新たに求めた分散

        if ( std::abs( v - new_v ) < threshold )
          return( true );

        v = new_v;
      }

      return( false );
    }

    /*
      FactorAnalysis::getFactorScore : 因子得点の線形式の係数 b を使い、各データの因子得点の推定値を計算して fs に返す

      b は因子得点の線形式の係数で m x p 行列、データは data_ を使い、n x p 行列である。
      但し、m : 因子の数 ; p : データ種の数 ; n : データ数 をそれぞれ表す。
    */
    void FactorAnalysis::getFactorScore( const Matrix< stat_type >& b, Matrix< stat_type >* fs ) const
    {
      assert( fs != nullptr );

      size_type p = classCnt(); // データ種の数
      if ( p == 0 ) return;

      size_type m = b.rows();     // 因子負荷量の数
      size_type n = data_.rows(); // データ数

      fs->assign( n, m );
      for ( size_type i = 0 ; i < n ; ++i ) {
        for ( size_type j = 0 ; j < m ; ++j ) {
          auto r1 = data_.row( i );
          auto r2 = b.row( j );
          (*fs)[i][j] = std::inner_product( r1, r1.end(), r2, stat_type() );
        }
      }
    }

  } // namespace statistics

} // namespace tadah_fussy
