namespace tadah_fussy
{
  namespace matrix_lib
  {
    /*
      CalcPPt : M = P(Pt) を求める
    */
    template< class T >
      void CalcPPt( const Matrix< T >& p, SymmetricMatrix< T >* m )
      {
        using size_type = typename Matrix< T >::size_type;

        m->assign( p.rows() );
        for ( size_type r = 0 ; r < p.rows() ; ++r ) {
          for ( size_type c = r ; c < p.rows() ; ++c ) {
            auto row1 = p[r];
            auto row2 = p[c];
            (*m)[r][c] = std::inner_product( row1, row1.end(), row2, T() );
          }
        }
      }

    /*
      CalcPtP : N = (Pt)P を求める
    */
    template< class T >
      void CalcPtP( const Matrix< T >& p, SymmetricMatrix< T >* n )
      {
        using size_type = typename Matrix< T >::size_type;

        n->assign( p.cols() );
        for ( size_type r = 0 ; r < p.cols() ; ++r ) {
          for ( size_type c = r ; c < p.cols() ; ++c ) {
            auto col1 = p.cbegin( r );
            auto col2 = p.cbegin( c );
            (*n)[r][c] = std::inner_product( col1, col1.end(), col2, T() );
          }
        }
      }

    /// @brief 特異値分解用関数オブジェクト
    ///
    /// 任意の行列 P に対し、対角行列 D と直交行列 U,V により P = UD(Vt) の形に分解することを特異値分解という。
    /// 但し、行列の添字 t は転置を表す。
    ///
    /// このクラスでは、
    ///
    /// - (Ut)P(Pt)U = D^2
    /// - (Vt)(Pt)PV = D^2
    /// - U = PVD^-1
    /// - V = (Pt)UD^-1
    ///
    /// であることを利用して ( ^2 は 2 乗、^-1 は逆行列を表す )、P の行数が列数以下なら P(Pt) を計算し、その固有値分解により U を求め、
    /// V = (Pt)UD^-1 を計算する。
    /// 逆に P の行数が列数より大きい場合は (Pt)P を計算し、その固有値分解により V を求め、U = PVD^-1 を計算する。
    /// 前者の場合、U は P の行数をサイズとする正方行列で、V は Pt の行列数と等しいサイズの行列になる。
    /// また、後者の場合、V は P の列数をサイズとする正方行列で、U は P の行列数と等しいサイズの行列になる。
    ///
    /// @param p 特異値分解を行う対象の行列 P
    /// @param u 直交行列 U を保持する変数へのポインタ
    /// @param d 対角行列 D を保持する変数へのポインタ
    /// @param v 直交行列 V を保持する変数へのポインタ
    /// @param e 対角化を行うときに、非対角成分をゼロと判断するためのしきい値
    /// @param maxCnt 対角化を行うときの反復処理最大回数
    template< class T >
      void SingularValueDecomposition( const Matrix< T >& p,
                                       Matrix< T >* u, std::vector< T >* d, Matrix< T >* v,
                                       T e, uint32_t maxCnt )
      {
        assert( u != nullptr && d != nullptr && v != nullptr );

        SquareMatrix< T > r;    // 上三角行列 R (対角成分が固有値になる)

        // 横長の行列
        if ( p.rows() <= p.cols() ) {
          SymmetricMatrix< T > m; // M = P(Pt)
          SquareMatrix< T > ev;    // M の固有ベクトル

          CalcPPt( p, &m ); // M = P(Pt) を求める

          // 対象行列の対角化 ( 対角行列の対角成分と直交行列 U を求める )
          math_lib::eigen_lib::Householder_DoubleShiftQR< T >( m, &r, &ev, e, maxCnt );
          *u = ev;
          //ev.clone( u );

          // szN : P の列数 : N = (Pt)P のサイズ
          typename Matrix< T >::size_type szN = p.cols();
          // szM : P の行数 : M = P(Pt), R, U のサイズ
          typename Matrix< T >::size_type szM = p.rows();

          // V = (Pt)U / λ
          v->assign( szN, szM );
          for ( typename Matrix< T >::size_type k = 0 ; k < szN ; ++k ) {
            for ( typename Matrix< T >::size_type j = 0 ; j < szM ; ++j ) {
              auto ptRow = p.cbegin( k );
              auto uCol = u->cbegin( j );
              (*v)[k][j] = std::inner_product( ptRow, ptRow.end(), uCol, T() );
              if ( r[j][j] != 0 ) (*v)[k][j] /= std::sqrt( std::abs( r[j][j] ) );
              if ( r[j][j] < 0 ) (*v)[k][j] *= -1;
            }
          }
          // 縦長の行列
        } else {
          SymmetricMatrix< T > n; // N = (Pt)P
          SquareMatrix< T > ev;    // N の固有ベクトル

          CalcPtP( p, &n ); // N = (Pt)P を求める

          // 対象行列の対角化 ( 対角行列の対角成分と直交行列 V を求める )
          math_lib::eigen_lib::Householder_DoubleShiftQR< T >( n, &r, &ev, e, maxCnt );
          *v = ev;

          // szN : P の列数 : N = (Pt)P, R, V のサイズ
          typename Matrix< T >::size_type szN = p.cols();
          // szM : P の行数 : M = P(Pt) のサイズ
          typename Matrix< T >::size_type szM = p.rows();

          // U = PV / λ
          u->assign( szM, szN );
          for ( typename Matrix< T >::size_type k = 0 ; k < szM ; ++k ) {
            for ( typename Matrix< T >::size_type j = 0 ; j < szN ; ++j ) {
              auto pRow = p[k];
              auto vCol = v->cbegin( j );
              (*u)[k][j] = std::inner_product( pRow, pRow.end(), vCol, T() );
              if ( r[j][j] != 0 ) (*u)[k][j] /= std::sqrt( std::abs( r[j][j] ) );
              if ( r[j][j] < 0 ) (*u)[k][j] *= -1;
            }
          }
        }

        // 固有値の抽出
        d->assign( r.size(), T() );
        for ( typename std::vector< T >::size_type i = 0 ; i < r.size() ; ++i )
          (*d)[i] = std::sqrt( std::abs( r[i][i] ) );
      }

  } // namespace matrix_lib

} // namespace tadah_fussy
