/**
   @file statistics.h
   @brief 確率・統計ルーチン定義用ヘッダ

   @author tadah_fussy
   @date 2015/03/15 コメントを Doxygen 用に変更 
   @date 2019/04/08 std::binder2nd の廃止に伴い std::function に切り替え
   @date 2019/04/19 エラーメッセージの見直し
   @date 2019/04/20 SignedRankDistribution にて n = 0 のときの確率が正しく計算されていなかった問題を修正
   @date 2019/04/22 Test_Wilcoxon にて p() が正しく計算されていなかった問題を修正
   @date 2019/04/26 コメントの追加
   @date 2019/04/27 MultipleRegressionCoefficient の連立方程式の解法を引数で任意に指定できるよう変更
   @date 2019/04/30 FactorAnalysis の連立方程式の解法を引数で任意に指定できるよう修正
   @date 2021/03/03 細かい部分の修正 ( 型など )
   @date 2021/03/21 名前空間の変更
**/
#ifndef STATISTICS_20210321_H
#define STATISTICS_20210321_H

#include <common/equation.h>
#include <common/eigen.h>
#include <common/gadget.h>

#include <cstdlib>
#include <cfloat>
#include <cmath>
#include <cassert>

#include <vector>
#include <map>
#include <iostream>
#include <sstream>
#include <algorithm>
#include <string>
#include <numeric>
#include <limits>
#include <stdexcept>

/**
   @brief 作成者 tadah_fussy
**/
namespace tadah_fussy
{
  /**
     @brief 確率・統計ルーチン用名前空間

     確率・統計ルーチン用の定数や関数をまとめた名前空間
  **/
  namespace statistics
  {
    /// @brief 近似値を求める時のデフォルトの精度
    const double DEFAULT_THRESHOLD = 1.0E-6;

    /// @brief 反復処理の最大回数デフォルト値
    const uint32_t DEFAULT_MAX_COUNT = 100;

    /************************************************************************
     補助用関数オブジェクト
    ************************************************************************/

    /* 一変数の変換 */

    /**
       @brief 逆数を求めるための関数オブジェクト

       調和平均 tadah_fussy::statistics::HarmonicMean の計算に利用する
    **/
    template< class T >
    struct Reciprocal
    {
      /// @brief 逆数を返す
      ///
      /// @note ゼロ除算に対するチェックはしていない。
      ///
      /// @param x 逆数を求める対象の数
      /// @return x の逆数
      T operator()( T x )
      { return( 1 / x ); }
    };

    /**
       @brief 指定した数 a との差の二乗を求める関数

       総和計算関数と組み合わせて、
       偏差平方和 tadah_fussy::statistics::DevSq
       標本分散 tadah_fussy::statistics::SampleVariance
       不偏分散 tadah_fussy::statistics::UnbiasedVariance
       の計算に利用する
    **/
    template< class T >
    class Deviation
    {
      T a_; // 数 a

    public:

      /// @brief 数 a を指定して構築
      ///
      /// @param a 値から減算する数 a
      explicit Deviation( T a )
        : a_( a ) {}

      /// @brief 数 x と a の差の二乗を求める
      ///
      /// @param x 変数
      /// @return x - a の二乗
      T operator()( T x )
      { return( ( x - a_ ) * ( x - a_ ) ); }
    };

    /**
       @brief 指定した数 a との差の絶対値を求める関数

       総和計算関数と組み合わせて、平均偏差 tadah_fussy::statistics::MeanDeviation の計算に利用する
    **/
    template< class T >
    class AbsDeviation
    {
      T a_; // 数 a

    public:

      /// @brief 数 a を指定して構築
      ///
      /// @param a 値から減算する数 a
      explicit AbsDeviation( T a )
        : a_( a ) {}

      /// @brief 数 x と a の差の絶対値を求める
      ///
      /// @param x 変数
      /// @return x - a の絶対値
      T operator()( const T& x )
      { return( ( x > a_ ) ? ( x - a_ ) : ( a_ - x ) ); }
    };

    /**
       @brief 標準得点のべき乗を求める関数

       総和計算関数と組み合わせて、標準化積率 tadah_fussy::statistics::StdMoment の計算に利用する
    **/
    template< class T >
    class Moment
    {
      T m_; // 平均
      T s_; // 標準偏差
      T r_; // 次数

    public:

      /// @brief 平均、標準偏差、次数を指定して構築
      ///
      /// @note 次数が正常か(ゼロ以上の整数であるか)はチェックしていない。
      ///
      /// @param m 平均
      /// @param s 標準偏差
      /// @param r 次数
      Moment( T m, T s, T r )
        : m_( m ), s_( s ), r_( r ) {}

      /// @brief 標準得点 ( x - m ) / s の r 乗を求める
      ///
      /// @param x 変数
      /// @return 標準得点 ( x - m ) / s の r 乗
      T operator()( const T& x )
      { return( std::pow( ( x - m_ ) / s_, r_ ) ); }
    };

    namespace distribution
    {
      /************************************************************************
        一変数確率分布用クラス
      ************************************************************************/

      /**
         @brief 一次元確率分布用インターフェース

         一次元確率密度関数を定義するためのインターフェース。
         平均と分散を返す関数 Base::average と Base::variance が定義されている。
      **/
      struct Base
      {
        /// @brief 仮想デストラクタ(何もしない)
        virtual ~Base() {}

        /// @brief 確率や統計量に対する型
        using stat_type = double;

        /// @brief カウンタ(データ数など)や自由度に対する型
        using counter_type = size_t;

        /// @brief 確率や統計量の無効値(NaN)
        static stat_type nan()
        { return( std::numeric_limits< stat_type >::quiet_NaN() ); }

        /// @brief 確率や統計量の無限大
        static stat_type infinity()
        { return( std::numeric_limits< stat_type >::infinity() ); }

        /// @brief 平均値を返す
        ///
        /// @return 平均値
        virtual stat_type average() const = 0;

        /// @brief 分散を返す
        ///
        /// @return 分散
        virtual stat_type variance() const = 0;
      };

    }

  } // namespace statistics

  namespace exception
  {
    /**
       @brief 単位区間 [ 0, 1 ] の範囲チェック用例外クラス

       値が単位区間 [ 0, 1 ] に含まれない場合に投げられる例外クラス。
       内部に不正と判断された値を保持する。
    **/
    class UnitInterval : public Domain
    {
      statistics::distribution::Base::stat_type invalidValue_; // 不正値

    public:

      /// @brief 不正と判断された値を指定して構築
      ///
      /// @param info 例外に関する情報
      /// @param invalidValue 不正値
      explicit UnitInterval( const std::string& info, statistics::distribution::Base::stat_type invalidValue )
        : Domain( info ), invalidValue_( invalidValue ) {}

      /// @brief 不正値を返す
      ///
      /// @return 不正値
      statistics::distribution::Base::stat_type invalidValue() const
      { return( invalidValue_ ); }

      /// @brief 不正値・指定範囲の出力
      ///
      /// @param os 出力ストリームへの参照
      /// @return なし
      virtual void printInfo( std::ostream& os ) const
      { os << "invalid value = " << invalidValue(); }
    };
  }

  namespace error_lib
  {
    /**
       @brief 単位区間 [ 0, 1 ] の範囲内にあるかを判定する関数オブジェクト

       確率や信頼度、危険率として適正な値かを判定する場合に利用する
    **/
    struct IsInUnitInterval
    {
      /// @brief 単位区間 [ 0, 1 ] の範囲内にあるかを判定する
      ///
      /// @param d 判定対象の値
      /// @return 単位区間内なら true を返す
      bool operator()( statistics::distribution::Base::stat_type d )
      {
        using stat_type = statistics::distribution::Base::stat_type;

        return( d >= stat_type( 0 ) && d <= stat_type( 1 ) );
      }
    };

    /// @brief 単位区間 [ 0, 1 ] の範囲内にあるかをチェックする
    ///
    /// 成り立っていればその値をそのまま返し、成り立っていない場合は例外 ExceptionUnitInterval を投げる。
    /// 値のフィルタリングに利用する。
    ///
    /// @param p チェックする値
    /// @param info 例外に関する情報
    /// @return なし
    inline statistics::distribution::Base::stat_type
    In01( statistics::distribution::Base::stat_type p, const std::string& info )
    { return( CheckTool( IsInUnitInterval(), exception::UnitInterval( info, p ), p ) ); }

  }

  namespace error_message
  {
    /// @brief 単位区間 [ 0, 1 ] の範囲外だった場合のエラー・メッセージ
    const std::string IN01 = "must be in unit interval [ 0, 1 ]";
  }

  namespace statistics
  {
    namespace distribution
    {
      /**
         @brief 一次元離散分布用インターフェース

         一次元離散確率密度関数を定義するためのインターフェース。
         確率変数は整数型 (Discrete::var_type) を取るものとする。
      **/
      class Discrete : public Base
      {
      public:

        /// @brief 確率変数に対する型
        using var_type = int32_t;

        /// @brief 確率変数 n に対する確率密度を返す
        ///
        /// @param n 確率変数
        /// @return 確率密度
        virtual stat_type operator[]( var_type n ) const = 0;

        /// @brief 区間 [s,e] における確率を返す
        ///
        /// @param s, e 区間
        /// @return 区間 [ s, e ] での確率
        stat_type p( var_type s, var_type e ) const;
      };

      /**
         @brief 一次元連続分布用インターフェース

         一次元連続確率密度関数を定義するためのインターフェース。
         確率変数は浮動小数点型で Base::stat_type と同等とする。
      **/
      class Continuous : public Base
      {
      public:

        /// @brief 確率変数 x における確率密度を返す
        ///
        /// @param x 確率変数
        /// @return 確率密度
        virtual stat_type operator[]( stat_type x ) const = 0;

        /// @brief 区間 ( -∞, a ] における下側確率を返す
        ///
        /// @param a 区間の上側の値
        /// @return 下側確率
        virtual stat_type lower_p( stat_type a ) const = 0;

        /// @brief 区間 [ a, b ] または [ b, a ] における確率を返す
        ///
        /// @param a 区間の境界値 1
        /// @param b 区間の境界値 2
        /// @return 区間 [ a, b ] の確率
        virtual stat_type p( stat_type a, stat_type b ) const
        { return( ( a > b ) ? lower_p( a ) - lower_p( b ) : lower_p( b ) - lower_p( a ) ); }

        /// @brief 確率が prob になる区間 [ 0, t ] に対して t を求める
        ///
        /// 区間の探索には二部探索を利用している。
        ///
        /// @param prob 区間 [ 0, t ] に対する確率
        /// @param threshold t を求める精度
        /// @return 求めた t の値
        stat_type binSearch( stat_type prob, stat_type threshold = DEFAULT_THRESHOLD ) const;
      };

      /* ----------------------------
         離散確率分布
         ---------------------------- */

      /**
         @brief 二項分布

         B( n ; p ) = nCk p^k q^(n-k)

         p : ある事象 A が一回の試行で起こる確率 ( 0 ≦ p ≦ 1 )\n
         q : ある事象 A が一回の試行で起こらない確率( = 1 - p )\n
         n : 独立試行の回数 ( n ≧ 0 )

         nCk は n 個の中から k 個を選択する組み合わせを表す。

         二項分布は、確率 p で起こる試行を n 回繰り返した時、k 回起こった時の確率を表す。
      **/
      class Binomial : public Discrete
      {
        stat_type p_; // 一回の試行で起こる確率
        var_type n_;  // 独立試行の回数

      public:

        /// @brief 一回の試行で事象が起こる確率 p と独立試行の回数 n を指定して構築
        ///
        /// - 0 <= p <= 1 である必要がある。そうでない場合は例外 exception::UnitInterval を投げる。
        /// - n >= 0 である必要がある。そうでない場合は例外 exception::NegativeNumber< Discrete::var_type > を投げる。
        ///
        /// @param p 一回の試行で事象が起こる確率
        /// @param n 独立試行の回数
        Binomial( stat_type p, var_type n )
          : p_( p ), n_( n )
        {
          const std::string FUNC_NAME = "tadah_fussy::statistics::distribution::Binomial constructor";

          error_lib::In01( p_, FUNC_NAME + " : p " + error_message::IN01 );
          error_lib::NNNum( n_, FUNC_NAME + " : n " + error_message::NNNUM );
        }

        virtual stat_type average() const; // 平均値
        virtual stat_type variance() const; // 分散

        // 確率変数 r に対する確率密度を返す
        virtual stat_type operator[]( var_type r ) const;
      };

      /**
         @brief ポアソン分布

         P( r ; λ ) = λ^r・e^(-λ) / r!

         λ : 事象の平均発生回数 ( λ >= 0 )

         ポアソン分布は、平均して(例えば一日あたり) λ 回発生する事象に対して r 回発生した時の確率を表す。
      **/
      class Poisson : public Discrete
      {
        stat_type lambda_; // 定数 λ

      public:

        /// @brief 定数 λ を指定して構築
        ///
        /// - λ >= 0 である必要がある。そうでない場合は例外 exception::NegativeNumber< Base::stat_type > を投げる。
        ///
        /// @param lambda 定数 λ ( λ >= 0 )
        explicit Poisson( stat_type lambda )
          : lambda_( lambda )
        {
          const std::string FUNC_NAME = "tadah_fussy::statistics::distribution::Poisson constructor";
          error_lib::NNNum( lambda_, FUNC_NAME + " : lambda " + error_message::NNNUM );
        }

        virtual stat_type average() const;  // 平均値
        virtual stat_type variance() const; // 分散

        // 確率変数 r に対する確率密度を返す
        virtual stat_type operator[]( var_type r ) const;
      };

      /**
         @brief 負の二項分布

         NB( r ; k,p ) = r+k-1Cr p^k q^r

         p : ある事象が一回の試行で起こる確率 ( 0 ≦ p ≦ 1 )\n
         q : ある事象が一回の試行で起こらない確率 ( = 1 - p )\n
         k : 事象の発生回数\n
         r : 事象の発生しなかった回数

         但し、nCk は n 個の中から k 個を選択する組み合わせ

         負の二項分布は、事象が k 回発生するまでに r 回は発生しなかったときの確率を表す。
         なお、もう一つの表し方がある ( NegativeBinomial_N を参照 )。
      **/
      class NegativeBinomial : public Discrete
      {
        stat_type p_; // 一回の試行で起こる確率
        var_type k_;  // 事象の発生回数

      public:

        /// @brief ある事象に対し、一回の試行で起こる確率 p と発生回数 k を指定して構築
        ///
        /// - 0 <= p <= 1 である必要がある。そうでない場合は例外 exception::UnitInterval を投げる。
        /// - p > 0 である必要がある。そうでない場合は例外 exception::NotPositiveNumber< Base::stat_type > を投げる。
        /// - k > 0 である必要がある。そうでない場合は例外 exception::NotPositiveNumber< Discrete::var_type > を投げる。
        ///
        /// @param p ある事象が一回の試行で起こる確率
        /// @param k 事象の発生回数
        explicit NegativeBinomial( stat_type p, var_type k = 1 )
          : p_( p ), k_( k )
        {
          const std::string FUNC_NAME = "tadah_fussy::statistics::distribution::NegativeBinomial constructor";

          error_lib::In01( p_, FUNC_NAME + " : p " + error_message::IN01 );
          error_lib::PNum( p_, FUNC_NAME + " : p " + error_message::PNUM );
          error_lib::PNum( k_, FUNC_NAME + " : k " + error_message::PNUM );
        }

        virtual stat_type average() const;  // 平均値
        virtual stat_type variance() const; // 分散

        // 確率変数 r に対する確率密度を返す
        virtual stat_type operator[]( var_type r ) const;
      };

      /**
         @brief 負の二項分布(2)

         NB( n ; k,p ) = n-1Ck-1 p^k q^n-k

         p : ある事象が一回の試行で起こる確率 ( 0 ≦ p ≦ 1 )\n
         q : ある事象が一回の試行で起こらない確率( = 1 - p )\n
         k : 事象の発生回数 \n
         n : 試行回数

         負の二項分布は、事象が k 回発生するまでに、n 回の試行が必要だったときの確率を表し、
         k = 1 のときは「幾何分布」になる。
         なお、もう一つの表し方がある ( NegativeBinomial を参照 )。
      **/
      class NegativeBinomial_N : public Discrete
      {
        stat_type p_; // 一回の試行で起こる確率
        var_type k_;  // 事象の発生回数

      public:

        /// @brief 一回の試行で起こる確率 p と事象の発生回数 k を指定して構築
        ///
        /// - 0 <= p <= 1 である必要がある。そうでない場合は例外 exception::UnitInterval を投げる。
        /// - p > 0 である必要がある。そうでない場合は例外 exception::NotPositiveNumber< Base::stat_type > を投げる。
        /// - k > 0 である必要がある。そうでない場合は例外 exception::NotPositiveNumber< Discrete::var_type > を投げる。
        ///
        /// @param p ある事象が一回の試行で起こる確率
        /// @param k 事象の発生回数
        explicit NegativeBinomial_N( stat_type p, var_type k = 1 )
          : p_( p ), k_( k )
        {
          const std::string FUNC_NAME = "tadah_fussy::statistics::distribution::NegativeBinomial_N constructor";

          error_lib::In01( p_, FUNC_NAME + " : p " + error_message::IN01 );
          error_lib::PNum( p_, FUNC_NAME + " : p " + error_message::PNUM );
          error_lib::PNum( k_, FUNC_NAME + " : k " + error_message::PNUM );
        }

        virtual stat_type average() const;  // 平均値
        virtual stat_type variance() const; // 分散

        // 確率変数 n に対する確率密度を返す
        virtual stat_type operator[]( var_type n ) const;
      };

      /**
         @brief 超幾何分布

         P( r; a,b,k ) = aCr bCk-r / a+bCk

         a,b : 二種類の要素の個数\n
         k : 要素から選択する個数\n
         r : 片側の要素が選択される個数

         超幾何分布は、二種類の要素から k 個の要素を選択した時、片側の要素が r 個選ばれる確率を表す。
      **/
      class Hypergeometric : public Discrete
      {
        var_type a_; // 要素 A の個数
        var_type b_; // 要素 B の個数
        var_type k_; // 要素から選択する個数

      public:

        /// @brief 各要素の個数 a, b と要素から選択する個数 k を指定して構築
        ///
        /// - a >= 0, b >= 0, k >= 0 である必要がある。そうでない場合は例外 exception::NegativeNumber< Discrete::var_type > を投げる。
        /// - k <= a + b である必要がある。そうでない場合は例外 exception::TooGreatNumber< Discrete::var_type > を投げる。
        ///
        /// @param a 要素 A の個数
        /// @param b 要素 B の個数
        /// @param k 要素から選択する個数
        Hypergeometric( var_type a, var_type b, var_type k )
          : a_( a ), b_( b ), k_( k )
        {
          const std::string FUNC_NAME = "tadah_fussy::statistics::distribution::Hypergeometric constructor";

          error_lib::NNNum( a_, FUNC_NAME + " : a " + error_message::NNNUM );
          error_lib::NNNum( b_, FUNC_NAME + " : b " + error_message::NNNUM );
          error_lib::NNNum( k_, FUNC_NAME + " : k " + error_message::NNNUM );
          error_lib::LE( k_, a_ + b_, FUNC_NAME + " : k " + error_message::LE + " a + b" );
        }

        virtual stat_type average() const;  // 平均値
        virtual stat_type variance() const; // 分散

        // 確率変数 r に対する確率密度を返す
        virtual stat_type operator[]( var_type r ) const;
      };

      /**
         @brief U-分布

         二つの値 x, y に対し、ある値 ( 例えば y ) の後ろにあるもう一方の値 ( x ) の個数を U とした時、
         全ての並べ方が等確率な場合に U の値に対する確率は以下の漸化式で求められる。

         U( m, n ; u ) = [ m / ( m + n ) ]・U( m - 1, n ; u - n ) + [ n / ( m + n ) ]・U( m, n - 1 ; u )

         m : 値 x の総数\n
         n : 値 y の総数

         例えば、yxxyyx は先頭の y の後ろに x が三つ、その他の y の後ろに x が一つずつあるので U = 5 になる。

         @note 漸化式を使って計算しているため、大きな値に対しては時間がかかる。
         @note U-分布は中心極限定理により正規分布に近似することができる。
      **/
      class U : public Discrete
      {
        // UMap用のキー
        struct UKey
        {
          // 二つの値の総数
          var_type m;
          var_type n;

          // コンストラクタ
          UKey( var_type m_, var_type n_ )
            : m( m_ ), n( n_ ) {}

          // 比較用演算子の多重定義
          bool operator<( const UKey& dest ) const
          {
            if ( m == dest.m )
              return( n < dest.n );
            else
              return( m < dest.m );
          }
        };

        // 過去に求めた U 値に対する確率密度を保持する配列用 Map
        using UMap = std::map< UKey, std::vector< stat_type > >;
        mutable UMap buff_;

        // 二つの値の総数
        var_type m_;
        var_type n_;

        // U( m, n ) = u となる確率を求める
        stat_type calcP( var_type m, var_type n, var_type u ) const;

        // U( m, n ) = u となる確率をバッファから取り出す
        stat_type getP( var_type m, var_type n, var_type u ) const;

        // U( m, n ) = u となる確率をバッファに登録する
        void setP( var_type m, var_type n, var_type u, stat_type p ) const;

      public:

        /// @brief 二値のそれぞれの総数 m, n を指定して構築
        ///
        /// - m >= 0, n >= 0 である必要がある。そうでない場合は例外 exception::NegativeNumber< Discrete::var_type > を投げる。
        ///
        /// @param m 値 1 の総数
        /// @param n 値 2 の総数
        U( var_type m, var_type n );

        virtual stat_type average() const;  // 平均値
        virtual stat_type variance() const; // 分散値

        // U( m, n ) = u となる確率を返す
        virtual stat_type operator[]( var_type u ) const;
      };

      /**
         @brief 符号付き順位和の確率分布

         二つの値 x, y に対してその差を求め、絶対値を昇順に並べ替えた時の順序を 1 から順に付ける。
         但し、差がゼロの要素は除外する。正値と負値のそれぞれに分けて順序和 Ss, Sm を求めたとき、
         値が等しい確率で発生する場合の s = Ss - Sm に対する確率は、漸化式

         Pn(s) = [ Pn-1( s - n ) + Pn-1( s + n ) ] / 2

         n : 値の総数(差がゼロの要素を除外する)
         s : 正値の順序和と負値の順序和の差

         で求められる。

         @note 漸化式を使って計算しているため、大きな値に対しては時間がかかる。
         @note この確率分布は中心極限定理により正規分布に近似することができる。
      **/
      class SignedRank : public Discrete
      {
        // 過去に求めた符号付き順位和に対する確率密度を保持する配列用 Map
        using ProbMap = std::map< var_type, stat_type >;
        using SRankMap = std::map< var_type, ProbMap >;
        mutable SRankMap buff_;

        var_type n_; // 値の総数(差がゼロの要素を除外する)

        // Pn( sRank ) の確率を求める
        stat_type calcP( var_type n, var_type sRank ) const;

        // Pn( sRank ) の確率をバッファから取り出す
        stat_type getP( var_type n, var_type sRank ) const;

        // U( m, n ) = u となる確率をバッファに登録する
        void setP( var_type n, var_type sRank, stat_type p ) const;

      public:

        /// @brief 値の総数 n を指定して構築
        ///
        /// 値の総数を実測値から取得する場合、二つの値の差がゼロの要素は除外する必要があることに注意。
        ///
        /// - n >= 0 である必要がある。そうでない場合は例外 exception::NegativeNumber< Discrete::var_type > を投げる。
        ///
        /// @param n 値の総数
        explicit SignedRank( var_type n )
          : n_( n )
        {
          const std::string FUNC_NAME = "tadah_fussy::statistics::distribution::SignedRank constructor";
          error_lib::NNNum( n_, FUNC_NAME + " : n " + error_message::NNNUM );
        }

        virtual stat_type average() const;  // 平均値
        virtual stat_type variance() const; // 分散値

        // Pn(sRank) となる確率を返す
        virtual stat_type operator[]( var_type sRank ) const;

      };

      /**
         @brief 各集団間の順位和の変動に対する確率分布

         複数の集団の要素の順位が全て互いに独立で等確率で発生するとき、各集団間の順位和の変動

         Sc = Σ( T[i]^2 / n[i] ) - T^2 / N

         n[i] : i 番目の集団の要素数\n
         T[i] : i 番目の集団の順位和\n
         N : 全要素数\n
         T : 全順位和 ( = N( N + 1 ) / 2 )

         を不偏分散

         σ^2 = N( N + 1 ) / 12

         割った値 H = Sc / σ^2 がとりうる確率密度を求める。

         H は要素数が多くなると自由度が ( 集団数 - 1 ) のカイ二乗分布に近似的に従う。

         @note 全要素数が多くなると処理は急激に重くなることに注意。10 個程度が目安。
         @note 離散確率分布 Discrete の派生クラスではない。平均・分散等を計算するメンバ関数は持たない。
         @note std::map の派生クラスであり、構築後は全ての確率変数に対する確率を保持した状態になる。
         通常は、構築後に反復子の開始から終了まで順番に値を取得して利用する。
         キーは浮動小数点数であり、近い値は同じキーであると見なされる
         ( 同値であると判断されるしきい値は構築時の引数 threshold でコントロールできる )。
      **/
      class H
        : public std::map< Base::stat_type, Base::stat_type, numeric_lib::ApproxLess< Base::stat_type > >
      {
        // 順位の二乗の平均を求めるための関数オブジェクト
        // 与えられた順列を使い、各集団ごとの順位の二乗平均を返す
        template< class T > class CalcRankSq;

      public:

        /// @brief 各集団の要素数を指定して構築
        ///
        /// - 各集団の要素数を保持する配列の指定範囲が有効かを内部でチェックしている。
        ///   有効でない場合は例外 exception::NegativeLength< std::iterator_traits< In >::difference_type > を投げる。
        /// - 要素数が負数のデータが見つかった場合は例外 exception::NegativeNumber< std::iterator_traits< In >::value_type > を投げる。
        /// - 要素数ゼロのデータは無視される。全要素数がゼロとなった場合は確率密度は生成されない。
        ///
        /// テンプレート引数の In はデータ列の反復子の型を表す。
        ///
        /// @param s 各集団の要素数を保持する配列の先頭位置
        /// @param e 各集団の要素数を保持する配列の末尾の次の位置
        /// @param threshold 確率変数に対して同値とみなす場合のしきい値
        template< class In >
        H( In s, In e, Base::stat_type threshold = DEFAULT_THRESHOLD );
      };

      /**
         @brief 各列(または行)の順位和の変動に対する確率分布

         行列の要素に対して各行(または列)ごとに順位付けを行った時、その順位が互いに独立で等確率で発生するとして、各列(または行)間の変動

         Sg = Σ( T[i]^2 - n( k + 1 ) / 2 ) / n

         T[i] : 列(または行) i の順位和\n
         n : 行(または列)数\n
         k : 列(または行)数

         を不偏分散

         σ^2 = N( N + 1 ) / 12

         割った値 Sg/σ^2 がとりうる確率密度を求める。

         Sg/σ^2 は要素数が多くなると自由度が ( 集団数 - 1 ) のカイ二乗分布に近似的に従う。

         @note 全要素数が多くなると処理は急激に重くなることに注意。
         @note 離散確率分布 Discrete の派生クラスではない。平均・分散等を計算するメンバ関数は持たない。
         @note std::map の派生クラスであり、構築後は全ての確率変数に対する確率を保持した状態になる。
         通常は、構築後に反復子の開始から終了まで順番に値を取得して利用する。キーは浮動小数点数
         であり、近い値は同じキーであると見なされる ( 同値であると判断されるしきい値は構築時の
         引数 threshold でコントロールできる )。
      **/
      class Sg
        : public std::map< Base::stat_type, Base::stat_type, numeric_lib::ApproxLess< Base::stat_type > >
      {
        // 指定された順列から順位を返す関数オブジェクト
        class GetPerm
        {
        public:

          // 集団に対する添字の型
          using class_index_type = std::vector< std::vector< Discrete::var_type > >::size_type;
          // 集団内のデータに対する添字の型
          using data_index_type = std::vector< Discrete::var_type >::size_type;

        private:

          const std::vector< std::vector< Discrete::var_type > >& permList_; // 順列のリスト
          data_index_type dataIdx_; // 各集団の何番目の要素が対象かを示すインデックス

        public:

          // 順列リストと要素の位置を指定して構築
          GetPerm( const std::vector< std::vector< Discrete::var_type > >& permList, data_index_type dataIdx )
            : permList_( permList ), dataIdx_( dataIdx ) {}

          // 指定した集団の順位を返す
          Discrete::var_type operator()( class_index_type classIdx )
          { return( permList_[classIdx][dataIdx_] ); }
        };

        // 順列のリスト permList 内の、各集団の順列に対する添字のリスト classIdx で
        // 指定された順列の各要素毎の和を計算し、その二乗和を求める
        Base::stat_type calcRankSqSum( const std::vector< std::vector< Discrete::var_type > >& permList,
                                       const std::vector< GetPerm::class_index_type >& classIdx );

        // classIdx を順列の次の組み合わせに更新する
        bool nextClassIdx( std::vector< GetPerm::class_index_type >::size_type index,
                           GetPerm::class_index_type maxPerm, std::vector< GetPerm::class_index_type >* classIdx );

      public:

        /// @brief 各集団の要素数を指定して構築
        ///
        /// - dataCnt >= 0、classCnt >= 0 である必要がある。そうでない場合は例外 ExceptionNegativeNumber< Discrete::var_type > を投げる。
        ///
        /// @param dataCnt 集団ごとの要素数
        /// @param classCnt 集団数
        /// @param threshold 確率変数に対して同値とみなす場合のしきい値
        Sg( Discrete::var_type dataCnt, Discrete::var_type classCnt,
            Base::stat_type threshold = DEFAULT_THRESHOLD );
      };

      /* ----------------------------
         連続確率分布
         ---------------------------- */

      /**
         @brief コーシー分布

         p( x ; μ,γ ) = 1 / πγ{ 1 + [ ( x - μ ) / γ ]^2 } = γ / π[ ( x - μ )^2 + γ^2 ]

         μ : 位置パラメータ ; γ : 尺度パラメータ
      **/
      class Cauchy : public Continuous
      {
        stat_type m_; // 位置パラメータ
        stat_type r_; // 尺度パラメータ

      public:

        /// @brief 位置・尺度パラメータを指定して構築
        ///
        /// - 尺度パラメータの条件 r > 0 を満たす必要がある。そうでない場合は例外 exception::NotPositiveNumber< Base::stat_type > を投げる。
        ///
        /// @param m 位置パラメータ
        /// @param r 尺度パラメータ ( r > 0 )
        explicit Cauchy( stat_type m = 0, stat_type r = 1 )
          : m_( m ), r_( r )
        {
          const std::string FUNC_NAME = "tadah_fussy::statistics::distribution::Cauchy constructor";
          error_lib::PNum( r_, FUNC_NAME + " : r " + error_message::PNUM );
        }

        // 確率変数 x における確率密度を返す
        virtual stat_type operator[]( stat_type x ) const;

        // 区間 ( -∞, a ] における確率を返す
        virtual stat_type lower_p( stat_type a ) const;

        // 区間  [ a, b ] または [ b, a ] における確率を返す
        virtual stat_type p( stat_type a, stat_type b ) const;

        virtual stat_type average() const;  // 平均値
        virtual stat_type variance() const; // 分散
      };

      /**
         @brief 指数分布

         P( t ; λ ) = λe^(-λt)

         但し、t < 0 に対しては P( t ; λ ) = 0 とする。
      **/
      class Exponential : public Continuous
      {
        stat_type lambda_; // 定数 λ

      public:

        /// @brief 定数 λ を指定して構築
        ///
        /// - 定数 λ の条件 λ > 0 を満たす必要がある。そうでない場合は例外 exception::NotPositiveNumber< Base::stat_type > を投げる。
        ///
        /// @param lambda 定数 λ ( λ > 0 )
        explicit Exponential( stat_type lambda )
          : lambda_( lambda )
        {
          const std::string FUNC_NAME = "tadah_fussy::statistics::distribution::Exponential constructor";
          error_lib::PNum( lambda_, FUNC_NAME + " : lambda " + error_message::PNUM );
        }

        virtual stat_type average() const;  // 平均値
        virtual stat_type variance() const; // 分散


        /// @brief 確率変数 t における確率密度を返す
        ///
        /// t < 0 のときは常にゼロを返す。
        ///
        /// @param t 確率変数
        /// @return 確率変数 t における確率密度
        virtual stat_type operator[]( stat_type t ) const;

        /// @brief 区間 (-∞,a] における確率を返す
        ///
        /// a < 0 のときは常にゼロを返す。
        ///
        /// @param a 区間の最大値
        /// @return 区間 ( -∞, a ] における確率
        virtual stat_type lower_p( stat_type a ) const;
      };

      /**
         @brief 一様分布

         P( t ; a,b ) = 1 / ( b - a ) ( a ≦ t ≦ b ) ; 0 (それ以外)
      **/
      class Uniform : public Continuous
      {
        // 確率変数の定義域
        stat_type a_;
        stat_type b_;

      public:

        /// @brief 確率変数の定義域を指定して構築
        ///
        /// a > b ならば定義域は [ b, a ] に自動変換する。
        ///
        /// @param a 確率変数の定義域 1
        /// @param b 確率変数の定義域 2
        Uniform( stat_type a, stat_type b )
          : a_( ( a > b ) ? b : a ), b_( ( a > b ) ? a : b ) {}

        virtual stat_type average() const; // 平均値
        virtual stat_type variance() const; // 分散

        /// @brief 確率変数 x における確率密度を返す
        ///
        /// a = b のとき、x = a ならば ProbDist::infinity() を返す。
        ///
        /// @param x 確率変数
        /// @return 確率変数 x における確率密度
        virtual stat_type operator[]( stat_type x ) const;

        // 区間 (-∞,a] における確率を返す
        virtual stat_type lower_p( stat_type a ) const;
      };

      /**
         @brief 正規分布

         N( μ, σ^2 ) = { 1 / sqrt( 2π )σ } exp( -( t - μ )^2 / 2σ^2 )

         μ : 平均 , σ^2 : 分散
      **/
      class Normal : public Continuous
      {
        stat_type mu_;  // 平均
        stat_type var_; // 分散

      public:

        /// 平均・標準偏差を指定して構築
        ///
        /// @note 二番目の引数は分散ではなく標準偏差であることに注意。誤って負数を入力しても成り立つようにするためだが、誤りに気が付かない恐れもある。
        ///
        /// @param mu 平均
        /// @param sigma 標準偏差
        Normal( stat_type mu, stat_type sigma )
          : mu_( mu ), var_( std::pow( sigma, 2.0 ) ) {}

        // 確率変数 x における確率密度を返す
        virtual stat_type operator[]( stat_type x ) const;

        // 区間 (-∞,a] における確率を返す
        virtual stat_type lower_p( stat_type a ) const;

        // 区間 [a,b] における確率を返す
        virtual stat_type p( stat_type a, stat_type b ) const;

        virtual stat_type average() const;  // 平均値
        virtual stat_type variance() const; // 分散
      };

      /**
         @brief カイ二乗分布

         Tn(t) = { 1 / 2^(n/2)Γ(n/2) } t^(n/2-1) e^(-t/2)

         n : 自由度

         t < 0 のときは Tn(t) = 0 になる。
      **/
      class ChiSquare : public Continuous
      {
        Base::counter_type n_;  // 自由度

      public:

        /// 自由度を指定して構築
        ///
        /// - 自由度 n の条件 n > 0 を満たす必要がある。
        ///   そうでない場合は例外 exception::NotPositiveNumber< Base::counter_type > を投げる。
        ///
        /// @param n 自由度
        explicit ChiSquare( Base::counter_type n )
          : n_( n )
        {
          const std::string FUNC_NAME = "tadah_fussy::statistics::distribution::ChiSquare constructor";
          error_lib::PNum( n_, FUNC_NAME + " : n " + error_message::PNUM );
        }

        /// @brief 確率変数 x における確率密度を返す
        ///
        /// 自由度が 1 の場合、x = 0 における確率密度として Base::infinity() を返す。
        ///
        /// @param x 確率変数
        /// @return 確率変数 x における確率密度
        virtual stat_type operator[]( stat_type x ) const;

        // 区間 (-∞,a] における確率を返す
        virtual stat_type lower_p( stat_type a ) const;

        virtual stat_type average() const;  // 平均値
        virtual stat_type variance() const; // 分散
      };

      /**
         @brief F-分布

         Gm,n(t) = m^(m/2) n^(n/2) t^(m/2-1) / B( m/2, n/2 )( mt + n )^((m+n)/2)

         m, n : 自由度
      **/
      class F : public Continuous
      {
        Base::counter_type m_;  // 自由度 M
        Base::counter_type n_;  // 自由度 N

      public:

        /// @brief 自由度を指定して構築
        ///
        /// - 自由度 m, n の条件 m > 0, n > 0 を満たす必要がある。
        ///   そうでない場合は例外 exception::NotPositiveNumber< Base::counter_type > を投げる。
        ///
        /// @param m 自由度 1
        /// @param n 自由度 2
        F( Base::counter_type m, Base::counter_type n )
          : m_( m ), n_( n )
        {
          const std::string FUNC_NAME = "tadah_fussy::statistics::distribution::F constructor";

          error_lib::PNum( m_, FUNC_NAME + " : m " + error_message::PNUM );
          error_lib::PNum( n_, FUNC_NAME + " : n " + error_message::PNUM );
        }

        // 確率変数 x における確率密度を返す
        virtual stat_type operator[]( stat_type x ) const;

        // 区間 (-∞,a] における確率を返す
        virtual stat_type lower_p( stat_type a ) const;

        virtual stat_type average() const;  // 平均値
        virtual stat_type variance() const; // 分散
      };

      /**
         @brief t-分布

         fn(t) = { 1 / √n B( 1 / 2, n / 2 ) } ( 1 + t^2 / n )^(-(n+1)/2)

         n : 自由度
      */
      class T : public Continuous
      {
        Base::counter_type n_;  // 自由度 N

      public:

        /// @brief 自由度を指定して構築
        ///
        /// - 自由度 n の条件 n > 0 を満たす必要がある。
        ///   そうでない場合は例外 exception::NotPositiveNumber< Base::counter_type > を投げる。
        ///
        /// @param n 自由度
        explicit T( Base::counter_type n )
          : n_( n )
        {
          const std::string FUNC_NAME = "tadah_fussy::statistics::distribution::T constructor";
          error_lib::PNum( n_, FUNC_NAME + " : n " + error_message::PNUM );
        }

        // 確率変数 x における確率密度を返す
        virtual stat_type operator[]( stat_type x ) const;

        // 区間 (-∞,a] における確率を返す
        virtual stat_type lower_p( stat_type a ) const;

        virtual stat_type average() const;  // 平均値
        virtual stat_type variance() const; // 分散
      };

      /************************************************************************
         多変数確率分布用クラス
      ************************************************************************/

      /**
         @brief 多変数確率分布用インターフェース

         確率変数は多変数となるため、std::vector を使って渡すように定義している。
         その型は MultiBase::array_type、添字の型は MultiBase::size_type と表現できる。
      **/
      template< class T >
      struct MultiBase
      {
        /// @brief 確率や統計量に対する型
        using stat_type = Base::stat_type;

        /// @brief 確率変数の配列に対する型
        using array_type = std::vector< T >;

        /// @brief 確率変数の添字に対する型
        using size_type = typename array_type::size_type;

        /// @brief 確率や統計量の無効値(NaN)
        static stat_type nan()
        { return( Base::nan() ); }

        /// @brief 確率や統計量の無限大
        static stat_type infinity()
        { return( Base::infinity() ); }

        /// @brief 仮想デストラクタ(何もしない)
        virtual ~MultiBase() {}

        /// @brief 平均値を求める(純粋仮想関数)
        ///
        /// @param i 対象の確率変数の添字
        /// @return 平均値
        virtual stat_type average( size_type i ) const = 0;

        /// @brief 分散を求める(純粋仮想関数)
        ///
        /// @param i 対象の確率変数の添字
        /// @return 分散
        virtual stat_type variance( size_type i ) const = 0;

        /// @brief 共分散を求める(純粋仮想関数)
        ///
        /// @param i 対象の確率変数の添字 1
        /// @param j 対象の確率変数の添字 2
        /// @return 共分散
        virtual stat_type covariance( size_type i, size_type j ) const = 0;

        /// @brief 確率変数の数(次元数)を返す(純粋仮想関数)
        ///
        /// @return 確率変数の数(次元数)
        virtual size_type size() const = 0;
      };

      /**
         @brief 多変数離散分布用インターフェース
      **/
      class MultiDiscrete : public MultiBase< Discrete::var_type >
      {
      public:

        /// @brief 確率変数に対する型
        using var_type = Discrete::var_type;

        /// @brief 確率変数ベクトル ( r1, r2, ... ) に対する同時確率密度を返す(純粋仮想関数)
        ///
        /// @param r 確率変数ベクトル
        /// @return 同時確率密度
        virtual stat_type operator[]( const array_type& r ) const = 0;

        /// @brief 区間 ( [s1,e1], [s2,e2] ... ) における確率を返す
        ///
        /// @param s 区間の始点を要素とするベクトル
        /// @param e 区間の終点を要素とするベクトル
        /// @return 区間に対する確率
        stat_type p( array_type s, array_type e ) const;

      private:

        // 確率を求める再起処理用サブルーチン
        //
        // p : 現在の確率変数ベクトル
        // s, e : 和を求める確率変数の範囲( s:開始 e:終了 )
        // curIdx : 和を求める範囲の開始位置( curIdx以降の和を求める )
        stat_type getProb( array_type& p, const array_type& s, const array_type& e, size_type curIdx ) const;
      };

      /**
         @brief 多変数連続分布用インターフェース
      **/
      struct MultiContinuous : public MultiBase< Base::stat_type >
      {
        /// @brief 確率変数ベクトル ( a1, a2, ... ) における同時確率密度を返す(純粋仮想関数)
        ///
        /// @param a 確率変数ベクトル
        /// @return 同時確率密度
        virtual stat_type operator[]( const array_type& a ) const = 0;

        /// @brief 区間 ( [a1,b1], [a2,b2], ... ) における確率を返す(純粋仮想関数)
        ///
        /// @param a 区間の端点を要素とするベクトル 1
        /// @param b 区間の端点を要素とするベクトル 2
        /// @return 区間に対する確率
        virtual stat_type p( const array_type& a, const array_type& b ) const = 0;
      };

      /**
         @brief 多項分布

         P(r1,r2,...rk) = ( n! / r1!r2!...rk! ) p1^r1 p2^r2 ... pk^rk

         ri : i 番目の事象の発生回数 <br>
         pi : i 番目の事象が発生する確率 <br>
         n  : 事象の発生回数の和

         i 番目の事象が発生する確率が pi であるとき、n 回の試行のうち各事象が ri 回発生した時の確率を表す。

         任意の i に対して ri >= 0 かつ pi >= 0 であり、pi の合計は 1 以下でなければならない
         ( 最後の事象は余事象として扱い、その確率は 1 - Sum( s, e ) として内部で求められる )。
         また、n > 0 である必要がある。
      **/
      class Multinomial : public MultiDiscrete
      {
        std::vector< stat_type > p_; // 各事象の発生確率(要素数が事象の数 k を表す)
        var_type n_;                 // 試行回数の合計

      public:

        /// @brief 各事象の発生確率と試行回数の合計を指定して実行
        ///
        /// - テンプレート引数の In はデータ列の反復子の型を表す。
        /// - 各事象の発生確率のデータ列の指定範囲が有効かを内部でチェックしている。
        ///   有効でない場合は例外 exception::NegativeLength< std::iterator_traits< In >::difference_type > を投げる。
        /// - n > 0 を満たす必要がある。そうでない場合は例外 exception::NotPositiveNumber< MultiDiscrete::var_type > を投げる。
        /// - 各事象の発生確率 [ s, e ) は全て 0 以上 1 以下である必要がある。また最後の事象は余事象として扱い、
        ///   その確率は 1 - MathLib::Sum( s, e ) として内部で求められるため、合計は 1 以下でなければならない。
        ///   そうでない場合、例外 exception::UnitInterval を投げる。
        ///
        /// @param s 各事象の発生確率のデータ列 開始位置
        /// @param e 各事象の発生確率のデータ列 末尾の次の位置
        /// @param n 試行回数の合計
        template< class In >
        Multinomial( In s, In e, var_type n );

        /// @brief 確率変数 r に対する確率密度を返す
        ///
        /// r が不正な状態だった場合は常に MultiBase::nan() を返す。
        ///
        /// @param r 確率変数ベクトル(余事象は含まない)
        /// @retval 同時確率密度
        /// @retval rの要素数が余事象を除く事象の数と等しくない場合はnanを返す
        /// @retval rの合計が試行回数より大きい場合はnanを返す
        /// @retval rに負数が含まれていた場合はnanを返す
        virtual stat_type operator[]( const array_type& r ) const;

        /// @brief 平均値を求める
        ///
        /// 添字が不正な場合は常に MultiBase::nan() を返す
        ///
        /// @param i 対象の確率変数の添字
        /// @return 平均値
        virtual stat_type average( size_type i ) const;

        /// @brief 分散を求める
        ///
        /// 添字が不正な場合は常に MultiBase::nan() を返す
        ///
        /// @param i 対象の確率変数の添字
        /// @return 分散
        virtual stat_type variance( size_type i ) const;

        /// @brief 共分散を求める
        ///
        /// 添字が不正な場合は常に MultiBase::nan() を返す
        ///
        /// @param i 対象の確率変数の添字 1
        /// @param j 対象の確率変数の添字 2
        /// @return 共分散
        virtual stat_type covariance( size_type i, size_type j ) const;

        /// @brief 確率変数の数を返す
        ///
        /// 余事象の発生確率と発生回数は他の事象から求めることができるため、次元は返り値より 1 小さくなる。
        ///
        /// @return 余事象を含めた確率変数の数
        virtual size_type size() const;
      };

      /**
         @brief 負の多項分布

         P(s,r1,r2,...rk) = { ( Σri + s - 1)! p0^s / ( s - 1 )! } Π( pi^ri / ri! )

         ri : i 番目の事象の発生回数
         pi : i 番目の事象が発生する確率
         s  : どの事象も発生しない(余事象の)回数
         p0 : どの事象も発生しない(余事象の)確率

         負の多項分布は余事象が s 回発生するまでの各事象の回数に対する確率であるため、余事象は必ず最低 1 回は発生する。
         従って、s と p0 は必ずゼロより大きい必要があり、そうでなければ成り立たない。
      **/
      class NegativeMultinomial : public MultiDiscrete
      {
        std::vector< stat_type > p_; // 各事象の発生回数(要素数が事象の数 k を表す)
        stat_type p0_;               // 事象の発生しない確率
        var_type s_;                 // 事象の発生しない回数

        void check(); // 確率密度関数として有効かをチェックする

      public:

        /// 各事象の発生確率と、どの事象も発生しない回数を指定して実行
        ///
        /// - テンプレート引数の In はデータ列の反復子の型を表す。
        /// - 各事象の発生確率のデータ列の指定範囲が有効かを内部でチェックしている。
        ///   有効でない場合は例外 exception::NegativeLength< std::iterator_traits< In >::difference_type > を投げる。
        /// - 余事象の発生回数 s は必ず 1 以上でなければならない。
        ///   そうでない場合は例外 exception::NotPositiveNumber< MultiDiscrete::var_type > を投げる。
        /// - 各事象の発生確率 [ ps, pe ) は全て 0 以上 1 以下である必要がある。そうでない場合、例外 exception::UnitInterval を投げる。
        /// - 余事象の発生確率は 1 - math_lib::Sum( ps, pe ) として内部で求められる。
        ///   この値も 0 以上 1 以下である必要があり、そうでない場合は例外 exception::UnitInterval を投げる。
        ///   また、余事象の発生確率がゼロの場合は前提条件が成り立たないため、発生確率の合計 Sum( ps, pe ) は 1 より小さくなければならない。
        ///   そうでない場合は例外 exception::NotPositiveNumber< MultiBase::stat_type > を投げる。
        ///
        /// @param ps 各事象の発生確率のデータ列 開始位置
        /// @param pe 各事象の発生確率のデータ列 末尾の次の位置
        /// @param s どの事象も発生しない回数
        template< class In >
        NegativeMultinomial( In ps, In pe, var_type s );

        /// @brief 確率変数 r に対する確率密度を返す
        ///
        /// @param r 確率変数ベクトル(余事象は含まない)
        /// @retval 同時確率密度
        /// @retval rの要素数が余事象を除く事象の数と等しくない場合はnanを返す
        /// @retval rに負数が含まれていた場合はnanを返す
        stat_type operator[]( const array_type& r ) const;

        /// @brief 平均値を求める
        ///
        /// 添字が不正な場合は常に MultiBase::nan() を返す
        ///
        /// @param i 対象の確率変数の添字
        /// @return 平均値
        stat_type average( size_type i ) const;

        /// @brief 分散を求める
        ///
        /// 添字が不正な場合は常に MultiBase::nan() を返す
        ///
        /// @param i 対象の確率変数の添字
        /// @return 分散
        stat_type variance( size_type i ) const;

        /// @brief 共分散を求める
        ///
        /// 添字が不正な場合は常に MultiBase::nan() を返す
        ///
        /// @param i 対象の確率変数の添字 1
        /// @param j 対象の確率変数の添字 2
        /// @return 共分散
        stat_type covariance( size_type i, size_type j ) const;

        /// @brief 確率変数の数(次元数)を返す
        ///
        /// @return 余事象を除いた確率変数の数(次元数)
        size_type size() const;
      };

    } // namespace distribution

    /************************************************************************
       ヒストグラム作成
    ************************************************************************/

    /**
       @brief 度数データ計算用クラス

       テンプレート引数の T は階級の型を、Frequency は度数の型をを表す。

       std::map< T, Frequency > からの公開派生であり std::map< T, Frequency > のメンバは全て利用できる
    **/
    template< class T, class Frequency >
    struct Histogram : public std::map< T, Frequency >
    {
      /// @brief 階級の型
      using class_type = T;
      /// @brief 度数の型
      using frequency_type = Frequency;

      /// @brief データ列 [ s, e ) から度数を計算する
      ///
      /// - テンプレート引数の In はデータ列の反復子の型を表す。
      /// - データ列の指定範囲が有効かを内部でチェックしている。
      ///   有効でない場合は例外 exception::NegativeLength< std::iterator_traits< In >::difference_type > を投げる。
      /// - 度数のリセットはしないため、前回計算した結果はそのまま反映される。
      /// - setNewData が true なら、新規に見つかった階級は自動的に追加されるのに対し、false のときは無視される。
      ///
      /// @param s データ列の先頭の反復子
      /// @param e データ列の末尾の次の反復子
      /// @param setNewData 階級のないデータが見つかったら階級を新たに追加する
      /// @return なし
      template< class In > void accumulate( In s, In e, bool setNewData );
    };

    namespace histogram
    {
      /**
         @brief 範囲を表すクラス

         テンプレート引数の T は範囲を表す数値の型。
         階級に値の範囲を利用したい場合に用いる。lower は最小値で範囲内になるのに対し、
         upper は最大値だが範囲外として扱う。従って [ lower, upper ) の範囲として表される。
      **/
      template< class T >
      struct RangeClass
      {
        T lower; ///< @brief 範囲の最小値(範囲内)
        T upper; ///< @brief 範囲の最大値(範囲外)

        /// @brief 範囲の最小値・最大値を指定して構築
        ///
        /// 二つの値を指定し、大小関係を自動で判断して最小・最大値を決める
        ///
        /// @param t1 範囲を表す値 1
        /// @param t2 範囲を表す値 2
        RangeClass( const T& t1, const T& t2 )
          : lower( std::min( t1, t2 ) ), upper( std::max( t1, t2 ) ) {}

        /// @brief 値を一つ指定して構築
        ///
        /// 範囲の幅はゼロになる
        ///
        /// @param t 対象の値
        RangeClass( const T& t )
          : lower( t ), upper( t ) {}

        /// @brief 大小判定用関数
        ///
        /// 自分の方が小さければ true を返す。
        /// 自分の範囲の最大値と相手の範囲の最小値を比較してその大小を判定する
        ///
        /// @note 自分が範囲に幅を持っている場合(一つの値を指定しているのではない場合)、
        ///       自分の範囲の最大値と相手の範囲の最小値が等しい場合も自分の方が小さいと判断される。
        /// @note 範囲に重複がある場合は等しいと判断されることに注意(どちらを左側にしても false になる)。
        ///
        /// @param dest 対象の範囲
        bool operator<( const RangeClass& dest ) const
        {
          if ( lower == upper ) // 自分が一つの値のみ指定している場合
            return( upper < dest.lower );

          return( upper <= dest.lower );
        }
      };

      /**
         @brief 階級名を文字列に変換して返す関数オブジェクト

         任意の値を整形して文字列として変換する。
         RangeClass に対しても専用の処理関数が定義されている。
      **/
      struct ClassName
      {
        /// @brief 文字列成形用フォーマットを保持する出力用文字列ストリーム
        ///
        /// 文字列を任意に整形したい場合はこのストリームに設定する
        std::ostringstream classFormat;

        /// @brief 値を文字列にして返す
        ///
        /// 値は classFormat の定義に従って整形される。
        ///
        /// @param value 変換対象の値
        /// @return 変換後の文字列
        template< class T >
        std::string operator()( T value ) const;

        /// @brief 値を文字列にして返す(RangeClass用)
        ///
        /// 範囲値は classFormat の定義に従って整形される。
        ///
        /// @param value 変換対象の値
        /// @return 変換後の文字列
        template< class T >
        std::string operator()( RangeClass< T > value ) const;
      };

      /**
         @brief ヒストグラムを出力する関数オブジェクト

         Histogram の度数を階級名順に出力する。
      **/
      template< class T, class Frequency >
      class Print
      {
      protected:

        /// @brief ヒストグラムの各階級にアクセスするための反復子
        using const_iterator = typename Histogram< T, Frequency >::const_iterator;

        /// @brief 度数と階級名の文字数の最大値を取得する
        ///
        /// 反復子 cit にある階級名・度数が現在の最大値より大きければ更新する。
        ///
        /// @param cit ヒストグラムの現在の階級位置(反復子)
        /// @param maxFreq 最大度数を取得する変数へのポインタ
        /// @param maxClassLen 階級名の文字数最大値を取得する変数へのポインタ
        /// @return なし
        void getMaxSize( const_iterator cit, std::streamsize* maxFreq, std::string::size_type* maxClassLen ) const;

        /// @brief ヒストグラム(一つの階級に対する度数グラフ)を文字列に変換して返す
        ///
        /// @param name 階級名
        /// @param freq 度数
        /// @param unitSize ヒストグラムの高さ
        /// @param maxClassLen 階級名の文字数最大値
        /// @return 変換後の文字列
        std::string classData( const T& name, const Frequency& freq, std::streamsize unitSize, std::string::size_type maxClassLen ) const;

      public:

        /// @brief 階級名を出力する関数オブジェクト
        ///
        /// 階級名を整形するために利用する。
        ClassName className;

        /// @brief ヒストグラムを出力する
        ///
        /// @param hist 対象のヒストグラムへの参照
        /// @param os 出力先ストリーム
        /// @param maxRange ヒストグラム高さの最大値
        /// @return なし
        void operator()( const Histogram< T, Frequency >& hist, std::ostream& os, std::streamsize maxRange = 40 ) const;
      };

      /**
         @brief パレート図を出力する関数オブジェクト

         Histogram の度数を度数の高い順に出力する。
      **/
      template< class T, class Frequency >
      class PrintPareto : Print< T, Frequency >
      {
        // ヒストグラムの各階級にアクセスするための反復子
        using const_iterator = typename Print< T, Frequency >::const_iterator;

        // 頻度順に並べた結果を保持する multimap
        using pareto_type = typename std::multimap< Frequency, T >;

        // 度数(maxFreq)と階級名の文字数(maxClassLen)の最大値を取得し、度数でソートしたコンテナ(pareto)を作成する
        void getMaxSize( const_iterator cit, std::streamsize* maxFreq,
                         std::string::size_type* maxClassLen, pareto_type* pareto ) const;

      public:

        /// @brief ヒストグラムを出力する
        ///
        /// @param hist 対象のヒストグラムへの参照
        /// @param os 出力先ストリーム
        /// @param maxRange ヒストグラム高さの最大値
        /// @return なし
        void operator()( const Histogram< T, Frequency >& hist, std::ostream& os, std::streamsize maxRange = 40 ) const;
      };

    } // namespace histogram

    /************************************************************************
        要約統計量オブジェクト
    ************************************************************************/

    /**
       @brief 代表的な要約統計量 ( モーメント 1 まで ) を保持するオブジェクト

       要約統計量 ( データ数・総和 ) をデータ列から計算し、
       保持しておくためのオブジェクト

       テンプレート引数の T は総和の型を表す。
    **/
    template< class T >
    class BaseMoment1
    {
    protected:

      distribution::Base::counter_type count_; ///< データ数
      T sum_;                                  ///< 総和

      /// @brief デストラクタ
      ///
      /// デストラクタのアクセス指定子は protected であり、派生クラスしか呼び出しができない。
      /// そのため、BaseMoment1 自身を構築することはできず、利用できるのはその派生クラスのみとなる。
      ~BaseMoment1() {}

    public:

      /// @brief 要素の型
      using value_type = T;

      /// @brief デフォルト・コンストラクタ
      BaseMoment1() : count_( 0 ), sum_( value_type() ) {}

      /// @brief データ列を指定して構築
      ///
      /// テンプレート引数の In はデータ列の反復子の型を表す。
      ///
      /// - データ列の指定範囲が有効かを内部でチェックしている。
      ///   有効でない場合は例外 exception::NegativeLength< std::iterator_traits< In >::difference_type > を投げる。
      ///
      /// @param s 対象のデータ列の開始位置の反復子
      /// @param e 対象のデータ列の末尾の次の位置の反復子
      template< class In >
      BaseMoment1( In s, In e );

      /// @brief データ数を返す
      ///
      /// @return データ数
      distribution::Base::counter_type count() const
      { return( count_ ); }

      /// @brief データの総和を返す
      ///
      /// @return データの総和
      value_type sum() const
      { return( sum_ ); }
    };

    /**
       @brief 代表的な要約統計量 ( モーメント 1 まで ) を保持するオブジェクト

       要約統計量 ( データ数・総和 ) をデータ列から計算し、
       保持しておくためのオブジェクト
    **/
    struct Moment1 : public BaseMoment1< distribution::Base::stat_type >
    {
      /// @brief デフォルト・コンストラクタ
      Moment1() : BaseMoment1< value_type >() {}

      /// @brief データ列を指定して構築
      ///
      /// テンプレート引数の In はデータ列の反復子の型を表す。
      ///
      /// - データ列の指定範囲が有効かを内部でチェックしている。
      ///   有効でない場合は例外 exception::NegativeLength< std::iterator_traits< In >::difference_type > を投げる。
      ///
      /// @param s 対象のデータ列の開始位置の反復子
      /// @param e 対象のデータ列の末尾の次の位置の反復子
      template< class In >
      Moment1( In s, In e )
        : BaseMoment1< value_type >( s, e ) {}

      /// @brief データの標本平均を返す
      ///
      /// データ数がゼロなら NaN を返す。
      ///
      /// @return データの標本平均
      value_type mean() const
      { return( ( count() == 0 ) ? distribution::Base::nan() : sum() / count() ); }

      /// @brief 他の要約統計量(データ数・総和)を加算する
      ///
      /// @param dest 加算する要約統計量への参照
      /// @return 自分自身への参照
      Moment1& operator+=( const Moment1& dest )
      {
        count_ += dest.count_;
        sum_ += dest.sum_;

        return( *this );
      }
    };

    /// @brief 二つの要約統計量 ( モーメント 1 ) を加算する
    ///
    /// @param m1 加算する要約統計量 1 への参照
    /// @param m2 加算する要約統計量 2 への参照
    /// @return 加算した結果
    inline Moment1 operator+( const Moment1& m1, const Moment1& m2 )
    {
      Moment1 res( m1 );
      return( res += m2 );
    }

    /**
       @brief 代表的な要約統計量 ( モーメント 1 まで ) のペアを保持するオブジェクト

       二種類の要約統計量 ( データ数・総和 ) をデータ列から同時に計算し、
       保持しておくためのオブジェクト
    **/
    class MomentPair1 : public BaseMoment1< std::pair< distribution::Base::stat_type, distribution::Base::stat_type > >
    {
      using stat_type = distribution::Base::stat_type;

    public:

      /// @brief デフォルト・コンストラクタ
      MomentPair1() : BaseMoment1< value_type >() {}

      /// @brief データ列を指定して構築
      ///
      /// テンプレート引数の In1, In2 はそれぞれ二つのデータ列の反復子の型を表す。
      ///
      /// - データ列 1 の指定範囲が有効かを内部でチェックしている。
      ///   有効でない場合は例外 exception::NegativeLength< std::iterator_traits< In1 >::difference_type > を投げる。
      /// - データ列 2 の指定範囲が有効かを内部でチェックしている。
      ///   有効でない場合は例外 exception::NegativeLength< std::iterator_traits< In2 >::difference_type > を投げる。
      /// - 二つのデータ列の長さが等しいかを内部でチェックしている。
      ///   等しくない場合は例外 exception::NotEqualLength< std::iterator_traits< In1 >::difference_type > を投げる。
      ///
      /// @param s1 対象のデータ列1の開始位置の反復子
      /// @param e1 対象のデータ列1の末尾の次の位置の反復子
      /// @param s2 対象のデータ列2の開始位置の反復子
      /// @param e2 対象のデータ列2の末尾の次の位置の反復子
      template< class In1, class In2 >
      MomentPair1( In1 s1, In1 e1, In2 s2, In2 e2 );

      /// @brief データ列を指定して構築
      ///
      /// テンプレート引数の In はデータ列の反復子の型を表す。
      ///
      /// - データ列の指定範囲が有効かを内部でチェックしている。
      ///   有効でない場合は例外 exception::NegativeLength< std::iterator_traits< In >::difference_type > を投げる。
      ///
      /// @note データ列は std::pair 型が要素であることを前提としている。
      ///
      /// @param s 対象のデータ列の開始位置の反復子
      /// @param e 対象のデータ列の末尾の次の位置の反復子
      template< class In >
      MomentPair1( In s, In e )
        : BaseMoment1< value_type >( s, e ) {}

      /// @brief データの標本平均を返す
      ///
      /// データ数がゼロなら NaN を返す。
      ///
      /// @return データの標本平均
      value_type mean() const
      { return( value_type( mean1(), mean2() ) ); }

      /// @brief データの 1 番目のパラメータの総和を返す
      ///
      /// @return データの 1 番目のパラメータの総和
      distribution::Base::stat_type sum1() const
      { return( sum().first ); }

      /// @brief データの 2 番目のパラメータの総和を返す
      ///
      /// @return データの 2 番目のパラメータの総和
      distribution::Base::stat_type sum2() const
      { return( sum().second ); }

      /// @brief データの 1 番目のパラメータの標本平均を返す
      ///
      /// データ数がゼロなら NaN を返す。
      ///
      /// @return データの 1 番目のパラメータの標本平均
      distribution::Base::stat_type mean1() const
      {
        return( ( count() == 0 ) ?
                distribution::Base::nan() : sum1() / count()
                );
      }

      /// @brief データの 2 番目のパラメータの標本平均を返す
      ///
      /// データ数がゼロなら NaN を返す。
      ///
      /// @return データの 2 番目のパラメータの標本平均
      distribution::Base::stat_type mean2() const
      {
        return( ( count() == 0 ) ?
                distribution::Base::nan() : sum2() / count()
                );
      }

      /// @brief 他の要約統計量(データ数・総和)を加算する
      ///
      /// @param dest 加算する要約統計量への参照
      /// @return 自分自身への参照
      MomentPair1& operator+=( const MomentPair1& dest )
      {
        count_ += dest.count_;
        sum_.first += dest.sum_.first;
        sum_.second += dest.sum_.second;

        return( *this );
      }
    };

    /// @brief 二つの要約統計量 ( モーメント 1 ) を加算する
    ///
    /// @param m1 加算する要約統計量 1 への参照
    /// @param m2 加算する要約統計量 2 への参照
    /// @return 加算した結果
    inline MomentPair1 operator+( const MomentPair1& m1, const MomentPair1& m2 )
    {
      MomentPair1 res( m1 );
      return( res += m2 );
    }

    /**
       @brief 代表的な要約統計量 ( モーメント 2 まで ) を保持するオブジェクト

       要約統計量 ( データ数・総和・偏差の平方和 ) をデータ列から計算し、
       保持しておくためのオブジェクト
    **/
    template< class MOMENT1 >
    class BaseMoment2
    {
    public:

      /// @brief 要素の型
      using value_type = typename MOMENT1::value_type;

    protected:

      MOMENT1 moment1_;  ///< 要約統計量 ( モーメント 1 )
      value_type sumSq_; ///< 平方和

      /// @brief デストラクタ
      ///
      /// デストラクタのアクセス指定子は protected であり、派生クラスしか呼び出しができない。
      /// そのため、BaseMoment2 自身を構築することはできず、利用できるのはその派生クラスのみとなる。
      ~BaseMoment2() {}

    public:

      /// @brief デフォルト・コンストラクタ
      BaseMoment2() : moment1_(), sumSq_( value_type() ) {}

      /// @brief データ列を指定して構築
      ///
      /// テンプレート引数の In はデータ列の反復子の型を表す。
      ///
      /// - データ列の指定範囲が有効かを内部でチェックしている。
      ///   有効でない場合は例外 exception::NegativeLength< std::iterator_traits< In >::difference_type > を投げる。
      ///
      /// @param s 対象のデータ列の開始位置の反復子
      /// @param e 対象のデータ列の末尾の次の位置の反復子
      template< class In >
      BaseMoment2( In s, In e ) : moment1_( s, e ) {}

      /// @brief データ数を返す
      ///
      /// @return データ数
      distribution::Base::counter_type count() const
      { return( moment1_.count() ); }

      /// @brief データの総和を返す
      ///
      /// @return データの総和
      value_type sum() const
      { return( moment1_.sum() ); }

      /// @brief データの標本平均を返す
      ///
      /// データ数がゼロなら NaN を返す。
      ///
      /// @return データの標本平均
      value_type mean() const
      { return( moment1_.mean() ); }

      /// @brief データの平方和を返す
      ///
      /// @return データの平方和
      value_type sumSq() const
      { return( sumSq_ ); }
    };

    /**
       @brief 代表的な要約統計量 ( モーメント 2 まで ) を保持するオブジェクト

       要約統計量 ( データ数・総和・偏差の平方和 ) をデータ列から計算し、
       保持しておくためのオブジェクト
    **/
    class Moment2 : public BaseMoment2< Moment1 >
    {
    public:

      /// @brief デフォルト・コンストラクタ
      Moment2() : BaseMoment2< Moment1 >() {}

      /// @brief データ列を指定して構築
      ///
      /// テンプレート引数の In はデータ列の反復子の型を表す。
      ///
      /// - データ列の指定範囲が有効かを内部でチェックしている。
      ///   有効でない場合は例外 ExceptionNegativeLength< std::iterator_traits< In >::difference_type > を投げる。
      ///
      /// @param s 対象のデータ列の開始位置の反復子
      /// @param e 対象のデータ列の末尾の次の位置の反復子
      template< class In >
      Moment2( In s, In e );

      /// @brief データの偏差の平方和を返す
      ///
      /// データ数がゼロなら NaN を返す。
      ///
      /// @return データの偏差の平方和
      value_type devSq() const
      {
        // Σ( x - m )^2 = Σx^2 - Nm^2
        //              = Σx^2 - s^2 / N
        //              = ( NΣx^2 - s^2 ) / N
        // 誤差を小さくするため最後の式を採用
        value_type res = ( sumSq() * count() - std::pow( sum(), 2.0 ) ) / count();
        if ( res < 0 ) res = 0; // 負数になった場合の処置

        return( res );
      }

      /// @brief データの標本分散を返す
      ///
      /// データ数がゼロなら NaN を返す。
      ///
      /// @return データの標本分散
      value_type variance() const
      { return( ( count() == 0 ) ? distribution::Base::nan() : devSq() / count() ); }

      /// @brief データの不偏分散を返す
      ///
      /// データ数が 1 以下なら NaN を返す。
      ///
      /// @return データの不偏分散
      value_type unbiasedVariance() const
      { return( ( count() < 2 ) ? distribution::Base::nan() : devSq() / ( count() - 1 ) ); }

      /// @brief 他の要約統計量 ( モーメント 1・平方和 ) を加算する
      ///
      /// @param dest 加算する要約統計量 ( モーメント 2 ) への参照
      /// @return 自分自身への参照
      Moment2& operator+=( const Moment2& dest )
      {
        moment1_ += dest.moment1_;
        sumSq_ += dest.sumSq_;

        return( *this );
      }
    };

    /// @brief 二つの要約統計量 ( モーメント 2 ) を加算する
    ///
    /// @param m1 加算する要約統計量 1 への参照
    /// @param m2 加算する要約統計量 2 への参照
    /// @return 加算した結果
    inline Moment2 operator+( const Moment2& m1, const Moment2& m2 )
    {
      Moment2 res( m1 );
      return( res += m2 );
    }

    /**
       @brief 代表的な要約統計量 ( モーメント 2 まで ) を保持するオブジェクト

       二種類の要約統計量 ( データ数・総和・偏差の平方和 ) をデータ列から同時に計算し、
       保持しておくためのオブジェクト
    **/
    class MomentPair2 : public BaseMoment2< MomentPair1 >
    {
      using stat_type = distribution::Base::stat_type;

      stat_type proSum_; // 二変数の積和

      // 無効値のペアを返す
      static value_type nan()
      { return( value_type( distribution::Base::nan(), distribution::Base::nan() ) ); }

      // 値の平方を二つ同時に計算する
      static value_type powPair( value_type p )
      {
        return( value_type
                ( std::pow( p.first, 2.0 ),
                  std::pow( p.second, 2.0 ) )
                );
      }

      // 二つの値の積を返す
      static stat_type product( value_type p )
      { return( p.first * p.second ); }

    public:

      /// @brief デフォルト・コンストラクタ
      MomentPair2() : BaseMoment2< MomentPair1 >(), proSum_( distribution::Base::stat_type() ) {}

      /// @brief データ列を指定して構築
      ///
      /// テンプレート引数の In1, In2 はそれぞれ二つのデータ列の反復子の型を表す。
      ///
      /// - データ列 1 の指定範囲が有効かを内部でチェックしている。
      ///   有効でない場合は例外 ExceptionNegativeLength< std::iterator_traits< In1 >::difference_type > を投げる。
      /// - データ列 2 の指定範囲が有効かを内部でチェックしている。
      ///   有効でない場合は例外 ExceptionNegativeLength< std::iterator_traits< In2 >::difference_type > を投げる。
      /// - 二つのデータ列の長さが等しいかを内部でチェックしている。
      ///   等しくない場合は例外 ExceptionNotEqualLength< std::iterator_traits< In1 >::difference_type > を投げる。
      ///
      /// @param s1 対象のデータ列の開始位置の反復子
      /// @param e1 対象のデータ列の末尾の次の位置の反復子
      /// @param s2 対象のデータ列の開始位置の反復子
      /// @param e2 対象のデータ列の末尾の次の位置の反復子
      template< class In1, class In2 >
      MomentPair2( In1 s1, In1 e1, In2 s2, In2 e2 );

      /// @brief データ列を指定して構築
      ///
      /// テンプレート引数の In はデータ列の反復子の型を表す。
      ///
      /// - データ列の指定範囲が有効かを内部でチェックしている。
      ///   有効でない場合は例外 ExceptionNegativeLength< std::iterator_traits< In >::difference_type > を投げる。
      ///
      /// @note データ列は std::pair を要素とするコンテナであることを前提としている。
      ///
      /// @param s 対象のデータ列の開始位置の反復子
      /// @param e 対象のデータ列の末尾の次の位置の反復子
      template< class In >
      MomentPair2( In s, In e );

      /// @brief データの 1 番目のパラメータの総計を返す
      ///
      /// @return データの 1 番目のパラメータの総計
      distribution::Base::stat_type sum1() const
      { return( sum().first ); }

      /// @brief データの 2 番目のパラメータの総計を返す
      ///
      /// @return データの 2 番目のパラメータの総計
      distribution::Base::stat_type sum2() const
      { return( sum().second ); }

      /// @brief データの 1 番目のパラメータの標本平均を返す
      ///
      /// データ数がゼロなら NaN を返す。
      ///
      /// @return データの 1 番目のパラメータの標本平均
      distribution::Base::stat_type mean1() const
      { return( mean().first ); }

      /// @brief データの 2 番目のパラメータの標本平均を返す
      ///
      /// データ数がゼロなら NaN を返す。
      ///
      /// @return データの 2 番目のパラメータの標本平均
      distribution::Base::stat_type mean2() const
      { return( mean().second ); }

      /// @brief データの 1 番目のパラメータの平方和を返す
      ///
      /// @return データの 1 番目のパラメータの平方和
      distribution::Base::stat_type sumSq1() const
      { return( sumSq().first ); }

      /// @brief データの 2 番目のパラメータの平方和を返す
      ///
      /// @return データの 2 番目のパラメータの平方和
      distribution::Base::stat_type sumSq2() const
      { return( sumSq().second ); }

      /// @brief データの偏差の平方和を返す
      ///
      /// データ数がゼロなら NaN を返す。
      ///
      /// @return データの偏差の平方和
      value_type devSq() const
      {
        if ( count() == 0 ) return( nan() );

        // Σ( x - m )^2 = Σx^2 - Nm^2
        //              = Σx^2 - s^2 / N
        //              = ( NΣx^2 - s^2 ) / N
        // 誤差を小さくするため最後の式を採用
        value_type res = value_type
          ( ( sumSq1() * count() - std::pow( sum1(), 2.0 ) ) / count(),
            ( sumSq2() * count() - std::pow( sum2(), 2.0 ) ) / count() );

        // 負数になった場合の処置
        if ( res.first < 0 ) res.first = 0;
        if ( res.second < 0 ) res.second = 0;

        return( res );
      }

      /// @brief データの 1 番目のパラメータに対し、偏差の平方和を返す
      ///
      /// データ数がゼロなら NaN を返す。
      ///
      /// @return データの偏差の平方和
      distribution::Base::stat_type devSq1() const
      {
        if ( count() == 0 ) return( distribution::Base::nan() );

        // Σ( x - m )^2 = Σx^2 - Nm^2
        //              = Σx^2 - s^2 / N
        //              = ( NΣx^2 - s^2 ) / N
        // 誤差を小さくするため最後の式を採用
        distribution::Base::stat_type res = ( sumSq1() * count() - std::pow( sum1(), 2.0 ) ) / count();

        // 負数になった場合の処置
        if ( res < 0 ) res = 0;

        return( res );
      }

      /// @brief データの 2 番目のパラメータに対し、偏差の平方和を返す
      ///
      /// データ数がゼロなら NaN を返す。
      ///
      /// @return データの偏差の平方和
      distribution::Base::stat_type devSq2() const
      {
        if ( count() == 0 ) return( distribution::Base::nan() );

        // Σ( x - m )^2 = Σx^2 - Nm^2
        //              = Σx^2 - s^2 / N
        //              = ( NΣx^2 - s^2 ) / N
        // 誤差を小さくするため最後の式を採用
        distribution::Base::stat_type res = ( sumSq2() * count() - std::pow( sum2(), 2.0 ) ) / count();

        // 負数になった場合の処置
        if ( res < 0 ) res = 0;

        return( res );
      }

      /// @brief 二つのデータの積和を返す
      ///
      /// @return 二つのデータの積和
      distribution::Base::stat_type proSum() const
      { return( proSum_ ); }

      /// @brief 二変数の偏差の積和を返す
      ///
      /// データ数がゼロなら NaN を返す。
      ///
      /// @return データの偏差の平方和
      distribution::Base::stat_type proDev() const
      {
        if ( count() == 0 ) return( distribution::Base::nan() );

        // Σ( x - mx )( y - my )
        //  = Σxy - Nmx・my
        //  = Σxy - sx・sy / N
        //  = ( NΣxy - sx・sy ) / N
        // 誤差を小さくするため最後の式を採用
        return( ( proSum() * count() - sum1() * sum2() ) / count() );
      }

      /// @brief データの標本分散を返す
      ///
      /// データ数がゼロなら NaN を返す。
      ///
      /// @return データの標本分散
      value_type variance() const
      {
        return( ( count() == 0 ) ?
                nan() :
                value_type( devSq1() / count(), devSq2() / count() )
                );
      }

      /// @brief データの 1 番目のパラメータに対し、標本分散を返す
      ///
      /// データ数がゼロなら NaN を返す。
      ///
      /// @return データの標本分散
      distribution::Base::stat_type var1() const
      {
        return( ( count() == 0 ) ?
                distribution::Base::nan() : devSq1() / count()
                );
      }

      /// @brief データの 2 番目のパラメータに対し、標本分散を返す
      ///
      /// データ数がゼロなら NaN を返す。
      ///
      /// @return データの標本分散
      distribution::Base::stat_type var2() const
      {
        return( ( count() == 0 ) ?
                distribution::Base::nan() : devSq2() / count()
                );
      }

      /// @brief データの不偏分散を返す
      ///
      /// データ数が 1 以下なら NaN を返す。
      ///
      /// @return データの不偏分散
      value_type unbiasedVariance() const
      {
        return( ( count() < 2 ) ?
                nan() :
                value_type( devSq1() / ( count() - 1 ), devSq2() / ( count() - 1 ) )
                );
      }

      /// @brief データの 1 番目のパラメータに対し、不偏分散を返す
      ///
      /// データ数が 1 以下なら NaN を返す。
      ///
      /// @return データの不偏分散
      distribution::Base::stat_type uVar1() const
      {
        return( ( count() < 2 ) ?
                distribution::Base::nan() : devSq1() / ( count() - 1 )
                );
      }

      /// @brief データの 2 番目のパラメータに対し、不偏分散を返す
      ///
      /// データ数が 1 以下なら NaN を返す。
      ///
      /// @return データの不偏分散
      distribution::Base::stat_type uVar2() const
      {
        return( ( count() < 2 ) ?
                distribution::Base::nan() : devSq2() / ( count() - 1 )
                );
      }

      /// @brief データの標本共分散を返す
      ///
      /// データ数がゼロなら NaN を返す。
      ///
      /// @return データの標本共分散
      distribution::Base::stat_type covariance() const
      {
        return( ( count() == 0 ) ?
                distribution::Base::nan() : proDev() / count()
                );
      }

      /// @brief 他の要約統計量 ( モーメント 1・平方和 ) を加算する
      ///
      /// @param dest 加算する要約統計量 ( モーメント 2 ) への参照
      /// @return 自分自身への参照
      MomentPair2& operator+=( const MomentPair2& dest )
      {
        moment1_ += dest.moment1_;
        sumSq_.first += dest.sumSq_.first;
        sumSq_.second += dest.sumSq_.second;
        proSum_ += dest.proSum_;

        return( *this );
      }
    };

    /// @brief 二つの要約統計量 ( モーメント 2 ) を加算する
    ///
    /// @param m1 加算する要約統計量 1 への参照
    /// @param m2 加算する要約統計量 2 への参照
    /// @return 加算した結果
    inline MomentPair2 operator+( const MomentPair2& m1, const MomentPair2& m2 )
    {
      MomentPair2 res( m1 );
      return( res += m2 );
    }

    /************************************************************************
       その他の統計量の計算
       ************************************************************************/

    /**
       @brief ピアソン残差を求めるための関数オブジェクト

       実測値 obj と期待値 exp と使い、ピアソン残差 ( obj - exp ) / ( exp )^(1/2) を計算する。
    **/
    template< class Obj, class Exp >
    struct PearsonResiduals
    {
      /// @brief ピアソン残差を求める
      ///
      /// - 期待値 exp が負数の場合、distribution::Base::nan() を返す。
      /// - 期待値 exp がゼロの場合、観測値 obj がゼロならばゼロを、そうでなければ distribution::Base::infinity() を返す。
      ///
      /// @param obj 観測値
      /// @param exp 期待値
      /// @return ピアソン残差
      distribution::Base::stat_type operator()( Obj obj, Exp exp )
      {
        if ( exp < Exp() )
          return( distribution::Base::nan() );
        if ( exp == Exp() )
          return( ( obj == Obj() ) ? distribution::Base::stat_type() : distribution::Base::infinity() );
        else
          return( ( obj - exp ) / std::sqrt( exp ) );
      }
    };

    /**
       @brief 区間推定用名前空間
    **/
    namespace interval_estimation
    {
      /**
         @brief 区間推定用パラメータ定義
      **/
      struct ConfidenceInterval
      {
        /// @brief 値の型
        using value_type = distribution::Base::stat_type;

        value_type lower; ///< 推定区間の下限
        value_type upper; ///< 推定区間の上限

        /// @brief 推定区間の上下限を指定して構築
        ///
        /// @param l 推定区間の下限
        /// @param u 推定区間の上限
        explicit ConfidenceInterval( value_type l = value_type(), value_type u = value_type() )
          : lower( l ), upper( u ) {}
      };

      /*
        平均値の区間推定
      */

      /**
         @brief t-分布による平均値の区間推定

         t-分布を確率分布として確率 p(-t≦x≦t) = b となる確率変数 t を求め、
         標本平均 mean、標準偏差(不偏分散の平方根) sigma、標本数 n を使って以下の式で信頼区間を求める。

         信頼区間 = mean ± sigma * t / √n
      **/
      class Mean_TDist
      {
        using stat_type = distribution::Base::stat_type;
        using counter_type = distribution::Base::counter_type;

        Moment2 summary_; // 推定するデータの要約統計量

      public:

        /// @brief 要約統計量 ( モーメント 2 ) を直接指定して構築
        ///
        /// - t-分布を使っているため標本の総数は 2 以上である必要がある。
        ///   そうでない場合は例外 exception::TooLittleNumber< distribution::Base::counter_type > を投げる。
        ///
        /// @param summary 要約統計量 ( モーメント 2 )
        explicit Mean_TDist( Moment2 summary )
          : summary_( summary )
        {
          const std::string FUNC_NAME = "tadah_fussy::statistics::interval_estimation::Mean_TDist constructor";
          error_lib::GE< counter_type >( summary_.count(), 2, FUNC_NAME + " : size of summary " + error_message::GE + " 2" );
        }

        /// @brief データ列を指定して構築
        ///
        /// - データ列の指定範囲が有効かを内部でチェックしている。
        ///   有効でない場合は例外 exception::NegativeLength< std::iterator_traits< In >::difference_type > を投げる。
        /// - t-分布を使っているため標本の総数は 2 以上である必要がある。
        ///   そうでない場合は例外 exception::TooLittleNumber< distribution::Base::counter_type > を投げる。
        ///
        /// テンプレート引数の In はデータ列の反復子の型を表す。
        ///
        /// @param s 対象のデータ列の開始位置の反復子
        /// @param e 対象のデータ列の末尾の次の位置の反復子
        template< class In >
        Mean_TDist( In s, In e )
          : summary_( s, e )
        {
          const std::string FUNC_NAME = "tadah_fussy::statistics::interval_estimation::Mean_TDist constructor";
          error_lib::GE< counter_type >( summary_.count(), 2, FUNC_NAME + " : size of sequence [ s, e ) " + error_message::GE + " 2" );
        }

        /// @brief t-分布による平均値の区間推定
        ///
        /// - 信頼度 b は 0 以上 1 以下でなければならない。そうでない場合は例外 ExceptionUnitInterval を投げる。
        ///
        /// @param b 信頼度
        /// @param threshold binSearch で処理する時のしきい値
        /// @return 信頼区間
        ConfidenceInterval operator()
        ( distribution::Base::stat_type b, distribution::Base::stat_type threshold = DEFAULT_THRESHOLD ) const;

        /// @brief 要約統計量 Moment2 を返す
        ///
        /// @return 要約統計量 Moment2
        Moment2 summary() const { return( summary_ ); }

        /// @brief t値を返す
        ///
        /// t 値は以下の式で求める
        /// t = ( m - μ )√N / u [ m:標本平均、μ:平均の基準値、N:データ数、u:不偏分散の平方根 ]
        ///
        /// @param mu 平均の基準値
        /// @return t値
        distribution::Base::stat_type t( distribution::Base::stat_type mu ) const
        { return( ( summary_.mean() - mu ) * std::sqrt( summary_.count() ) /
                  std::sqrt( summary_.unbiasedVariance() ) ); }

        /// @brief t値の絶対値が t より大きくなる確率 ( p 値 ) を求める
        ///
        /// @param t t値の上下限
        /// @return p 値
        distribution::Base::stat_type p( distribution::Base::stat_type t ) const;
      };

      /**
         @brief 標準正規分布による平均値の区間推定

         標準正規分布を確率分布として確率 p(-t≦x≦t) = b となる確率変数 z を求め、
         標本平均 mean、標準偏差 sigma、標本数 n を使って以下の式で信頼区間を求める。

         信頼区間 = mean ± sigma * z / √n

         sigma は母集団の標準偏差で既知とする。
      **/
      class Mean_NormDist
      {
        using stat_type = distribution::Base::stat_type;

        Moment1 summary_;           // 推定するデータの要約統計量
        stat_type sigma_; // 母集団の標準偏差
        distribution::Normal nDist_;  // 標準正規分布

      public:

        /// @brief 要約統計量 ( モーメント 1 ) と母集団の標準偏差を直接指定して構築
        ///
        /// - 標準偏差 sigma が負数だった場合は例外 exception::NegativeNumber< distribution::Base::stat_type > を投げる。
        /// - データの総数がゼロの場合は例外 exception::NotPositiveNumber< distribution::Base::counter_type > を投げる。
        ///
        /// @param summary 要約統計量 ( モーメント 1 )
        /// @param sigma 母集団の標準偏差
        Mean_NormDist( Moment1 summary, distribution::Base::stat_type sigma )
          : summary_( summary ), sigma_( sigma ), nDist_( 0, 1 )
        {
          const std::string FUNC_NAME = "tadah_fussy::statistics::interval_estimation::Mean_NormDist constructor";

          error_lib::NNNum( sigma_, FUNC_NAME + " : sigma " + error_message::NNNUM );
          error_lib::PNum( summary_.count(), FUNC_NAME + " : size of summary " + error_message::PNUM );
        }

        /// @brief データ列と母集団の標準偏差を指定して構築
        ///
        /// - データ列の指定範囲が有効かを内部でチェックしている。
        ///   有効でない場合は例外 exception::NegativeLength< std::iterator_traits< In >::difference_type > を投げる。
        /// - 標準偏差 sigma が負数だった場合は例外 exception::NegativeNumber< distribution::Base::stat_type > を投げる。
        /// - データの総数がゼロの場合は例外 exception::NotPositiveNumber< distribution::Base::counter_type > を投げる。
        ///
        /// テンプレート引数の In はデータ列の反復子の型を表す。
        ///
        /// @param s 対象のデータ列の開始位置の反復子
        /// @param e 対象のデータ列の末尾の次の位置の反復子
        /// @param sigma 母集団の標準偏差
        template< class In >
        Mean_NormDist( In s, In e, distribution::Base::stat_type sigma )
          : summary_( s, e ), sigma_( sigma ), nDist_( 0, 1 )
        {
          const std::string FUNC_NAME = "tadah_fussy::statistics::interval_estimation::Mean_NormDist constructor";

          error_lib::NNNum( sigma_, FUNC_NAME + " : sigma " + error_message::NNNUM );
          error_lib::PNum( summary_.count(), FUNC_NAME + " : size of sequence [ s, e ) " + error_message::PNUM );
        }

        /// @brief 標準正規分布による平均値の区間推定
        ///
        /// - 信頼度 b は 0 以上 1 以下でなければならない。そうでない場合は例外 exception::UnitInterval を投げる。
        ///
        /// @param b 信頼度
        /// @param threshold binSearch で処理する時のしきい値
        /// @return 信頼区間
        ConfidenceInterval operator()
        ( distribution::Base::stat_type b, distribution::Base::stat_type threshold = DEFAULT_THRESHOLD ) const;

        /// @brief 要約統計量 Moment1 を返す
        ///
        /// @return 要約統計量 Moment1
        Moment1 summary() const { return( summary_ ); }

        /// @brief z値を返す
        ///
        /// z 値は以下の式で求める
        /// z = ( m - μ )√N / s [ m:標本平均、μ:平均の基準値、N:データ数、u:母集団の標準偏差 ]
        ///
        /// @param mu 平均の基準値
        /// @return z値
        distribution::Base::stat_type z( distribution::Base::stat_type mu ) const
        { return( ( summary_.mean() - mu ) * std::sqrt( static_cast< stat_type >( summary_.count() ) ) / sigma_ ); }

        /// @brief z値の絶対値が z より大きくなる確率 ( p 値 ) を求める
        ///
        /// @param z z値の上下限
        /// @return p 値
        distribution::Base::stat_type p( distribution::Base::stat_type z ) const
        { return( 0.5 - nDist_.p( 0, std::abs( z ) ) ); }
      };

      /**
         @brief F-分布による平均差の区間推定

         F-分布を確率分布として確率 p(-t≦x≦t) = b となる確率変数 f を求め、
         二つの母集団の標本平均 m1, m2、標本分散 V1, V2 標本数 N1, N2 を使って以下の式で平均差 Δμ の信頼区間を求める。

         信頼区間 = ( m1 - m2 ) ± { f[( 1 / N1 + 1 / N2 )( N1・V1 + N2・V2 ) ] / ( N1 + N2 - 2 ) }^(1/2)
      **/
      class MeanDiff
      {
        using stat_type = distribution::Base::stat_type;
        using counter_type = distribution::Base::counter_type;

        Moment2 summary1_; // データ 1 の要約統計量
        Moment2 summary2_; // データ 2 の要約統計量

        // データ 1 のデータ数を返す
        counter_type count1() const { return( summary1_.count() ); }
        // データ 1 の平均値を返す
        stat_type mean1() const { return( summary1_.mean() ); }
        // データ 1 の標本分散を返す
        stat_type variance1() const { return( summary1_.variance() ); }

        // データ 2 のデータ数を返す
        counter_type count2() const { return( summary2_.count() ); }
        // データ 2 の平均値を返す
        stat_type mean2() const { return( summary2_.mean() ); }
        // データ 2 の標本分散を返す
        stat_type variance2() const { return( summary2_.variance() ); }

      public:

        /// @brief 二つのデータの要約統計量 ( モーメント 2 ) を直接指定して構築
        ///
        /// - 二つのデータの総数はそれぞれ正値でなければならない。
        ///   そうでない場合は例外 exception::NotPositiveNumber< distribution::Base::counter_type > を投げる。
        /// - 二つのデータの総数の合計は 3 以上でなければならない。
        ///   そうでない場合は例外 ExceptionTooLittleNumber< distribution::Base::counter_type > を投げる。
        ///
        /// @param summary1 データ 1 の要約統計量 ( モーメント 2 )
        /// @param summary2 データ 2 の要約統計量 ( モーメント 2 )
        MeanDiff( Moment2 summary1, Moment2 summary2 )
          : summary1_( summary1 ), summary2_( summary2 )
        {
          const std::string FUNC_NAME = "tadah_fussy::statistics::interval_estimation::MeanDiff constructor";

          error_lib::PNum( count1(), FUNC_NAME + " : size of summary1 " + error_message::PNUM );
          error_lib::PNum( count2(), FUNC_NAME + " : size of summary2 " + error_message::PNUM );
          error_lib::GE< counter_type >( count1() + count2(), 3,
                                         FUNC_NAME + " : sum of size of summary1 and summary2 " + error_message::GE + " 3" );
        }

        /// @brief 二つのデータ列を指定して構築
        ///
        /// - データ列 1 の指定範囲が有効かを内部でチェックしている。
        ///   有効でない場合は例外 exception::NegativeLength< std::iterator_traits< In1 >::difference_type > を投げる。
        /// - データ列 2 の指定範囲が有効かを内部でチェックしている。
        ///   有効でない場合は例外 exception::NegativeLength< std::iterator_traits< In2 >::difference_type > を投げる。
        /// - 二つのデータの総数はそれぞれ正値でなければならない。
        ///   そうでない場合は例外 exception::NotPositiveNumber< tadah_fussy::distribution::Base::counter_type > を投げる。
        /// - 二つのデータの総数の合計は 3 以上でなければならない。
        ///   そうでない場合は例外 exception::TooLittleNumber< tadah_fussy::distribution::Base::counter_type > を投げる。
        ///
        /// @param s1 データ列 1 の開始位置の反復子
        /// @param e1 データ列 1 の末尾の次の位置の反復子
        /// @param s2 データ列 2 の開始位置の反復子
        /// @param e2 データ列 2 の末尾の次の位置の反復子
        template< class In1, class In2 >
        MeanDiff( In1 s1, In1 e1, In2 s2, In2 e2 )
          : summary1_( s1, e1 ), summary2_( s2, e2 )
        {
          const std::string FUNC_NAME = "tadah_fussy::statistics::interval_estimation::MeanDiff constructor";

          error_lib::PNum( count1(), FUNC_NAME + " : size of summary1 " + error_message::PNUM );
          error_lib::PNum( count2(), FUNC_NAME + " : size of summary2 " + error_message::PNUM );
          error_lib::GE< distribution::Base::counter_type >( count1() + count2(), 3,
                                                             FUNC_NAME + " : sum of size of summary1 and summary2 " + error_message::GE + " 3" );
        }

        /// @brief 平均差の区間推定
        ///
        /// - 信頼度 b は 0 以上 1 以下でなければならない。そうでない場合は例外 exceptio::nUnitInterval を投げる。
        ///
        /// @param b 信頼度
        /// @param threshold binSearch で処理する時のしきい値
        /// @return 信頼区間
        ConfidenceInterval operator()
        ( distribution::Base::stat_type b, distribution::Base::stat_type threshold = DEFAULT_THRESHOLD ) const;

        /// @brief データ 1 の要約統計量 Moment2 を返す
        ///
        /// @return データ 1 の要約統計量 Moment2
        Moment2 summary1() const { return( summary1_ ); }

        /// @brief データ 2 の要約統計量 Moment2 を返す
        ///
        /// @return データ 2 の要約統計量 Moment2
        Moment2 summary2() const { return( summary2_ ); }
      };

      /**
         @brief 分散の区間推定 ( 自由度 N - 1 のカイ二乗分布を利用 )

         カイ二乗分布を確率分布として確率 p(0≦x≦Xa) = ( 1 - b ) / 2, p(Xb≦x≦+∞) = ( 1 - b ) / 2
         となる確率変数 Xa, Xb を求め、母集団の標本分散 V と標本数 N を使って以下の式で分散 σ^2 の信頼区間を求める。

         NV / Xb ≦ 信頼区間 ≦ NV / Xa

         データ列のみを使い、母集団の平均は未知として推定を行うためカイ二乗分布の自由度は N - 1 とする。
      **/
      class VarN_1
      {
        using counter_type = distribution::Base::counter_type;

        Moment2 summary_; // 推定するデータの要約統計量

      public:

        /// @brief 要約統計量 ( モーメント 2 ) を直接指定して構築
        ///
        /// N 個のデータのみを指定した場合は確率密度関数に自由度 N - 1 のカイ二乗分布を使う。
        ///
        /// - データの総数は 1 より多くなければならない。
        ///   そうでない場合は例外 exception::TooLittleNumber< distribution::Base::counter_type > を投げる。
        ///
        /// @param summary 要約統計量 ( モーメント 2 )
        VarN_1( Moment2 summary )
          : summary_( summary )
        {
          const std::string FUNC_NAME = "tadah_fussy::statistics::interval_estimation::VarN_1 constructor";
          error_lib::GT< counter_type >( summary_.count(), 1, FUNC_NAME + " : size of summary " + error_message::GT + " 1" );
        }

        /// @brief データ列を指定して構築
        ///
        /// N 個のデータのみを指定した場合は確率密度関数に自由度 N - 1 のカイ二乗分布を使う。
        ///
        /// - データ列の指定範囲が有効かを内部でチェックしている。
        ///   有効でない場合は例外 exception::NegativeLength< std::iterator_traits< In >::difference_type > を投げる。
        /// - データの総数は 1 より多くなければならない。
        ///   そうでない場合は例外 exception::TooLittleNumber< distribution::Base:::counter_type > を投げる。
        ///
        /// @param s 対象のデータ列の開始位置の反復子
        /// @param e 対象のデータ列の末尾の次の位置の反復子
        template< class In >
        VarN_1( In s, In e )
          : summary_( s, e )
        {
          const std::string FUNC_NAME = "tadah_fussy::statistics::interval_estimation::VarN_1 constructor";
          error_lib::GT< counter_type >( summary_.count(), 1, FUNC_NAME + " : size of sequence [ s, e ) " + error_message::GT + " 1" );
        }

        /// @brief 区間推定を行う
        ///
        /// - 信頼度 b は 0 以上 1 以下でなければならない。そうでない場合は例外 ExceptionUnitInterval を投げる。
        ///
        /// @param b 信頼度
        /// @param threshold binSearch で処理する時のしきい値
        /// @return 信頼区間
        ConfidenceInterval operator()( distribution::Base::stat_type b, distribution::Base::stat_type threshold = DEFAULT_THRESHOLD ) const;

        /// @brief 要約統計量 Moment2 を返す
        ///
        /// @return 要約統計量 Moment2
        Moment2 summary() const { return( summary_ ); }
      };

      /**
         @brief 分散の区間推定 ( 自由度 N のカイ二乗分布を利用 )

         カイ二乗分布を確率分布として確率 p(0≦x≦Xa) = ( 1 - b ) / 2, p(Xb≦x≦+∞) = ( 1 - b ) / 2
         となる確率変数 Xa, Xb を求め、母集団の平均差の平方和 S^2 を使って以下の式で分散 σ^2 の信頼区間を求める。

         S^2 / Xb ≦ 信頼区間 ≦ S^2 / Xa

         データ列と母集団の平均を指定する。母集団の平均は既知として推定を行うためカイ二乗分布の自由度は N とする。
      **/
      class VarN
      {
        using stat_type = distribution::Base::stat_type;
        using counter_type = distribution::Base::counter_type;

        counter_type count_; // データ数
        stat_type sumSq_;    // 母集団の平均との差の平方和

      public:

        /// @brief データ列を指定して構築
        ///
        /// 母集団の平均を指定した場合は確率密度関数に自由度 N のカイ二乗分布を使う。
        ///
        /// - データ列の指定範囲が有効かを内部でチェックしている。
        ///   有効でない場合は例外 exception::NegativeLength< std::iterator_traits< In >::difference_type > を投げる。
        /// - データの総数は 0 より多くなければならない。
        ///   そうでない場合は例外 exception::NotPositiveNumber< distribution::Base::counter_type > を投げる。
        ///
        /// @param s 対象のデータ列の開始位置の反復子
        /// @param e 対象のデータ列の末尾の次の位置の反復子
        /// @param mu 母集団の平均
        template< class In >
        VarN( In s, In e, distribution::Base::stat_type mu );

        /// @brief 区間推定を行う
        ///
        /// - 信頼度 b は 0 以上 1 以下でなければならない。そうでない場合は例外 ExceptionUnitInterval を投げる。
        ///
        /// @param b 信頼度
        /// @param threshold binSearch で処理する時のしきい値
        /// @return 信頼区間
        ConfidenceInterval operator()( distribution::Base::stat_type b, distribution::Base::stat_type threshold = DEFAULT_THRESHOLD ) const;

        /// @brief データ数を返す
        ///
        /// @return データ数
        distribution::Base::counter_type count() const { return( count_ ); }

        /// @brief 母集団の平均との差の平方和を返す
        ///
        /// @return 母集団の平均との差の平方和
        distribution::Base::stat_type sumSq() const { return( sumSq_ ); }
      };

      /**
         @brief 分散比の区間推定

         F-分布を確率分布として確率 p(0≦x≦fa) = ( 1 - b ) / 2, p(fb≦x≦+∞) = ( 1 - b ) / 2 となるような
         確率変数 fa, fb から二つのデータの不偏分散比の信頼区間を [ fa, fb ] として求めることができる。
      **/
      class VarRatio
      {
        using stat_type = distribution::Base::stat_type;
        using counter_type = distribution::Base::counter_type;

        counter_type n1_; // 母集団 1 の標本数
        counter_type n2_; // 母集団 2 の標本数

      public:

        /// @brief 二つの母集団の標本数を指定して構築
        ///
        /// - 二つの標本数はそれぞれ 2 以上でなければならない。
        ///   そうでない場合は例外 exception::TooLittleNumber< distribution::Base::counter_type > を投げる。
        ///
        /// @param n1 母集団 1 の標本数
        /// @param n2 母集団 2 の標本数
        VarRatio( distribution::Base::counter_type n1, distribution::Base::counter_type n2 )
          : n1_( n1 ), n2_( n2 )
        {
          const std::string FUNC_NAME = "tadah_fussy::statistics::interval_estimation::VarRatio constructor";

          error_lib::GE< counter_type >( n1_, 2, FUNC_NAME + " : n1 " + error_message::GE + " 2" );
          error_lib::GE< counter_type >( n2_, 2, FUNC_NAME + " : n2 " + error_message::GE + " 2" );
        }

        /// @brief 区間推定を行う
        ///
        /// - 信頼度 b は 0 以上 1 以下でなければならない。そうでない場合は例外 ExceptionUnitInterval を投げる。
        ///
        /// @param b 信頼度
        /// @param threshold binSearch で処理する時のしきい値
        /// @return 信頼区間
        ConfidenceInterval operator()( distribution::Base::stat_type b, distribution::Base::stat_type threshold = DEFAULT_THRESHOLD ) const;
      };

      /**
         @brief 百分率の区間推定

         interval_estimation::Percent_SS と interval_estimation::Percent_LS を利用するための基底クラス。
         単独で利用することはできない。
      **/
      class Percent
      {
      protected:

        using stat_type = distribution::Base::stat_type;
        using counter_type = distribution::Base::counter_type;

      private:

        counter_type count_; // 総試行回数
        counter_type r_;     // 試行に成功した回数

      protected:

        /// @brief デストラクタ
        ///
        /// @note デストラクタのアクセス指定子は protected であり、派生クラスしか呼び出しができない。
        /// そのため、Percent のポインタや参照から派生クラスを利用することはできない。
        ~Percent() {}

      public:

        /// @brief 総試行回数と成功した回数を指定して構築
        ///
        /// - 総試行回数と成功回数はゼロより大きくなければならない。
        ///   そうでない場合は例外 exception::NotPositiveNumber< distribution::Base::counter_type > を投げる。
        /// - 総試行回数は成功回数以上でなければならない。
        ///   そうでない場合は例外 exception::TooLittleNumber< distribution::Base::counter_type > を投げる。
        ///
        /// @param trialCount 総試行回数
        /// @param successCount 試行に成功した回数
        Percent( distribution::Base::counter_type trialCount, distribution::Base::counter_type successCount )
          : count_( trialCount ), r_( successCount )
        {
          const std::string FUNC_NAME = "tadah_fussy::statistics::interval_estimation::Percent constructor";

          error_lib::PNum( count_, FUNC_NAME + " : trialCount " + error_message::PNUM );
          error_lib::PNum( r_, FUNC_NAME + " : successCount " + error_message::PNUM );
          error_lib::GE( count_, r_, FUNC_NAME + " : trialCount " + error_message::GE + " successCount" );
        }

        /// @brief 総試行回数を返す
        ///
        /// @return 総試行回数
        distribution::Base::counter_type count() const { return( count_ ); }

        /// @brief 試行に成功した回数を返す
        ///
        /// @return 試行に成功した回数
        distribution::Base::counter_type r() const { return( r_ ); }

        /// @brief 百分率の区間推定
        ///
        /// - 信頼度 b は 0 以上 1 以下でなければならない。そうでない場合は例外 exception::UnitInterval を投げる。
        ///
        /// @param b 信頼度
        /// @param threshold binSearch で t を求める時のしきい値
        /// @return 信頼区間
        virtual ConfidenceInterval operator()( distribution::Base::stat_type b, distribution::Base::stat_type threshold ) const = 0;
      };

      /**
         @brief 百分率の区間推定(小標本)

         総試行回数を N、成功回数を r としたとき、百分率の信頼区間 [ pa, pb ] は自由度 m, n の F-分布 Gm,n(x) を使って

         ∫{0→n・pa/m(1-pa)}Gm,n(x)dx\n
         ∫{n・pb/m(1-pb)→∞}Gm,n(x)dx

         但し m = 2( r + 1 )、n = 2( N - r )

         で求められる。この推定は、N が比較的小さい場合に用いられる。
      **/
      struct Percent_SS : public Percent
      {
        /// @brief 総試行回数と成功した回数を指定して構築
        ///
        /// - 総試行回数と成功回数はゼロより大きくなければならない。
        ///   そうでない場合は例外 exception::NotPositiveNumber< distribution::Base::counter_type > を投げる。
        /// - 総試行回数は成功回数以上でなければならない。
        ///   そうでない場合は例外 exception::TooLittleNumber< distribution::Base::counter_type > を投げる。
        ///
        /// @param trialCount 総試行回数
        /// @param successCount 試行に成功した回数
        Percent_SS( distribution::Base::counter_type trialCount, distribution::Base::counter_type successCount )
          : Percent( trialCount, successCount ) {}

        // 区間推定
        virtual ConfidenceInterval operator()( distribution::Base::stat_type b, distribution::Base::stat_type threshold = DEFAULT_THRESHOLD ) const;
      };

      /**
         @brief 百分率の区間推定(大標本)

         標準正規分布 N( 0, 1 ) を使い、信頼度βに対して ∫{-ta→ta} N( 0, 1 ) dx = β を満たす ta を求めたとき、
         百分率 p の信頼区間は

         r / N - ta[ r( N - r ) / N^3 ]^(1/2) < p < r / N + ta[ r( N - r ) / N^3 ]^(1/2)

         となる。この推定は、N が比較的大きい場合に用いられる。
      **/
      struct Percent_LS : public Percent
      {
        /// @brief 総試行回数と成功した回数を指定して構築
        ///
        /// - 総試行回数と成功回数はゼロより大きくなければならない。
        ///   そうでない場合は例外 exception::NotPositiveNumber< distribution::Base::counter_type > を投げる。
        /// - 総試行回数は成功回数以上でなければならない。
        ///   そうでない場合は例外 exception::TooLittleNumber< distribution::Base::counter_type > を投げる。
        ///
        /// @param trialCount 総試行回数
        /// @param successCount 試行に成功した回数
        Percent_LS( distribution::Base::counter_type trialCount, distribution::Base::counter_type successCount )
          : Percent( trialCount, successCount ) {}

        // 区間推定
        virtual ConfidenceInterval operator()( distribution::Base::stat_type b, distribution::Base::stat_type threshold = DEFAULT_THRESHOLD ) const;
      };

    } // namespace interval_estimation

    /**
       @brief 検定用名前空間
    **/
    namespace test
    {
      /**
         @brief t-分布を使った一標本の平均値検定

         基準値 m0 と母集団の平均 μ を比較・検定するためのクラス

         内部で interval_estimation::Mean_TDist を使って区間推定を行い検定結果を得る。
      **/
      class OneSampleMean_TDist
      {
        interval_estimation::Mean_TDist iEst_; // 平均値の区間推定用クラス

      public:

        /// @brief 要約統計量 ( モーメント 2 ) を直接指定して構築
        ///
        /// - t-分布を使っているため標本の総数は 2 以上である必要がある。
        ///   そうでない場合は例外 exception::TooLittleNumber< distribution::Base::counter_type > を投げる。
        ///
        /// @param summary 要約統計量 ( モーメント 2 )
        OneSampleMean_TDist( Moment2 summary ) : iEst_( summary ) {}

        /// @brief データ列を指定して構築
        ///
        /// - データ列の指定範囲が有効かを内部でチェックしている。
        ///   有効でない場合は例外 exception::NegativeLength< std::iterator_traits< In >::difference_type > を投げる。
        /// - t-分布を使っているため標本の総数は 2 以上である必要がある。
        ///   そうでない場合は例外 exception::TooLittleNumber< distribution::Base::counter_type > を投げる。
        ///
        /// @param s 対象のデータ列の開始位置の反復子
        /// @param e 対象のデータ列の末尾の次の位置の反復子
        template< class In >
        OneSampleMean_TDist( In s, In e ) : iEst_( s, e ) {}

        /// @brief 危険率 a で平均値 m0 を片側検定 ( μ < m0 ) する。
        ///
        /// 標本平均 m の信頼区間より m0 が大きければ仮説を棄却する。
        ///
        /// - 危険率 a は 0 以上 1 以下でなければならない。そうでない場合は例外 exception::UnitInterval を投げる。
        ///
        /// @param m0 検定対象の平均値
        /// @param a 危険率
        /// @param mp 母集団平均の信頼区間上限を代入する変数へのポインタ(NULLならば無視)
        /// @param threshold binSearch で t を求める時のしきい値
        /// @return 仮説を棄却する場合は true を返す
        bool oneSideLess( distribution::Base::stat_type m0, distribution::Base::stat_type a,
                          distribution::Base::stat_type* mp,
                          distribution::Base::stat_type threshold = DEFAULT_THRESHOLD );

        /// @brief 危険率 a で平均値 m0 を片側検定 ( μ > m0 ) する。
        ///
        /// 標本平均 m の信頼区間より m0 が小さければ仮説を棄却する。
        ///
        /// - 危険率 a は 0 以上 1 以下でなければならない。そうでない場合は例外 exception::UnitInterval を投げる。
        ///
        /// @param m0 検定対象の平均値
        /// @param a 危険率
        /// @param mp 母集団平均の信頼区間下限を代入する変数へのポインタ(NULLならば無視)
        /// @param threshold binSearch で t を求める時のしきい値
        /// @return 仮説を棄却する場合は true を返す
        bool oneSideMore( distribution::Base::stat_type m0, distribution::Base::stat_type a,
                          distribution::Base::stat_type* mp,
                          distribution::Base::stat_type threshold = DEFAULT_THRESHOLD );

        /// @brief 危険率 a で平均値 m0 を両側検定 ( μ ≠ m0 ) する。
        ///
        /// 標本平均 m の信頼区間内に m0 がなければ仮説を棄却する。
        ///
        /// - 危険率 a は 0 以上 1 以下でなければならない。そうでない場合は例外 exception::UnitInterval を投げる。
        ///
        /// @param m0 検定対象の平均値
        /// @param a 危険率
        /// @param cip 母集団平均の信頼区間を代入する変数へのポインタ(NULLならば無視)
        /// @param threshold binSearch で t を求める時のしきい値
        /// @return 仮説を棄却する場合は true を返す
        bool twoSide( distribution::Base::stat_type m0, distribution::Base::stat_type a,
                      interval_estimation::ConfidenceInterval* cip,
                      distribution::Base::stat_type threshold = DEFAULT_THRESHOLD );

        /// @brief 平均値 m0 から t 値を求める
        ///
        /// 式 ( 標本平均 - m )√N / σ を使い、引数として渡した値 m0 を t 値に変換する。
        /// σ は不偏分散の平方根が使われる。
        ///
        /// @param m0 平均値
        /// @return t値
        distribution::Base::stat_type t( distribution::Base::stat_type m0 ) const { return( iEst_.t( m0 ) ); }

        /// @brief t値の絶対値が x より大きくなる確率 ( p 値 ) を求める
        ///
        /// @param t t値
        /// @return p 値
        distribution::Base::stat_type p( distribution::Base::stat_type t ) const { return( iEst_.p( t ) ); }
      };

      /**
         @brief 標準正規分布を使った一標本の平均値検定

         基準値 m0 と母集団の平均 μ を比較・検定するためのクラス

         内部で intervalEstimation::Mean_NormDist を使って区間推定を行い検定結果を得る。
      **/
      class OneSampleMean_NormDist
      {
        interval_estimation::Mean_NormDist iEst_; // 平均値の区間推定用クラス

      public:

        /// @brief 要約統計量 ( モーメント 1 ) と母集団の標準偏差を直接指定して構築
        ///
        /// - 標準偏差 sigma が負数だった場合は例外 exception::NegativeNumber< distribution::Base::stat_type > を投げる。
        /// - データの総数がゼロの場合は例外 exception::NotPositiveNumber< distribution::Base::counter_type > を投げる。
        ///
        /// @param summary 要約統計量 ( モーメント 1 )
        /// @param sigma 母集団の標準偏差
        OneSampleMean_NormDist( Moment1 summary, distribution::Base::stat_type sigma )
          : iEst_( summary, sigma ) {}

        /// @brief データ列と母集団の標準偏差を指定して構築
        ///
        /// 検定は標準正規分布を使って行われる
        ///
        /// - データ列の指定範囲が有効かを内部でチェックしている。
        ///   有効でない場合は例外 exception::NegativeLength< std::iterator_traits< In >::difference_type > を投げる。
        /// - 標準偏差 sigma が負数だった場合は例外 exception::NegativeNumber< distribution::Base::stat_type > を投げる。
        /// - データの総数がゼロの場合は例外 exception::NotPositiveNumber< distribution::Base::counter_type > を投げる。
        ///
        /// @param s 対象のデータ列の開始位置の反復子
        /// @param e 対象のデータ列の末尾の次の位置の反復子
        /// @param sigma 母集団の標準偏差
        template< class In >
        OneSampleMean_NormDist( In s, In e, distribution::Base::stat_type sigma )
          : iEst_( s, e, sigma ) {}

        /// @brief 危険率 a で平均値 m0 を片側検定 ( μ < m0 ) する。
        ///
        /// 標本平均 m の信頼区間より m0 が大きければ仮説を棄却する。
        ///
        /// - 危険率 a は 0 以上 1 以下でなければならない。そうでない場合は例外 exception::UnitInterval を投げる。
        ///
        /// @param m0 検定対象の平均値
        /// @param a 危険率
        /// @param mp 母集団平均の信頼区間上限を代入する変数へのポインタ(NULLならば無視)
        /// @param threshold binSearch で t を求める時のしきい値
        /// @return 仮説を棄却する場合は true を返す
        bool oneSideLess( distribution::Base::stat_type m0, distribution::Base::stat_type a,
                          distribution::Base::stat_type* mp,
                          distribution::Base::stat_type threshold = DEFAULT_THRESHOLD );

        /// @brief 危険率 a で平均値 m0 を片側検定 ( μ > m0 ) する。
        ///
        /// 標本平均 m の信頼区間より m0 が小さければ仮説を棄却する。
        ///
        /// - 危険率 a は 0 以上 1 以下でなければならない。そうでない場合は例外 exception::UnitInterval を投げる。
        ///
        /// @param m0 検定対象の平均値
        /// @param a 危険率
        /// @param mp 母集団平均の信頼区間下限を代入する変数へのポインタ(NULLならば無視)
        /// @param threshold binSearch で t を求める時のしきい値
        /// @return 仮説を棄却する場合は true を返す
        bool oneSideMore( distribution::Base::stat_type m0, distribution::Base::stat_type a,
                          distribution::Base::stat_type* mp,
                          distribution::Base::stat_type threshold = DEFAULT_THRESHOLD );

        /// @brief 危険率 a で平均値 m0 を両側検定 ( μ ≠ m0 ) する。
        ///
        /// 標本平均 m の信頼区間内に m0 がなければ仮説を棄却する。
        ///
        /// - 危険率 a は 0 以上 1 以下でなければならない。そうでない場合は例外 exception::UnitInterval を投げる。
        ///
        /// @param m0 検定対象の平均値
        /// @param a 危険率
        /// @param cip 母集団平均の信頼区間を代入する変数へのポインタ(NULLならば無視)
        /// @param threshold binSearch で t を求める時のしきい値
        /// @return 仮説を棄却する場合は true を返す
        bool twoSide( distribution::Base::stat_type m0, distribution::Base::stat_type a,
                      interval_estimation::ConfidenceInterval* cip,
                      distribution::Base::stat_type threshold = DEFAULT_THRESHOLD );

        /// @brief 平均値 m0 から z 値を求める
        ///
        /// 式 ( 標本平均 - m )√N / σ を使い、引数として渡した値 m0 を z 値に変換する。
        /// σ は構築時に引数として渡した母集団の標準偏差がそのまま使われる。
        ///
        /// @param m0 平均値
        /// @return 確率変数 ( t値・z値 )
        distribution::Base::stat_type z( distribution::Base::stat_type m0 ) const
        { return( iEst_.z( m0 ) ); }

        /// @brief z値の絶対値が z より大きくなる確率 ( p 値 ) を求める
        ///
        /// @param z z値の上下限
        /// @return p 値
        distribution::Base::stat_type p( distribution::Base::stat_type z ) const
        { return( iEst_.p( z ) ); }
      };

      /**
         @brief t-分布を使った二標本の検定

         分散比の検定には interval_estimation::VarRatio が、平均差の検定には interval_estimation::MeanDiff が用いられる。
      **/
      class TwoSample
      {
        Moment2 summary1_; // 分布 1 の要約統計量
        Moment2 summary2_; // 分布 2 の要約統計量

      public:

        /// @brief 二つの分布の要約統計量 ( モーメント 2 ) を直接指定して構築
        ///
        /// @param summary1 分布 1 の要約統計量 ( モーメント 2 )
        /// @param summary2 分布 2 の要約統計量 ( モーメント 2 )
        TwoSample( Moment2 summary1, Moment2 summary2 )
          : summary1_( summary1 ), summary2_( summary2 ) {}

        /// @brief 二つの標本のデータ列を指定して構築
        ///
        /// - データ列 1 の指定範囲が有効かを内部でチェックしている。
        ///   有効でない場合は例外 exception::NegativeLength< std::iterator_traits< In1 >::difference_type > を投げる。
        /// - データ列 2 の指定範囲が有効かを内部でチェックしている。
        ///   有効でない場合は例外 exception::NegativeLength< std::iterator_traits< In2 >::difference_type > を投げる。
        ///
        /// @param s1 標本 1 のデータ列の開始位置
        /// @param e1 標本 1 のデータ列の末尾の次の位置
        /// @param s2 標本 2 のデータ列の開始位置
        /// @param e2 標本 2 のデータ列の末尾の次の位置
        template< class In1, class In2 >
        TwoSample( In1 s1, In1 e1, In2 s2, In2 e2 )
          : summary1_( s1, e1 ), summary2_( s2, e2 ) {}

        /// @brief 二標本の分散比を検定する
        ///
        /// - 二つの標本数はそれぞれ 2 以上でなければならない。
        ///   そうでない場合は例外 exception::TooLittleNumber< distribution::Base::counter_type > を投げる。
        /// - 危険率 a は 0 以上 1 以下でなければならない。そうでない場合は例外 exception::UnitInterval を投げる。
        ///
        /// @param a 危険率
        /// @param cip 分散比の推定区間を返す変数のポインタ(NULLならば無視)
        /// @param threshold binSearch で棄却域を求める時のしきい値
        /// @return True ... 仮説を棄却 , False ... 仮説を保留
        bool testVarRatio( distribution::Base::stat_type a, interval_estimation::ConfidenceInterval* cip,
                           distribution::Base::stat_type threshold = DEFAULT_THRESHOLD );

        /// @brief 二標本の平均差を検定する
        ///
        /// - 二つのデータの総数はそれぞれ正値でなければならない。
        ///   そうでない場合は例外 exception::NotPositiveNumber< distribution::Base::counter_type > を投げる。
        /// - 二つのデータの総数の合計は 3 以上でなければならない。
        ///   そうでない場合は例外 exception::TooLittleNumber< distribution::Base::counter_type > を投げる。
        /// - 危険率 a は 0 以上 1 以下でなければならない。そうでない場合は例外 exception::UnitInterval を投げる。
        ///
        /// @param a 危険率
        /// @param cip 分散比の推定区間を返す変数のポインタ(NULLならば無視)
        /// @param threshold binSearch で棄却域を求める時のしきい値
        /// @return True ... 仮説を棄却 , False ... 仮説を保留
        bool testMeanDiff( distribution::Base::stat_type a, interval_estimation::ConfidenceInterval* cip,
                           distribution::Base::stat_type threshold = DEFAULT_THRESHOLD );
      };

      /**
         @brief 二項検定

         小標本用に interval_estimation::Percent_SS が、大標本用に interval_estimation::Percent_LS が用いられる。
      **/
      class Binomial
      {
        distribution::Base::counter_type count_; // 総試行回数
        distribution::Base::counter_type r_;     // 試行の成功した回数

        // 百分率の区間推定 iEst を使って p を危険率 a で検定する
        bool test( interval_estimation::Percent& iEst,
                   distribution::Base::stat_type p, distribution::Base::stat_type a, interval_estimation::ConfidenceInterval* cip,
                   distribution::Base::stat_type threshold );

      public:

        /// @brief 総試行回数と成功した回数を指定して構築
        ///
        /// - 総試行回数と成功回数はゼロより大きくなければならない。
        ///   そうでない場合は例外 exception::NotPositiveNumber< distribution::Base::counter_type > を投げる。
        /// - 総試行回数は成功回数以上でなければならない。
        ///   そうでない場合は例外 exception::TooLittleNumber< distribution::Base::::counter_type > を投げる。
        ///
        /// @param trialCount 総試行回数
        /// @param successCount 成功した回数
        Binomial( distribution::Base::counter_type trialCount, distribution::Base::counter_type successCount )
          : count_( trialCount ), r_( successCount )
        {
          const std::string FUNC_NAME = "tadah_fussy::statistics::test::Binomial constructor";

          error_lib::PNum( count_, FUNC_NAME + " : trialCount " + error_message::PNUM );
          error_lib::PNum( r_, FUNC_NAME + " : successCount " + error_message::PNUM );
          error_lib::GE( count_, r_, FUNC_NAME + " : trialCount " + error_message::GE + " successCount" );
        }

        /// @brief 二項検定(小標本)
        ///
        /// - 百分率 p と危険率 a は 0 以上 1 以下でなければならない。そうでない場合は例外 exception::UnitInterval を投げる。
        ///
        /// @param p 検定対象の百分率
        /// @param a 危険率
        /// @param cip 百分率の推定区間を返す変数のポインタ(NULLならば無視)
        /// @param threshold binSearch で棄却域を求める時のしきい値
        /// @return 仮説を棄却する場合は true を返す
        bool testSS( distribution::Base::stat_type p, distribution::Base::stat_type a, interval_estimation::ConfidenceInterval* cip,
                     distribution::Base::stat_type threshold = DEFAULT_THRESHOLD );

        /// @brief 二項検定(大標本)
        ///
        /// - 百分率 p と危険率 a は 0 以上 1 以下でなければならない。そうでない場合は例外 exception::UnitInterval を投げる。
        ///
        /// @param p 検定対象の百分率
        /// @param a 危険率
        /// @param cip 百分率の推定区間を返す変数のポインタ(NULLならば無視)
        /// @param threshold binSearch で棄却域を求める時のしきい値
        /// @return 仮説を棄却する場合は true を返す
        bool testLS( distribution::Base::stat_type p, distribution::Base::stat_type a, interval_estimation::ConfidenceInterval* cip,
                     distribution::Base::stat_type threshold = DEFAULT_THRESHOLD );
      };

      /**
         @brief カイ二乗検定 (適合度検定)

         一回の試行 E に対し、k 個の事象 ( E1, E2, ... Ek ) の中の一つだけが必ず起こるとしたとき、
         それぞれの事象 Ei ( i = 1, 2, ... k ) が発生する確率が pi、発生回数が ri ならば

         χ^2 = Σi{1→k}( ( ri - N・pi )^2 / ( N・pi ) )

         は自由度 k - 1 の χ2-分布に従う。これを利用して、データの分布が指定した確率密度関数に適合しているかを検定する。
      **/
      class ChiSquare
      {
        using stat_type = distribution::Base::stat_type;
        using counter_type = distribution::Base::counter_type;

        std::vector< stat_type > obj_; // 観測値
        std::vector< stat_type > exp_; // 期待値

        stat_type chiSquare_; // カイ二乗値
        stat_type p_;         // カイ二乗値に対する p 値

        // p 値の計算
        void calcP();

      public:

        /// @brief 各事象が等しい確率で発生するかを検定するためのコンストラクタ
        ///
        /// - 事象の総数は 1 より大きくなければならない。
        ///   そうでない場合は例外 exception::TooLittleNumber< distribution::Base::counter_type > を投げる。
        ///
        /// @param hist 各事象の出現頻度(ヒストグラム)
        template< class T >
        ChiSquare( const Histogram< T, distribution::Base::counter_type >& hist );

        /// @brief 離散確率分布と適合するかを検定するためのコンストラクタ
        ///
        /// - 事象の総数は 1 より大きくなければならない。
        ///   そうでない場合は例外 exception::TooLittleNumber< distribution::Base::counter_type > を投げる。
        ///
        /// @param dist 離散確率分布
        /// @param hist 各事象の出現頻度(ヒストグラム)
        ChiSquare( const distribution::Discrete& dist, const Histogram< distribution::Discrete::var_type, distribution::Base::counter_type >& hist );

        /// @brief 連続確率分布と適合するかを検定するためのコンストラクタ
        ///
        /// - 事象の総数は 1 より大きくなければならない。
        ///   そうでない場合は例外 exception::TooLittleNumber< distribution::Base::counter_type > を投げる。
        ///
        /// @param dist 連続確率分布
        /// @param hist 各事象の出現頻度(ヒストグラム)
        ChiSquare( const distribution::Continuous& dist, const Histogram< histogram::RangeClass< distribution::Base::stat_type >, distribution::Base::counter_type >& hist );

        /// @brief カイ二乗検定(適合度検定)を行う
        ///
        /// - 危険率 a は 0 以上 1 以下でなければならない。そうでない場合は例外 ExceptionUnitInterval を投げる。
        ///
        /// @param a 危険率
        /// @param threshold binSearch で棄却域を求める時のしきい値
        /// @return 仮説を棄却する場合は true を返す
        bool operator()( distribution::Base::stat_type a, distribution::Base::stat_type threshold = DEFAULT_THRESHOLD );

        /// @brief カイ二乗値を返す
        ///
        /// @return カイ二乗値
        distribution::Base::stat_type chiSquare() const { return( chiSquare_ ); }

        /// @brief p 値を返す
        ///
        /// @return p 値
        distribution::Base::stat_type p() const { return( p_ ); }
      };

      /**
         @brief カイ二乗検定 (独立性検定)

         一回の試行 E に対し、k 個の事象 ( A1, A2, ... Ak ) と l 個の事象 ( B1, B2, ... Bl ) の中のそれぞれ一つだけが必ず起こるとする。
         Ai と Bj がすべて独立で、Ai ∩ Bj の発生回数を rij したとき、

         χ^2 = Σi{1→k}( Σj{1→l}( ( rij - ai・bj / N )^2 / ( ai・bj / N ) ) )

         但し、

         N = Σi{1→k}( Σj{1→l}( rij ) )

         ai = Σj{1→l}( rij )

         bj = Σi{1→k}( rij )

         は自由度 ( k - 1 )( l - 1 ) の χ2-分布に従う。これを利用して行・列の間での独立性を検定する。
      **/
      class ChiSquare_Independence
      {
        using stat_type = distribution::Base::stat_type;
        using counter_type = distribution::Base::counter_type;

        // 対象の頻度データ
        Matrix< stat_type > exp_; // 頻度データの期待値

        stat_type chiSquare_; // カイ二乗値
        stat_type p_;         // p 値

        // 期待値を計算する
        void calcExp( const Matrix< counter_type >& data );

      public:

        /// @brief 頻度データと行列のサイズを指定して構築
        ///
        /// - 実測値の行・列のサイズはどちらも 1 より大きくなければならない。
        ///   そうでない場合は例外 exception::TooLittleNumber< distribution::Base::counter_type > を投げる。
        ///
        /// @param data 頻度データの実測値
        ChiSquare_Independence( const Matrix< distribution::Base::counter_type >& data );

        /// @brief カイ二乗検定(独立性検定)を行う
        ///
        /// - 危険率 a は 0 以上 1 以下でなければならない。そうでない場合は例外 ExceptionUnitInterval を投げる。
        ///
        /// @param a 危険率
        /// @param threshold binSearch で棄却域を求める時のしきい値
        /// @return 仮説を棄却する場合は true を返す
        bool operator()( distribution::Base::stat_type a, distribution::Base::stat_type threshold = DEFAULT_THRESHOLD );

        /// @brief カイ二乗値を返す
        ///
        /// @return カイ二乗値
        distribution::Base::stat_type chiSquare() const { return( chiSquare_ ); }

        /// @brief p 値を返す
        ///
        /// @return p 値
        distribution::Base::stat_type p() const { return( p_ ); }
      };

      /**
         @brief フィッシャーの直接確率検定 ( Fisher's Exact Test )

         2 x 2 対応表の場合、カイ二乗検定ではなく多項分布を利用した検定にすることで精度を上げることができる。
      **/
      class FishersExact
      {
        using stat_type = distribution::Base::stat_type;
        using counter_type = distribution::Base::counter_type;

        counter_type rSum0_; // 行 0 の合計
        counter_type rSum1_; // 行 1 の合計
        counter_type cSum0_; // 列 0 の合計
        counter_type cSum1_; // 列 1 の合計
        counter_type total_; // 総計
        stat_type p_; // p 値

        // 四つの数の階乗の積を計算する
        stat_type mulOfFact( counter_type n0, counter_type n1,
                             counter_type n2, counter_type n3 )
        { return( std::tgamma( static_cast< stat_type >( n0 + 1 ) ) * std::tgamma( static_cast< stat_type >( n1 + 1 ) ) *
                  std::tgamma( static_cast< stat_type >( n2 + 1 ) ) * std::tgamma( static_cast< stat_type >( n3 + 1 ) )
                  );
        }

      public:

        /// @brief 2 x 2 対応表の各頻度を指定して構築
        ///
        /// @param d00 ( a0, b0 ) の頻度
        /// @param d10 ( a1, b0 ) の頻度
        /// @param d01 ( a0, b1 ) の頻度
        /// @param d11 ( a1, b1 ) の頻度
        FishersExact( distribution::Base::counter_type d00, distribution::Base::counter_type d10,
                      distribution::Base::counter_type d01, distribution::Base::counter_type d11 );

        /// @brief 検定結果を返す
        ///
        /// - 危険率 a は 0 以上 1 以下でなければならない。そうでない場合は例外 exception::UnitInterval を投げる。
        ///
        /// @param a 危険率
        /// @return 仮説を棄却する場合は true を返す
        bool operator()( distribution::Base::stat_type a ) const
        {
          const std::string FUNC_NAME = "tadah_fussy::statistics::test::FishersExact::operator()";

          error_lib::In01( a, FUNC_NAME + " : a " + error_message::IN01 );

          return( p_ < a );
        }

        /// @brief p 値を返す
        ///
        /// @return p 値
        distribution::Base::stat_type p() const { return( p_ ); }
      };

      /**
         @brief 一元配置分散分析 ( One-way ANOVA ) 用クラス

         一元配置分散分析 ( One-way ANOVA ) を実装するためのクラス。各集団のデータを vector による二次元配列
         の形で渡して構築し、必要なパラメータを計算する。
      **/
      class OneWayANOVA
      {
        using stat_type = distribution::Base::stat_type;
        using counter_type = distribution::Base::counter_type;

        counter_type classCnt_; // 集団の数
        std::vector< Moment1 > classSum_; // 各集団のデータ数と平均
        Moment1 totalSum_;                // 全体のデータ数と平均

        stat_type sc_; // 各集団間の変動
        stat_type se_; // 集団内の変動
        stat_type f_;  // f 値

      public:

        /// @brief 各集団のデータ列を渡して構築
        ///
        /// データは vector (可変長配列) を要素とする vector であり、要素の vector は要素数が異なっていても構わない。
        /// 一つの要素が各集団のデータを表す。
        ///
        /// - 集団数は 1 より大きくなければならない。そうでない場合は例外 exception::TooLittleNumber< distribution::Base::counter_type > を投げる。
        /// - 要素数ゼロの集団があってはならない。そうでない場合は例外 exception::NotPositiveNumber< distribution::Base::counter_type > を投げる。
        /// - 全データの総数は、集団数より多くなければならない。各集団に対してデータが一つしかない場合はこれが成り立たなくなる。
        ///   この場合は例外 exception::TooLittleNumber< distribution::Base::counter_type > を投げる。
        ///
        /// @param data 各集団のデータを持つ配列への参照
        OneWayANOVA( const std::vector< std::vector< distribution::Base::stat_type > >& data );

        /// @brief 集団の個数 k を返す
        ///
        /// @return 集団の個数 k
        distribution::Base::counter_type classCnt() const
        { return( classCnt_ ); }

        /// @brief 指定した添字の集団の集計値 ( モーメント 1 ) を返す
        ///
        /// 集団の個数以上の添字を指定した場合は std::out_of_range 例外を投げる。
        ///
        /// @return 指定した添字の集団の集計値
        Moment1 classSum( size_t i ) const
        { return( classSum_.at( i ) ); }

        /// @brief データの総集計値 ( モーメント 1 ) を返す
        ///
        /// @return データの総集計値
        Moment1 totalSum() const
        { return( totalSum_ ); }

        /// @brief 各集団間の変動 ssB を返す
        ///
        /// 各集団の要素数を ni、平均を mi とし、全体の平均を m としたとき、
        ///
        /// ssB = Σi{1→k}( ni( mi - m )^2 )
        ///
        /// で計算できる。
        ///
        /// @return 各集団間の変動 ssB
        distribution::Base::stat_type ssB() const
        { return( sc_ ); }

        /// @brief 集団内の変動 ssW を返す
        ///
        /// 各集団の要素数を ni、平均を mi、各要素を xij としたとき、
        ///
        /// ssW = Σi{1→k}( Σj{1→ni}( ( xij - mi )^2 ) )
        ///
        /// で計算できる。
        ///
        /// @return 集団内の変動 ssW
        distribution::Base::stat_type ssW() const
        { return( se_ ); }

        /// @brief 全体の変動 ssT を返す
        ///
        /// 各集団の要素数を ni、各要素を xij とし、全体の平均を m としたとき、
        ///
        /// ssT = Σi{1→k}( Σj{1→ni}( ( xij - m )^2 ) )
        ///
        /// で計算できる。
        ///
        /// @return 全体の変動 ssT
        distribution::Base::stat_type ssT() const
        { return( ssB() + ssW() ); }

        /// @brief f 値を返す
        ///
        /// f = varB / varW で計算できる。
        ///
        /// @return f 値
        distribution::Base::stat_type f() const
        { return( f_ ); }

        /// @brief 指定した危険率 での f 値の棄却域を返す
        ///
        /// - 危険率 a は 0 以上 1 以下でなければならない。そうでない場合は例外 exception::UnitInterval を投げる。
        ///
        /// @param a 危険率
        /// @param threshold binSearch で棄却域を求める時のしきい値
        /// @return f 値の棄却域
        distribution::Base::stat_type cr_f( distribution::Base::stat_type a, distribution::Base::stat_type threshold = DEFAULT_THRESHOLD ) const;

        /// @brief 各集団間の変動 ssB に対する自由度を返す
        ///
        /// ssB に対する自由度は k - 1 となる。
        ///
        /// @return 各集団間の変動 ssB に対する自由度
        distribution::Base::counter_type dofB() const
        { return( classCnt() - 1 ); }

        /// @brief 集団内の変動 ssW に対する自由度を返す
        ///
        /// ssW に対する自由度は N - k となる。
        ///
        /// @return 集団内の変動 ssW に対する自由度
        distribution::Base::counter_type dofW() const
        { return( totalSum().count() - classCnt() ); }

        /// @brief 全体の変動 ssT に対する自由度を返す
        ///
        /// ssT に対する自由度は N - 1 となる。
        ///
        /// @return 全体の変動 ssT に対する自由度
        distribution::Base::counter_type dofT() const
        { return( totalSum().count() - 1 ); }

        /// @brief 各集団間の不偏分散 varB を返す
        ///
        /// varB = ssB / ( k - 1 ) で計算できる。
        ///
        /// @return 各集団間の不偏分散 varB
        distribution::Base::stat_type varB() const
        { return( ssB() / dofB() ); }

        /// @brief 集団内の不偏分散 varW を返す
        ///
        /// varW = ssW / ( N - k ) で計算できる。
        ///
        /// @return 集団内の不偏分散 varW
        distribution::Base::stat_type varW() const
        { return( ssW() / dofW() ); }

        /// @brief f 値以上になる確率 ( p 値 ) を返す
        ///
        /// 全データが等しい場合( ssB = ssW = ssT = 0 の場合)は NaN を返す。
        ///
        /// @return p 値
        distribution::Base::stat_type p() const
        {
          if ( std::isnan( f() ) )
            return( distribution::Base::nan() );

          distribution::F fDist( dofB(), dofW() );

          return( 1.0 - fDist.lower_p( f() ) );
        }

        /// @brief 寄与率 (相関比の二乗) R2 を返す
        ///
        /// R2 = ssB / ssT で計算できる。ssT = ssB + ssW より、R2 が小さければ ssB は相対的に小さいことになる。
        /// したがって、集団間の変動よりも集団内の変動が大きいことになり、集団間の平均のばらつきは小さいことを意味する。
        /// 逆に R2 が大きければ、集団内の変動よりも集団間の変動が大きいことになり、各集団の平均に差があることになる。
        ///
        /// 全データが等しい場合( ssB = ssW = ssT = 0 の場合)は NaN を返す。
        ///
        /// @return 寄与率
        distribution::Base::stat_type sqR() const
        { return( ( ssB() + ssW() == 0 ) ?  distribution::Base::nan() : ssB() / ( ssB() + ssW() ) ); }

        /// @brief 一元配置分散分析 ( One-way ANOVA ) を行う
        ///
        /// - 危険率 a は 0 以上 1 以下でなければならない。そうでない場合は例外 ExceptionUnitInterval を投げる。
        ///
        /// @note 全データが等しい場合( ssB = ssW = ssT = 0 の場合)は false を返すことに注意。
        ///
        /// @param a 危険率
        /// @param threshold binSearch で棄却域を求める時のしきい値
        /// @return 仮説を棄却する場合は true を返す
        bool operator()( distribution::Base::stat_type a, distribution::Base::stat_type threshold = DEFAULT_THRESHOLD ) const
        { return( cr_f( a, threshold ) < f() ); }
      };

      /**
         @brief 二元配置分散分析 ( Two-way ANOVA ) 用クラス

         二元配置分散分析 ( Two-way ANOVA ) を実装するためのクラス。各集団のデータを Matrix による行列の形で
         渡して構築し、必要なパラメータを計算する。
      **/
      class TwoWayANOVA
      {
        using stat_type = distribution::Base::stat_type;
        using counter_type = distribution::Base::counter_type;
        // 行列のサイズを表す型
        using size_type = Matrix< stat_type >::size_type;

        counter_type rows_;
        counter_type cols_;

        std::vector< stat_type > mr_; // 行毎の平均
        std::vector< stat_type > mc_; // 列毎の平均
        Moment2 m_; // 全体のモーメント1

        stat_type sr_; // 行間変動
        stat_type sc_; // 列間変動
        stat_type st_; // 全体の変動
        stat_type se_; // 誤差変動

        void calcMean( const Matrix< stat_type >& data ); // 平均値の算出
        void calcVar( const Matrix< stat_type >& data );  // 変動の算出

        // 危険率 a での f 値の棄却域を返す
        //
        // dof は F-分布の一番目の自由度で、行間の f 値なら dofR()、列間の f 値なら dofC() になる。
        stat_type cr_f( counter_type dof, stat_type a, stat_type threshold ) const;

        // F 値以上になる確率 ( p 値 ) を返す
        //
        // dof は F-分布の一番目の自由度で、行間の f 値なら dofR()、列間の f 値なら dofC() になる。
        // f は F 値で、行間の f 値なら fR()、列間の f 値なら fC() になる。
        stat_type p( counter_type dof, stat_type f ) const;

      public:

        /* コンストラクタ */

        /// @brief データを指定して実行
        ///
        /// - 行・列のサイズは 1 より大きくなければならない。
        ///   そうでない場合は例外 exception::TooLittleNumber< distribution::Base::counter_type > を投げる。
        ///
        /// @param data データ(行列の形式)
        TwoWayANOVA( const Matrix< distribution::Base::stat_type >& data );

        /* サイズ */

        /// @brief 行のサイズ l を返す
        ///
        /// @return 行のサイズ l
        distribution::Base::counter_type rows() const { return( rows_ ); }
        /// @brief 列のサイズ k を返す
        ///
        /// @return 列のサイズ k
        distribution::Base::counter_type cols() const { return( cols_ ); }

        /* 平均値 */

        /// @brief 全データの平均を返す
        ///
        /// @return 全データの平均
        distribution::Base::stat_type totalMean() const { return( m_.mean() ); }
        /// @brief 行平均を返す
        ///
        /// - 行番号が範囲外の場合は std::out_of_range を投げる。
        ///
        /// @param r 行番号
        /// @return r 番目の行の平均
        distribution::Base::stat_type rowMean( distribution::Base::counter_type r ) const { return( mr_.at( r ) ); }
        /// @brief 列平均を返す
        ///
        /// - 列番号が範囲外の場合は std::out_of_range を投げる。
        ///
        /// @param c 列番号
        /// @return c 番目の列の平均
        distribution::Base::stat_type colMean( distribution::Base::counter_type c ) const { return( mc_.at( c ) ); }

        /* 変動 */

        /// @brief 行間変動 ssR を返す
        ///
        /// 各行の平均を mr.、全体の平均を m としたとき、
        ///
        /// ssR = kΣr{1→l}( ( mr. - m )^2 )
        ///
        /// で計算できる。
        ///
        /// @return 行間変動 ssR
        distribution::Base::stat_type ssR() const { return( sr_ ); }
        /// @brief 列間変動 ssC を返す
        ///
        /// 各列の平均を m.c、全体の平均を m としたとき、
        ///
        /// ssC = lΣc{1→k}( ( m.c - m )^2 )
        ///
        /// で計算できる。
        ///
        /// @return 列間変動 ssC
        distribution::Base::stat_type ssC() const { return( sc_ ); }
        /// @brief 全体の変動 ssT を返す
        ///
        /// 各行列の要素を xrc、全体の平均を m としたとき、
        ///
        /// ssT = Σr{1→l}( Σc{1→k}( ( xrc - m )^2 ) )
        ///
        /// で計算できる。
        ///
        /// @return 全体の変動 ssT
        distribution::Base::stat_type ssT() const { return( st_ ); }
        /// @brief 誤差変動 ssE を返す
        ///
        /// 各行列の要素を xrc、各行の平均を mr.、各列の平均を m.c、全体の平均を m としたとき、
        ///
        /// ssE = Σr{1→l}( Σc{1→k}( ( xrc - mr. - m.c + m )^2 ) )
        ///
        /// で計算できる。
        ///
        /// @return 誤差変動 ssE
        distribution::Base::stat_type ssE() const { return( se_ ); }

        /* 自由度 */

        /// @brief 行間変動 ssR に対する自由度を返す
        ///
        /// ssR に対する自由度は l - 1 になる。
        ///
        /// @return 行間変動 ssR に対する自由度
        distribution::Base::counter_type dofR() const { return( rows() - 1 ); }
        /// @brief 列間変動 ssC に対する自由度を返す
        ///
        /// ssC に対する自由度は k - 1 になる。
        ///
        /// @return 列間変動 ssC に対する自由度
        distribution::Base::counter_type dofC() const { return( cols() - 1 ); }
        /// @brief 全体の変動 ssT に対する自由度を返す
        ///
        /// ssT に対する自由度は kl - 1 になる。
        ///
        /// @return 全体の変動 ssT に対する自由度
        distribution::Base::counter_type dofT() const { return( rows() * cols() - 1 ); }
        /// @brief 誤差変動 ssE に対する自由度を返す
        ///
        /// ssE に対する自由度は ( k - 1 )( l - 1 ) になる。
        ///
        /// @return 誤差変動 ssE に対する自由度
        distribution::Base::counter_type dofE() const { return( ( rows() - 1 ) * ( cols() - 1 ) ); }

        /* 不偏分散 */

        /// @brief 行間変動の不偏分散を返す
        ///
        /// varR = ssR / ( l - 1 ) で計算できる。
        ///
        /// @return 行間変動の不偏分散 varR
        distribution::Base::stat_type varR() const { return( ssR() / dofR() ); }
        /// @brief 列間変動の不偏分散を返す
        ///
        /// varC = ssC / ( k - 1 ) で計算できる。
        ///
        /// @return 列間変動の不偏分散 varC
        distribution::Base::stat_type varC() const { return( ssC() / dofC() ); }
        /// @brief 誤差変動の不偏分散を返す
        ///
        /// varE = ssE / ( k - 1 )( l - 1 ) で計算できる。
        ///
        /// @return 誤差変動の不偏分散 varE
        distribution::Base::stat_type varE() const { return( ssE() / dofE() ); }

        /* F 値 */

        /// @brief 行間変動に対する F 値を返す
        ///
        /// fR = varR / varE で計算できる。
        ///
        /// @return 行間変動に対する F 値
        distribution::Base::stat_type fR() const { return( ( ssR() == 0 && ssE() == 0 ) ?  distribution::Base::nan() : ssR() * dofC() / ssE() ); } // 行間変動
        /// @brief 列間変動に対する F 値を返す
        ///
        /// fC = varC / varE で計算できる。
        ///
        /// @return 列間変動に対する F 値
        distribution::Base::stat_type fC() const { return( ( ssC() == 0 && ssE() == 0 ) ?  distribution::Base::nan() : ssC() * dofR() / ssE() ); } // 列間変動

        /* F 値の棄却域 */

        /// @brief 指定した危険率 での行間変動に対する F 値の棄却域を返す
        ///
        /// - 危険率 a は 0 以上 1 以下でなければならない。そうでない場合は例外 exception::UnitInterval を投げる。
        ///
        /// @param a 危険率
        /// @param threshold binSearch で棄却域を求める時のしきい値
        /// @return F 値の棄却域
        distribution::Base::stat_type cr_fR( distribution::Base::stat_type a, distribution::Base::stat_type threshold = DEFAULT_THRESHOLD ) const
        { return( cr_f( dofR(), a, threshold ) ); }
        /// @brief 指定した危険率 での列間変動に対する F 値の棄却域を返す
        ///
        /// - 危険率 a は 0 以上 1 以下でなければならない。そうでない場合は例外 ExceptionUnitInterval を投げる。
        ///
        /// @param a 危険率
        /// @param threshold binSearch で棄却域を求める時のしきい値
        /// @return F 値の棄却域
        distribution::Base::stat_type cr_fC( distribution::Base::stat_type a, distribution::Base::stat_type threshold = DEFAULT_THRESHOLD ) const
        { return( cr_f( dofC(), a, threshold ) ); }

        /* p 値 */

        /// @brief 行間変動に対して F 値以上になる確率 ( p 値 ) を返す
        ///
        /// @return p 値
        distribution::Base::stat_type pR() const
        { return( p( dofR(), fR() ) ); }
        /// @brief 列間変動に対して F 値以上になる確率 ( p 値 ) を返す
        ///
        /// @return p 値
        distribution::Base::stat_type pC() const
        { return( p( dofC(), fC() ) ); }

        /* 寄与率 */

        /// @brief 行間変動における寄与率 (相関比の二乗) R2 を返す
        ///
        /// R2 = ssR / ( ssR + ssE ) で計算できる。R2 が大きければ行間の変動が大きいことを意味する。
        ///
        /// @return 行間変動における寄与率
        distribution::Base::stat_type sqR_R() const { return( ( ssR() + ssE() == 0 ) ?  distribution::Base::nan() : ssR() / ( ssR() + ssE() ) ); }
        /// @brief 列間変動における寄与率 (相関比の二乗) R2 を返す
        ///
        /// R2 = ssC / ( ssC + ssE ) で計算できる。R2 が大きければ列間の変動が大きいことを意味する。
        ///
        /// @return 列間変動における寄与率
        distribution::Base::stat_type sqR_C() const { return( ( ssC() + ssE() == 0 ) ?  distribution::Base::nan() : ssC() / ( ssC() + ssE() ) ); }

        /* 検定 */

        /// @brief 行間変動に対して危険率 a で検定を行う
        ///
        /// - 危険率 a は 0 以上 1 以下でなければならない。そうでない場合は例外 exception::UnitInterval を投げる。
        ///
        /// @param a 危険率
        /// @param threshold binSearch で棄却域を求める時のしきい値
        /// @return 仮説を棄却する場合は true を返す
        bool testR( distribution::Base::stat_type a, distribution::Base::stat_type threshold = DEFAULT_THRESHOLD ) const
        { return( cr_fR( a, threshold ) < fR() ); }
        /// @brief 列間変動に対して危険率 a で検定を行う
        ///
        /// - 危険率 a は 0 以上 1 以下でなければならない。そうでない場合は例外 ExceptionUnitInterval を投げる。
        ///
        /// @param a 危険率
        /// @param threshold binSearch で棄却域を求める時のしきい値
        /// @return 仮説を棄却する場合は true を返す
        bool testC( distribution::Base::stat_type a, distribution::Base::stat_type threshold = DEFAULT_THRESHOLD ) const
        { return( cr_fC( a, threshold ) < fC() ); }
      };

      /**
         @brief 反復のある二元配置分散分析 ( Two-way Repeated-Measures ANOVA ) 用クラス

         反復のある二元配置分散分析 ( Two-way Repeated-Measures ANOVA ) を実装するためのクラス。
         各集団のデータ(行列)は、反復測定などの理由で複数存在している。
      **/
      class RepMeasANOVA
      {
        using stat_type = distribution::Base::stat_type;
        using counter_type = distribution::Base::counter_type;
        // データ列 (二次元配列)
        using mat_type = Matrix< stat_type >;
        // 行列のサイズを表す型
        using size_type = mat_type::size_type;

        mat_type data_;

        counter_type classCnt_; // データ列の数(測定回数)
        counter_type cols_;     // 1 データあたりの列数

        std::vector< stat_type > mr_; // 行毎の平均
        std::vector< stat_type > mc_; // 列毎の平均
        mat_type mrc_;                // 行列毎の平均
        stat_type m_;                 // 全体の平均

        stat_type sr_;  // 行間変動
        stat_type sc_;  // 列間変動
        stat_type st_;  // 全体の変動
        stat_type se_;  // 誤差変動
        stat_type src_; // 交互作用

        void calcMean(); // 平均値の算出
        void calcVar();  // 変動の算出

        // 危険率 a での f 値の棄却域を返す
        //
        // dof は F-分布の一番目の自由度で、行間の f 値なら dofR()、列間の f 値なら dofC()、交互作用なら dofRC() になる。
        stat_type cr_f( counter_type dof, stat_type a, stat_type threshold ) const;

        // F 値以上になる確率 ( p 値 ) を返す
        //
        // dof は F-分布の一番目の自由度で、行間の f 値なら dofR()、列間の f 値なら dofC()、交互作用なら dofRC() になる。
        // f は F 値で、行間の f 値なら fR()、列間の f 値なら fC()、交互作用なら fRC() になる。
        stat_type p( counter_type dof, stat_type f ) const;

      public:

        /// @brief データ列と集団の数、行列のサイズを指定して実行
        ///
        /// データ列は、各集団(行列)単位で並び、各集団に対しては行方向から先にデータが並ぶと見なされる。
        /// 従って、R 行 C 列のサイズの集団に対し、i 番目の集団の j 行 k 列のデータは
        ///
        /// s[ i*R*C + k*R + j ]
        ///
        /// になる。
        ///
        /// - 集団の数、行・列のサイズは全て 1 より大きくなければならない。
        ///   そうでない場合は例外 exception::TooLittleNumber< distribution::Base::counter_type > を投げる。
        /// - データ列の開始・終了位置の指定が正しくない(位置が逆転している)場合、
        ///   例外 exception::NegativeLength< std::iterator_traits< In >::difference_type > を投げる。
        /// - 指定した集団の数、行・列のサイズに対し、データ列のサイズは一致していなければならない。
        ///   そうでない場合は例外 exception::NotEqualLength< std::iterator_traits< In >::difference_type > を投げる。
        /// - 交互作用と誤差変動がともにゼロである場合(全データが等しい場合)、交互作用に対する F 値は NaN になる。
        ///   このとき、交互作用の検定結果は必ず保留となることに注意。
        ///
        /// @param s データ列の開始位置
        /// @param e データ列の末尾の次の位置
        /// @param classCnt 集団の数
        /// @param rowSize 行のサイズ
        /// @param colSize 列のサイズ
        template< class In >
        RepMeasANOVA( In s, In e, distribution::Base::counter_type classCnt,
                      distribution::Base::counter_type rowSize, distribution::Base::counter_type colSize );

        /* サイズ */

        /// @brief 集団の数 ( 測定回数 ) n を返す
        ///
        /// @return 集団の数 ( 測定回数 ) n
        distribution::Base::counter_type classCnt() const { return( classCnt_ ); }
        /// @brief 行のサイズを l 返す
        ///
        /// @return 行のサイズ l
        distribution::Base::counter_type rows() const { return( data_.rows() ); }
        /// @brief 列のサイズ k を返す
        ///
        /// @return 列のサイズ k
        distribution::Base::counter_type cols() const { return( cols_ ); }

        /* 平均 */

        /// @brief 全データの平均を返す
        ///
        /// @return 全データの平均
        distribution::Base::stat_type totalMean() const { return( m_ ); }
        /// @brief 行平均を返す
        ///
        /// @param r 行番号
        /// @return r 番目の行の平均
        distribution::Base::stat_type rowMean( distribution::Base::counter_type r ) const
        { return( mr_.at( r ) ); }
        /// @brief 列平均を返す
        ///
        /// @param c 列番号
        /// @return c 番目の列の平均
        distribution::Base::stat_type colMean( distribution::Base::counter_type c ) const
        { return( mc_.at( c ) ); }
        /// @brief 行列ごとの平均を返す
        ///
        /// @param r 行番号
        /// @param c 列番号
        /// @return ( r, c ) 番目の行列の平均
        distribution::Base::stat_type elemMean( distribution::Base::counter_type r,  distribution::Base::counter_type c ) const
        {
          return( mrc_.row( r ).at( c ) );
        }

        /* 変動 */

        /// @brief 行間変動 ssR を返す
        ///
        /// 各行の平均を mr..、全体の平均を m としたとき、
        ///
        /// ssR = knΣr{1→l}( ( mr.. - m )^2 )
        ///
        /// で計算できる。
        ///
        /// @return 行間変動 ssR
        distribution::Base::stat_type ssR() const { return( sr_ ); }
        /// @brief 列間変動 ssC を返す
        ///
        /// 各列の平均を m.c.、全体の平均を m としたとき、
        ///
        /// ssC = lnΣc{1→k}( ( m.c. - m )^2 )
        ///
        /// で計算できる。
        ///
        /// @return 列間変動 ssC
        distribution::Base::stat_type ssC() const { return( sc_ ); }
        /// @brief 全体の変動 ssT を返す
        ///
        /// 行 r、列 c の i 番目の要素を xrci、全体の平均を m としたとき、
        ///
        /// ssT = Σr{1→l}( Σc{1→k}( Σi{1→n}( ( xrci - m )^2 ) ) )
        ///
        /// で計算できる。
        ///
        /// @return 全体の変動 ssT
        distribution::Base::stat_type ssT() const { return( st_ ); }
        /// @brief 誤差変動 ssE を返す
        ///
        /// 各行列の i 番目の要素を xrci、平均を mrc. としたとき、
        ///
        /// ssE = Σr{1→l}( Σc{1→k}( Σi{1→n}( ( xrci - mrc. )^2 ) ) )
        ///
        /// で計算できる。
        ///
        /// @return 誤差変動 ssE
        distribution::Base::stat_type ssE() const { return( se_ ); }
        /// @brief 交互作用 ssRC を返す
        ///
        /// 各行の平均を mr..、各列の平均を m.c.、各行列の平均を mrc.、全体の平均を m としたとき、
        ///
        /// ssRC = nΣr{1→l}( Σc{1→k}( ( mrc. - mr.. - m.c. + m )^2 ) )
        ///
        /// で計算できる。
        ///
        /// @return 交互作用 ssRC
        distribution::Base::stat_type ssRC() const { return( src_ ); }

        /* 自由度 */

        /// @brief 行間変動 ssR に対する自由度を返す
        ///
        /// 行間変動 ssR に対する自由度は l - 1 になる。
        ///
        /// @return 行間変動 ssR に対する自由度
        distribution::Base::counter_type dofR() const { return( rows() - 1 ); }
        /// @brief 列間変動 ssC に対する自由度を返す
        ///
        /// 列間変動 ssC に対する自由度は k - 1 になる。
        ///
        /// @return 列間変動 ssC に対する自由度
        distribution::Base::counter_type dofC() const { return( cols() - 1 ); }
        /// @brief 全体の変動 ssT に対する自由度を返す
        ///
        /// 全体の変動 ssT に対する自由度は kln - 1 になる。
        ///
        /// @return 全体の変動 ssT に対する自由度
        distribution::Base::counter_type dofT() const { return( rows() * cols() * classCnt() - 1 ); }
        /// @brief 誤差変動 ssE に対する自由度を返す
        ///
        /// 誤差変動 ssE に対する自由度は kl( n - 1 ) になる。
        ///
        /// @return 誤差変動 ssE に対する自由度
        distribution::Base::counter_type dofE() const { return( rows() * cols() * ( classCnt() - 1 ) ); }
        /// @brief 交互作用 ssRC に対する自由度を返す
        ///
        /// 交互作用 ssRC に対する自由度は ( k - 1 )( l - 1 ) になる。
        ///
        /// @return 交互作用 ssRC に対する自由度
        distribution::Base::counter_type dofRC() const { return( ( rows() - 1 ) * ( cols() - 1 ) ); }

        /* 不偏分散 */

        /// @brief 行間変動の不偏分散 varRを返す
        ///
        /// varR = ssR / ( l - 1 ) で計算できる。
        ///
        /// @return 行間変動の不偏分散 varR
        distribution::Base::stat_type varR() const { return( ssR() / dofR() ); }
        /// @brief 列間変動の不偏分散 varCを返す
        ///
        /// varC = ssC / ( k - 1 ) で計算できる。
        ///
        /// @return 列間変動の不偏分散 varC
        distribution::Base::stat_type varC() const { return( ssC() / dofC() ); }
        /// @brief 誤差変動の不偏分散 varE を返す
        ///
        /// varE = ssE / kl( n - 1 ) で計算できる。
        ///
        /// @return 誤差変動の不偏分散 varE
        distribution::Base::stat_type varE() const { return( ssE() / dofE() ); }
        /// @brief 交互作用の不偏分散 varRC を返す
        ///
        /// varRC = ssRC / ( k - 1 )( l - 1 ) で計算できる。
        ///
        /// @return 交互作用の不偏分散 varRC
        distribution::Base::stat_type varRC() const { return( ssRC() / dofRC() ); }

        /* F 値 */

        /// @brief 行間変動に対する F 値 fR を返す
        ///
        /// fR = varR / varE で計算できる。
        ///
        /// @return 行間変動に対する F 値 fR
        distribution::Base::stat_type fR() const { return( ( ssR() == 0 && ssE() == 0 ) ? distribution::Base::nan() : ssR() * dofE() / ( ssE() * dofR() ) ); }
        /// @brief 列間変動に対する F 値 fC を返す
        ///
        /// fC = varC / varE で計算できる。
        ///
        /// @return 列間変動に対する F 値 fC
        distribution::Base::stat_type fC() const { return( ( ssC() == 0 && ssE() == 0 ) ? distribution::Base::nan() : ssC() * dofE() / ( ssE() * dofC() ) ); }
        /// @brief 交互作用に対する F 値 fRC を返す
        ///
        /// fRC = varRC / varE で計算できる。
        ///
        /// @return 交互作用に対する F 値 fRC
        distribution::Base::stat_type fRC() const { return( ( ssRC() == 0 && ssE() == 0 ) ? distribution::Base::nan() : ssRC() * dofE() / ( ssE() * dofRC() ) ); }

        /* F 値の棄却域 */

        /// @brief 指定した危険率 での行間変動に対する F 値の棄却域を返す
        ///
        /// - 危険率 a は 0 以上 1 以下でなければならない。そうでない場合は例外 exception::UnitInterval を投げる。
        ///
        /// @param a 危険率
        /// @param threshold binSearch で棄却域を求める時のしきい値
        /// @return F 値の棄却域
        distribution::Base::stat_type cr_fR( distribution::Base::stat_type a, distribution::Base::stat_type threshold = DEFAULT_THRESHOLD ) const
        { return( cr_f( dofR(), a, threshold ) ); }
        /// @brief 指定した危険率 での列間変動に対する F 値の棄却域を返す
        ///
        /// - 危険率 a は 0 以上 1 以下でなければならない。そうでない場合は例外 exception::UnitInterval を投げる。
        ///
        /// @param a 危険率
        /// @param threshold binSearch で棄却域を求める時のしきい値
        /// @return F 値の棄却域
        distribution::Base::stat_type cr_fC( distribution::Base::stat_type a, distribution::Base::stat_type threshold = DEFAULT_THRESHOLD ) const
        { return( cr_f( dofC(), a, threshold ) ); }
        /// @brief 指定した危険率 での交互作用に対する F 値の棄却域を返す
        ///
        /// - 危険率 a は 0 以上 1 以下でなければならない。そうでない場合は例外 exception::UnitInterval を投げる。
        ///
        /// @param a 危険率
        /// @param threshold binSearch で棄却域を求める時のしきい値
        /// @return F 値の棄却域
        distribution::Base::stat_type cr_fRC( distribution::Base::stat_type a, distribution::Base::stat_type threshold = DEFAULT_THRESHOLD ) const
        { return( cr_f( dofRC(), a, threshold ) ); }

        /* p 値 */

        /// @brief 行間変動に対して F 値以上になる確率 ( p 値 ) を返す
        ///
        /// @return p 値
        distribution::Base::stat_type pR() const
        { return( p( dofR(), fR() ) ); }
        /// @brief 列間変動に対して F 値以上になる確率 ( p 値 ) を返す
        ///
        /// @return p 値
        distribution::Base::stat_type pC() const
        { return( p( dofC(), fC() ) ); }
        /// @brief 交互作用に対して F 値以上になる確率 ( p 値 ) を返す
        ///
        /// @return p 値
        distribution::Base::stat_type pRC() const
        { return( p( dofRC(), fRC() ) ); }

        /* 寄与率 */

        /// @brief 行間変動における寄与率 (相関比の二乗) R2 を返す
        ///
        /// R2 = ssR / ( ssR + ssE ) で計算できる。R2 が大きければ行間の変動が大きいことを意味する。
        ///
        /// @return 行間変動における寄与率
        distribution::Base::stat_type sqR_R() const { return( ( ssR() + ssE() == 0 ) ? distribution::Base::nan() : ssR() / ( ssR() + ssE() ) ); }
        /// @brief 列間変動における寄与率 (相関比の二乗) R2 を返す
        ///
        /// R2 = ssC / ( ssC + ssE ) で計算できる。R2 が大きければ列間の変動が大きいことを意味する。
        ///
        /// @return 列間変動における寄与率
        distribution::Base::stat_type sqR_C() const { return( ( ssC() + ssE() == 0 ) ? distribution::Base::nan() : ssC() / ( ssC() + ssE() ) ); }
        /// @brief 交互作用における寄与率 (相関比の二乗) R2 を返す
        ///
        /// R2 = ssRC / ( ssRC + ssE ) で計算できる。R2 が大きければ交互作用が大きいことを意味する。
        ///
        /// @return 交互作用における寄与率
        distribution::Base::stat_type sqR_RC() const { return( ( ssRC() + ssE() == 0 ) ? distribution::Base::nan() : ssRC() / ( ssRC() + ssE() ) ); }

        /* 検定 */

        /// @brief 行間変動に対して危険率 a で検定を行う
        ///
        /// - 危険率 a は 0 以上 1 以下でなければならない。そうでない場合は例外 exception::UnitInterval を投げる。
        ///
        /// @param a 危険率
        /// @param threshold binSearch で棄却域を求める時のしきい値
        /// @return 仮説を棄却する場合は true を返す
        bool testR( distribution::Base::stat_type a, distribution::Base::stat_type threshold = DEFAULT_THRESHOLD ) const
        { return( cr_fR( a, threshold ) < fR() ); }
        /// @brief 列間変動に対して危険率 a で検定を行う
        ///
        /// - 危険率 a は 0 以上 1 以下でなければならない。そうでない場合は例外 exception::UnitInterval を投げる。
        ///
        /// @param a 危険率
        /// @param threshold binSearch で棄却域を求める時のしきい値
        /// @return 仮説を棄却する場合は true を返す
        bool testC( distribution::Base::stat_type a, distribution::Base::stat_type threshold = DEFAULT_THRESHOLD ) const
        { return( cr_fC( a, threshold ) < fC() ); }
        /// @brief 交互作用に対して危険率 a で検定を行う
        ///
        /// - 危険率 a は 0 以上 1 以下でなければならない。そうでない場合は例外 exception::UnitInterval を投げる。
        ///
        /// @param a 危険率
        /// @param threshold binSearch で棄却域を求める時のしきい値
        /// @return 仮説を棄却する場合は true を返す
        bool testRC( distribution::Base::stat_type a, distribution::Base::stat_type threshold = DEFAULT_THRESHOLD ) const
        { return( cr_fRC( a, threshold ) < fRC() ); }
      };

      /**
         @brief 共分散分析 ( ANCOVA ) 用クラス

         従属変数に影響を及ぼす独立変数(共変数)に対し、その影響を取り除いた上で集団間の差異を検定する。
         集団内の変動に対する F 値 fC、交互作用に対する F 値 fI がともに小さければ、共変数の影響はなく、
         通常の一元配置分散分析法を利用すれば充分であることになる。
         fC が大きい場合、従属変数は共変数の影響を受けており、fI が小さければその回帰式の傾きはほぼ等しい。
         よって、共変数の影響を補正した集団間差に対する F 値 fB を使って差異があるかを検定することができる。
      **/
      class ANCOVA
      {
        // 確率変数・統計量用変数の型
        using stat_type = distribution::Base::stat_type;
        using counter_type = distribution::Base::counter_type;

        counter_type classCnt_;               // 集団の数
        std::vector< MomentPair2 > classSum_; // 各集団のデータ数と平均
        MomentPair2 totalSum_;                // 全体のデータ数と平均

        stat_type sb_; // 各集団間の変動
        stat_type sc_; // 集団内の変動
        stat_type si_; // 交互作用

        // 危険率 a での f 値の棄却域を返す
        //
        // dof は F-分布の一番目の自由度で、集団内の F 値なら dofC()、交互作用の F 値なら dofI()、補正した集団間の F 値なら dofB() になる。
        stat_type cr_f( counter_type dof, stat_type a, stat_type threshold ) const;

        // F 値以上になる確率 ( p 値 ) を返す
        //
        // dof は F-分布の一番目の自由度で、集団内の F 値なら dofC()、交互作用の F 値なら dofI()、補正した集団間の F 値なら dofB() になる。
        // f は F 値で、集団内の F 値なら fC()、交互作用の F 値なら fI()、補正した集団間の変動なら fB() になる。
        stat_type p( counter_type dof, stat_type f ) const;

      public:

        /// @brief 集団の値の型
        ///
        /// 独立変数(共変数)と従属変数のペアを表す
        using value_type = std::pair< distribution::Base::stat_type, distribution::Base::stat_type >;

        /// @brief 各集団のデータ列を渡して構築
        ///
        /// データは vector (可変長配列) を要素とする vector であり、要素の vector は要素数が異なっていても構わない。
        /// 一つの要素が各集団のデータを表す。データは x, y のペア ( std::pair ) であることに注意。
        ///
        /// - 集団の数は 1 より大きくなければならない。
        ///   そうでない場合は例外 exception::TooLittleNumber< distribution::Base::counter_type > を投げる。
        /// - 全集団のデータ数の合計は集団の数の二倍より大きくなければならない。
        ///   そうでない場合は例外 exception::TooLittleNumber< distribution::Base::counter_type > を投げる。
        /// - 残差 ( ssE() ) の他に、集団内・交互作用・補正した集団間の変動 ( ssC(), ssI(), ssCorB() ) がゼロまたは NaN になると、
        ///   それぞれに対応する F 値 ( fC(), fI(), fB() ) は NaN になる。このとき、検定結果は保留となることに注意。
        ///
        /// @param data 各集団のデータを持つ配列への参照
        ANCOVA( const std::vector< std::vector< value_type > >& data );

        /* サイズ・集約値(モーメント 2) */

        /// @brief 集団の数 p を返す
        ///
        /// @return 集団の数 p
        distribution::Base::counter_type classCnt() const { return( classCnt_ ); }
        /// @brief 指定した添字の集団の集計値 ( モーメント 2 ) を返す
        ///
        /// 集団の個数以上の添字を指定した場合は out_of_range 例外を投げる。
        ///
        /// @return 指定した添字の集団の集計値
        MomentPair2 classSum( size_t i ) const
        { return( classSum_.at( i ) ); }
        /// @brief データの総集計値 ( モーメント 2 ) を返す
        ///
        /// @return データの総集計値
        MomentPair2 totalSum() const
        { return( totalSum_ ); }

        /* 変動 */

        /// @brief 集団間の変動 ssB を返す
        ///
        /// 各集団の要素数を ni、y に対する平均を my,i とし、y の全体の平均を my としたとき、
        ///
        /// ssB = Σi{1→p}( ni( my,i - my )^2 )
        ///
        /// で計算できる。
        ///
        /// @return 各集団間の変動 ssB
        distribution::Base::stat_type ssB() const
        { return( sb_ ); }
        /// @brief 集団内の変動 ssC を返す
        ///
        /// 各集団の要素数を ni、共分散を sxy,i、x の分散を vx,i としたとき、
        ///
        /// ssC = [ Σi{1→p}( ni・sxy,i ) ]^2 / Σi{1→p}( ni・vx,i )
        ///
        /// で計算できる。
        ///
        /// @return 集団内の変動 ssC
        distribution::Base::stat_type ssC() const
        { return( sc_ ); }
        /// @brief 交互作用 ssI を返す
        ///
        /// 各集団の要素数を ni、共分散を sxy,i、x の分散を vx,i としたとき、
        ///
        /// ssI = Σi{1→p}( ni・sxy,i^2 / vx,i ) - ssW
        ///
        /// で計算できる。
        ///
        /// @return 交互作用 ssI
        distribution::Base::stat_type ssI() const
        { return( si_ ); }
        /// @brief 補正した集団間の変動 ssCorB を返す
        ///
        /// 全要素の要素数を N 、共分散を sxy、x の分散を vx としたとき、
        ///
        /// ssCorB = ssC + ssB - N・sxy^2 / vx
        ///
        /// で計算できる。各集団の独立変数と従属変数に共通の傾きを持つ回帰式が成り立つと仮定したとき、
        /// 補正した集団間の変動はその影響を取り除いた上での集団間の差異を表す。
        ///
        /// @return 補正した各集団間の変動 ssCorB
        distribution::Base::stat_type ssCorB() const
        {
          stat_type rev = std::pow( totalSum_.proDev(), 2 ) / totalSum_.devSq1();
          return( ( sb_ + sc_ <= rev ) ? 0 : sb_ + sc_ - rev );
        }
        /// @brief 残差 ssE を返す
        ///
        /// 各集団の要素数を ni、y の実測値を yi,j、予測値を y^i,j としたとき、
        ///
        /// ssE = Σi{1→p}( Σj{1→ni}( ( yi,j - y^i,j )^2 ) )
        ///
        /// で計算できる。
        ///
        /// @return 残差 ssE
        distribution::Base::stat_type ssE() const
        { return( ( totalSum_.devSq2() <= sb_ + sc_ + si_ ) ? 0 : totalSum_.devSq2() - sb_ - sc_ - si_ ); }
        /// @brief 全体の変動 ssT を返す
        ///
        /// 各集団の要素数を ni、y の実測値を yi,j とし、y の全要素の平均を my としたとき、
        ///
        /// ssT = Σi{1→p}( Σj{1→ni}( ( yi,j - my )^2 ) )
        ///
        /// で計算できる。
        ///
        /// @return 全体の変動 ssT
        distribution::Base::stat_type ssT() const
        { return( totalSum_.devSq2() ); }

        /* 自由度 */

        /// @brief 各集団間の変動 ssB に対する自由度を返す
        ///
        /// 各集団間の変動 ssB に対する自由度は p - 1 になる。補正した集団間の変動に対しても同じ値となる。
        ///
        /// @return 各集団間の変動 ssB に対する自由度を返す
        distribution::Base::counter_type dofB() const
        { return( classCnt_ - 1 ); }
        /// @brief 集団内の変動 ssC に対する自由度を返す
        ///
        /// 集団内の変動 ssC に対する自由度は常に 1 になる
        ///
        /// @return 集団内の変動 ssC に対する自由度を返す
        static distribution::Base::counter_type dofC()
        { return( 1 ); }
        /// @brief 交互作用 ssI に対する自由度を返す
        ///
        /// 交互作用 ssI に対する自由度は p - 1 になる
        ///
        /// @return 交互作用 ssI に対する自由度を返す
        distribution::Base::counter_type dofI() const
        { return( classCnt_ - 1 ); }
        /// @brief 残差 ssE に対する自由度を返す
        ///
        /// 残差 ssE に対する自由度は N - 2p になる。但し、N は全要素数を表す。
        ///
        /// @return 残差 ssE に対する自由度を返す
        distribution::Base::counter_type dofE() const
        { return( totalSum_.count() - 2 * classCnt_ ); }
        /// @brief 全体の変動 ssT に対する自由度を返す
        ///
        /// 全体の変動 ssT に対する自由度は N - 1 になる。但し、N は全要素数を表す。
        ///
        /// @return 全体の変動 ssT に対する自由度を返す
        distribution::Base::counter_type dofT() const
        { return( totalSum_.count() - 1 ); }

        /* 不偏分散 */

        /// @brief 各集団間の不偏分散 varB を返す
        ///
        /// varB = ssB / ( p - 1 ) で計算できる。
        ///
        /// @return 各集団間の不偏分散 varB
        distribution::Base::stat_type varB() const { return( ssB() / dofB() ); }
        /// @brief 集団内の不偏分散 varC を返す
        ///
        /// varC = ssC である。
        ///
        /// @return 集団内の不偏分散 varC
        distribution::Base::stat_type varC() const { return( ssC() / dofC() ); }
        /// @brief 交互作用の不偏分散 varI を返す
        ///
        /// varI = ssI / ( p - 1 ) で計算できる。
        ///
        /// @return 交互作用の不偏分散 varI
        distribution::Base::stat_type varI() const { return( ssI() / dofI() ); }
        /// @brief 補正した各集団間の不偏分散 varCorBを返す
        ///
        /// varCorB = ssCorB / ( p - 1 ) で計算できる。
        ///
        /// @return 補正した各集団間の不偏分散 varCorB
        distribution::Base::stat_type varCorB() const { return( ssCorB() / dofB() ); }
        /// @brief 残差の不偏分散 varE を返す
        ///
        /// varE = ssE / ( N - 2p ) で計算できる。但し、N は全要素の数である。
        ///
        /// @return 残差の不偏分散 varE
        distribution::Base::stat_type varE() const { return( ssE() / dofE() ); }

        /* F 値 */

        /// @brief 集団内の変動 ssC に対する F 値 fC を返す
        ///
        /// fC = varC / varE で計算できる。
        ///
        /// @return 集団内の変動に対する F 値 fC
        distribution::Base::stat_type fC() const
        { return( ( ssC() == 0 && ssE() == 0 ) ? distribution::Base::nan() : ssC() * dofE() / ssE() ); }
        /// @brief 交互作用 ssI に対する F 値 fI を返す
        ///
        /// fI = varI / varE で計算できる。
        ///
        /// @return 交互作用に対する F 値 fI
        distribution::Base::stat_type fI() const
        { return( ( ssI() == 0 && ssE() == 0 ) ? distribution::Base::nan() : ssI() * dofE() / ssE() * dofI() ); }
        /// @brief 補正した集団間の変動 ssCorB に対する F 値 fB を返す
        ///
        /// fB = varCorB / varE で計算できる。
        ///
        /// @return 補正した集団間の変動に対する F 値 fB
        distribution::Base::stat_type fB() const
        { return( ( ssCorB() == 0 && ssE() == 0 ) ? distribution::Base::nan() : ssCorB() * dofE() / ssE() * dofB() ); }

        /* 回帰式 */

        /// @brief 指定した添字の集団の回帰式の傾きと切片を返す
        ///
        /// 集団の個数以上の添字を指定した場合は out_of_range 例外を投げる。
        ///
        /// @param i 対象の集団の添字
        /// @param a 傾きを取得する変数へのポインタ
        /// @param b 切片を取得する変数へのポインタ
        /// @return なし
        void classRegExp( size_t i, distribution::Base::stat_type* a, distribution::Base::stat_type* b ) const
        {
          if ( a != nullptr )
            *a = classSum_.at( i ).proDev() / classSum_.at( i ).devSq1();
          if ( b != nullptr )
            *b = ( classSum_.at( i ).mean2() * classSum_.at( i ).devSq1() - classSum_.at( i ).mean1() * classSum_.at( i ).proDev() ) / classSum_.at( i ).devSq1();
        }

        /// @brief 全データに対する回帰式の傾きと切片を返す
        ///
        /// @param a 傾きを取得する変数へのポインタ
        /// @param b 切片を取得する変数へのポインタ
        /// @return なし
        void totalRegExp( distribution::Base::stat_type* a, distribution::Base::stat_type* b ) const
        {
          if ( a != nullptr )
            *a = totalSum_.proDev() / totalSum_.devSq1();
          if ( b != nullptr )
            *b = ( totalSum_.mean2() * totalSum_.devSq1() - totalSum_.mean1() * totalSum_.proDev() ) / totalSum_.devSq1();
        }

        /* F 値の棄却域 */

        /// @brief 指定した危険率 での集団内の変動に対する F 値の棄却域を返す
        ///
        /// - 危険率 a は 0 以上 1 以下でなければならない。そうでない場合は例外 exception::UnitInterval を投げる。
        ///
        /// @param a 危険率
        /// @param threshold binSearch で棄却域を求める時のしきい値
        /// @return F 値の棄却域
        distribution::Base::stat_type cr_fC( distribution::Base::stat_type a, distribution::Base::stat_type threshold = DEFAULT_THRESHOLD ) const
        { return( cr_f( dofC(), a, threshold ) ); }
        /// @brief 指定した危険率 での交互作用に対する F 値の棄却域を返す
        ///
        /// - 危険率 a は 0 以上 1 以下でなければならない。そうでない場合は例外 exception::UnitInterval を投げる。
        ///
        /// @param a 危険率
        /// @param threshold binSearch で棄却域を求める時のしきい値
        /// @return F 値の棄却域
        distribution::Base::stat_type cr_fI( distribution::Base::stat_type a, distribution::Base::stat_type threshold = DEFAULT_THRESHOLD ) const
        { return( cr_f( dofI(), a, threshold ) ); }
        /// @brief 指定した危険率 での補正した集団間の変動に対する F 値の棄却域を返す
        ///
        /// - 危険率 a は 0 以上 1 以下でなければならない。そうでない場合は例外 exception::UnitInterval を投げる。
        ///
        /// @param a 危険率
        /// @param threshold binSearch で棄却域を求める時のしきい値
        /// @return F 値の棄却域
        distribution::Base::stat_type cr_fB( distribution::Base::stat_type a, distribution::Base::stat_type threshold = DEFAULT_THRESHOLD ) const
        { return( cr_f( dofB(), a, threshold ) ); }

        /* p 値 */

        /// @brief 集団内の変動に対して F 値以上になる確率 ( p 値 ) を返す
        ///
        /// @return p 値
        distribution::Base::stat_type pC() const
        { return( p( dofC(), fC() ) ); }
        /// @brief 交互作用に対して F 値以上になる確率 ( p 値 ) を返す
        ///
        /// @return p 値
        distribution::Base::stat_type pI() const
        { return( p( dofI(), fI() ) ); }
        /// @brief 補正した集団間の変動に対して F 値以上になる確率 ( p 値 ) を返す
        ///
        /// @return p 値
        distribution::Base::stat_type pB() const
        { return( p( dofB(), fB() ) ); }

        /* 検定 */

        /// @brief 集団内の変動に対して危険率 a で検定を行う
        ///
        /// - 危険率 a は 0 以上 1 以下でなければならない。そうでない場合は例外 exception::UnitInterval を投げる。
        ///
        /// @param a 危険率
        /// @param threshold binSearch で棄却域を求める時のしきい値
        /// @return 仮説を棄却する場合は true を返す
        bool testC( distribution::Base::stat_type a, distribution::Base::stat_type threshold = DEFAULT_THRESHOLD ) const
        { return( cr_fC( a, threshold ) < fC() ); }
        /// @brief 交互作用に対して危険率 a で検定を行う
        ///
        /// - 危険率 a は 0 以上 1 以下でなければならない。そうでない場合は例外 exception::UnitInterval を投げる。
        ///
        /// @param a 危険率
        /// @param threshold binSearch で棄却域を求める時のしきい値
        /// @return 仮説を棄却する場合は true を返す
        bool testI( distribution::Base::stat_type a, distribution::Base::stat_type threshold = DEFAULT_THRESHOLD ) const
        { return( cr_fI( a, threshold ) < fI() ); }
        /// @brief 補正した集団間の変動に対して危険率 a で検定を行う
        ///
        /// - 危険率 a は 0 以上 1 以下でなければならない。そうでない場合は例外 exception::UnitInterval を投げる。
        ///
        /// @param a 危険率
        /// @param threshold binSearch で棄却域を求める時のしきい値
        /// @return 仮説を棄却する場合は true を返す
        bool testB( distribution::Base::stat_type a, distribution::Base::stat_type threshold = DEFAULT_THRESHOLD ) const
        { return( cr_fB( a, threshold ) < fB() ); }
      };

    } // namespace test

    /**
       順序統計量
    **/

    /// @brief 順序統計量の確率密度関数
    ///
    /// 確率密度関数 p(x) に従う確率変数 x から N 個の標本を抽出したとき、
    /// j 番目の順序統計量 x(j) の確率密度関数は
    ///
    /// { N! / ( j - 1 )!( N - j )! }・{ ∫{-∞→x} p(t) dt }^(j-1)・{ ∫{x→∞} p(t) dt }^(N-j)・p(x)
    ///
    /// で求められる。
    ///
    /// @param x 確率密度を求める対象の確率変数
    /// @param n 抽出した標本の数
    /// @param j 対象となる順序
    /// @param dist 対象の連続確率密度関数
    /// @return j 番目の確率変数 x(j) が x となるときの確率密度
    distribution::Base::stat_type OrderedPDF( distribution::Base::stat_type x, distribution::Base::counter_type n, distribution::Base::counter_type j, distribution::Continuous& dist );

    namespace test
    {
      /**
         @brief マン・ホイットニーの U-検定
      **/
      class MannWhitney
      {
        using stat_type = distribution::Base::stat_type;
        using counter_type = distribution::Base::counter_type;

        counter_type m_; // データ 1 の個数
        counter_type n_; // データ 2 の個数
        stat_type u_; // データから得られた U 値

        // U 値が中央値を超えている場合(分布の右側)はtrue
        bool isRightSide_;

      public:

        /// @brief 二種類のデータの個数とそこから得られた U 値を指定して構築
        ///
        /// @param m データ 1 の個数
        /// @param n データ 2 の個数
        /// @param u データから得られた U 値
        MannWhitney( distribution::Base::counter_type m, distribution::Base::counter_type n, distribution::Base::stat_type u )
          : m_( m ), n_( n ), u_( u ),
            isRightSide_( u_ >= stat_type( m_ * n_ ) / 2.0 )
        {
          if ( isRightSide_ )
            u_ = static_cast< stat_type >( m_ * n_ ) - u_;
        }

        /// @brief 二つのデータ列と比較用関数オブジェクトを指定して構築
        ///
        /// - データ列 1 の長さが負数(開始と終了が逆)なら例外 exception::NegativeLength< std::iterator_traits< In1 >::difference_type > を投げる。
        /// - データ列 2 の長さが負数(開始と終了が逆)なら例外 exception::NegativeLength< std::iterator_traits< In2 >::difference_type > を投げる。
        ///
        /// @param s1, e1 データ列 1 の先頭と末尾の次の位置
        /// @param s2, e2 データ列 2 の先頭と末尾の次の位置
        /// @param less データ列 1 内でのデータ比較用関数オブジェクト ( [ s1, e1 ) のソートに利用 )
        /// @param cmp データ列 1 とデータ 2 の要素を比較するための関数オブジェクト
        template< class In1, class In2, class Less, class Cmp >
        MannWhitney( In1 s1, In1 e1, In2 s2, In2 e2, Less less, Cmp cmp );

        /// @brief 二つのデータ列を指定して構築
        ///
        /// - データ列 1 の長さが負数(開始と終了が逆)なら例外 exception::NegativeLength< std::iterator_traits< In1 >::difference_type > を投げる。
        /// - データ列 2 の長さが負数(開始と終了が逆)なら例外 exception::NegativeLength< std::iterator_traits< In2 >::difference_type > を投げる。
        ///
        /// @param s1, e1 データ 1 の先頭と末尾の次の位置
        /// @param s2, e2 データ 2 の先頭と末尾の次の位置
        template< class In1, class In2 >
        MannWhitney( In1 s1, In1 e1, In2 s2, In2 e2 );

        /// @brief データ 1 の個数を返す
        ///
        /// @brief データ 1 の個数
        distribution::Base::counter_type count1() const
        { return( m_ ); }

        /// @brief データ 2 の個数を返す
        ///
        /// @brief データ 2 の個数
        distribution::Base::counter_type count2() const
        { return( n_ ); }

        /// @brief U 値を返す
        ///
        /// @brief U 値
        distribution::Base::stat_type u() const
        { return( u_ ); }

        /// @brief U 値の偏り
        ///
        /// U 値が中央値に対して左側(小さい側)であれば、データ 1 の分布はデータ 2 よりも左側であることを意味する。
        ///
        /// @return U 値から得られる確率が中央値に対して右側なら true を返す。
        bool isRightSide() const
        { return( isRightSide_ ); }

        /// @brief U 値に対する p 値を返す
        ///
        /// p 値は、U 値が左に偏っていれば下側確率、右に偏っていれば上側確率を返す。
        /// そのため、返り値は 0.5 を超えない。
        ///
        /// @return p 値
        distribution::Base::stat_type p() const;

        /// @brief 危険率 a で検定を行う
        ///
        /// @param a 危険率
        /// @return 仮説を棄却する場合は true を返す
        bool test( distribution::Base::stat_type a ) const
        { return( p() < a / 2.0 ); }
      };

      /**
         @brief ウィルコクソンの符号順位検定 ( Wilcoxon Signed-rank Test )
      **/
      class Wilcoxon
      {
        using stat_type = distribution::Base::stat_type;
        using counter_type = distribution::Base::counter_type;

        counter_type n_;  // 値の総数
        stat_type sRank_; // 符号付き順位和

      public:

        /// @brief 二種類の値の個数とそこから得られた符号付き順位和を指定して構築
        ///
        /// @param n 値の個数(差分がゼロのものを除く)
        /// @param sRank 符号付き順位和
        Wilcoxon( distribution::Base::counter_type n, distribution::Base::stat_type sRank )
          : n_( n ), sRank_( sRank ) {}

        /// @brief 二つのデータ列と比較用関数オブジェクトを指定して構築
        ///
        /// - データ列 1 の長さが負数(開始と終了が逆)なら、
        ///   例外 exception::NegativeLength< std::iterator_traits< In1 >::difference_type > を投げる。
        /// - データ列 2 の長さが負数(開始と終了が逆)なら、
        ///   例外 exception::NegativeLength< std::iterator_traits< In2 >::difference_type > を投げる。
        /// - データ列 1 とデータ列 2 の長さが一致していない場合、
        ///   例外 exception::NotEqualLength< std::iterator_traits< In1 >::difference_type > を投げる。
        ///
        /// @note less は差分の絶対値を比較することを想定している。処理の中では絶対値変換を行なっていないことに注意。
        /// @note eq は、差分どうしが等しいかをチェックする他に、差分がゼロかをチェックする時にも利用される。
        /// @note 差分に対する型は x データ列の要素の型が採用される。x, y の型が異なる場合は注意。
        ///       また、x データ列の型が符号なしの場合は正しい結果が得られない。
        ///
        /// @param s1, e1 データ列 1
        /// @param s2, e2 データ列 2
        /// @param less データ列 1 とデータ列 2 の差分の絶対値を比較するための関数オブジェクト ( 差分のソートで利用 )
        /// @param eq データ列 1 とデータ列 2 の差分が等しいかを判定するための関数オブジェクト
        template< class In1, class In2, class Less, class Eq >
        Wilcoxon( In1 s1, In1 e1, In2 s2, In2 e2, Less less, Eq eq );

        /// @brief 二つのデータ列と比較用関数オブジェクトを指定して構築
        ///
        /// - データ列 1 の長さが負数(開始と終了が逆)なら、
        ///   例外 exception::NegativeLength< std::iterator_traits< In1 >::difference_type > を投げる。
        /// - データ列 2 の長さが負数(開始と終了が逆)なら、
        ///   例外 exception::NegativeLength< std::iterator_traits< In2 >::difference_type > を投げる。
        /// - データ列 1 とデータ列 2 の長さが一致していない場合、
        ///   例外 exception::NotEqualLength< std::iterator_traits< In1 >::difference_type > を投げる。
        ///
        /// @note 差分に対する型は x データ列の要素の型が採用される。x, y の型が異なる場合は注意。
        ///       また、x データ列の型が符号なしの場合は正しい結果が得られない。
        ///
        /// @param s1, e1 データ列 1
        /// @param s2, e2 データ列 2
        template< class In1, class In2 >
        Wilcoxon( In1 s1, In1 e1, In2 s2, In2 e2 );

        /// @brief 符号付き順位和の偏り
        ///
        /// 正値に対する順位和の方が大きい ( sRank > 0 ) 場合、データ 1 の方が分布は右側にある傾向となる。
        ///
        /// @return 正値に対する順位和の方が大きいなら true を返す。
        bool isRightSide() const
        { return( sRank_ > 0 ); }

        /// @brief 符号付き順位和に対する p 値を返す
        ///
        /// @return p 値
        distribution::Base::stat_type p() const;

        /// @brief 危険率 a で検定を行う
        ///
        /// @param a 危険率
        /// @return 仮説を棄却する場合は true を返す
        bool test( distribution::Base::stat_type a ) const
        { return( p() < a / 2.0 ); }
      };

      /**
         @brief クラスカル・ワリス検定 ( Kruskal Wallis Test )
      **/
      class KruskalWallis
      {
        // 統計量の型
        using stat_type = distribution::Base::stat_type;
        using counter_type = distribution::Base::counter_type;

        // 配列の二乗平均を返す関数オブジェクト
        // H 値を計算するために利用する
        struct CalcAveSq;

        counter_type classCnt_; // 集団の数
        counter_type totalCnt_; // 要素の総計
        stat_type h_;           // H 値 (各集団間の変動/順位の不偏分散)
        stat_type r_;           // R 値 (等しい順位があった場合の不偏分散の補正項)

        // データ列 data から順位 rank を作成する。
        // R 値と要素の総計も同時に計算する
        template< class T, class Less >
        static void createRank
        ( const std::vector< std::vector< T > >& data, Less less,
          std::vector< std::vector< stat_type > >* rank, stat_type* r, counter_type* totalCnt );

        // H 値を求める
        void calcH( const std::vector< std::vector< stat_type > >& rank );

        // 初期化ルーチン
        template< class T, class Less >
        void init( const std::vector< std::vector< T > >& data, Less less );

      public:

        /// @brief 元のデータ列を指定して構築
        ///
        /// テンプレート引数の T はコンストラクタに引き渡すデータ列の型を、
        /// Less はデータ列を順位に変換するときの大小比較用関数オブジェクトをそれぞれ表す。
        /// データは配列の配列で各要素が一つの集団を表す。
        ///
        /// - 自由度は 1 以上である必要があるため、集団の数は 1 より大きくなければならない。
        ///   そうでない場合は例外 exception::TooLittleNumber< distribution::Base::counter_type > を投げる。
        /// - H 値の計算で要素数での除算があるため、要素数ゼロの集団があってはならない。
        ///   そうでない場合は例外 exception::NotPositiveNumber< distribution::Base::counter_type > を投げる。
        /// - 全データが等しい場合、補正項 ( R 値 ) はゼロになる。このとき、H 値は∞として扱われる。
        ///
        /// @param data データ列(配列の配列)への参照
        /// @param less データ列の要素どうしを比較するための関数オブジェクト
        template< class T, class Less >
        KruskalWallis( const std::vector< std::vector< T > >& data, Less less )
        { init( data, less ); }

        /// @brief 元のデータ列を指定して構築
        ///
        /// テンプレート引数の T はコンストラクタに引き渡すデータ列の型を表す。
        /// データ列を順位に変換するときの大小比較用関数オブジェクトとしては std::less< T > が使われる。
        /// データは配列の配列で各要素が一つの集団を表す。
        ///
        /// - 自由度は 1 以上である必要があるため、集団の数は 1 より大きくなければならない。
        ///   そうでない場合は例外 exception::TooLittleNumber< distribution::Base::counter_type > を投げる。
        /// - H 値の計算で要素数での除算があるため、要素数ゼロの集団があってはならない。
        ///   そうでない場合は例外 exception::NotPositiveNumber< distribution::Base::counter_type > を投げる。
        /// - 全データが等しい場合、補正項 ( R 値 ) はゼロになる。このとき、H 値は∞として扱われる。
        ///
        /// @param data データ列(配列の配列)への参照
        template< class T >
        KruskalWallis( const std::vector< std::vector< T > >& data )
        { init( data, std::less< T >() ); }

        /// @brief 集団の数を返す
        ///
        /// @return 集団の数
        distribution::Base::counter_type classCnt() const
        { return( classCnt_ ); }

        /// @brief 自由度を返す
        ///
        /// H 値はこの自由度を持つカイ二乗分布に従う。
        ///
        /// @return 自由度
        distribution::Base::counter_type dof() const
        { return( classCnt_ - 1 ); }

        /// @brief 全集団の要素数の総計を返す
        ///
        /// @return 全集団の要素数の総計
        distribution::Base::counter_type totalCnt() const
        { return( totalCnt_ ); }

        /// @brief H 値を返す
        ///
        /// H 値の計算式は以下の通り。
        ///
        /// H = Sc / σ^2 = 12Σ( T[i]^2 / n[i] ) / N( N + 1 ) - 3( N + 1 )
        ///
        /// 但し T[i] と n[i] はそれぞれ集団 i の順位和と要素数、N は全集団の要素数の総計を表す。
        /// また Sc は集団間の変動、σ^2 は順位の不偏分散 ( = N( N + 1 ) / 12 ) を意味する。
        /// なお、同順位の要素があった場合、補正項 R 値で除算して補正が行われることに注意。
        ///
        /// @return H 値
        stat_type h() const
        { return( h_ ); }

        /// @brief 補正項 ( R 値 ) を返す
        ///
        /// R 値は同順位の要素が存在しない場合の不偏分散に対する、同順位が存在する場合の不偏分散の比率を表し、
        /// 以下の計算式で計算することができる。
        ///
        /// R = 1 - Σ( m[i]^3 - m[i] ) / ( N^3 - N )
        ///
        /// 但し m[i] は順位の等しい要素のグループ i の要素数、N は全集団の要素数の総計を表す。
        ///
        /// 同順位の要素がある場合、H 値を R 値で割ることで補正をかけることができる。
        ///
        /// @return R 値
        stat_type r() const
        { return( r_ ); }

        /// @brief H 値に対する p 値(上側確率)を返す
        ///
        /// @return p 値
        stat_type p() const
        { return( 1.0 - distribution::ChiSquare( dof() ).p( 0, h() ) ); }

        /// @brief 指定した危険率 a での H 値の棄却域を返す
        ///
        /// - 危険率 a は 0 以上 1 以下でなければならない。そうでない場合は例外 exception::UnitInterval を投げる。
        ///
        /// @param a 危険率
        /// @param threshold binSearch で棄却域を求める時のしきい値
        /// @return H 値の棄却域
        stat_type cr_H( distribution::Base::stat_type a, distribution::Base::stat_type threshold = DEFAULT_THRESHOLD ) const
        {
          const std::string FUNC_NAME = "tadah_fussy::statistics::test::KruskalWallis::cr_H";

          error_lib::In01( a, FUNC_NAME + " : a " + error_message::IN01 );

          return( distribution::ChiSquare( dof() ).binSearch( 1.0 - a, threshold ) );
        }

        /// @brief 危険率 a で検定を行う
        ///
        /// - 危険率 a は 0 以上 1 以下でなければならない。そうでない場合は例外 exception::UnitInterval を投げる。
        ///
        /// @param a 危険率
        /// @param threshold binSearch で棄却域を求める時のしきい値
        /// @return 仮説を棄却する場合は true を返す
        bool test( distribution::Base::stat_type a, distribution::Base::stat_type threshold = DEFAULT_THRESHOLD ) const
        { return( cr_H( a, threshold ) < h() ); }
      };

      /**
         @brief フリードマン検定 ( Friedman Test )
      **/
      class Friedman
      {
        // 統計量の型
        using stat_type = distribution::Base::stat_type;
        // 順位リストの型
        using rank_type = Matrix< stat_type >;
        // 順位データの型
        using size_type = rank_type::size_type;
        // 集合・要素数の型
        using counter_type = distribution::Base::counter_type;

        rank_type rank_; // 順位リスト
        stat_type r_;    // 補正値 R
        stat_type x2_;   // カイ二乗値

        counter_type classCnt_; // 集合の数
        counter_type dataCnt_;  // 集合内の要素数

        // 順位データの作成
        template< class In, class Less >
        void createRank( In s, In e, Less less );

        // カイ二乗値の計算
        void calcX2();

      public:

        /// @brief データ列と集団数・要素数を指定して構築
        ///
        /// - 集団の数はゼロより大きくなければならない。そうでない場合は例外 exception::NotPositiveNumber< distribution::Base::counter_type > を投げる。
        /// - 各集団の要素数は 1 より大きくなければならない。そうでない場合は例外 exception::TooLittleNumber< distribution::Base::counter_type > を投げる。
        /// - データ列の開始・終了位置の指定が正しくない(位置が逆転している)場合、
        ///   例外 exception::NegativeLength< std::iterator_traits< In >::difference_type > を投げる。
        /// - 指定した集団の数とその要素数に対し、データ列のサイズは一致していなければならない。
        ///   そうでない場合は例外 exception::NotEqualLength< std::iterator_traits< In >::difference_type > を投げる。
        /// - 集団内の要素が全て等しい場合、順位も全て等しくなる。全集団がそのような状態になったとき、
        ///   補正項 ( R 値 ) はゼロとなるので、カイ二乗値は∞とする。
        ///
        /// @param s データ列の開始位置
        /// @param e データ列の末尾の次の位置
        /// @param less データどうしを比較する関数オブジェクト
        /// @param classCnt 集団の数
        /// @param dataCnt 集団内の要素数
        template< class In, class Less >
        Friedman( In s, In e, Less less,  distribution::Base::counter_type classCnt, distribution::Base::counter_type dataCnt )
          : rank_( classCnt, dataCnt ), r_( 0 ), x2_( 0 ),
            classCnt_( classCnt ), dataCnt_( dataCnt )
        {
          const std::string FUNC_NAME = "tadah_fussy::staistics::test::Friedman constructor";

          error_lib::PNum( classCnt_, FUNC_NAME + " : classCnt " + error_message::PNUM );
          error_lib::GT< counter_type >( dataCnt_, 1, FUNC_NAME + " : dataCnt " + error_message::GT + " 1" );

          createRank( s, e, less );
          calcX2();
        }

        /// @brief 指定した集団・要素位置の順位を返す
        ///
        /// - 範囲外の番号を指定した場合は std::out_of_range を投げる。
        ///
        /// @param classIdx 集団の番号
        /// @param dataIdx 要素の番号
        /// @return 順位
        distribution::Base::stat_type rank( distribution::Base::counter_type classIdx, distribution::Base::counter_type dataIdx )
        {
          return( rank_.row( classIdx ).at( dataIdx ) );
        }

        /// @brief 集団内の要素数を返す
        ///
        /// @return 集団内の要素数
        distribution::Base::counter_type dataCnt() const
        { return( dataCnt_ ); }

        /// @brief 集団の数を返す
        ///
        /// @return 集団の数
        distribution::Base::counter_type classCnt() const
        { return( classCnt_ ); }

        /// @brief 自由度を返す
        ///
        /// カイ二乗値はこの自由度を持つカイ二乗分布に従う。
        ///
        /// @return 自由度 ( 集団の数が 1 以下ならゼロ )
        distribution::Base::counter_type dof() const
        { return( dataCnt_ - 1 ); }

        /// @brief 全集団の要素数の総計を返す
        ///
        /// @return 全集団の要素数の総計
        distribution::Base::counter_type totalCnt() const
        { return( dataCnt_ * classCnt_ ); }

        /// @brief 集団間変動と不偏分散の比 Sg / σ^2 (カイ二乗値) を返す
        ///
        /// Χ^2 = Sg / σ^2 の計算式は以下の通り。
        ///
        /// Χ^2 = Sg / σ^2 = [ 12 / nk( k + 1 ) ]Σ( T[i]^2 ) - 3n( k + 1 )
        ///
        /// 但し T[i] は集団 i の順位和、n は集団数、k は集団内の要素数をそれぞれ表す。
        /// なお、同順位の要素があった場合、補正項 R 値で除算して補正が行われることに注意。
        ///
        /// @return カイ二乗値(集団数・要素数のいずれかがゼロなら必ずゼロを返す)
        distribution::Base::stat_type x2() const
        { return( x2_ ); }

        /// @brief 補正項 ( R 値 ) を返す
        ///
        /// R 値は同順位の要素が存在しない場合の不偏分散に対する、同順位が存在する場合の不偏分散の比率を表し、
        /// 以下の計算式で計算することができる。
        ///
        /// R = 1 - Σ( m[i]^3 - m[i] ) / n( k^3 - k )
        ///
        /// 但し m[i] は順位の等しい要素のグループ i の要素数、n は集団数、k は集団内の要素数をそれぞれ表す。
        ///
        /// 同順位の要素がある場合、カイ二乗値を R 値で割ることで補正をかけることができる。
        ///
        /// @return R 値(集団数・要素数のいずれかがゼロなら必ずゼロを返す)
        distribution::Base::stat_type r() const
        { return( r_ ); }

        /// @brief カイ二乗値に対する p 値(上側確率)を返す
        ///
        /// @return p 値
        distribution::Base::stat_type p() const
        { return( 1.0 - distribution::ChiSquare( dof() ).p( 0, x2() ) ); }

        /// @brief 指定した危険率 a でのカイ二乗値の棄却域を返す
        ///
        /// - 危険率 a は 0 以上 1 以下でなければならない。そうでない場合は例外 exceptionUnitInterval を投げる。
        ///
        /// @param a 危険率
        /// @param threshold binSearch で棄却域を求める時のしきい値
        /// @return カイ二乗値の棄却域
        distribution::Base::stat_type cr_x2( distribution::Base::stat_type a, distribution::Base::stat_type threshold = DEFAULT_THRESHOLD ) const
        {
          const std::string FUNC_NAME = "tadah_fussy::statistics::test::Friedman::cr_x2";

          error_lib::In01( a, FUNC_NAME + " : a " + error_message::IN01 );

          return( distribution::ChiSquare( dof() ).binSearch( 1.0 - a, threshold ) );
        }

        /// @brief 危険率 a で検定を行う
        ///
        /// @param a 危険率
        /// @param threshold binSearch で棄却域を求める時のしきい値
        /// @return 仮説を棄却する場合は true を返す
        bool test( distribution::Base::stat_type a, distribution::Base::stat_type threshold = DEFAULT_THRESHOLD ) const
        { return( cr_x2( a, threshold ) < x2() ); }
      };

    } // namespace test

    /**
       @brief 回帰分析
    **/
    namespace regression
    {
      /**
         @brief 単回帰分析
      **/
      class Singular
      {
        using stat_type = distribution::Base::stat_type;
        using counter_type = distribution::Base::counter_type;
        using ConfidenceInterval = interval_estimation::ConfidenceInterval;
        using Continuous = distribution::Continuous;

        stat_type est_a_;     // 傾きの推定量
        stat_type est_b_;     // 切片の推定量
        MomentPair2 summary_; // 統計量
        stat_type se_;        // 誤差項の平方和

        // 回帰直線の区間推定
        void iEst_RegCoef( const Continuous& dist, stat_type var, stat_type b,
                           ConfidenceInterval* interval_a, ConfidenceInterval* interval_b,
                           stat_type threshold = DEFAULT_THRESHOLD ) const;

        // 従属変数(またはその期待値)の区間推定
        ConfidenceInterval iEst_DepVar( const Continuous& dist, stat_type var, stat_type b,
                                        stat_type x0, stat_type expFlag,
                                        stat_type threshold = DEFAULT_THRESHOLD ) const;

      public:

        /// @brief 対象の x, y データ列を指定して構築
        ///
        /// - データ列 x の指定範囲が有効かを内部でチェックしている。
        ///   有効でない場合は例外 exception::NegativeLength< std::iterator_traits< In1 >::difference_type > を投げる。
        /// - データ列 y の指定範囲が有効かを内部でチェックしている。
        ///   有効でない場合は例外 exception::NegativeLength< std::iterator_traits< In2 >::difference_type > を投げる。
        /// - 二つのデータ列 x, y の長さが等しいかを内部でチェックしている。
        ///   等しくない場合は例外 exception::NotEqualLength< std::iterator_traits< In1 >::difference_type > を投げる。
        /// - ( x, y ) データの数は 2 より多くなければならない。
        ///   そうでない場合は例外 exception::TooLittleNumber< distribution::Base::counter_type > を投げる。
        ///
        /// @param xs データ列 x の開始位置
        /// @param xe データ列 x の末尾の次の位置
        /// @param ys データ列 y の開始位置
        /// @param ye データ列 y の末尾の次の位置
        template< class In1, class In2 >
        Singular( In1 xs, In1 xe, In2 ys, In2 ye );

        /// @brief 回帰係数(傾き)の推定値を返す
        ///
        /// @return 回帰係数(傾き)の推定値
        distribution::Base::stat_type a() const { return( est_a_ ); }

        /// @brief 回帰係数(切片)の推定値を返す
        ///
        /// @return 回帰係数(切片)の推定値
        distribution::Base::stat_type b() const { return( est_b_ ); }

        /// @brief データ列から得られる統計量を返す
        ///
        /// x,yデータ列から計算された統計量を返す。この中には
        /// それぞれの平均・偏差の平方和・標本分散・不偏分散と、
        /// 二値の積和・偏差の積和・共分散が含まれる。
        ///
        /// @return 統計量
        MomentPair2 summary() const { return( summary_ ); }

        /// @brief 従属変数の推定値を返す
        ///
        /// @param x 当てはめる独立変数
        /// @return 従属変数の推定値
        distribution::Base::stat_type expY( distribution::Base::stat_type x ) const { return( a() * x + b() ); }

        /// @brief 予測値の偏差の平方和 ssy^ を返す
        ///
        /// 予測値の偏差の平方和 ssy^ は以下の式で求められる。
        ///
        /// ssy^ = Σ( y^i - my^ )^2
        ///
        /// 但し、y^i は従属変数 Y の i 番目の予測値、my^ は予測値の標本平均を表す。
        ///
        /// @return 予測値の偏差の平方和 ssy^
        distribution::Base::stat_type ssExpY() const { return( summary_.devSq2() - se_ ); }

        /// @brief 予測値に対する自由度を返す
        ///
        /// 予測値に対する自由度は常に 1 に等しい
        ///
        /// @return 予測値に対する自由度 (常に 1 を返す)
        static distribution::Base::counter_type dofExpY() { return( 1 ); }

        /// @brief 予測値の不偏分散 vy^ を返す
        ///
        /// 予測値の不偏分散 vy^ は偏差の平方和 ssy^ と等しくなる。
        ///
        /// @return 予測値の不偏分散
        distribution::Base::stat_type varExpY() const { return( ssExpY() ); }

        /// @brief 誤差項の平方和 sse を返す
        ///
        /// 回帰式からのズレ量(誤差項)の平方和 sse は以下の式で求められる。
        ///
        /// sse = Σ( yi - y^i )^2
        ///
        /// 但し、yi は従属変数 Y の i 番目の実測値、y^i はその予測値を表す。
        ///
        /// @return 誤差項の分散の不偏推定量 sse
        distribution::Base::stat_type ssE() const { return( se_ ); }

        /// @brief 誤差項に対する自由度を返す
        ///
        /// 予測値に対する自由度は ( データ数 - 2 ) に等しい
        ///
        /// @return 誤差項に対する自由度
        distribution::Base::counter_type dofE() const { return( summary_.count() - 2 ); }

        /// @brief 誤差項の分散の不偏推定量を返す
        ///
        /// 回帰式からのズレ量(誤差項)の分散の不偏推定量 ve は以下の式で求められる。
        ///
        /// ve = sse / ( N - 2 )
        ///
        /// 但し、N はデータ数を表す。
        ///
        /// @return 誤差項の分散の不偏推定量
        distribution::Base::stat_type varE() const { return( se_ / dofE() ); }

        /// @brief 観測値の偏差の平方和 ssy を返す
        ///
        /// 観測値の偏差の平方和 ssy は以下の式で求められる。
        ///
        /// ssy = Σ( yi - my )^2
        ///
        /// 但し、y^i は従属変数 Y の i 番目の観測値、my^ は観測値の標本平均を表す。
        ///
        /// @return 観測値の偏差の平方和 ssy
        distribution::Base::stat_type ssY() const { return( summary_.devSq2() ); }

        /// @brief 観測値に対する自由度を返す
        ///
        /// 観測値に対する自由度は常に ( データ数 - 1 ) に等しい
        ///
        /// @return 観測値に対する自由度
        distribution::Base::counter_type dofY() const { return( summary_.count() - 1 ); }

        /// @brief f値を返す
        ///
        /// f値は、誤差項の不偏分散に対する予測値の不偏分散の比 vy^ / ve で求められる。
        ///
        /// @return f値
        distribution::Base::stat_type f() const { return( varExpY() / varE() ); }

        /// @brief 指定した危険率 での F 値の棄却域を返す
        ///
        /// - 危険率 a は 0 以上 1 以下でなければならない。そうでない場合は例外 exception::UnitInterval を投げる。
        ///
        /// @param a 危険率
        /// @param threshold binSearch で棄却域を求める時のしきい値
        /// @return F 値の棄却域
        distribution::Base::stat_type cr_f( distribution::Base::stat_type a, distribution::Base::stat_type threshold = DEFAULT_THRESHOLD ) const
        {
          const std::string FUNC_NAME = "tadah_fussy::statistics::regression::Singular::cr_f";

          error_lib::In01( a, FUNC_NAME + " : a " + error_message::IN01 );

          return( distribution::F( dofExpY(), dofE() ).binSearch( 1.0 - a, threshold ) );
        }

        /// @brief F 値に対する p 値(上側確率)を返す
        ///
        /// @return p 値
        distribution::Base::stat_type p() const
        { return( 1.0 - distribution::F( dofExpY(), dofE() ).p( 0, f() ) ); }

        /// @brief 重相関係数を返す
        ///
        /// 重相関係数は、従属変数の観測値の偏差平方和に対する期待値の偏差平方和の比率の平方根 ( ssy^ / ssy )^(1/2) である。
        /// 重相関係数の平方は「決定係数」または「寄与率」と呼ばれる。
        /// この値は、回帰分析によって説明できる部分の比率を表す指標となる。
        ///
        /// @return 重相関係数
        distribution::Base::stat_type mcc() const
        { return( std::sqrt( ssExpY() / summary_.devSq2() ) ); }

        /*
          区間推定
        */

        /// @brief 回帰直線の区間推定(誤差項の分散が既知の時)
        ///
        /// - 信頼度 b は 0 以上 1 以下でなければならない。そうでない場合は例外 exception::UnitInterval を投げる。
        ///
        /// @param var 誤差項の分散
        /// @param b 信頼度
        /// @param interval_a 傾きの信頼区間を保持する変数へのポインタ
        /// @param interval_b 切片の信頼区間を保持する変数へのポインタ
        /// @param threshold binSearchでtを求める時のしきい値
        /// @return なし
        void iEst_RegCoef( distribution::Base::stat_type var, distribution::Base::stat_type b,
                           interval_estimation::ConfidenceInterval* interval_a, interval_estimation::ConfidenceInterval* interval_b,
                           distribution::Base::stat_type threshold = DEFAULT_THRESHOLD ) const
        { iEst_RegCoef( distribution::Normal( 0, 1 ), var, b, interval_a, interval_b, threshold ); }

        /// @brief 回帰直線の区間推定(誤差項の分散が未知の時)
        ///
        /// - 信頼度 b は 0 以上 1 以下でなければならない。そうでない場合は例外 exception::UnitInterval を投げる。
        ///
        /// @param b 信頼度
        /// @param interval_a 傾きの信頼区間を保持する変数へのポインタ
        /// @param interval_b 切片の信頼区間を保持する変数へのポインタ
        /// @param threshold binSearchでtを求める時のしきい値
        /// @return なし
        void iEst_RegCoef( distribution::Base::stat_type b,
                           interval_estimation::ConfidenceInterval* interval_a, interval_estimation::ConfidenceInterval* interval_b,
                           distribution::Base::stat_type threshold = DEFAULT_THRESHOLD ) const
        {
          iEst_RegCoef( distribution::T( dofE() ), varE(), b, interval_a, interval_b, threshold );
        }

        /// @brief 従属変数の期待値の区間推定(誤差項の分散が既知の時)
        ///
        /// - 信頼度 b は 0 以上 1 以下でなければならない。そうでない場合は例外 exception::UnitInterval を投げる。
        ///
        /// @param var 誤差項の分散
        /// @param b 信頼度
        /// @param x : 独立変数
        /// @param threshold binSearchでtを求める時のしきい値
        /// @return 求めた信頼区間
        interval_estimation::ConfidenceInterval iEst_DepVarExp( distribution::Base::stat_type var, distribution::Base::stat_type b,
                                                                distribution::Base::stat_type x,
                                                                distribution::Base::stat_type threshold = DEFAULT_THRESHOLD ) const
        { return( iEst_DepVar( distribution::Normal( 0, 1 ), var, b, x, 0, threshold ) ); }

        /// @brief 従属変数の期待値の区間推定(誤差項の分散が未知の時)
        ///
        /// - 信頼度 b は 0 以上 1 以下でなければならない。そうでない場合は例外 exception::UnitInterval を投げる。
        ///
        /// @param b 信頼度
        /// @param x : 独立変数
        /// @param threshold binSearchでtを求める時のしきい値
        /// @return 求めた信頼区間
        interval_estimation::ConfidenceInterval iEst_DepVarExp( distribution::Base::stat_type b, distribution::Base::stat_type x,
                                                                distribution::Base::stat_type threshold = DEFAULT_THRESHOLD ) const
        {
          return( iEst_DepVar( distribution::T( dofE() ), varE(), b, x, 0, threshold ) );
        }

        /// @brief 従属変数の区間推定(誤差項の分散が既知の時)
        ///
        /// - 信頼度 b は 0 以上 1 以下でなければならない。そうでない場合は例外 exception::UnitInterval を投げる。
        ///
        /// @param var 誤差項の分散
        /// @param b 信頼度
        /// @param x : 独立変数
        /// @param threshold binSearchでtを求める時のしきい値
        /// @return 求めた信頼区間
        interval_estimation::ConfidenceInterval iEst_DepVar( distribution::Base::stat_type var, distribution::Base::stat_type b,
                                                             distribution::Base::stat_type x,
                                                             distribution::Base::stat_type threshold = DEFAULT_THRESHOLD ) const
        { return( iEst_DepVar( distribution::Normal( 0, 1 ), var, b, x, 1, threshold ) ); }

        /// @brief 従属変数の区間推定(誤差項の分散が未知の時)
        ///
        /// - 信頼度 b は 0 以上 1 以下でなければならない。そうでない場合は例外 exception::UnitInterval を投げる。
        ///
        /// @param b 信頼度
        /// @param x : 独立変数
        /// @param threshold binSearchでtを求める時のしきい値
        /// @return 求めた信頼区間
        interval_estimation::ConfidenceInterval iEst_DepVar( distribution::Base::stat_type b, distribution::Base::stat_type x,
                                                             distribution::Base::stat_type threshold = DEFAULT_THRESHOLD ) const
        {
          return( iEst_DepVar( distribution::T( dofE() ), varE(), b, x, 1, threshold ) );
        }
      };

      /**
         @brief 重回帰モデル
      **/
      class Multiple
      {
        using stat_type = distribution::Base::stat_type;
        using counter_type = distribution::Base::counter_type;
        using ConfidenceInterval = interval_estimation::ConfidenceInterval;
        using Continuous = distribution::Continuous;

        std::vector< Moment2 > summary_x_; // x[i] の平均
        Moment2 summary_y_;                // y の平均
        stat_type se_; // 誤差項の平方和

        std::vector< stat_type > est_a_; // 回帰係数 ai の推定量 ( i = 1, 2, ... )
        stat_type est_a0_;               // 回帰係数 a0 の推定量
        std::vector< stat_type > est_y_; // 回帰係数 ai の推定量 ( i = 1, 2, ... )

        SquareMatrix< stat_type > s_inv_; // 共分散行列の逆行列

        // 回帰係数の推定
        template< class X, class In, class Solver >
        void estimateCoef( const Matrix< X >& x, In ys, In ye, Solver solver );

        // 誤差項の平方和を求める
        template< class X, class In >
        void calcSSE( const Matrix< X >& x, In ys );

        // 回帰直線の区間推定
        void iEst_RegCoef( const Continuous& dist, stat_type var, stat_type b,
                           std::vector< ConfidenceInterval >* interval_a,
                           stat_type threshold = DEFAULT_THRESHOLD ) const;

        // 従属変数(またはその期待値)の区間推定
        ConfidenceInterval iEst_DepVar( const Continuous& dist, stat_type var, stat_type b,
                                        const std::vector< stat_type >& x0, stat_type expFlag,
                                        stat_type threshold = DEFAULT_THRESHOLD ) const;

        // 偏相関係数算出用の標本を作成する
        template< class X >
        static void createSampleOfPartialCC_( const Matrix< X >& x,  counter_type k,
                                              std::vector< X >* xk, Matrix< X >* xl );

        // 偏相関係数を算出する
        template< class X, class In >
        static stat_type calcPartialCC_( const Multiple& mrc_y, const Multiple& mrc_xk,
                                         const std::vector< X >& xk, const Matrix< X >& xl, In ys );
        template< class X, class In >
        static stat_type calcPartialCC_( const Multiple& mrc_y,
                                         const std::vector< X >& xk, const Matrix< X >& xl, In ys, In ye );

      public:

        /// @brief データ列を指定して構築
        ///
        /// 独立変数ベクトル(X)は、一行に一つのベクトルの要素が、列ごとに各ベクトルが並んだ行列で表される。
        /// ベクトルの数(標本数)は従属変数(Y)の要素数で決まる。そのため、X行列の行数とYの要素数は一致
        /// させる必要がある。
        ///
        /// - 独立変数ベクトルの数(標本数)とその中の要素数はどちらもゼロより大きくなければならない。
        ///   そうでない場合は例外 exception::NotPositiveLength< Matrix< X >::size_type > を投げる。
        /// - 従属変数の要素の範囲が不正な(開始と終了が逆になっている)場合、
        ///   例外 exception::NegativeLength< std::iterator_traits< In >::difference_type > を投げる。
        /// - デザイン行列の行数と従属変数の要素数が一致していなければならない。
        ///   そうでない場合、例外 exception::NotEqualLength< Matrix< X >::size_type > を投げる。
        /// - 共分散行列の逆行列が得られなかった場合、また回帰係数の推定値が得られなかった(連立方程式の計算に失敗した)場合、
        ///   例外 std::runtime_error を投げる。
        /// - 標本数が、独立変数ベクトルより小さい場合は回帰係数の推定値の計算ができない(逆行列が得られない)が、
        ///   誤差の関係で計算に成功して例外が投げられない可能性がある。この時の動作は未定義である。
        /// - 標本数が、独立変数ベクトルより大きくても、共分散行列が特異な場合は逆行列が得られない。
        ///   しかし、この場合も誤差の関係で計算に成功して例外が投げられない可能性があり、
        ///   この時の動作は未定義である。
        ///
        /// @param x 独立変数ベクトル(X)を保持する行列
        /// @param ys 従属変数(Y)のデータ列開始位置
        /// @param ye 従属変数(Y)のデータ列末尾の次の位置
        /// @param solver 連立方程式の解法
        template< class X, class In, class Solver >
        Multiple( const Matrix< X >& x, In ys, In ye, Solver solver );

        /// @brief 独立変数ベクトルの要素数 p を返す
        ///
        /// @note 定数項は数に含まれないので注意
        ///
        /// @return (定数項を除く)独立変数の数 p
        distribution::Base::counter_type varSize() const { return( summary_x_.size() ); }

        /// @brief 標本数 N を返す
        ///
        /// @return 標本数 N
        distribution::Base::counter_type sampleSize() const { return( summary_y_.count() ); }

        /// @brief 回帰係数の推定値を返す
        ///
        /// 独立変数ベクトルの要素数より大きな添字を指定した場合は std::out_of_range 例外を投げる。
        ///
        /// @param i 係数(項)の番号 (0は定数項の番号を表す)
        /// @return 回帰係数の推定値
        stat_type a( distribution::Base::counter_type i ) const
        { return( ( i == 0 ) ? est_a0_ : est_a_.at( i - 1 ) ); }

        /// @brief xデータ列から得られる統計量を返す
        ///
        /// xデータ列から計算された統計量を返す。
        /// この中にはそれぞれの平均・偏差の平方和・標本分散・不偏分散が含まれる。
        ///
        /// 独立変数ベクトルの要素数以上の添字を指定した場合は std::out_of_range 例外を投げる。
        ///
        /// @param i 独立変数ベクトルの要素番号
        /// @return 対象の変数に対する統計量
        Moment2 summary_x( distribution::Base::counter_type i ) const
        { return( summary_x_.at( i ) ); }

        /// @brief yデータ列から得られる統計量を返す
        ///
        /// yデータ列から計算された統計量を返す。
        /// この中にはそれぞれの平均・偏差の平方和・標本分散・不偏分散が含まれる。
        ///
        /// @return yデータ列に対する統計量
        Moment2 summary_y() const
        { return( summary_y_ ); }

        /// @brief 共分散行列の逆行列の要素を返す
        ///
        /// 独立変数ベクトルの要素数以上の添字を指定した場合は std::out_of_range 例外を投げる。
        ///
        distribution::Base::stat_type s_inv( distribution::Base::counter_type r, distribution::Base::counter_type c ) const
        { return( s_inv_.row( r ).at( c ) ); }

        /// @brief 従属変数の推定値を返す
        ///
        /// - 引数の独立変数ベクトルの範囲が不正な(開始と終了が逆になっている)場合、
        ///   例外 exception::NegativeLength< std::iterator_traits< In >::difference_type > を投げる。
        /// - 引数の独立変数ベクトルのサイズは、独立変数ベクトルの要素数と一致していなければならない。
        ///     そうでない場合、例外 exception::NotEqualLength< std::iterator_traits< In >::difference_type > を投げる。
        ///
        /// @param xs 当てはめる独立変数ベクトルの開始位置
        /// @param xe 当てはめる独立変数ベクトルの末尾の次の位置
        /// @return 従属変数の推定値
        template< class In >
        stat_type expY( In xs, In xe ) const;

        /// @brief 予測値の偏差の平方和 ssy^ を返す
        ///
        /// 予測値の偏差の平方和 ssy^ は以下の式で求められる。
        ///
        /// ssy^ = Σ( y^i - my^ )^2
        ///
        /// 但し、y^i は従属変数 Y の i 番目の予測値、my^ は予測値の標本平均を表す。
        ///
        /// @return 予測値の偏差の平方和 ssy^
        distribution::Base::stat_type ssExpY() const { return( summary_y_.devSq() - se_ ); }

        /// @brief 予測値に対する自由度を返す
        ///
        /// 予測値に対する自由度は p に等しい。
        ///
        /// @return 予測値に対する自由度
        distribution::Base::counter_type dofExpY() const { return( varSize() ); }

        /// @brief 予測値の不偏分散 vy^ を返す
        ///
        /// vy^ = ssy^ / p で計算できる。
        ///
        /// @return 予測値の不偏分散 vy^
        distribution::Base::stat_type varExpY() const { return( ssExpY() / varSize() ); }

        /// @brief 誤差項の平方和を返す
        ///
        /// 回帰式からのズレ量(誤差項)の平方和 sse は以下の式で求められる。
        ///
        /// sse = Σ( yi - y^i )^2
        ///
        /// 但し、yi は従属変数 Y の i 番目の実測値、y^i は予測値を表す。
        ///
        /// @return 誤差項の平方和 sse
        distribution::Base::stat_type ssE() const { return( se_ ); }

        /// @brief 誤差項に対する自由度を返す
        ///
        /// 誤差項に対する自由度は N - p - 1 となる。但し、N <= p + 1 である場合はゼロを返す。
        ///
        /// @return 誤差項に対する自由度
        distribution::Base::counter_type dofE() const { return( ( sampleSize() <= varSize() + 1 ) ? 0 : sampleSize() - varSize() - 1 ); }

        /// @brief 誤差項の分散の不偏推定量 ve を返す
        ///
        /// 回帰式からのズレ量(誤差項)の分散の不偏推定量 ve は以下の式で求められる。
        ///
        /// ve = sse / ( N - p - 1 )
        ///
        /// 但し、N <= p + 1 である場合は NaN を返す。
        ///
        /// @return 誤差項の分散の不偏推定量
        distribution::Base::stat_type varE() const { return( ( dofE() == 0 ) ? distribution::Base::nan() : ssE() / dofE() ); }

        /// @brief 観測値の偏差の平方和 ssy を返す
        ///
        /// 観測値の偏差の平方和 ssy は以下の式で求められる。
        ///
        /// ssy = Σ( yi - my )^2
        ///
        /// 但し、yi は従属変数 Y の i 番目の観測値、my は観測値の標本平均を表す。
        ///
        /// @return 観測値の偏差の平方和 ssy
        distribution::Base::stat_type ssY() const { return( summary_y_.devSq() ); }

        /// @brief 観測値に対する自由度を返す
        ///
        /// 観測値に対する自由度は N - 1 に等しい
        ///
        /// @return 観測値に対する自由度
        distribution::Base::counter_type dofY() const { return( ( sampleSize() <= 1 ) ? 0 : sampleSize() - 1 ); }

        /// @brief f値を返す
        ///
        /// f値は、誤差項の不偏分散に対する予測値の不偏分散の比 vy^ / ve で求められる。
        ///
        /// @return f値
        distribution::Base::stat_type f() const { return( varExpY() / varE() ); }

        /// @brief 指定した危険率 での F 値の棄却域を返す
        ///
        /// 誤差項に対する自由度 dofE がゼロの場合は常に NaN を返す。
        ///
        /// - 危険率 a は 0 以上 1 以下でなければならない。そうでない場合は例外 exception::UnitInterval を投げる。
        ///
        /// @param a 危険率
        /// @param threshold binSearch で棄却域を求める時のしきい値
        /// @return F 値の棄却域
        distribution::Base::stat_type cr_f( distribution::Base::stat_type a, distribution::Base::stat_type threshold = DEFAULT_THRESHOLD ) const
        {
          const std::string FUNC_NAME = "MultipleRegressionCoefficient::cr_f";

          error_lib::In01( a, FUNC_NAME + " : a " + error_message::IN01 );

          if ( dofE() == 0 ) return(  distribution::Base::nan() );

          return( distribution::F( dofExpY(), dofE() ).binSearch( 1.0 - a, threshold ) );
        }

        /// @brief F 値に対する p 値(上側確率)を返す
        ///
        /// 誤差項に対する自由度 dofE がゼロの場合は常に NaN を返す。
        ///
        /// @return p 値
        distribution::Base::stat_type p() const
        {
          if ( dofE() == 0 ) return( distribution::Base::nan() );

          return( 1.0 - distribution::F( dofExpY(), dofE() ).p( 0, f() ) );
        }

        /// @brief 重相関係数を返す
        ///
        /// 重相関係数は、従属変数の観測値の偏差平方和に対する期待値の偏差平方和の比率の平方根 ( ssy^ / ssy )^(1/2) である。
        /// 重相関係数の平方は「決定係数」または「寄与率」と呼ばれる。
        /// この値は、回帰分析によって説明できる部分の比率を表す指標となる。
        ///
        /// @return 重相関係数
        distribution::Base::stat_type mcc() const
        { return( std::sqrt( ssExpY() / ssY() ) ); }

        /*
          区間推定
        */

        /// @brief 回帰直線の区間推定(誤差項の分散が既知の時)
        ///
        /// - 信頼度 b は 0 以上 1 以下でなければならない。そうでない場合は例外 exception::UnitInterval を投げる。
        ///
        /// @param var 誤差項の分散
        /// @param b 信頼度
        /// @param interval_a 回帰係数の信頼区間を保持するベクトルへのポインタ
        /// @param threshold binSearchでtを求める時のしきい値
        /// @return なし
        void iEst_RegCoef( distribution::Base::stat_type var, distribution::Base::stat_type b,
                           std::vector< interval_estimation::ConfidenceInterval >* interval_a,
                           distribution::Base::stat_type threshold = DEFAULT_THRESHOLD ) const
        { iEst_RegCoef( distribution::Normal( 0, 1 ), var, b, interval_a, threshold ); }

        /// @brief 回帰直線の区間推定(誤差項の分散が未知の時)
        ///
        /// - 信頼度 b は 0 以上 1 以下でなければならない。そうでない場合は例外 exception::UnitInterval を投げる。
        /// - 誤差項に対する自由度 dofE はゼロより大きくなければならない。
        ///   そうでない場合は例外 exception::NotPositiveNumber< distribution::Base::counter_type > を投げる。
        ///
        /// @param b 信頼度
        /// @param interval_a 回帰係数の信頼区間を保持するベクトルへのポインタ
        /// @param threshold binSearchでtを求める時のしきい値
        /// @return なし
        void iEst_RegCoef( distribution::Base::stat_type b, std::vector< interval_estimation::ConfidenceInterval >* interval_a,
                           distribution::Base::stat_type threshold = DEFAULT_THRESHOLD ) const
        { iEst_RegCoef( distribution::T( dofE() ), varE(), b, interval_a, threshold ); }

        /// @brief 従属変数の期待値の区間推定(誤差項の分散が既知の時)
        ///
        /// - 引数の独立変数ベクトルの範囲が不正な(開始と終了が逆になっている)場合、
        ///   例外 exception::NegativeLength< std::iterator_traits< In >::difference_type > を投げる。
        /// - 引数の独立変数ベクトルのサイズは、独立変数ベクトルの要素数と一致していなければならない。
        ///   そうでない場合、例外 exception::NotEqualLength< std::iterator_traits< In >::difference_type > を投げる。
        /// - 信頼度 b は 0 以上 1 以下でなければならない。そうでない場合は例外 exception::UnitInterval を投げる。
        ///
        /// @param var 誤差項の分散
        /// @param b 信頼度
        /// @param xs : 独立変数ベクトルの開始位置
        /// @param xe : 独立変数ベクトルの末尾の次の位置
        /// @param threshold binSearchでtを求める時のしきい値
        /// @return 求めた信頼区間
        template< class In >
        interval_estimation::ConfidenceInterval iEst_DepVarExp( distribution::Base::stat_type var, distribution::Base::stat_type b, In xs, In xe,
                                                                distribution::Base::stat_type threshold = DEFAULT_THRESHOLD ) const
        {
          const std::string FUNC_NAME = "tadah_fussy::statistics::regression::Multiple::iEst_DepVarExp";

          error_lib::CheckArrayLength( xs, xe, varSize(),
                                       FUNC_NAME + " : [ xs, xe ) " + error_message::NEGATIVE_LENGTH,
                                       FUNC_NAME + " : length of [ xs, xe ) " + error_message::NOT_EQUAL_LENGTH + "varSize()" );

          std::vector< stat_type > x( xs, xe );

          return( iEst_DepVar( distribution::Normal( 0, 1 ), var, b, x, 0, threshold ) );
        }

        /// @brief 従属変数の期待値の区間推定(誤差項の分散が未知の時)
        ///
        /// - 引数の独立変数ベクトルの範囲が不正な(開始と終了が逆になっている)場合、
        ///   例外 exception::NegativeLength< std::iterator_traits< In >::difference_type > を投げる。
        /// - 引数の独立変数ベクトルのサイズは、独立変数ベクトルの要素数と一致していなければならない。
        ///   そうでない場合、例外 exception::NotEqualLength< std::iterator_traits< In >::difference_type > を投げる。
        /// - 信頼度 b は 0 以上 1 以下でなければならない。そうでない場合は例外 exception::UnitInterval を投げる。
        /// - 誤差項に対する自由度 dofE はゼロより大きくなければならない。
        ///   そうでない場合は例外 exception::NotPositiveNumber<  distribution::Base::counter_type > を投げる。
        ///
        /// @param b 信頼度
        /// @param xs : 独立変数ベクトルの開始位置
        /// @param xe : 独立変数ベクトルの末尾の次の位置
        /// @param threshold binSearchでtを求める時のしきい値
        /// @return 求めた信頼区間
        template< class In >
        interval_estimation::ConfidenceInterval iEst_DepVarExp( distribution::Base::stat_type b, In xs, In xe,
                                                                distribution::Base::stat_type threshold = DEFAULT_THRESHOLD ) const
        {
          const std::string FUNC_NAME = "tadah_fussy::statistics::regression::Multiple::iEst_DepVarExp";

          error_lib::PNum( dofE(), FUNC_NAME + " : DOF of error " + error_message::PNUM + " ( check dofE() )" ); // データ数は 3 以上必要
          error_lib::CheckArrayLength( xs, xe, varSize(),
                                       FUNC_NAME + " : [ xs, xe ) " + error_message::NEGATIVE_LENGTH,
                                       FUNC_NAME + " : length of [ xs, xe ) " + error_message::NOT_EQUAL_LENGTH + "varSize()" );

          std::vector< stat_type > x( xs, xe );

          return( iEst_DepVar( distribution::T( dofE() ), varE(), b, x, 0, threshold ) );
        }

        /// @brief 従属変数の区間推定(誤差項の分散が既知の時)
        ///
        /// - 引数の独立変数ベクトルの範囲が不正な(開始と終了が逆になっている)場合、
        ///   例外 exception::NegativeLength< std::iterator_traits< In >::difference_type > を投げる。
        /// - 引数の独立変数ベクトルのサイズは、独立変数ベクトルの要素数と一致していなければならない。
        ///   そうでない場合、例外 exception::NotEqualLength< std::iterator_traits< In >::difference_type > を投げる。
        /// - 信頼度 b は 0 以上 1 以下でなければならない。そうでない場合は例外 exception::UnitInterval を投げる。
        ///
        /// @param var 誤差項の分散
        /// @param b 信頼度
        /// @param xs : 独立変数ベクトルの開始位置
        /// @param xe : 独立変数ベクトルの末尾の次の位置
        /// @param threshold binSearchでtを求める時のしきい値
        /// @return 求めた信頼区間
        template< class In >
        interval_estimation::ConfidenceInterval iEst_DepVar( distribution::Base::stat_type var, distribution::Base::stat_type b, In xs, In xe,
                                                             distribution::Base::stat_type threshold = DEFAULT_THRESHOLD ) const
        {
          const std::string FUNC_NAME = "tadah_fussy::statistics::regression::Multiple::iEst_DepVar";

          error_lib::CheckArrayLength( xs, xe, varSize(),
                                       FUNC_NAME + " : [ xs, xe ) " + error_message::NEGATIVE_LENGTH,
                                       FUNC_NAME + " : length of [ xs, xe ) " + error_message::NOT_EQUAL_LENGTH + "varSize()" );

          std::vector< stat_type > x( xs, xe );

          return( iEst_DepVar( distribution::Normal( 0, 1 ), var, b, x, 1, threshold ) );
        }

        /// @brief 従属変数の区間推定(誤差項の分散が未知の時)
        ///
        /// - 引数の独立変数ベクトルの範囲が不正な(開始と終了が逆になっている)場合、例外 exception::NegativeLength を投げる。
        /// - 引数の独立変数ベクトルのサイズは、独立変数ベクトルの要素数と一致していなければならない。
        ///   そうでない場合、例外 exception::NotEqualLength< std::iterator_traits< In >::difference_type > を投げる。
        /// - 信頼度 b は 0 以上 1 以下でなければならない。そうでない場合は例外 exception::UnitInterval を投げる。
        /// - 誤差項に対する自由度 dofE はゼロより大きくなければならない。そうでない場合は例外 exception::NotPositiveNumber を投げる。
        ///
        /// @param b 信頼度
        /// @param xs : 独立変数ベクトルの開始位置
        /// @param xe : 独立変数ベクトルの末尾の次の位置
        /// @param threshold binSearchでtを求める時のしきい値
        /// @return 求めた信頼区間
        template< class In >
        interval_estimation::ConfidenceInterval iEst_DepVar( distribution::Base::stat_type b, In xs, In xe,
                                                             distribution::Base::stat_type threshold = DEFAULT_THRESHOLD ) const
        {
          const std::string FUNC_NAME = "tadah_fussy::statistics::regression::Multiple::iEst_DepVar";

          error_lib::PNum( dofE(), FUNC_NAME + " : DOF of error " + error_message::PNUM + " ( check dofE() )" ); // データ数は 3 以上必要
          error_lib::CheckArrayLength( xs, xe, varSize(),
                                       FUNC_NAME + " : [ xs, xe ) " + error_message::NEGATIVE_LENGTH,
                                       FUNC_NAME + " : length of [ xs, xe ) " + error_message::NOT_EQUAL_LENGTH + "varSize()" );

          std::vector< stat_type > x( xs, xe );

          return( iEst_DepVar( distribution::T( dofE() ), varE(), b, x, 1, threshold ) );
        }

        /*
          偏相関係数
        */

        /// @brief 偏相関係数を算出する
        ///
        /// 独立変数ベクトルの k 番目の要素に対する偏相関係数 r は、以下の式で計算することができる。
        ///
        /// r = Σ( u[i]v[i] ) / ( Σ u[i]^2・Σ v[i]^2 )^(1/2)
        ///
        /// 但し、u[i] は、従属変数 Y に対し、k 番目の要素を除いた独立変数ベクトルから求められた予測値と実測値との差分を、
        /// v[i] は、k 番目の独立変数に対し、k 番目の要素を除いた独立変数ベクトルから求められた予測値と実測値との差分を
        /// それぞれ表す。これは、k 番目の要素だけの Y との相関係数と考えることができる。
        ///
        /// - 独立変数ベクトルの中の対象の要素番号 k が範囲外である場合、
        ///   例外 exception::TooGreatNumber<  distribution::Base::counter_type > を投げる。
        /// - 独立変数ベクトルの数(標本数)とその中の要素数はどちらもゼロより大きくなければならない。
        ///   そうでない場合は例外 exception::NotPositiveLength< Matrix< X >::size_type > を投げる。
        /// - 従属変数の要素の範囲が不正な(開始と終了が逆になっている)場合、
        ///   例外 exception::NegativeLength< std::iterator_traits< In >::difference_type > を投げる。
        /// - デザイン行列の行数と従属変数の要素数が一致していなければならない。
        ///   そうでない場合、例外 exception::NotEqualLength< Matrix< X >::size_type > を投げる。
        /// - 共分散行列の逆行列が得られなかった場合、また回帰係数の推定値が得られなかった(連立方程式の計算に失敗した)場合、
        ///   例外 std::runtime_error を投げる。
        /// - 標本数が、独立変数ベクトルより小さい場合は回帰係数の推定値の計算ができない(逆行列が得られない)が、
        ///   誤差の関係で計算に成功して例外が投げられない可能性がある。この時の動作は未定義である。
        /// - 標本数が、独立変数ベクトルより大きくても、共分散行列が特異な場合は逆行列が得られない。
        ///   しかし、この場合も誤差の関係で計算に成功して例外が投げられない可能性があり、
        ///   この時の動作は未定義である。
        ///
        /// @note 異なる計算方法もある。tadah_fussy::statistics::regression::Multiple::calcPartialCC2 を参照。
        ///
        /// @param x 独立変数ベクトルの集合(行方向に各独立変数ベクトルが並ぶ)
        /// @param ys 従属変数列の開始位置
        /// @param ye 従属変数列の末尾の次の位置
        /// @param k 独立変数ベクトルの中の対象の要素番号
        /// @param solver 連立方程式の解法
        /// @return 偏相関係数
        template< class X, class In, class Solver >
        static distribution::Base::stat_type calcPartialCC( const Matrix< X >& x, In ys, In ye, distribution::Base::counter_type k, Solver solver )
        {
          const std::string FUNC_NAME = "tadah_fussy::statistics::regression::Multiple::calcPartialCC";

          std::vector< X > xk; // k 番目の独立変数の標本
          Matrix< X > xl;      // k 番目を除いた独立変数の標本

          error_lib::LT< counter_type >( k, x.cols(), FUNC_NAME + " : k " + error_message::LT + " x.cols()" );

          createSampleOfPartialCC_( x, k, &xk, &xl );

          Multiple mrc_y( xl, ys, ye, solver ); //  y と xl の重回帰モデル
          Multiple mrc_xk( xl, xk.begin(), xk.end(), solver );  // xk と xl の重回帰モデル

          return( calcPartialCC_( mrc_y, mrc_xk, xk, xl, ys ) );
        }

        /// @brief 偏相関係数を算出する
        ///
        /// 独立変数ベクトルの k 番目の要素 xk に対する偏相関係数 r は、以下の式で計算することができる。
        ///
        /// r = ( Syk - Σyk ) / [ ( Syy - Σyy )^(1/2) ( Skk - Σkk )^(1/2) ]
        ///
        /// 但し、Syk は従属変数 Y と k 番目の要素との共分散、Syy と Skk はそれぞれ従属変数 Y と k 番目の要素の分散を表す。
        /// また、k 番目の要素を除く独立変数ベクトルを xl とし、xl の共分散行列の逆行列を inv(Vl)、
        /// xl の各要素と Y の共分散からなるベクトルを Syl、xl の各要素と xk との共分散からなるベクトルを Skl としたとき、
        ///
        /// Σyk = ( Syl, inv(Vl)Skl ) ; Σyy = ( Syl, inv(Vl)Syl ) ; Σkk = ( Skl, inv(Vl)Skl )
        ///
        /// となる。
        ///
        /// - 独立変数ベクトルの中の対象の要素番号 k が範囲外である場合、
        ///   例外 exception::TooGreatNumber< distribution::Base::counter_type > を投げる。
        /// - 独立変数ベクトルの数(標本数)とその中の要素数はどちらもゼロより大きくなければならない。
        ///   そうでない場合は例外 exception::NotPositiveLength< Matrix< X >::size_type > を投げる。
        /// - 従属変数の要素の範囲が不正な(開始と終了が逆になっている)場合、
        ///   例外 exception::NegativeLength< std::iterator_traits< In >::difference_type > を投げる。
        /// - デザイン行列の行数と従属変数の要素数が一致していなければならない。
        ///   そうでない場合、例外 exception::NotEqualLength< Matrix< X >::size_type > を投げる。
        /// - 共分散行列の逆行列が得られなかった場合、また回帰係数の推定値が得られなかった(連立方程式の計算に失敗した)場合、
        ///   例外 std::runtime_error を投げる。
        /// - 標本数が、独立変数ベクトルより小さい場合は回帰係数の推定値の計算ができない(逆行列が得られない)が、
        ///   誤差の関係で計算に成功して例外が投げられない可能性がある。この時の動作は未定義である。
        /// - 標本数が、独立変数ベクトルより大きくても、共分散行列が特異な場合は逆行列が得られない。
        ///   しかし、この場合も誤差の関係で計算に成功して例外が投げられない可能性があり、
        ///   この時の動作は未定義である。
        ///
        /// @note 異なる計算方法もある。tadah_fussy::statistics::regression::Multiple::calcPartialCC を参照。
        ///
        /// @param x 独立変数ベクトルの集合(行方向に各独立変数ベクトルが並ぶ)
        /// @param ys 従属変数列の開始位置
        /// @param ye 従属変数列の末尾の次の位置
        /// @param k 独立変数ベクトルの中の対象の要素番号
        /// @param solver 連立方程式の解法
        /// @return 偏相関係数
        template< class X, class In, class Solver >
        static distribution::Base::stat_type calcPartialCC2( const Matrix< X >& x, In ys, In ye, distribution::Base::counter_type k, Solver solver )
        {
          const std::string FUNC_NAME = "tadah_fussy::statistics::regression::Multiple::calcPartialCC2";

          std::vector< X > xk; // k 番目の独立変数の標本
          Matrix< X > xl;      // k 番目を除いた独立変数の標本

          error_lib::LT< counter_type >( k, x.cols(), FUNC_NAME + " : k " + error_message::LT + " x.cols()" );

          createSampleOfPartialCC_( x, k, &xk, &xl );

          Multiple mrc_y( xl, ys, ye, solver );   //  y と xl の重回帰モデル

          return( calcPartialCC_( mrc_y, xk, xl, ys, ye ) );
        }
      };

    } // namespace regression

    /************************************************************************
          相関係数
    ************************************************************************/

    namespace interval_estimation
    {
      /**
         @brief 相関係数の区間推定
      **/
      class CorrCoef
      {
        // 相関係数のデータ型
        using stat_type = distribution::Base::stat_type;

        MomentPair2 summary_; // x, y データ列から得られる統計量

      public:

        /// @brief 対象の x, y データ列を指定して構築
        ///
        /// - データ列 x の指定範囲が有効かを内部でチェックしている。
        ///   有効でない場合は例外 exception::NegativeLength< std::iterator_traits< In1 >::difference_type > を投げる。
        /// - データ列 y の指定範囲が有効かを内部でチェックしている。
        ///   有効でない場合は例外 exception::NegativeLength< std::iterator_traits< In2 >::difference_type > を投げる。
        /// - 二つのデータ列の長さが等しいかを内部でチェックしている。
        ///   等しくない場合は例外 exception::NotEqualLength< std::iterator_traits< In1 >::difference_type > を投げる。
        ///
        /// @param xs データ列 x の開始位置
        /// @param xe データ列 x の末尾の次の位置
        /// @param ys データ列 y の開始位置
        /// @param ye データ列 y の末尾の次の位置
        template< class In1, class In2 >
        CorrCoef( In1 xs, In1 xe, In2 ys, In2 ye )
        { summary_ = MomentPair2( xs, xe, ys, ye ); }

        /// @brief 標本相関係数を返す
        ///
        /// @return 標本相関係数
        distribution::Base::stat_type cor() const
        { return(
                 ( summary_.var1() * summary_.var2() == 0 ) ?
                 distribution::Base::nan() :
                 summary_.covariance() / std::sqrt( summary_.var1() * summary_.var2() )
                 );
        }

        /// @brief z値を返す
        ///
        /// 標本相関係数 r に対し、z 値 z(r) は
        ///
        /// z(r) = ln( ( 1 + r ) / ( 1 - r ) ) / 2
        ///
        /// で求めることができる。z 値は正規分布 N( z(ρ), 1 / ( N - 3 ) ) に近似的に従う。
        /// ここでρは母相関係数、N はデータ数をそれぞれ表す。
        ///
        /// @return z値
        distribution::Base::stat_type z() const
        { return( std::log( ( 1.0 + cor() ) / ( 1.0 - cor() ) ) / 2.0 ); }

        /// @brief 相関係数の区間推定量を返す
        ///
        /// - 信頼度 b は 0 以上 1 以下でなければならない。そうでない場合は例外 exception::UnitInterval を投げる。
        /// - データ数が 3 以下だった場合、例外 exception::TooLittleNumber<  distribution::Base::!counter_type > を投げる。
        ///
        /// @param b 信頼度
        /// @param threshold binSearchでのしきい値
        /// @return 相関係数の区間推定量
        interval_estimation::ConfidenceInterval operator()( distribution::Base::stat_type b, distribution::Base::stat_type threshold = DEFAULT_THRESHOLD );
      };

    } // namespace interval_estimation

    /**
       @brief 順位相関係数
    **/
    class RankCorrCoef
    {
      using stat_type = distribution::Base::stat_type;      // 順位相関係数のデータ型
      using rank_type = std::vector< stat_type >; // 順位に変換する配列の型
      using size_type = rank_type::size_type;     // rank_type のサイズを表す型

      rank_type x_; // X の順位データ
      rank_type y_; // Y の順位データ

    public:

      /// @brief データ列を順位データに変換する
      ///
      /// テンプレート引数の Less はデータを比較するための関数オブジェクトの型、
      /// In はコンテナの iterator の型であることが前提となる。
      ///
      /// - データ列の指定範囲が有効かを内部でチェックしている。
      ///   有効でない場合は例外 exception::NegativeLength< std::iterator_traits< In >::difference_type > を投げる。
      ///
      /// @param s データ列の開始位置
      /// @param e データ列の末尾の次の位置
      /// @param rank 順位を登録する配列へのポインタ
      /// @param less データ比較用の関数オブジェクト
      /// @return なし
      template< class In, class Less >
      static void getRank( In s, In e, std::vector< distribution::Base::stat_type >* rank, Less less );

      /// @brief データ列を順位データに変換する
      ///
      /// In はコンテナの iterator の型であることが前提となる。
      /// データの比較に std::less を利用する。
      ///
      /// - データ列の指定範囲が有効かを内部でチェックしている。
      ///   有効でない場合は例外 exception::NegativeLength< std::iterator_traits< In >::difference_type > を投げる。
      ///
      /// @param s データ列の開始位置
      /// @param e データ列の末尾の次の位置
      /// @param rank 順位を登録する配列へのポインタ
      /// @return なし
      template< class In >
      static void getRank( In s, In e, std::vector< distribution::Base::stat_type >* rank )
      { getRank( s, e, rank, std::less< typename std::iterator_traits< In >::value_type >() ); }

      /// @brief データ列から構築
      ///
      /// テンプレート引数の In1, In2 はそれぞれ X, Y データ列の反復子の型を表し、コンテナの iterator の型であることが前提となる。
      /// また、Less1, Less2 は X, Y それぞれについてデータを比較するための関数オブジェクトの型を表す。
      ///
      /// - Xデータ列の指定範囲が有効かを内部でチェックしている。
      ///   有効でない場合は例外 exception::NegativeLength< std::iterator_traits< In1 >::difference_type > を投げる。
      /// - Yデータ列の指定範囲が有効かを内部でチェックしている。
      ///   有効でない場合は例外 exception::NegativeLength< std::iterator_traits< In2 >::difference_type > を投げる。
      /// - 二つのデータ列の長さが等しいかを内部でチェックしている。
      ///   等しくない場合は例外 exception::NotEqualLength< std::iterator_traits< In1 >::difference_type > を投げる。
      ///
      /// @param xs Xデータ列の開始位置
      /// @param xe Xデータ列の末尾の次の位置
      /// @param ys Yデータ列の開始位置
      /// @param ye Yデータ列の末尾の次の位置
      /// @param less1 Xデータの比較用関数オブジェクト
      /// @param less2 Yデータの比較用関数オブジェクト
      template< class In1, class In2, class Less1, class Less2 >
      RankCorrCoef( In1 xs, In1 xe, In2 ys, In2 ye, Less1 less1, Less2 less2 );

      /// @brief データ列から構築
      ///
      /// テンプレート引数の In1, In2 はそれぞれ X, Y データ列の反復子の型を表し、コンテナの iterator の型であることが前提となる。
      /// データの比較に std::less を利用する。
      ///
      /// - Xデータ列の指定範囲が有効かを内部でチェックしている。
      ///   有効でない場合は例外 exception::NegativeLength< std::iterator_traits< In1 >::difference_type > を投げる。
      /// - Yデータ列の指定範囲が有効かを内部でチェックしている。
      ///   有効でない場合は例外 exception::NegativeLength< std::iterator_traits< In2 >::difference_type > を投げる。
      /// - 二つのデータ列の長さが等しいかを内部でチェックしている。
      ///   等しくない場合は例外 exception::NotEqualLength< std::iterator_traits< In1 >::difference_type > を投げる。
      ///
      /// @param xs Xデータ列の開始位置
      /// @param xe Xデータ列の末尾の次の位置
      /// @param ys Yデータ列の開始位置
      /// @param ye Yデータ列の末尾の次の位置
      template< class In1, class In2 >
      RankCorrCoef( In1 xs, In1 xe, In2 ys, In2 ye );

      /// @brief スピアマンの順位相関係数(ρ)を返す
      ///
      /// @return スピアマンの順位相関係数(ρ)
      stat_type spearmanRankCorrCoef() const;

      /// @brief ケンドールの順位相関係数(τ)を返す
      ///
      /// @return ケンドールの順位相関係数(τ)
      stat_type kendallRankCorrCoef() const;
    };

    /************************************************************************
    因子分析
    ************************************************************************/

    /**
       @brief 因子分析用クラス

       複数のデータ種を持つ集団を { zj,i } とする。j はデータ種の番号、i はデータ番号を表す。
       次のような統計的モデルを考える。

       zj,i = Σk( aj,k・fk,i ) + ej,i

       fk,j は「共通因子」と呼ばれ、各データに対する fk,j の値は「因子得点」という。
       また、aj,k は共通因子の zj,i への重みを表し「因子負荷量」という。k は共通因子に対する番号になる。
       ej,i は共通因子以外の各データ独自の変動量であり「独自因子」という。

       因子分析では、データを共通因子と独自因子の二つで表現・解釈する。
    **/
    class FactorAnalysis
    {
    public:

      using stat_type = distribution::Base::stat_type;        ///< 確率変数や統計量の型
      using size_type = SquareMatrix< stat_type >::size_type; ///< サイズの型

    private:

      Matrix< stat_type > data_;            // 正規化したデータ列
      SymmetricMatrix< stat_type > cov_;    // 共分散行列
      std::vector< stat_type > eigenValue_; // 固有値

      // 因子負荷量 a の二乗の転置行列を求めて at に返す
      // 転置を行う時に共通性 h で除算する処理を併せて行う。
      static void transpose( const Matrix< stat_type >& a, const std::vector< stat_type >& h, Matrix< stat_type >* at );

      // 因子負荷量の二乗の転置行列 at を使い、分散を最大にする回転角を返す
      static stat_type calcMaxVarAngle( const Matrix< stat_type >& at, size_type i0, size_type i1 );

    public:

      /// @brief データ列を渡して構築
      ///
      /// データ列は一列に同種のデータが ( 一行が一つの独立変数ベクトルとして ) 並ぶとして処理する。
      ///
      /// - データの正規化を内部で行なっているため、失敗した場合は例外 std::runtime_error を投げる。標準偏差がゼロの場合にこの例外が発生する。
      /// - 固有値計算に失敗した場合、例外 std::runtime_error を投げる。
      ///
      /// @param data データ列
      /// @param threshold 固有値計算時の反復処理において解の精度に対するしきい値
      /// @param maxCnt 固有値計算時の反復処理における最大処理回数
      FactorAnalysis( const Matrix< stat_type >& data, stat_type threshold = DEFAULT_THRESHOLD, uint32_t maxCnt = DEFAULT_MAX_COUNT );

      /// @brief データ種の数を返す
      ///
      /// @return データ種の数
      size_type classCnt() const
      { return( cov_.size() ); }

      /// @brief 正規化データから求めた共分散行列の要素を返す
      ///
      /// @param r 行番号
      /// @param c 列番号
      /// @return r 行 c 列の共分散行列の要素
      stat_type cov( size_type r, size_type c ) const
      { return( cov_.row( r ).at( c ) ); }

      /// @brief 固有値を取得する
      ///
      /// @param eigenValue 固有値を取得する配列のポインタ
      /// @return なし
      void getEigenValue( std::vector< stat_type >* eigenValue ) const
      {
        assert( eigenValue != nullptr );
        eigenValue->assign( eigenValue_.begin(), eigenValue_.end() );
      }

      /// @brief 固有値を取得する
      ///
      /// @param eigenValue 固有値を取得するヒストグラムのポインタ
      /// @return なし
      void getEigenValue( Histogram< size_t, stat_type >* eigenValue ) const
      {
        assert( eigenValue != nullptr );
        eigenValue->clear();
        for ( size_t i = 0 ; i < eigenValue_.size() ; ++i )
          (*eigenValue)[i] = eigenValue_[i];
      }

      /// @brief SMC 法による独自因子の分散の初期値の算出
      ///
      /// ある種別のデータとそれ以外のデータの間の重相関係数の二乗 ( Squared Multiple Correlation ; SMC ) を求め、
      /// 因子負荷量の推定における独自因子の分散の初期値として利用する。
      ///
      /// @param d 計算した独自因子の分散の初期値を取得する配列へのポインタ
      /// @param solver 連立方程式の解法
      /// @return なし
      template< class Solver >
      void calcSMC( std::vector< stat_type >* d, Solver solver ) const;

      /// @brief 因子負荷量の推定(主因子分析法)
      ///
      /// - 固有値計算に失敗した場合、例外 std::runtime_error を投げる。
      ///
      /// @param m 因子数
      /// @param a 求めた因子負荷量を返す行列へのポインタ (一列に同じ共通因子の因子負荷量が並ぶ)
      /// @param d 求めた独自因子の分散を返す配列へのポインタ(あらかじめ初期値を登録しておく)
      /// @param iterThreshold 反復法のしきい値
      /// @param qrThreshold 固有値計算時のしきい値
      /// @param iterMaxCnt 反復法の最大処理回数
      /// @param qrMaxCnt 固有値計算時の最大計算回数
      /// @return 固有値計算/反復法で最大処理回数をオーバーしたら false を返す
      bool estimateFactorLoading( size_type m, Matrix< stat_type >* a, std::vector< stat_type >* d,
                                  stat_type iterThreshold = DEFAULT_THRESHOLD, stat_type qrThreshold = DEFAULT_THRESHOLD,
                                  uint32_t iterMaxCnt = DEFAULT_MAX_COUNT, uint32_t qrMaxCnt = DEFAULT_MAX_COUNT ) const;

      /// @brief 共通性による因子負荷量の標準化
      ///
      /// 共通性は、独自因子 ej,i の分散値を 1 から引いた値である。この値は分散の中で共通因子が占める割合を表す。
      /// 基準 Varimax 法では、因子負荷量を共通性で割った値を使う。共通性で割らずにそのまま回転を行った場合は素 Varimax 法となる。
      /// 共通性が大きい因子負荷量ほど回転に与える影響は大きくなるため、通常は標準化した上で Varimax 法を適用したほうがよい。
      ///
      /// @param d 独自因子の分散(FactorAnalysis::estimateFactorLoadingにて得られた値)
      /// @param a 標準化する因子負荷量へのポインタ(FactorAnalysis::estimateFactorLoadingにて得られた値)
      /// @return なし
      static void normalizeByCommunality( const std::vector< stat_type >& d, Matrix< stat_type >* a );

      /// @brief 因子負荷量の回転(Varimax法)
      ///
      /// @param a 処理対象の因子負荷量へのポインタ
      /// @param threshold 反復法のしきい値
      /// @param maxCnt 反復法の最大処理回数
      /// @return 反復法で最大処理回数をオーバーしたら false を返す
      static bool varimax( Matrix< stat_type >* a, stat_type threshold = DEFAULT_THRESHOLD, uint32_t maxCnt = DEFAULT_MAX_COUNT );

      /// @brief 因子負荷量 a を使い、因子得点の線形式の係数を b に求める
      ///
      /// 独自因子がゼロのときの因子得点の推定値を f~k,i としたとき、
      ///
      /// f~k,i = Σj( bk,j・zj,i )
      ///
      /// と表せるような係数 bk,j を求める。
      ///
      /// 因子負荷量 a は FactorAnalysis::estimateFactorLoading や FactorAnalysis::varimax を使って求めた行列をそのまま使うことを想定している。
      ///
      /// - 共分散行列の逆行列の計算に失敗した場合、例外 std::runtime_error を投げる。
      ///
      /// @param a 因子負荷量
      /// @param b 求めた因子得点の線形式の係数を返す行列へのポインタ
      /// @param solver 連立方程式の解法
      /// @return なし
      template< class Solver >
      void estimateFScoreCoef( const Matrix< stat_type >& a, Matrix< stat_type >* b, Solver solver ) const;

      /// @brief 因子得点の線形式の係数 b を使い、各データの因子得点の推定値を計算して fs に返す
      ///
      /// @param b 因子得点の線形式の係数(m x p 行列)
      /// @param fs 各データの因子得点の推定値を返す行列へのポインタ(n x p行列にリサイズされる)
      void getFactorScore( const Matrix< stat_type >& b, Matrix< stat_type >* fs ) const;
    };

  } // namespace statistics

} // namespace tadah_fussy

#include "statistics_inline.h"

#endif
